program vismo_test1
   use vismo
 
   implicit none
   integer, parameter :: LXSIZE = 256
   integer, parameter :: LYSIZE = 256
   integer, parameter :: LZSIZE = 256
   integer, parameter :: dp = selected_real_kind(14)

   real(kind=dp), allocatable, dimension(:, :, :), target :: scalar, scalar2
   real(kind=dp), allocatable, dimension(:, :, :), target :: vectx, vecty, vectz

   real(kind=dp), allocatable, dimension(:), target :: ptx, pty, ptz, pts
   integer, allocatable, dimension(:), target :: ptq

   integer, parameter :: particlenum = 10

   real(kind=dp), dimension(3) :: origin, dx
   integer, dimension(3) :: n
   real(kind=dp), dimension(3) :: loccorner
   integer, dimension(3) :: ln, rln, lx0

   integer :: i, myrank, psize, comm
   real(kind=dp) :: tb, te, t0, t1, t
   real(kind=dp) :: tbmin, temax

   allocate(scalar(LXSIZE, LYSIZE, LZSIZE), scalar2(LXSIZE, LYSIZE, LZSIZE))
   allocate(vectx(LXSIZE, LYSIZE, LZSIZE),vecty(LXSIZE, LYSIZE, LZSIZE),vectz(LXSIZE, LYSIZE, LZSIZE))
   allocate(ptx(particlenum), pty(particlenum),ptz(particlenum),pts(particlenum),ptq(particlenum))
   
   print *, "making data"
   myrank = 0; psize = 1; comm=0
   ! --------- Making Data ---------
   call makeCrdData(origin, n, dx, loccorner, ln, rln, lx0, myrank)
   call makeFldData(scalar2, vectx, vecty, vectz, loccorner, origin, dx)
   call makeFldData_old(scalar, vectx, vecty, vectz, loccorner, origin, dx)
   call makePTCLData(ptx, pty, ptz, ptq, pts, loccorner, dx)

!  scalar2(1:LXSIZE, 1:LYSIZE, 1:LZSIZE) = scalar(1:LXSIZE, 1:LYSIZE, 1:LZSIZE)
   ! ------ End of Making Data -----
     print *, "vismo Init"
     call vismo__init("configtest.vsm")

     print *, "vismo SetData"
     call vismo__addScalar(scalar, LXSIZE, LYSIZE, LZSIZE)
     call vismo__addScalar(scalar2, LXSIZE, LYSIZE, LZSIZE)
     call vismo__addVector(vectx, vecty, vectz, LXSIZE, LYSIZE, LZSIZE)
!     call vismo__addPtcl(ptx,pty,ptz,ptq,particlenum)
     call vismo__addPtclCol(ptx,pty,ptz,ptq,pts, particlenum)

     print *, "vismo SetCoord"
     call vismo__initCoords(.true.) ! Uniform = .true., Rect = .false.
     call vismo__setUniCoord(n, origin, dx)
     call vismo__setLocalUniCoord(ln, lx0, rln, loccorner, dx)

     print *, "vismo PrepareVis"
     call vismo__prepareVis

   ! < computation loop>

     print *, "vismo Visualization"
     call vismo__visualization(0) ! arg is timestep (integer)

   ! <end of computation loop>

     print *, "vismo Finalize"
     call vismo__finalize

   if(myrank == 0) then
!     print *,"total time:", te-tb
     !print *,"making data time", t1-t0
   end if

   deallocate(scalar, scalar2, vectx, vecty, vectz, ptx,pty,ptz,pts,ptq)

 contains

subroutine makePTCLData(ptx, pty, ptz, ptq, pts, lcorner, dx)
   real(kind=dp), dimension(particlenum), intent(out) :: ptx, pty, ptz, pts
   integer, dimension(particlenum), intent(out) :: ptq
   real(kind=dp), dimension(3), intent(in) :: lcorner, dx
   real(kind=dp), dimension(3) :: center
   real(kind=dp) :: r, theta, z, zlen, hxlen,s
   real(kind=dp), parameter :: pi = 3.14159
   integer :: i
!   print *,"ptcldata, loccorner", lcorner, dx

   ptq = 0
   center(1) = lcorner(1) + dx(1)*LXSIZE/2
   center(2) = lcorner(2) + dx(2)*LYSIZE/2
   center(3) = lcorner(3) + dx(3)
   zlen = dx(3)*(LZSIZE-1)
   hxlen = dx(1)*(LXSIZE-1)/2.0_dp
   
   r = 0.0_dp
   theta = 0.0_dp
   z = 0.0_dp
   s = 0.0_dp
   do i=1, particlenum
     ptx(i) = center(1) + r*cos(theta)
     pty(i) = center(2) + r*sin(theta)
     ptz(i) = center(3) + z
     pts(i) = s
!     print *,"",ptx(i), pty(i),ptz(i)
     r = r + hxlen/particlenum*0.8_dp
     theta = theta + 2.0_dp*pi/particlenum
     z = z + zlen/particlenum*0.8_dp
     s = s + 1.0_dp/particlenum

   end do

end subroutine makePTCLData

 subroutine makeCrdData(origin, n, dx, loccorner, ln, rln, lx0, myrank)
   real(kind=dp), dimension(3), intent(out) :: origin, dx, loccorner
   integer, dimension(3), intent(out) :: n, ln, rln, lx0
   integer, intent(in) :: myrank

   ! whole data
   dx(1:3) = 0.01_dp * (2048/LXSIZE)
   
   origin(1) = -dx(1) * LXSIZE/2 
   origin(2) = -dx(2) * LYSIZE/2 
   origin(3) = -dx(3) * LZSIZE/2

   n(1) = LXSIZE; n(2) = LYSIZE; n(3) = LZSIZE

   ! local data
   if(myrank == 0) then
     loccorner(1:3) = origin(1:3)
     rln(1) = LXSIZE; rln(2) = LYSIZE; rln(3) = LZSIZE
     lx0(1:3) = 1;
     ln(1)  = LXSIZE; ln(2) = LYSIZE; ln(3)  = LZSIZE
   end if

 end subroutine makeCrdData

 subroutine makeFldData(scal, vecx, vecy, vecz, lcorner, origin, dx)
   real(kind=dp), dimension(LXSIZE, LYSIZE, LZSIZE), intent(out) :: scal
   real(kind=dp), dimension(LXSIZE, LYSIZE, LZSIZE), intent(out) :: vecx, vecy, vecz
   real(kind=dp), dimension(3), intent(in) :: lcorner, origin, dx

   integer :: i, j, k
   real(kind=dp) :: abs_v, exp_v, z_liner
   real(kind=dp) :: crdx, crdy, crdz
   real(kind=dp) :: minv = 1.0D+24, maxv=-1.0D+24

   do k=1, LZSIZE
     do j=1, LYSIZE
       do i=1, LXSIZE
         crdx = lcorner(1)+dx(1)*(i-1)
         crdy = lcorner(2)+dx(2)*(j-1)
         crdz = lcorner(3)+dx(3)*(k-1)
         abs_v = sqrt(crdx*crdx + crdy*crdy)
         if(abs_v < 0.000001 .or. abs_v /= abs_v) abs_v=0.000001

         exp_v = exp(-0.1_dp*abs_v)
         z_liner = (-(crdz-origin(3))/(dx(3)*LZSIZE) *2.0_dp +1)*0.5_dp

         vecx(i, j, k) = (-crdy - crdx* z_liner) /abs_v*exp_v;
         vecy(i, j, k) = ( crdx - crdy* z_liner) /abs_v*exp_v;
         vecz(i, j, k) = 0.25_dp * exp_v;


         scal(i, j, k) = sqrt(crdx*crdx+crdy*crdy+crdz*crdz)

         if(scal(i,j,k) > maxv) maxv = scal(i,j,k)
         if(scal(i,j,k) < minv) minv = scal(i,j,k)

       end do
     end do
   end do
   print *, "min max ", minv, maxv
 end subroutine makeFldData

 subroutine makeFldData_old(scal, vecx, vecy, vecz, lcorner, origin, dx)
   real(kind=dp), dimension(LXSIZE, LYSIZE, LZSIZE), intent(out) :: scal
   real(kind=dp), dimension(LXSIZE, LYSIZE, LZSIZE), intent(out) :: vecx, vecy, vecz
   real(kind=dp), dimension(3), intent(in) :: lcorner, origin, dx

   integer :: i, j, k
   real(kind=dp) :: abs_v, exp_v, z_liner
   real(kind=dp) :: crdx, crdy, crdz
   real(kind=dp) :: minv = 1.0D+24, maxv=-1.0D+24

   do k=1, LZSIZE
     do j=1, LYSIZE
       do i=1, LXSIZE
         crdx = lcorner(1)+dx(1)*(i-1)
         crdy = lcorner(2)+dx(2)*(j-1)
         crdz = lcorner(3)+dx(3)*(k-1)
         abs_v = sqrt(crdx*crdx + crdy*crdy)
         if(abs_v < 0.000001 .or. abs_v /= abs_v) abs_v=0.000001

         exp_v = exp(-0.1_dp*abs_v)
         z_liner = (-(crdz-origin(3))/(dx(3)*LZSIZE) *2.0_dp +1)*0.5_dp

         vecx(i, j, k) = (-crdy - crdx* z_liner) /abs_v*exp_v;
         vecy(i, j, k) = ( crdx - crdy* z_liner) /abs_v*exp_v;
         vecz(i, j, k) = 0.25_dp * exp_v;


         scal(i, j, k) = sqrt(vecx(i, j, k) * vecx(i, j, k) + &
            vecy(i, j, k) * vecy(i, j, k) + vecz(i, j, k) * vecz(i, j, k))

!         if(scal(i,j,k) > maxv) maxv = scal(i,j,k)
!         if(scal(i,j,k) < minv) minv = scal(i,j,k)

       end do
     end do
   end do
!   print *, "min max ", minv, maxv
 end subroutine makeFldData_old
end program vismo_test1

