#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#include "vismo.h"

   const int LXSIZE = 256;
   const int LYSIZE = 256;
   const int LZSIZE = 34; /* (1, 2,...., 60, 61, 62), (60, 61, 62, ..., 121)  60, 61, 62 are overlapped*/

   const int mpisize = 8;
   const int maxmpisize = 64;
   const int particlenum = 10;

   double *scalar, *scalar2;
   double *vectx, *vecty, *vectz;
   double *ptx, *pty, *ptz, *pts;
   int *ptq;

   double  origin[3], dx[3];
   int n[3];
   double  loccorner[3];
   int ln[3], rln[3], lx0[3];

void makePTCLData(double *ptx, double *pty, double *ptz, int *ptq, double *pts, double *lcorner, double *dx){
   double center[3];
   double  r, theta, z, zlen, hxlen,s;
   const double pi = 3.14159;
   int i;

   ptq = 0;
   center[0] = lcorner[0] + dx[0]*LXSIZE/2;
   center[1] = lcorner[1] + dx[1]*LYSIZE/2;
   center[2] = lcorner[2] + dx[2];
   zlen = dx[2]*(LZSIZE-1);
   hxlen = dx[0]*(LXSIZE-1)/2.0;
    
   r = 0.0;
   theta = 0.0;
   z = 0.0;
   s = 0.0;
   for(i=0; i<particlenum; i++){
     ptx[i] = center[0] + r*cos(theta);
     pty[i] = center[1] + r*sin(theta);
     ptz[i] = center[2] + z;
     pts[i] = s;
     r = r + hxlen/particlenum*0.8;
     theta = theta + 2.0*pi/particlenum;
     z = z + zlen/particlenum*0.8;
     s = s + 1.0/particlenum;
   }

}

 void makeCrdData(double *origin, int *n, double *dx, double *loccorner, int *ln, int *rln, int *lx0, int myrank){
   dx[0] = 0.01 * (2048/LXSIZE);
   dx[1] = 0.01 * (2048/LXSIZE);
   dx[2] = 0.01 * (2048/LXSIZE);
   
   origin[0] = -dx[0] * LXSIZE/2; 
   origin[1] = -dx[1] * LYSIZE/2; 
   origin[2] = -dx[2] * (LZSIZE-2)*(mpisize/2);

   n[0] = LXSIZE; n[1] = LYSIZE; n[2] = 2048*mpisize/maxmpisize;

   if(myrank == 0) {
     loccorner[0] = origin[0];
     loccorner[1] = origin[1];
     loccorner[2] = origin[2];
     rln[0] = LXSIZE; rln[1] = LYSIZE; rln[2] = 32;
     lx0[0]=1; lx0[1]=1; lx0[2] = 2;
     ln[0]  = LXSIZE; ln[1] = LYSIZE; ln[2]  = 34;
   } else if(myrank == mpisize-1) {
     loccorner[0] = origin[0];
     loccorner[1] = origin[1];
     loccorner[2] = origin[2] + dx[2]*(32 + 32*(myrank-1));
     rln[0] = LXSIZE; rln[1] = LYSIZE; rln[2] = 32;
     lx0[0]=1; lx0[1]=1; lx0[2] = 1;
     ln[0]  = LXSIZE; ln[1] = LYSIZE; ln[2]  = 34;
   } else {
     loccorner[0] = origin[0];
     loccorner[1] = origin[1];
     loccorner[2] = origin[2] +dx[2]*(32 + 32*(myrank-1));
     rln[0] = LXSIZE; rln[1] = LYSIZE; rln[2] = 33;
     lx0[0]=1; lx0[1]=1; lx0[2] = 1;
     ln[0]  = LXSIZE; ln[1] = LYSIZE; ln[2]  = 34;
   }

 }

void makeFldData(double *scal, double *vecx, double *vecy, double *vecz, double *lcorner, double *origin, double *dx){
   int i, j, k;
   double abs_v, exp_v, z_liner;
   double  crdx, crdy, crdz;
   double  minv = 1.0E+24, maxv=-1.0E+24;

   for(k=0; k<LZSIZE;k++){
     for(j=0; j<LYSIZE;j++){
       for(i=0;i<LXSIZE;i++){
         crdx = lcorner[0]+dx[0]*i;
         crdy = lcorner[1]+dx[1]*j;
         crdz = lcorner[2]+dx[2]*k;
         abs_v = sqrt(crdx*crdx + crdy*crdy);
         if(abs_v < 0.000001 || abs_v != abs_v) abs_v=0.000001;

         exp_v = exp(-0.1*abs_v);
         z_liner = (-(crdz-origin[2])/(dx[2]*mpisize*32) *2.0 +1)*0.5;

         vecx[k*LXSIZE*LYSIZE+j*LXSIZE+i] = (-crdy - crdx* z_liner) /abs_v*exp_v;
         vecy[k*LXSIZE*LYSIZE+j*LXSIZE+i] = ( crdx - crdy* z_liner) /abs_v*exp_v;
         vecz[k*LXSIZE*LYSIZE+j*LXSIZE+i] = 0.25 * exp_v;

         scal[k*LXSIZE*LYSIZE+j*LXSIZE+i] = sqrt(crdx*crdx+crdy*crdy+crdz*crdz);
      }
     }
    }
 }

 void makeFldData_old(double *scal, double *vecx, double *vecy, double *vecz, double *lcorner, double *origin, double *dx){
   int i, j, k;
   double abs_v, exp_v, z_liner;
   double  crdx, crdy, crdz;
   double  minv = 1.0E+24, maxv=-1.0E+24;

   for(k=0; k<LZSIZE;k++){
     for(j=0; j<LYSIZE;j++){
       for(i=0;i<LXSIZE;i++){
         crdx = lcorner[0]+dx[0]*i;
         crdy = lcorner[1]+dx[1]*j;
         crdz = lcorner[2]+dx[2]*k;
         abs_v = sqrt(crdx*crdx + crdy*crdy);
         if(abs_v < 0.000001 || abs_v != abs_v) abs_v=0.000001;

         exp_v = exp(-0.1*abs_v);
         z_liner = (-(crdz-origin[2])/(dx[2]*mpisize*32) *2.0 +1)*0.5;

         vecx[k*LXSIZE*LYSIZE+j*LXSIZE+i] = (-crdy - crdx* z_liner) /abs_v*exp_v;
         vecy[k*LXSIZE*LYSIZE+j*LXSIZE+i] = ( crdx - crdy* z_liner) /abs_v*exp_v;
         vecz[k*LXSIZE*LYSIZE+j*LXSIZE+i] = 0.25 * exp_v;


         scal[k*LXSIZE*LYSIZE+j*LXSIZE+i] = sqrt(vecx[k*LXSIZE*LYSIZE+j*LXSIZE+i] * vecx[k*LXSIZE*LYSIZE+j*LXSIZE+i] +
           vecy[k*LXSIZE*LYSIZE+j*LXSIZE+i] * vecy[k*LXSIZE*LYSIZE+j*LXSIZE+i] + 
           vecz[k*LXSIZE*LYSIZE+j*LXSIZE+i] * vecz[k*LXSIZE*LYSIZE+j*LXSIZE+i]);

      }
     }
    }
 }


 int main(int argc, char **argv){
   int i;
   int myrank, psize, mpicomm;
   int step, gsize[3], ptnum;
   double  tb, te, t0, t1, t;
   double  tbmin, temax;
   int visex[15]; /*vsmMaxVisKind = 15 in vismo_const.90*/
   
   scalar = (double *)malloc(sizeof(double)*LXSIZE*LYSIZE*LZSIZE);
   scalar2 = (double *)malloc(sizeof(double)*LXSIZE*LYSIZE*LZSIZE);
   vectx = (double *)malloc(sizeof(double)*LXSIZE*LYSIZE*LZSIZE);
   vecty = (double *)malloc(sizeof(double)*LXSIZE*LYSIZE*LZSIZE);
   vectz = (double *)malloc(sizeof(double)*LXSIZE*LYSIZE*LZSIZE);
   ptx = (double *)malloc(sizeof(double)*particlenum);
   pty = (double *)malloc(sizeof(double)*particlenum);
   ptz = (double *)malloc(sizeof(double)*particlenum);
   pts = (double *)malloc(sizeof(double)*particlenum);
   ptq = (int *)malloc(sizeof(int)*particlenum);

   MPI_Init(&argc,&argv); 
   MPI_Comm_size(MPI_COMM_WORLD,&psize); 
   MPI_Comm_rank(MPI_COMM_WORLD,&myrank); 

   if(mpisize != psize) {
      printf("mpisize /= psize %d %d \n", mpisize, psize);
      MPI_Finalize();
      exit(1);
   }

   if(myrank == 0) printf("vismo with C code\n");
   if(myrank == 0) printf("making data\n");

   t0 = MPI_Wtime();
   /* --------- Making Data ---------*/
   makeCrdData(origin, n, dx, loccorner, ln, rln, lx0, myrank);
   makeFldData(scalar2, vectx, vecty, vectz, loccorner, origin, dx);
   makeFldData_old(scalar, vectx, vecty, vectz, loccorner, origin, dx);
   makePTCLData(ptx, pty, ptz, ptq, pts, loccorner, dx);

   /* ------ End of Making Data -----*/
   t1 = MPI_Wtime();

   MPI_Barrier(MPI_COMM_WORLD);
   tb = MPI_Wtime();

     if(myrank == 0) printf("vismo Init\n");
     vismo__init(MPI_COMM_WORLD, myrank, psize, "configtest.vsm");

     if(myrank == 0) printf("vismo SetData\n");
     gsize[0] = LXSIZE; gsize[1]=LYSIZE; gsize[2]=LZSIZE;
     vismo__addscalarv(scalar, gsize);
     vismo__addscalarv(scalar2, gsize);
     vismo__addvectorv(vectx, vecty, vectz, gsize);
/*     call vsmAddPtcl(ptx,pty,ptz,ptq,particlenum);*/
     ptnum = particlenum;
     vismo__addptclcol(ptx,pty,ptz,ptq,pts, ptnum);

     if(myrank == 0) printf("vismo SetCoord\n");
     vismo__initcoords(vismo__uniform); /* Uniform = 1., Rect = 0*/
     vismo__setunicoord(n, origin, dx);
     vismo__setlocalunicoord(ln, lx0, rln, loccorner, dx);

     if(myrank == 0) printf("vismo PrepareVis\n");
     vismo__preparevis();

   /* < computation loop>*/

     if(myrank == 0) printf("vismo Visualization\n");
     step = 0;
     vismo__visualization(step);/* ! arg is timestep (int)*/
     /*
     for(i=0;i<15;i++){
         if(i%2==0) {
             visex[i] = 1;
          } else {
             visex[i] = 0; 
          }
     }
     vismo__visualizationv(step, visex);
     */
   /* <end of computation loop>*/

     if(myrank == 0) printf("vismo Finalize\n");
     vismo__finalize();

   MPI_Barrier(MPI_COMM_WORLD);
   te = MPI_Wtime();

   if(myrank == 0){
     printf("total time: %f\n", te-tb);
   }

   free(scalar); free(scalar2); free(vectx); free(vecty); free(vectz);
   free(ptx); free(pty); free(ptz); free(pts); free(ptq);
   MPI_Finalize();
}

