module vismo
#ifndef CLANG
#ifndef SERIAL
 use mpi
#endif
#endif
 use vsmConst
 use vsmFunction
 use vsmData
 use vsmParamReader
#ifndef SERIAL
 use vsmParamBroadcast
#endif
 use vsmCoordinates
 use vsmOctree
 use vsmOctree4Iso
 use vsmRayCaster
 use vsmArrows
 use vsmStreamLine
 use vsmParticle
 use vsmFrame
#ifndef SERIAL
 use vsmBSwap
 use vsmBTree
#endif
 use vsmLightData
 use vsmCameraData
 use vsmColorMap
 use vsmImage
 use vsmParam
 use vsmDepthQ
!$ use omp_lib !changed
 implicit none
 private

 integer, parameter :: FNSIZE=256
 character(FNSIZE), save :: vsmFileName

interface vismo__addScalar
  module procedure vismo__AddScalarS, vismo__AddScalarV
end interface vismo__addScalar

interface vismo__addVector
  module procedure vismo__AddVectorS, vismo__AddVectorV
end interface vismo__addVector

 public vismo__init, vismo__initCoords, vismo__setUniCoord, vismo__setLocalUniCoord, vismo__prepareVis
 public vismo__addScalar, vismo__addVector, vismo__addPtcl
 public vismo__visualization, vismo__finalize
 public vismo__setRectCoord, vismo__setLocalRectCoord, vismo__addPtclCol, vismo__setBSwap
 public vismo__init4c, vismo__initCoords4c,vismo__visualization4c,vismo__visualization4cv
 public vismo__addScalarS, vismo__addScalarV, vismo__addVectorS, vismo__addVectorV

 logical, parameter, public :: vismo__uniform = .true.
 logical, parameter, public :: vismo__rect = .false.
 contains

 ! Initialization of vismo
 subroutine vismo__init4c(rank, psize, vsmfile)
   integer, intent(in) :: rank, psize
   integer :: comm, ierr
   character(*), intent(in) :: vsmfile
#ifndef SERIAL
   comm = 0
   call vismo__init(comm, rank, psize, vsmfile)
#else
   comm = 0
   call vismo__init(vsmfile)
#endif
 end subroutine

#ifndef SERIAL
 subroutine vismo__init(comm, rank, psize, vsmfile)
   integer, intent(in) :: comm, rank, psize
   character(*), intent(in) :: vsmfile
   character(4) :: ascrank
   integer :: onum
   logical :: fexist
#else
 subroutine vismo__init(vsmfile)
   integer :: comm, rank, psize
   character(*), intent(in) :: vsmfile
   character(4) :: ascrank
   integer :: onum
   logical :: fexist
   comm = 0; rank = 0; psize = 1
#endif

   onum = 1
   write(ascrank, '(I4.4)') rank 

   vsmFileName = vsmfile

   call vsmLogStart('vsmlog'//ascrank//'.txt')
   call vsmLogPrint("vsmInit is called")
   ! initialization of parameters
!$ onum = omp_get_max_threads() !changed

   call vsmInitVisParam(comm, rank, psize, onum)
   call vsmLogPrint("MPI Processes:", psize)
   call vsmLogPrint("OpenMP threads:", onum) !changed
   ! initialization of variables
   call vsmInitColmap(vsmMaxScalar, vsmMaxVector, vsmMaxPtcl)
   call vsmInitCamera
#ifndef SERIAL
   call vsmInitBSwap(rank, psize)
   call vsmInitBTree(rank, psize)
#endif
   call vsmInitData
   call vsmInitOctree
   call vsmInitOctree4Iso
   call vsmInitLight
   call vsmInitBGColor
   call vsmInitFrame
   ! reading and broadcasting parameters
   call vsmInitParamReader
   if (rank == 0) then
     call vsmInvestigateFiles(trim(vsmFileName), fexist)
     if(.not.fexist) then
       call vsmError("Could not find all the config files. be careful about INCLUDE")
     end if

     call vsmReadParam(trim(vsmFileName), .true.)
   end if
#ifndef SERIAL
   call vsmParamBcast
#endif
   call vsmLogPrint("vsmInit ends")
 end subroutine vismo__init

 ! ---- Setting data
 subroutine vismo__addScalarS(scalar, nx, ny, nz)
   integer, intent(in) :: nx, ny, nz
#ifndef FLOATDATA
   real(kind=vdp), dimension(nx,ny,nz), target, intent(in) :: scalar
#else
   real, dimension(nx,ny,nz), target, intent(in) :: scalar
#endif
   call vsmLogPrint("vsmAddScalar is called")
   call vsmAddScalData(scalar, nx, ny, nz)
   call vsmLogPrint("vsmAddScalar ends")
 end subroutine vismo__addScalarS

 subroutine vismo__addScalarV(scalar, nx)
   integer, dimension(3), intent(in) :: nx
#ifndef FLOATDATA
   real(kind=vdp), dimension(nx(1),nx(2),nx(3)), target, intent(in) :: scalar
#else
   real, dimension(nx(1),nx(2),nx(3)), target, intent(in) :: scalar
#endif
   call vsmLogPrint("vsmAddScalarV is called")
   call vsmAddScalData(scalar, nx(1),nx(2),nx(3))
   call vsmLogPrint("vsmAddScalarV ends")
 end subroutine vismo__addScalarV

 subroutine vismo__addVectorS(vect_x, vect_y, vect_z, nx, ny, nz)
   integer, intent(in) :: nx, ny, nz
#ifndef FLOATDATA
   real(kind=vdp), dimension(nx,ny,nz), target, intent(in) :: vect_x, vect_y, vect_z
#else
   real, dimension(nx,ny,nz), target, intent(in) :: vect_x, vect_y, vect_z
#endif
   call vsmLogPrint("vsmAddVectorS is called")
   call vsmAddVectData(vect_x, vect_y, vect_z, nx, ny, nz)
   call vsmLogPrint("vsmAddVectorS ends")
 end subroutine vismo__addVectorS

 subroutine vismo__addVectorV(vect_x, vect_y, vect_z, nx)
   integer, dimension(3), intent(in) :: nx
#ifndef FLOATDATA
   real(kind=vdp), dimension(nx(1),nx(2),nx(3)), target, intent(in) :: vect_x, vect_y, vect_z
#else
   real, dimension(nx(1),nx(2),nx(3)), target, intent(in) :: vect_x, vect_y, vect_z
#endif
   call vsmLogPrint("vsmAddVectorV is called")
   call vsmAddVectData(vect_x, vect_y, vect_z, nx(1),nx(2),nx(3))
   call vsmLogPrint("vsmAddVectorV ends")
 end subroutine vismo__addVectorV

 subroutine vismo__addPtcl(p_x, p_y, p_z, q, n)
   integer, intent(in) :: n
#ifndef FLOATDATA
   real(kind=vdp), dimension(n), target, intent(in) :: p_x, p_y, p_z
#else
   real, dimension(n), target, intent(in) :: p_x, p_y, p_z
#endif
   integer, dimension(n), target, intent(in) :: q
   call vsmLogPrint("vsmAddPtcl is called")
   call vsmAddPtclData(p_x, p_y, p_z, q, n)
   call vsmLogPrint("vsmAddPtcl ends")
 end subroutine vismo__addPtcl

 subroutine vismo__addPtclCol(p_x, p_y, p_z, q, v, n)
   integer, intent(in) :: n
#ifndef FLOATDATA
   real(kind=vdp), dimension(n), target, intent(in) :: p_x, p_y, p_z,v
#else
   real, dimension(n), target, intent(in) :: p_x, p_y, p_z,v
#endif
   integer, dimension(n), target, intent(in) :: q
   call vsmLogPrint("vsmAddPtclCol is called")
   call vsmAddPtclDataCol(p_x, p_y, p_z, q, v, n)
   call vsmLogPrint("vsmAddPtclCol ends")
 end subroutine vismo__addPtclCol
 ! ---- End of data

 ! ---- Setting coordinates ----
 subroutine vismo__initCoords4c(uni)
  integer, intent(in) :: uni
  logical :: uniform
  if(uni == 0) then
    uniform = .false.
  else if(uni == 1) then
    uniform = .true.
  else
    call vsmError("init coord : uni = 0 or 1")
  end if
  call vsmLogPrint("vsmSetInitCoords is called")
  call vsmInitCoord(uniform)
  call vsmLogPrint("vismo__SetInitCoords ends")
 end subroutine vismo__initCoords4c

 subroutine vismo__initCoords(uniform)
  logical, intent(in) :: uniform
  call vsmLogPrint("vsmSetInitCoords is called")
  call vsmInitCoord(uniform)
  call vsmLogPrint("vsmSetInitCoords ends")
 end subroutine vismo__initCoords

 ! << Uniform>>
 subroutine vismo__setUniCoord(nx, origin, dx)
  integer, dimension(3), intent(in) :: nx
  real(kind=vdp), dimension(3), intent(in) :: origin
  real(kind=vdp), dimension(3), intent(in) :: dx
  call vsmLogPrint("vsmSetUniCoord is called")
  call vsmSetUniform(nx, origin, dx)
  call vsmLogPrint("vsmSetUniCoord ends")
 end subroutine vismo__setUniCoord

 subroutine vismo__setLocalUniCoord(nx, lx0, lnx, origin, dx)
  integer, dimension(3), intent(in) :: nx, lx0, lnx
  real(kind=vdp), dimension(3), intent(in) :: origin
  real(kind=vdp), dimension(3), intent(in) :: dx
  call vsmLogPrint("vsmSetLocalUniCoord is called")
  call vsmSetLUniform(nx, lx0, lnx, origin, dx)
  call vsmLogPrint("vsmSetLocalUniCoord ends")
 end subroutine vismo__setLocalUniCoord

 ! << Rect >>
 subroutine vismo__setRectCoord(nx, coordx, coordy, coordz)
  integer, dimension(3), intent(in) :: nx
  real(kind=vdp), dimension(nx(1)), intent(in) :: coordx
  real(kind=vdp), dimension(nx(2)), intent(in) :: coordy
  real(kind=vdp), dimension(nx(3)), intent(in) :: coordz

  call vsmLogPrint("vsmSetRectCoord is called")
  call vsmSetRect(nx, coordx, coordy, coordz)
  call vsmLogPrint("vsmSetRectCoord ends")
 end subroutine vismo__SetRectCoord

 subroutine vismo__setLocalRectCoord(nx, lx0, lnx, coordx, coordy, coordz)
  integer, dimension(3), intent(in) :: nx, lx0, lnx
  real(kind=vdp), dimension(nx(1)), intent(in) :: coordx
  real(kind=vdp), dimension(nx(2)), intent(in) :: coordy
  real(kind=vdp), dimension(nx(3)), intent(in) :: coordz

  call vsmLogPrint("vsmSetLocalRectCoord is called")
  call vsmSetLRect(nx, lx0, lnx, coordx, coordy, coordz)
  call vsmLogPrint("vsmSetLocalRectCoord ends")
 end subroutine vismo__setLocalRectCoord

 ! --- End of coordinates

 ! Preparing visualization
 ! should be called after both coord and data settings
 subroutine vismo__preparevis
!   logical :: root
   integer :: i, calcdpflag
   call vsmLogPrint("vsmPrepareVis is called")

   call vsmInitImage ! 2014/12/2
   call vsmOctPrepare
   call vsmOctDivide(0)
   call vsmOctPrepare4Iso
   call vsmOctDivide4Iso(0)
!   root = .true.
!   call vsmOctDivide(vsmOctant(0), root)

   call vsmArrowsPrepare
   call vsmCheckParam
   call vsmPrepareCoord
   call vsmSetDataMinMax
   calcdpflag = 0
   do i=1, vsmNumVisKind
     if(vsmVisParam(i)%sline_num > 0 .or. vsmVisParam(i)%arrow_num > 0) then
       calcdpflag=calcdpflag+1
     end if 
   end do
   if(calcdpflag > 0) then
     do i=1, vsmCameraNum
       call vsmCalcDP(vsmCamera(i)%dp, i)
     end do
   end if
   call vsmLogPrint("vsmPrepareVis end")
 end subroutine vismo__prepareVis

 ! Visualization
 subroutine vsmVectVis(vnum, cnum, dp)
  integer, intent(in) :: vnum, cnum
  real(kind=vdp), intent(in) :: dp

  call vsmLogPrint("------ vsmVectVis is called")
  select case (vsmPCmode)
    case (0)
      call vsmDrawStreamLinesIM(vnum, cnum, dp)
      call vsmDrawArrowsIM(vnum, cnum)
    case (1)
    case (2)
    case default
      call vsmError("pc mode is strange : vsmVectVis")
  end select  
  call vsmLogPrint("------ vsmVectVis ends")
 end subroutine vsmVectVis

 subroutine vsmScalVis(vnum, cnum, time, bon, dp)
  integer, intent(in) :: vnum, cnum, time
  logical, intent(in) :: bon
  real(kind=vdp), intent(in) :: dp
  call vsmLogPrint("------ vsmScalVis is called")
  call vsmScalarVisualization(vnum, cnum, time, bon, dp)
  call vsmLogPrint("------ vsmScalVis ends")
 end subroutine vsmScalVis

 subroutine vsmPtclVis(vnum, cnum)
  integer, intent(in) :: vnum, cnum
  call vsmLogPrint("------ vsmPtclVis is called")
  select case (vsmPCmode)
    case (0)
      call vsmDrawParticlesIM(vnum, cnum)
    case (1)
    case (2)
    case default
      call vsmError("pc mode is strange : vsmPtclVis")
  end select  
  call vsmLogPrint("------ vsmPtclVis ends")
 end subroutine vsmPtclVis

 subroutine vsmIsBlendOn(visnum, bon)
   integer, intent(in)  :: visnum
   logical, intent(out) :: bon
   real :: alpha
   integer :: i

   bon = .false.
   vsmVisParam(visnum)%iso_trans = .false.
   vsmVisParam(visnum)%iso_solid = .false.
   vsmVisParam(visnum)%slice_trans = .false.
   vsmVisParam(visnum)%slice_solid = .false.

   ! volume rendering
   if(vsmVisParam(visnum)%volren_num > 0) then
     bon = .true.
   end if

   ! isosurface
   do i=1, vsmVisParam(visnum)%iso_num
     alpha = 1.0
     call vsmGetIsoAlpha(visnum, i, alpha)
     if(alpha < vsmErtValue) then
       bon = .true.
       vsmVisParam(visnum)%iso_trans = .true.
     else
       vsmVisParam(visnum)%iso_solid = .true.
     end if
   end do

   ! slice
   do i=1, vsmVisParam(visnum)%slice_num
     alpha = 1.0
     call vsmGetSliceAlpha(visnum, i, alpha)
     if(alpha < vsmErtValue) then
       bon = .true.
       vsmVisParam(visnum)%slice_trans = .true.
     else
       vsmVisParam(visnum)%slice_solid = .true.
     end if
   end do
 end subroutine vsmIsBlendOn

 subroutine vismo__setBSwap(midx, ndv, idx)
   integer, dimension(3), intent(in) :: midx, ndv
   integer, dimension(ndv(1),ndv(2),ndv(3)), intent(in) :: idx
   call vsmLogPrint("setBSwap")
!   call vsmSetBinarySwap(midx, ndv, idx)
   print *,"vsmSetBSwap is not required any more."
   call vsmLogPrint("setBSwap ends")
 end subroutine vismo__setBSwap

 subroutine vismo__visualization4c(time)
   integer, intent(in) :: time
   call vismo__visualization(time)
 end subroutine vismo__visualization4c

 subroutine vismo__visualization4cv(time, vex)
   integer, intent(in) :: time
   integer, dimension(vsmMaxVisKind), intent(in) :: vex
   logical, dimension(vsmMaxVisKind) :: visex
   integer :: i
   visex = .false.
   do i=1, vsmNumViskind
     if(vex(i) == 1) then
       visex(i) = .true.
     end if
   end do
   call vismo__visualization(time, visex)
 end subroutine vismo__visualization4cv

 subroutine vismo__visualization(time, visex)
   integer, intent(in) :: time
   character(4) :: viskind, camnum, rank 
   character(9) :: timestep
   integer :: i, j
   integer :: comm, myrank, psize, ierr
   integer :: UnNum, iostatus
   integer :: onum, tonum

   logical :: blend
   logical, save :: first = .true.

   real(kind=vdp) :: dp
   logical, dimension(vsmMaxVisKind), optional, intent(in) :: visex
   logical, dimension(vsmMaxVisKind) :: visexcute

   call vsmLogPrint("vsmVisualization is called")
   call vsmGetMPIVars(comm, myrank, psize)

   if(present(visex)) then
      visexcute(1:vsmMaxVisKind) = visex(1:vsmMaxVisKind)
   else
      visexcute(1:vsmMaxVisKind) = .true.
   end if

    onum = 1
!$ onum = omp_get_max_threads() !changed
   if(first) then
     first = .false.
     if (myrank == 0) then
        UnNum = vsmSearchUnitNum("DataInfo")
        open(UnNum, file=trim(vsmDataInfoFile), action='write', &
            iostat=iostatus)

        if(iostatus > 0) then
           call vsmError("cant open data info file")
        end if
        
        write(UnNum, '(1A)', advance='no') 'COORDMIN '
        do i=1, size(vsmCoord%origin)
           write(UnNum, '(1F16.10,1X)', advance='no') vsmCoord%origin(i)
        end do
        write(UnNum, *) ''
        
        write(UnNum, '(1A)', advance='no') 'COORDMAX '
        do i=1, size(vsmCoord%cend)
           write(UnNum, '(1F16.10,1X)', advance='no') vsmCoord%cend(i)
        end do
        write(UnNum, *) ''

        write(UnNum, '(1A,1F16.10)') "RAYINTV ", vsmIntval%ray_interval
        write(UnNum, '(1A,1F16.10)') "SLINEDH ", vsmIntval%dh

        write(UnNum, '(1A,1I5)') "SCALNUM ", vsmDataInfo%NumOfScal
        write(UnNum, '(1A,1I5)') "VECTNUM ", vsmDataInfo%NumOfVect
        write(UnNum, '(1A,1I5)') "PTCLNUM ", vsmDataInfo%NumOfPtcl

        close(UnNum)

      end if

   end if
   ! updating octree for scalar visualization
   call vsmint2asc9(time, timestep)

   if(vsmDbg) then
     call vsmLogPrint("viskindnum:", vsmNumVisKind)
     call vsmLogPrint("cameranum:", vsmCameraNum)
   end if

   do j=1, vsmNumViskind
     if(.not.visexcute(j)) cycle
     blend = .false.

     call vsmLogPrint("viskindloop:", j)

     call vsmIsBlendOn(j, blend)
     if((blend .or. vsmMultipleCameras) .and. vsmPCmode /= 0) then
       call vsmLogPrint("PCmode is made 0 because of blendmode or multiple cameras")
       vsmPCmode = 0
     end if

     call vsmint2asc4(j, viskind)
        call vsmSetOctComDat(j)
        call vsmUpdateOctree(0)
        call vsmSetOctComDat4Iso(j)
        call vsmUpdateOctree4Iso(0)

     do i=1, vsmCameraNum

       call vsmint2asc4(i, camnum)
       call vsmLogPrint("cameraloop:", i)
       ! initialization of camera and image
       call vsmInitScreen(i)

       if(.not.vsmCamera(i)%pers) then
         dp = vsmImageData%dp
       else
         dp = vsmCamera(i)%dp
       end if

       if(vsmVisParam(j)%sline_num > 0 .or. vsmVisParam(j)%arrow_num > 0 .or. vsmVisParam(j)%ptcl_num > 0) then
         call vsmClearImage4Solid

!         if(onum > vsmMaxOMP4Solid) then
! !$ call omp_set_num_threads(vsmMaxOMP4Solid)
!         end if

         ! particle visualization
         call vsmPtclVis(j, i)
         ! vector visualization
         call vsmVectVis(j, i, dp)

!         if(onum > vsmMaxOMP4Solid) then
! !$ call omp_set_num_threads(onum)
!         end if

         ! merge local buffer
         call vsmMergeBuf4Solid
       else
         call vsmClearImage
       end if

      if(vsmPCmode /= 2) then
        call vsmDrawFrame(j, i, .false.)
      end if
#ifndef SERIAL
      if(blend .or. (vsmVisParam(j)%slice_num == 0 .and. vsmVisParam(j)%iso_num == 0)) then
        if(vsmBSwapParam%bswap) then
         call vsmBinarySwapIM(.true., i, j, time)
        else
         call vsmBinarySwap4SolidIM(i, j, time)
        end if
      end if
#endif
      if(.not. vsmVisParam(j)%slice_solid .and. .not. vsmVisParam(j)%iso_solid) then
        if(vsmVisParam(j)%depthq) call vsmDepthCueingImage
      end if

      if(vsmVisParam(j)%slice_num > 0 .or. vsmVisParam(j)%iso_num > 0 .or. vsmVisParam(j)%volren_num > 0) then
        if(blend) then
          call vsmLogPrint("clear raysegment")
          call vsmClearRaySegment
          call vsmClearRaySegmentR
        end if
         ! scalar visualization
        call vsmScalVis(j, i, time, blend, dp)

        if(blend) then
           select case (vsmPCmode)
             case (0)
#ifndef SERIAL
               call vsmBinarySwapIMRS(i, j, time)
#else
               call vsmBlendRaySegmentF(0)
#endif
               call vsmAlphaBlending(1, vsmImageData%rimgsize)

             case (1)

             case (2)
               
             case default
               call vsmError("pc mode is strange : vsmVisualization scalblend")

           end select
        end if

      end if

       ! saving image
      if(myrank == 0 .and. vsmPCmode /= 2) then
        call vsmBGBlending
        call vsmWriteImage(vsmImageFile//'v'//viskind//'c'//camnum//'t'//timestep)
      end if
     end do
   end do
!   call vsmInitTube
   call vsmLogPrint("vsmVisualization ends")
 end subroutine vismo__visualization

 ! Finalizing vismo
 subroutine vismo__finalize
   call vsmLogPrint("vsmFinalize is called")
   call vsmFinalizeImage
#ifndef SERIAL
   call vsmFinalizeBSwap
   call vsmFinalizeBTree
#endif
   call vsmFinalizeColmap
   call vsmFinalizeCoord
   call vsmFinalizeOctree
   call vsmFinalizeOctree4Iso
   call vsmLogPrint("vsmFinalize ends")
   call vsmLogEnd
 end subroutine vismo__finalize

end module vismo

