module vsmArrows
 use vsmConst
 use vsmFunction
 use vsmColor
 use vsmData
 use vsmCoordinates
 use vsmParam
 use vsmImage
 use vsmIntpol
 use vsmSolidDrawer

 implicit none
 private

 type vismo_ArrowComDat
   real(kind=vdp), dimension(3, 2) :: wd_vertx, ld_vertx
   real(kind=vdp) :: dc
 end type vismo_ArrowComDat

 type(vismo_ArrowComDat),save :: vsmArrowsComDat

 public vsmArrowsPrepare, vsmDrawArrowsIM
 contains

 subroutine vsmArrowsPrepare
   vsmArrowsComDat%dc = ((vsmCoord%cend(1) - vsmCoord%origin(1))/vsmCoord%nx(1) + &
                     (vsmCoord%cend(2) - vsmCoord%origin(2))/vsmCoord%nx(2) + &
                     (vsmCoord%cend(3) - vsmCoord%origin(3))/vsmCoord%nx(3) )/3.0_vdp

   vsmArrowsComDat%wd_vertx(1:3, 1) = vsmCoord%origin(1:3)
   vsmArrowsComDat%wd_vertx(1:3, 2) = vsmCoord%cend(1:3)

   vsmArrowsComDat%ld_vertx(1:3, 1) = vsmLocalCoord%lorigin(1:3)
   vsmArrowsComDat%ld_vertx(1:3, 2) = vsmLocalCoord%lcend(1:3)
 end subroutine vsmArrowsPrepare

 subroutine vsmDrawArrowsIM(vis_num, cam_num)
   integer, intent(in) :: vis_num, cam_num
   integer :: j, mode
   do j=1, vsmVisParam(vis_num)%arrow_num
       call vsmDrawArrowsSLIM(vis_num, cam_num, j)
   end do
 end subroutine vsmDrawArrowsIM

 subroutine vsmDrawArrowsSLIM(vis_num, cam_num, j)
   integer, intent(in) :: vis_num, cam_num, j

   integer :: i, x, y, xc, yc
   integer :: nvec
   integer, dimension(3) :: trans_bool
   real(kind=vdp), dimension(3) :: arpos
   real(kind=vdp) :: rad, rad15, leng, trans, dens, dc, sdc
   real(kind=vdp), dimension(3) :: temp_v, pos
   real(kind=vdp), dimension(3, 8) :: trori, trlori
   real(kind=vdp), dimension(3) :: max_wd, min_wd, max_ld, min_ld
   real(kind=vdp), dimension(4) :: sc
   real(kind=vdp), dimension(3, 3) :: matrix, invmatrix
   real(kind=vdp), dimension(3, 8) :: wd_vertx, ld_vertx

   real(kind=vdp), dimension(3) :: dx, vec_val
   real(kind=vdp), dimension(8) :: vx
   real(kind=vdp), dimension(3, 8) :: grid_val
   integer, dimension(3) :: ci
   real(kind=vdp), dimension(3) :: pos1, pos2
   real(kind=vdp) :: maxv, ratio
   real :: crt
   logical :: normalize
   integer, save :: prev_visnum = 0

   if(vsmPCmode /=0) call vsmError("DrawArrowsIM is called though vsmPCmode is not 0")
   call vsmLogPrint("draw arrows")
   crt = 1.0 - vsmARConeRatio
   crt = vsmScalef01(crt)

   dc = vsmArrowsComDat%dc
   sdc = dc/15.0_vdp
   wd_vertx(1:3, 1) = vsmArrowsComDat%wd_vertx(1:3,1)
   wd_vertx(1:3, 8) = vsmArrowsComDat%wd_vertx(1:3,2)

   wd_vertx(2:3, 2) = wd_vertx(2:3, 1)
   wd_vertx(1, 2) = wd_vertx(1, 8)

   wd_vertx(1, 3) = wd_vertx(1, 1)
   wd_vertx(2, 3) = wd_vertx(2, 8)
   wd_vertx(3, 3) = wd_vertx(3, 1)

   wd_vertx(1, 4) = wd_vertx(1, 8)
   wd_vertx(2, 4) = wd_vertx(2, 8)
   wd_vertx(3, 4) = wd_vertx(3, 1)

   wd_vertx(1, 5) = wd_vertx(1, 1)
   wd_vertx(2, 5) = wd_vertx(2, 1)
   wd_vertx(3, 5) = wd_vertx(3, 8)

   wd_vertx(1, 6) = wd_vertx(1, 8)
   wd_vertx(2, 6) = wd_vertx(2, 1)
   wd_vertx(3, 6) = wd_vertx(3, 8)

   wd_vertx(1, 7) = wd_vertx(1, 1)
   wd_vertx(2, 7) = wd_vertx(2, 8)
   wd_vertx(3, 7) = wd_vertx(3, 8)

   ld_vertx(1:3, 1) = vsmArrowsComDat%ld_vertx(1:3,1)
   ld_vertx(1:3, 8) = vsmArrowsComDat%ld_vertx(1:3,2)

   ld_vertx(2:3, 2) = ld_vertx(2:3, 1)
   ld_vertx(1, 2) = ld_vertx(1, 8)

   ld_vertx(1, 3) = ld_vertx(1, 1)
   ld_vertx(2, 3) = ld_vertx(2, 8)
   ld_vertx(3, 3) = ld_vertx(3, 1)

   ld_vertx(1, 4) = ld_vertx(1, 8)
   ld_vertx(2, 4) = ld_vertx(2, 8)
   ld_vertx(3, 4) = ld_vertx(3, 1)

   ld_vertx(1, 5) = ld_vertx(1, 1)
   ld_vertx(2, 5) = ld_vertx(2, 1)
   ld_vertx(3, 5) = ld_vertx(3, 8)

   ld_vertx(1, 6) = ld_vertx(1, 8)
   ld_vertx(2, 6) = ld_vertx(2, 1)
   ld_vertx(3, 6) = ld_vertx(3, 8)

   ld_vertx(1, 7) = ld_vertx(1, 1)
   ld_vertx(2, 7) = ld_vertx(2, 8)
   ld_vertx(3, 7) = ld_vertx(3, 8)

   !if(vsmDbg) then
   !  call vsmLogPrint("dc = ", dc)
   !  call vsmLogPrint("wd 1 = ", wd_vertx(1:3, 1))
   !  call vsmLogPrint("wd 2 = ", wd_vertx(1:3, 2))
   !  call vsmLogPrint("ld 1 = ", ld_vertx(1:3, 1))
   !  call vsmLogPrint("ld 2 = ", ld_vertx(1:3, 2))
   !end if

   trans_bool = 0
!   do j=1, vsmVisParam(vis_num)%arrow_num
     sc(1:4)   = vsmVisParam(vis_num)%arrowparam(j)%sc(1:4) 
     nvec      = vsmVisParam(vis_num)%arrowparam(j)%nvec
     rad       = vsmVisParam(vis_num)%arrowparam(j)%radius
     leng      = vsmVisParam(vis_num)%arrowparam(j)%length
     dens      = vsmVisParam(vis_num)%arrowparam(j)%density
     rad15 = rad*1.5_vdp
     maxv = vsmVectData(nvec)%maxvalue

!     if(vsmDbg) then
!       call vsmLogPrint("sc = ", sc)
!       call vsmLogPrint("vec = ", nvec)
!       call vsmLogPrint("rad = ", rad)
!       call vsmLogPrint("leng = ", leng)
!       call vsmLogPrint("dens = ", dens)
!     end if

     call vsmPlaneTrans(sc, trans, trans_bool)

     !call vsmLogPrint("trans =", trans)
     !call vsmLogPrint("trans_bool = ", trans_bool)

     call vsmRotateMatrix(sc, matrix)
     call vsmRotateInvMatrix(sc, invmatrix)

     do i=1, 8
       call vsmTransVertx(wd_vertx(1:3, i), temp_v, trans, trans_bool)
       call vsmMatrixCrossVect(matrix, temp_v, trori(:, i))
       call vsmLogPrint("wd 1 ->", temp_v)
       call vsmLogPrint(" ->", trori(:,i))

       call vsmTransVertx(ld_vertx(1:3, i), temp_v, trans, trans_bool)
       call vsmMatrixCrossVect(matrix, temp_v, trlori(:, i))
       call vsmLogPrint("ld 1 ->", temp_v)
       call vsmLogPrint(" ->", trlori(:,i))
     end do

     max_wd(1:3) = max(trori(1:3,1),trori(1:3,2),trori(1:3,3),trori(1:3,4), &
                       trori(1:3,5),trori(1:3,6),trori(1:3,7), trori(1:3,8)) - sdc

     min_wd(1:3) = min(trori(1:3,1),trori(1:3,2),trori(1:3,3),trori(1:3,4), &
                       trori(1:3,5),trori(1:3,6),trori(1:3,7), trori(1:3,8)) + sdc

     max_ld(1:3) = max(trlori(1:3,1),trlori(1:3,2),trlori(1:3,3),trlori(1:3,4), &
                       trlori(1:3,5),trlori(1:3,6),trlori(1:3,7), trlori(1:3,8))

     min_ld(1:3) = min(trlori(1:3,1),trlori(1:3,2),trlori(1:3,3),trlori(1:3,4), &
                       trlori(1:3,5),trlori(1:3,6),trlori(1:3,7), trlori(1:3,8))

     call vsmLogPrint("max_wd", max_wd)
     call vsmLogPrint("min_wd", min_wd)
     call vsmLogPrint("max_ld", max_ld)
     call vsmLogPrint("min_ld", min_ld)

!     print *, "max_wd", max_wd
!     print *, "min_wd", min_wd
!     print *, "max_ld", max_ld
!     print *, "min_ld", min_ld

     ! the plane never cross local data
     if(min_ld(3) > 0.0_vdp .or. max_ld(3) < 0.0_vdp) return

     arpos(2) = min_wd(2) - dc*2.0_vdp; arpos(3) = 0.0_vdp
     yc = int((max_wd(2) - arpos(2))/(dc/dens))
     normalize = .not. vsmVisParam(vis_num)%arrowparam(j)%variable
     if(.not.normalize) then
       if(maxv <= 0.0_vdp) then
         call vsmError("maxv is less than 0.0, vector's colormap may not be set :: vsmDrawArrows")
       end if
     end if
!$omp parallel do private(x,y,xc, temp_v, pos, dx, ci, vx, grid_val, vec_val, pos1, pos2, ratio) firstprivate(arpos)
     do y = 1, yc !do while(arpos(2) < max_wd(2)) ! y-direc
       !arpos(2) = arpos(2) + dc/dens
       arpos(2) = min_wd(2) - dc*2.0_vdp + dc/dens*y
       arpos(1) = min_wd(1) - dc*2.0_vdp
       xc = int((max_wd(1) - arpos(1))/(dc/dens))
       do x=1, xc !do while(arpos(1) < max_wd(1)) ! x-direc
         arpos(1) = arpos(1) + dc/dens
         if(arpos(1) < max_ld(1) .and. arpos(1) >= min_ld(1) .and. &
                  arpos(2) < max_ld(2) .and. arpos(2) >= min_ld(2)) then

           call vsmMatrixCrossVect(invmatrix, arpos, temp_v)
           !call vsmLogPrint("arpos  -> temp_v", arpos)
           call vsmTransVertx(temp_v, pos, -trans, trans_bool)
           !call vsmLogPrint("temp_v -> ", pos)

           if(.not.vsmIsInsidel(pos)) cycle

           call vsmFindCoord(pos, dx, ci)
           call vsmSetFrac(dx, vx)
           call vsmSetVectGridVal(nvec, ci, grid_val)
           call vsmGetVectVal(vx, grid_val, vec_val, normalize)

           if(.not.normalize)then
             ratio = sqrt(dot_product(vec_val, vec_val))/maxv
             ratio = vsmScaled01(ratio)
           else
             ratio = 1.0_vdp
           end if

           call vsmNormVec(vec_val)
           pos1(1:3) = pos(1:3) - vec_val(1:3)*leng*ratio/2.0_vdp
           pos2(1:3) = pos1(1:3) + vec_val(1:3)*leng*ratio*crt

           call vsmDrawCylinder(pos1, pos2, rad, cam_num, vis_num, &
                               vsmArAmb, vsmArDif, vsmArSpec, vsmArwShin)

           pos1(1:3) = pos2(1:3)
           pos2(1:3) = pos1(1:3) + vec_val(1:3)*leng*ratio*(1.0-crt)

           call vsmDrawCone(pos1, pos2, rad15, cam_num, vis_num, vsmArcAmb, &
                              vsmArcDif, vsmArcSpec, vsmArwShin)

         end if
       end do
     end do

!   end do

   call vsmLogPrint("draw arrows ends")

 end subroutine vsmDrawArrowsSLIM
 
end module vsmArrows

