module vsmBSwap
#ifndef CLANG
 use mpi
#endif
 use vsmConst
 use vsmFunction
 use vsmParam
 use vsmImage

 implicit none
 private

 type(vismo_bswap), save, public :: vsmBSwapParam

public vsmInitBSwap, vsmFinalizeBSwap, vsmBinarySwapIMRS, vsmBinarySwapIM, vsmBinarySwap4SolidIM

 contains

 subroutine vsmFinalizeBSwap
  deallocate(vsmBSwapParam%partner, vsmBSwapParam%above)
  if(vsmBSwapParam%bswap) then
    if(allocated(vsmBSwapParam%index2rank)) deallocate(vsmBSwapParam%index2rank)
  end if
 end subroutine vsmFinalizeBSwap

 subroutine vsmInitBSwap(rank, psize)
   integer, intent(in) :: rank, psize
   integer :: i
   integer :: allocstate

   call vsmLogPrint("vsmInitBSwap is called")
   call vsmCheckPowerOf2(psize, vsmBSwapParam%numstage, vsmBSwapParam%possible)
   vsmBSwapParam%myrank = rank
   vsmBSwapParam%psize = psize
   vsmBSwapParam%bswap = .false.
   vsmBSwapParam%myindex(1:3) = 0
   vsmBSwapParam%ndv(1:3) = 1

   allocate(vsmBSwapParam%partner(vsmBSwapParam%numstage), stat=allocstate)
   if(allocstate /= 0) call vsmError("can't allocate bswap partner")

   allocate(vsmBSwapParam%above(vsmBSwapParam%numstage), stat=allocstate)
   if(allocstate /= 0) call vsmError("can't allocate bswap above")

   do i=1, vsmBSwapParam%numstage
     vsmBSwapParam%partner(i) = -1
     vsmBSwapParam%above(i)= .false.
   end do

   call vsmDvProcess4BS(rank, psize)

   call vsmLogPrint("vsmInitBSwap ends")
 end subroutine vsmInitBSwap

 subroutine vsmDvProcess4BS(rank, psize)
  integer, intent(in) :: rank, psize
  integer :: nstage
  integer :: NumBSProcess
  integer :: prank
  integer :: st, ed, dv
  logical :: a

  nstage = vsmBSwapParam%numstage

  NumBSProcess = 2**nstage

  if(psize == NumBSProcess) vsmBSwapParam%bswap = .true.
  if(rank > NumBSProcess-1) return

  st = 1; ed = NumBSProcess; dv = ed

  do
   if(dv > 1) then
    dv = dv/2
    if(rank+1 < st+dv) then
      prank = rank + dv
      a=.true.
      ed = st+dv-1
    else
      prank = rank - dv
      a=.false.
      st = st+dv
    end if

    vsmBSwapParam%partner(nstage) = prank
    vsmBSwapParam%above(nstage) = a
    call vsmLogPrint("binary swap 4 solid: stage = ", nstage)
    call vsmLogPrint("binary swap 4 solid : myrank = ", rank)
    call vsmLogPrint("binary swap 4 solid : parter rank = ", prank)
    nstage = nstage - 1
    if(nstage == 0) exit
   end if
  end do
 end subroutine vsmDvProcess4BS

 subroutine vsmSetBinarySwap(midx, ndv, idx)
   integer, dimension(3), intent(in) :: midx, ndv
   integer, dimension(ndv(1),ndv(2),ndv(3)), intent(in) :: idx
   integer :: allocstate

   vsmBSwapParam%myindex(1:3) = midx(1:3)
   vsmBSwapParam%ndv(1:3) = ndv(1:3)

   if(vsmDbg) then
     call vsmLogPrint("vsmSetBSwap is called")
     call vsmLogPrint("myindex:",midx)
     call vsmLogPrint("ndv:",ndv)
   end if

   if(vsmBSwapParam%possible) then

     allocate(vsmBSwapParam%index2rank(ndv(1),ndv(2),ndv(3)), stat=allocstate)
     if(allocstate /= 0) call vsmError("can't allocate bswap index2rank")

     vsmBSwapParam%index2rank(1:ndv(1),1:ndv(2),1:ndv(3)) = idx

     call vsmDvData4BS

     vsmBSwapParam%bswap = .true.

   else

     call vsmLogPrint("Binary Swap is impossible. Number of Processes is not a power of 2")

   end if

   call vsmLogPrint("vsmSetBSwap ends")

 end subroutine vsmSetBinarySwap

 subroutine vsmDvData4BS
  integer :: nstage
  integer :: st_x, st_y, st_z
  integer :: ed_x, ed_y, ed_z
  integer :: dv_x, dv_y, dv_z
  integer, dimension(3) :: pindex
  logical :: a

  call vsmLogPrint("vsmDvData4BS is called")
  call vsmLogPrint("edx, edy,edz=",vsmBSwapParam%ndv)
  nstage = vsmBSwapParam%numstage
  st_x = 1; st_y = 1; st_z = 1
  ed_x = vsmBSwapParam%ndv(1)
  ed_y = vsmBSwapParam%ndv(2)
  ed_z = vsmBSwapParam%ndv(3)
  dv_x = ed_x; dv_y = ed_y; dv_z = ed_z
  do
   ! x direction
   if(dv_x > 1) then
    dv_x = dv_x/2
    if(vsmBSwapParam%myindex(1) < st_x+dv_x) then
      pindex(1) = vsmBSwapParam%myindex(1) + dv_x
      a=.true.
      ed_x = st_x+dv_x-1
    else
      pindex(1) = vsmBSwapParam%myindex(1) - dv_x
      a=.false.
      st_x = st_x+dv_x
    end if
    pindex(2) = vsmBSwapParam%myindex(2)
    pindex(3) = vsmBSwapParam%myindex(3)

    vsmBSwapParam%partner(nstage) = vsmBSwapParam%index2rank(pindex(1),pindex(2),pindex(3))
    vsmBSwapParam%above(nstage) = a
    call vsmLogPrint("binary swap x-direc: stage = ", nstage)
    call vsmLogPrint("binary swap x-direc: myindex = ", vsmBSwapParam%myindex)
    call vsmLogPrint("binary swap x-direc: parter index = ", pindex)
    call vsmLogPrint("binary swap x-direc: parter rank = ", vsmBSwapParam%partner(nstage))
    nstage = nstage - 1
    if(nstage == 0) exit
   end if

   ! y direction
   if(dv_y > 1) then
    dv_y = dv_y/2
    if(vsmBSwapParam%myindex(2) < st_y+dv_y) then
      pindex(2) = vsmBSwapParam%myindex(2) + dv_y
      a=.true.
      ed_y = st_y+dv_y-1
    else
      pindex(2) = vsmBSwapParam%myindex(2) - dv_y
      a=.false.
      st_y = st_y+dv_y
    end if
    pindex(1) = vsmBSwapParam%myindex(1)
    pindex(3) = vsmBSwapParam%myindex(3)
    vsmBSwapParam%partner(nstage) = vsmBSwapParam%index2rank(pindex(1),pindex(2),pindex(3))
    vsmBSwapParam%above(nstage) = a
    call vsmLogPrint("binary swap y-direc: stage = ", nstage)
    call vsmLogPrint("binary swap y-direc: myindex = ", vsmBSwapParam%myindex)
    call vsmLogPrint("binary swap y-direc: parter index = ", pindex)
    call vsmLogPrint("binary swap y-direc: parter rank = ", vsmBSwapParam%partner(nstage))
    nstage = nstage - 1
    if(nstage == 0) exit
   end if

   ! z direction
   if(dv_z > 1) then
    dv_z = dv_z/2
    if(vsmBSwapParam%myindex(3) < st_z+dv_z) then
      pindex(3) = vsmBSwapParam%myindex(3) + dv_z
      a=.true.
      ed_z = st_z+dv_z-1
    else
      pindex(3) = vsmBSwapParam%myindex(3) - dv_z
      a=.false.
      st_z = st_z+dv_z
    end if
    pindex(1) = vsmBSwapParam%myindex(1)
    pindex(2) = vsmBSwapParam%myindex(2)
    vsmBSwapParam%partner(nstage) = vsmBSwapParam%index2rank(pindex(1),pindex(2),pindex(3))
    vsmBSwapParam%above(nstage) = a
    call vsmLogPrint("binary swap z-direc: stage = ", nstage)
    call vsmLogPrint("binary swap z-direc: myindex = ", vsmBSwapParam%myindex)
    call vsmLogPrint("binary swap z-direc: parter index = ", pindex)
    call vsmLogPrint("binary swap z-direc: parter rank = ", vsmBSwapParam%partner(nstage))
    nstage = nstage-1
    if(nstage == 0) exit
   end if
  end do

 end subroutine vsmDvData4BS

 subroutine vsmCheckImageRegionB(startpixel, endpixel)
   integer, intent(out) :: startpixel, endpixel
   integer :: i, numpxls
!   integer :: comm, myrank, psize
!   call vsmGetMPIVars(comm, myrank, psize)

   startpixel = 1
   endpixel   = vsmImageData%pxlsize

   do i=0, vsmImageData%pxlsize-1
     if(vsmImageData%image(i*5+5) < vsmFarAwayd10) then
       startpixel = i
       exit
     end if
   end do

   do i=vsmImageData%pxlsize-1, 0, -1
     if(vsmImageData%image(i*5+5) < vsmFarAwayd10) then
       endpixel = i
       exit
     end if
   end do

   numpxls = endpixel - startpixel + 1
   if(numpxls < 0) then
   end if
 end subroutine vsmCHeckImageRegionB

 subroutine vsmBinarySwapIM(solid, cam, vis, tim)
   logical, intent(in) :: solid
   integer, intent(in) :: cam, vis, tim
   integer :: comm, myrank, psize, ierr, master
#ifndef CLANG
   integer, dimension(mpi_status_size) :: stat
#endif
   integer :: pixels, pixel5
   integer :: i, st_pixel, ed_pixel, prank
   integer, dimension(2) :: spixl
   integer :: counts, countr, length
   integer :: tag, count, dest, src

   character(4) :: viskind, camnum, rank 
   character(7) :: timestep

   if(vsmPCmode /= 0) call vsmError("binaryswapIM is called though pcmode is not 0")
   call vsmLogPrint("----- Binary Swap")
   call vsmGetMPIVars(comm, myrank, psize)
   tag = 0
   st_pixel = 1
   ed_pixel = vsmImageData%rimgsize !vsmImageData%width * vsmImageData%height ! send receive data*5

   call vsmCheckImageL(st_pixel, ed_pixel)
!   do i=vsmBSwapParam%numstage, 1, -1
   do i=1, vsmBSwapParam%numstage, 1 ! 2015/9/27
     call vsmLogPrint("binary swap stage = ", i)
     prank = vsmBSwapParam%partner(i)
     length = ed_pixel - st_pixel +1

     call vsmLogPrint("bswap parter ", prank)
     if(mod(length,2) /=0) call vsmError("bswap: length is not even")

     call vsmClearRImage
     if(vsmBSwapParam%above(i)) then
       countr = length/2
       !counts = countr
       counts = length - countr

       count = counts*5
#ifndef CLANG
       call mpi_send(vsmImageData%image((st_pixel+countr-1)*5+1), count, mpi_real, prank, tag, comm, ierr)
#else
       call VSMMPISENDFLOAT(vsmImageData%image((st_pixel+countr-1)*5+1), count, prank)
#endif
       count = countr*5
#ifndef CLANG
       call mpi_recv(vsmImageData%image_r, count, mpi_real, prank, tag, comm, stat, ierr)
#else
       call VSMMPIRECVFLOAT(vsmImageData%image_r, count, prank)
#endif

       !st_pixel = st_pixel
       ed_pixel = st_pixel+countr-1
     else
       counts = length/2
       !countr = counts
       countr = length - counts
       count = countr*5
#ifndef CLANG 
       call mpi_recv(vsmImageData%image_r, count, mpi_real, prank, tag, comm, stat, ierr)
#else
       call VSMMPIRECVFLOAT(vsmImageData%image_r, count, prank)
#endif
       count = counts*5
#ifndef CLANG 
       call mpi_send(vsmImageData%image((st_pixel-1)*5+1), count, mpi_real, prank, tag, comm, ierr)
#else
       call VSMMPISENDFLOAT(vsmImageData%image((st_pixel-1)*5+1), count, prank) 
#endif
       !ed_pixel = ed_pixel
       st_pixel = st_pixel+counts
     end if

     call vsmLogPrint("stpixel ", st_pixel)
     call vsmLogPrint("edpixel ", ed_pixel)
     if(solid) then
       call vsmSolidBlending(st_pixel, ed_pixel)
     else
       call vsmAlphaBlending(st_pixel, ed_pixel)
     end if
     call vsmCheckImageL(st_pixel, ed_pixel)
   end do

   call vsmLogPrint("start gathering")

   if(myrank == 0) then

     do i=1, psize-1
       count = 2
#ifndef CLANG
       call mpi_recv(spixl, count, mpi_integer, i, tag, comm, stat, ierr)
#else
       call VSMMPIRECVINT(spixl, count, i)
#endif
       count = spixl(2)*5
#ifndef CLANG
       call mpi_recv(vsmImageData%image((spixl(1)-1)*5+1), count, mpi_real, i, tag, comm, stat, ierr)
#else
       call VSMMPIRECVFLOAT(vsmImageData%image((spixl(1)-1)*5+1), count, i)
#endif
     end do

   else
     spixl(1) = st_pixel
     spixl(2) = ed_pixel-st_pixel+1
     dest = 0
     count = 2
#ifndef CLANG
     call mpi_send(spixl, count, mpi_integer, dest, tag, comm, ierr)
#else
     call VSMMPISENDINT(spixl, count, dest)
#endif
     count = spixl(2)*5
#ifndef CLANG
     call mpi_send(vsmImageData%image((st_pixel-1)*5+1), count, mpi_real, dest, tag, comm, ierr)
#else
     call VSMMPISENDFLOAT(vsmImageData%image((st_pixel-1)*5+1), count, dest)
#endif
   end if

   if(solid) then
     src = 0
     count = vsmImageData%rimgsize*5
#ifndef CLANG
     call mpi_bcast(vsmImageData%image, count, mpi_real, src, comm, ierr)
#else
     call VSMMPIBCASTFLOAT(vsmImageData%image, count, src)
#endif
   end if

   call vsmLogPrint("----- Binary Swap ends")
 end subroutine vsmBinarySwapIM


 subroutine vsmBinarySwapIMRS(cam, vis, tim)
   integer, intent(in) :: cam, vis, tim
   integer :: comm, myrank, psize, ierr, master
#ifndef CLANG
   integer, dimension(mpi_status_size) :: stat
#endif
   integer :: i, st_pixel, ed_pixel, prank
   integer, dimension(2) :: spixl
   integer :: counts, countr, length, fstage, numr, nums, soffset
   integer :: exproc, nbsproc
   integer :: tag, src, dest, count

   character(4) :: viskind, camnum, rank 
   character(7) :: timestep

   if(vsmPCmode /= 0) call vsmError("binaryswapIM is called though pcmode is not 0")
   tag = 0
   call vsmLogPrint("----- Binary Swap IMRS")
   call vsmGetMPIVars(comm, myrank, psize)

   if(vsmDbg) then
   call vsmLogPrint("vsmDbg raysegment")
   call vsmBlendRaySegmentF(0)
   call vsmLogPrint("vsmDbg raysegment ends")
   select case(myrank)
   case (0)
     call vsmWriteImage('image0')
   case(1)
     call vsmWriteImage('image1')
   case(2)
     call vsmWriteImage('image2')
   case(3)
     call vsmWriteImage('image3')
   case(4)
     call vsmWriteImage('image4')
   case(5)
     call vsmWriteImage('image5')
   case(6)
     call vsmWriteImage('image6')
   case(7)
     call vsmWriteImage('image7')
   end select
   end if

  nbsproc = 2**vsmBSwapParam%numstage

   if(psize > nbsproc) then

     exproc = psize - nbsproc
     length = vsmImageData%rimgsize
     nums = length*vsmRSD*vsmRSBNum
     call  vsmClearRaySegmentR

     if(myrank < exproc) then
       prank = myrank + nbsproc
       countr = length/2
       counts = length - countr
       nums = counts*vsmRSD*vsmRSBNum
       count = 1
#ifndef CLANG
       call mpi_send(nums, count, mpi_integer, prank, tag, comm, ierr)
#else
       call VSMMPISENDINT(nums, count, prank)
#endif
       call vsmLogPrint("data sent max",vsmRSD*vsmRSBNum*vsmImageData%rimgsize)
       call vsmLogPrint("data sent",nums)

       call vsmLogPrint("data send from",countr*vsmRSBNum*vsmRSD+1)
       call vsmLogPrint("data send to",countr*vsmRSBNum*vsmRSD+nums)
#ifndef CLANG
       call mpi_send(vsmImageData%rsdata(countr*vsmRSBNum*vsmRSD+1), nums, &
            mpi_integer, prank, tag, comm, ierr)
#else
       call VSMMPISENDINT(vsmImageData%rsdata(countr*vsmRSBNum*vsmRSD+1), nums, prank)
#endif

       call vsmLogPrint("rsdata sent")
      count = 1
#ifndef CLANG
       call mpi_recv(numr, count, mpi_integer, prank, tag, comm, stat, ierr)
       call vsmLogPrint("data receive", numr)
       call mpi_recv(vsmImageData%rsdata_r, numr, mpi_integer, prank, tag, comm, stat, ierr)
#else
       call VSMMPIRECVINT(numr, count, prank)
       call vsmLogPrint("data receive", numr)
       call VSMMPIRECVINT(vsmImageData%rsdata_r, numr, prank)
#endif
       call vsmLogPrint("data received")
       st_pixel = st_pixel
       ed_pixel = st_pixel+countr-1
       soffset = 1
       call vsmLogPrint("st_pixel", st_pixel)
       call vsmLogPrint("ed_pixel", ed_pixel)
     else if(myrank > nbsproc-1) then
       prank = myrank - nbsproc
       counts = length/2
       countr = length - counts
       nums = counts*vsmRSD*vsmRSBNum
       count = 1
#ifndef CLANG
       call mpi_recv(numr, count, mpi_integer, prank, tag, comm, stat, ierr)
#else
       call VSMMPIRECVINT(numr, count, prank)
#endif
       call vsmLogPrint("data received max",vsmRSD*vsmRSBNum*vsmImageData%rimgsize)
       call vsmLogPrint("data received",numr)
#ifndef CLANG
       call mpi_recv(vsmImageData%rsdata_r, numr, mpi_integer, prank, tag, comm, stat, ierr)
#else
       call VSMMPIRECVINT(vsmImageData%rsdata_r, numr, prank)
#endif
       call vsmLogPrint("rs data received",numr)
       call vsmLogPrint("all data received")
       count = 1
#ifndef CLANG
       call mpi_send(nums, count, mpi_integer, prank, tag, comm, ierr)
       call vsmLogPrint("data send",nums)
       call mpi_send(vsmImageData%rsdata, nums, mpi_integer, prank, tag, comm, ierr)
#else
       call VSMMPISENDINT(nums, count, prank)
       call vsmLogPrint("data send",nums)
       call VSMMPISENDINT(vsmImageData%rsdata, nums, prank)
#endif
       call vsmLogPrint("data sent")

       st_pixel = st_pixel+counts
       ed_pixel = ed_pixel
       soffset = counts+1
       call vsmLogPrint("st_pixel", st_pixel)
       call vsmLogPrint("ed_pixel", ed_pixel)
     end if

     if(myrank < exproc .or. myrank > nbsproc-1) then
       call vsmLogPrint("check segment ")
       call vsmCheckRaySegment
       call vsmLogPrint("copy segment ")
       call vsmCopyRaySegment(soffset, countr, numr, 1)
       call vsmLogPrint("check segment ")
       call vsmCheckRaySegment

       call vsmLogPrint("clear segment ")
       call vsmClearRaySegment
       call vsmLogPrint("Merge segment ")
       call vsmMergeRaySegmentBN(countr)

       if(myrank < exproc) then
#ifndef CLANG
         call mpi_recv(vsmImageData%rsdata(countr*vsmRSBNum*vsmRSD+1), nums, &
              mpi_integer, prank, tag, comm, stat, ierr)
#else
         call VSMMPIRECVINT(vsmImageData%rsdata(countr*vsmRSBNum*vsmRSD+1), nums, prank)
#endif
       else if(myrank > nbsproc-1) then
#ifndef CLANG
         call mpi_send(vsmImageData%rsdata_r, numr, mpi_integer, prank, tag, comm, ierr)
#else
         call VSMMPISENDINT(vsmImageData%rsdata_r, numr, prank)
#endif
       end if
     end if
   end if

#ifndef CLANG
   call mpi_barrier(comm, ierr)
#else
   call VSMMPIBARRIER()
#endif

   if(myrank < nbsproc) then

     st_pixel = 1
     ed_pixel = vsmImageData%rimgsize !vsmImageData%width * vsmImageData%height ! send receive data*5

     fstage = vsmBSwapParam%numstage
     do i=1, fstage
       call vsmLogPrint("binary swap stage = ", i)
       prank = vsmBSwapParam%partner(i)
       length = ed_pixel - st_pixel +1

       call vsmLogPrint("bswap partner ", prank)

       call  vsmClearRaySegmentR
       call vsmLogPrint("RaySegmentR cleared")
       if(vsmBSwapParam%above(i)) then
         countr = length/2
         counts = length - countr
         nums = counts*vsmRSD*vsmRSBNum*(2**(i-1))
         count = 1
#ifndef CLANG
         call mpi_send(nums, count, mpi_integer, prank, tag, comm, ierr)
#else
         call VSMMPISENDINT(nums, count, prank)
#endif
         call vsmLogPrint("data sent max",vsmRSD*vsmRSBNum*vsmImageData%rimgsize)
         call vsmLogPrint("data sent",nums)
         call vsmLogPrint("data send from",countr*vsmRSBNum*vsmRSD*(2**(i-1))+1)
         call vsmLogPrint("data send to",countr*vsmRSBNum*vsmRSD*(2**(i-1))+nums)
#ifndef CLANG
         call mpi_send(vsmImageData%rsdata(countr*vsmRSBNum*vsmRSD*(2**(i-1))+1), nums, &
            mpi_integer, prank, tag, comm, ierr)
#else
         call VSMMPISENDINT(vsmImageData%rsdata(countr*vsmRSBNum*vsmRSD*(2**(i-1))+1), nums, prank)
#endif
         call vsmLogPrint("rsdata sent")
         count = 1
#ifndef CLANG
         call mpi_recv(numr, count, mpi_integer, prank, tag, comm, stat, ierr)
         call vsmLogPrint("data receive", numr)
         call mpi_recv(vsmImageData%rsdata_r, numr, mpi_integer, prank, tag, comm, stat, ierr)
#else
         call VSMMPIRECVINT(numr, count, prank)
         call vsmLogPrint("data receive", numr)
         call VSMMPIRECVINT(vsmImageData%rsdata_r, numr, prank)
#endif
         call vsmLogPrint("data received")
         st_pixel = st_pixel
         ed_pixel = st_pixel+countr-1
         soffset = 1
         call vsmLogPrint("st_pixel", st_pixel)
         call vsmLogPrint("ed_pixel", ed_pixel)

       else
         counts = length/2
         countr = length - counts
         nums = counts*vsmRSD*vsmRSBNum*(2**(i-1))
         count = 1
#ifndef CLANG
         call mpi_recv(numr, count, mpi_integer, prank, tag, comm, stat, ierr)
         call vsmLogPrint("data received max",vsmRSD*vsmRSBNum*vsmImageData%rimgsize)
         call vsmLogPrint("data received",numr)
         call mpi_recv(vsmImageData%rsdata_r, numr, mpi_integer, prank, tag, comm, stat, ierr)
#else
         call VSMMPIRECVINT(numr, count, prank)
         call vsmLogPrint("data received max",vsmRSD*vsmRSBNum*vsmImageData%rimgsize)
         call vsmLogPrint("data received",numr)
         call VSMMPIRECVINT(vsmImageData%rsdata_r, numr, prank)
#endif

         call vsmLogPrint("rs data received",numr)
         call vsmLogPrint("all data received")

#ifndef CLANG
         call mpi_send(nums, count, mpi_integer, prank, tag, comm, ierr)
         call vsmLogPrint("data send",nums)
         call mpi_send(vsmImageData%rsdata, nums, mpi_integer, prank, tag, comm, ierr)
#else
         call VSMMPISENDINT(nums, count, prank)
         call vsmLogPrint("data send",nums)
         call VSMMPISENDINT(vsmImageData%rsdata, nums, prank)
#endif
         call vsmLogPrint("data sent")

         st_pixel = st_pixel+counts
         ed_pixel = ed_pixel
         soffset = counts+1
         call vsmLogPrint("st_pixel", st_pixel)
         call vsmLogPrint("ed_pixel", ed_pixel)
       end if
       call vsmLogPrint("check segment ")
       call vsmCheckRaySegment
       call vsmLogPrint("copy segment ")
       call vsmCopyRaySegment(soffset, countr, numr, i)
       call vsmLogPrint("check segment ")
       call vsmCheckRaySegment

       call vsmLogPrint("clear segment ")
       call vsmClearRaySegment
       call vsmLogPrint("Merge segment ")
       call vsmMergeRaySegment(i, countr)

       call vsmLogPrint("stpixel ", st_pixel)
       call vsmLogPrint("edpixel ", ed_pixel)
       call vsmCheckRaySegment

!     call vsmCheckImageL(st_pixel, ed_pixel)
     end do
     call vsmLogPrint("start blend ray segment")
     call vsmBlendRaySegmentF(fstage)
     call vsmLogPrint("start gathering")

     if(myrank == 0) then
       do i=1, nbsproc-1
#ifndef CLANG
         count = 2
         call mpi_recv(spixl, count, mpi_integer, i, tag, comm, stat, ierr)
         count = spixl(2)*5
         call mpi_recv(vsmImageData%image((spixl(1)-1)*5+1), count, mpi_real, i, tag, comm, stat, ierr)
#else
         count = 2
         call VSMMPIRECVINT(spixl, count, i)
         count = spixl(2)*5
         call VSMMPIRECVFLOAT(vsmImageData%image((spixl(1)-1)*5+1), count, i)
#endif
       end do
     else
       spixl(1) = st_pixel
       spixl(2) = ed_pixel-st_pixel+1
       dest = 0
#ifndef CLANG
       count = 2
       call mpi_send(spixl, count, mpi_integer, dest, tag, comm, ierr)
       count = spixl(2)*5
       call mpi_send(vsmImageData%image((st_pixel-1)*5+1), count, mpi_real, dest, tag, comm, ierr)
#else
       count = 2
       call VSMMPISENDINT(spixl, count, dest)
       count = spixl(2)*5
       call VSMMPISENDFLOAT(vsmImageData%image((st_pixel-1)*5+1), count, dest)
#endif

     end if
   end if ! if(myrank < nbsproc)

   call vsmLogPrint("----- Binary Swap ends")
 end subroutine vsmBinarySwapIMRS

 subroutine vsmBinarySwap4SolidIM(cam, vis, tim)
   integer, intent(in) :: cam, vis, tim
   integer :: comm, myrank, psize, ierr, master
#ifndef CLANG
   integer, dimension(mpi_status_size) :: stat
#endif
   integer :: pixels, pixel5
   integer :: i, st_pixel, ed_pixel, prank
   integer :: nbsproc, exproc
   integer, dimension(2) :: spixl
   integer :: counts, countr, length
   integer :: count, tag, src, dest

   character(4) :: viskind, camnum, rank 
   character(7) :: timestep

   if(vsmPCmode /= 0) call vsmError("BSwap4solidIm is called though pcmode is not 0")
   tag = 0
   call vsmLogPrint("----- Binary Swap for Solid")
   call vsmGetMPIVars(comm, myrank, psize)

   nbsproc = 2**vsmBSwapParam%numstage

   if(psize > nbsproc) then

     exproc = psize - nbsproc

     if(myrank < exproc) then
       prank = myrank + nbsproc
       count = vsmImageData%rimgsize*5
#ifndef CLANG
       call mpi_recv(vsmImageData%image_r, count, mpi_real, prank, tag, comm, stat, ierr)
#else
       call VSMMPIRECVFLOAT(vsmImageData%image_r, count, prank)
#endif
       call vsmSolidBlending(1, vsmImageData%rimgsize)
     else if(myrank > nbsproc-1) then
       prank = myrank - nbsproc
       count = vsmImageData%rimgsize*5
#ifndef CLANG
       call mpi_send(vsmImageData%image, count, mpi_real, prank, tag, comm, ierr)
#else
       call VSMMPISENDFLOAT(vsmImageData%image, count, prank)
#endif
     end if
   end if
#ifndef CLANG
   call mpi_barrier(comm, ierr)
#else
   call VSMMPIBARRIER()
#endif
   if(myrank < nbsproc) then

     st_pixel = 1
     ed_pixel = vsmImageData%rimgsize !vsmImageData%width * vsmImageData%height ! send receive data*5

     call vsmCheckImageL(st_pixel, ed_pixel)
!     do i=vsmBSwapParam%numstage, 1, -1
     do i=1, vsmBSwapParam%numstage
       call vsmLogPrint("binary swap for solid stage = ", i)
       prank = vsmBSwapParam%partner(i)
       length = ed_pixel - st_pixel +1
       call vsmClearRImage
       if(vsmBSwapParam%above(i)) then
         countr = length/2
         !counts = countr
         counts = length - countr
#ifndef CLANG
         count = counts*5
         call mpi_send(vsmImageData%image((st_pixel+countr-1)*5+1), count, mpi_real, prank, tag, comm, ierr)
         count = countr*5
         call mpi_recv(vsmImageData%image_r, count, mpi_real, prank, tag, comm, stat, ierr)
#else
         count = counts*5
         call VSMMPISENDFLOAT(vsmImageData%image((st_pixel+countr-1)*5+1), count, prank)
         count = countr*5
         call VSMMPISENDFLOAT(vsmImageData%image_r, count, prank)
#endif
         !st_pixel = st_pixel
         ed_pixel = st_pixel+countr-1
       else
         counts = length/2
         !countr = counts
         countr = length - counts
#ifndef CLANG
         count = countr*5
         call mpi_recv(vsmImageData%image_r, count, mpi_real, prank, tag, comm, stat, ierr)
         count = counts*5
         call mpi_send(vsmImageData%image((st_pixel-1)*5+1), count, mpi_real, prank, tag, comm, ierr)
#else
         count = countr*5
         call VSMMPIRECVFLOAT(vsmImageData%image_r, count, prank)
         count = counts*5
         call VSMMPISENDFLOAT(vsmImageData%image((st_pixel-1)*5+1), count, prank)
#endif
         !ed_pixel = ed_pixel
         st_pixel = st_pixel+counts
       end if

       call vsmSolidBlending(st_pixel, ed_pixel)
       call vsmCheckImageL(st_pixel, ed_pixel)
     end do

     call vsmLogPrint("start gathering")

     if(myrank == 0) then
       do i=1, nbsproc-1
#ifndef CLANG
         count = 2
         call mpi_recv(spixl, count, mpi_integer, i, tag, comm, stat, ierr)
         count = spixl(2)*5
         call mpi_recv(vsmImageData%image((spixl(1)-1)*5+1), count, mpi_real, i, tag, comm, stat, ierr)
#else
         count = 2
         call VSMMPIRECVINT(spixl, count, i)
         count = spixl(2)*5
         call VSMMPIRECVFLOAT(vsmImageData%image((spixl(1)-1)*5+1), count, i)
#endif
       end do
     else
       spixl(1) = st_pixel
       spixl(2) = ed_pixel-st_pixel+1
       dest = 0
#ifndef CLANG
       count = 2
       call mpi_send(spixl, count, mpi_integer, dest, tag, comm, ierr)
       count = spixl(2)*5
       call mpi_send(vsmImageData%image((st_pixel-1)*5+1), count, mpi_real, dest, tag, comm, ierr)
#else
       count = 2
       call VSMMPISENDINT(spixl, count, dest)
       count = spixl(2)*5
       call VSMMPISENDFLOAT(vsmImageData%image((st_pixel-1)*5+1), count, dest)
#endif
     end if

   end if
   src = 0
   count = vsmImageData%rimgsize*5
#ifndef CLANG
   call mpi_bcast(vsmImageData%image, count, mpi_real, src, comm, ierr)
#else
   call VSMMPIBCASTFLOAT(vsmImageData%image, count, src)
#endif
   !----
   call vsmLogPrint("----- Binary Swap for Solid ends")
 end subroutine vsmBinarySwap4SolidIM

end module vsmBSwap

