module vsmConst
 ! big endian for outputting bmp
 logical, parameter :: bigend = .false.

 !for double precision
 integer, parameter :: vdp = selected_real_kind(14)

 !for 2bytes integer (short), from -32768 to 32767
 integer, parameter :: isp = selected_int_kind(4)

 !debug
 logical, parameter :: vsmDbg = .false.

 !debug
 logical, parameter :: vsmReInitMode = .false.

 !pi
 real(kind=vdp) :: pi = 3.1415926535897932384_vdp

 integer, parameter :: int_max = 2147483647

 !for OpenMP
 integer, parameter :: vsmMaxOMP = 512
 integer, parameter :: vsmMaxOMP4Solid = 11 ! currently maximum 11

 !for file open
! integer, parameter :: vsmUnitNumErr = 6 ! Fortran's standard error num.. maybe
! integer, parameter :: vsmUnitNumLog = 35 ! logfiles
 integer, parameter :: vsmUnitNum =  36! for reading param files (from  vsmUnitNum,... if using include), image and datainfo
 integer, parameter :: vsmUnitMaxNum = 99 ! Fortran's Max unit number .. maybe
 integer, parameter :: vsmUnitNumPCSline = 37 ! point cloud data:stream lines
 integer, parameter :: vsmUnitNumPCArrows = 38 ! arrows
 integer, parameter :: vsmUnitNumPCParticle = 39 ! particles
 integer, parameter :: vsmUnitNumPCIso = 40 ! isosurface
 integer, parameter :: vsmUnitNumPCSlice = 41 ! slice

 ! image
 character(5), parameter :: vsmImageFile = 'image' ! imagev[vis]c[camera]t[time].ppm
 integer, parameter :: vsmImgKind = 2 ! 0:BMP, 1:PPM, 2:PNG

 ! config change record
 character(16), parameter :: vsmConfigChangeFile = 'ConfigChange.vsm'

 character(15), parameter :: vsmDataInfoFile = 'vsmDataInfo.vsm'

 !for camera
 integer, parameter :: vsmMaxCamera = 256

 !for light
 integer, parameter :: vsmMaxLight = 4

 !image
 real(kind=vdp), parameter :: vsmFaraway = 1.0D+20
 real(kind=vdp), parameter :: vsmFarawayd10 = 1.0D+19

 integer, parameter :: vsmPbLen = 512
 !integer, parameter :: vsmPbLen = 10000
 integer, parameter :: vsmPackConst = 900

 !isosurface
 integer, parameter :: vsmMaxIsosurf = 5 ! more than 5, less than 20
 !slice
 integer, parameter :: vsmMaxSlice = 5 ! more than 4
 !stream line
 integer, parameter :: vsmMaxSeed = 50
 integer, parameter :: vsmMaxSlineSteps = 40000
 real(kind=vdp), parameter :: vsmIntegStride = 0.2_vdp ! min grid size * vsmIntegStride = Runge-Kutta integration interval
 !arrow : arrows
 integer, parameter :: vsmMaxArrows = 5
 real, parameter :: vsmARConeRatio = 0.25 ! ratio of arrow's cone >0.0 .and. < 1.0
 integer, parameter :: vsmARMaxSeed = 50
 
 !volume rendering
 integer, parameter :: vsmMaxVolren = 2
 integer, parameter :: vsmMaxRayCount = 1000000
 real, parameter :: vsmErtValue = 0.9999
 real(kind=vdp), parameter :: vsmRayStrideVR = 1.0_vdp ! min grid size * vsmRayStride = sampling interval
 real(kind=vdp), parameter :: vsmRayStride   = 1.0_vdp ! min grid size * vsmRayStride = sampling interval

 integer, parameter :: vsmMaxColmap = 256

 integer, parameter :: vsmMaxVisKind = 15

 ! data
 integer, parameter :: vsmMaxScalar = 5
 integer, parameter :: vsmMaxVector = 5
 integer, parameter :: vsmMaxPtcl   = 2

 ! Octree
 integer, parameter :: vsmMaxOctLevel = 3
 integer, parameter :: vsmMaxOctNum   = 1 + 8 + 8*8 + 8*8*8
 integer, dimension(0:vsmMaxOctLevel) :: vsmOctNum= (/1, 1+8, 1+8+8*8, 1+8+8*8+8*8*8/)
 real, parameter :: vsmSkipAlpha = 0.0001

 !Transformation
 type vismo_transdata
   real(kind=vdp), dimension(3,3) :: matrix
   real(kind=vdp), dimension(3,3) :: invmatrix
   real(kind=vdp), dimension(3)   :: pos
 end type vismo_transdata

 !Point Cloud
! integer, parameter :: vsmPCIsoC = 3 ! width*height*
 integer, parameter :: vsmPCArrowC = 1000! vsmMaxArrows*
 integer, parameter :: vsmPCParticleC = 10000 
 character(6), parameter :: vsmPcFile = 'pcdata' ! 
 integer, parameter :: vsmPCISOCrossPointPerPixel = 6 ! more than 6
 integer, parameter :: vsmPCISOElements = 3
 integer, parameter :: vsmPCSLICEElements = 3
 
 ! For Vectorization :: SX-ACE, etc...
 integer, parameter :: vsmVeclen = 256
 integer, parameter :: vsmNoVeclen = 4
 integer, parameter :: vsmVecMaxOpenMP = 4
  integer, parameter :: vsmMaxNumArrows = vsmVecLen*8+1 ! max number of arrows per an image for vector machine
 integer, parameter :: vsmVecMaxOMP = 4
 integer, parameter :: vsmOMPVecLen = vsmVecLen*vsmVecMaxOpenMP


 integer, parameter :: vsmRSBNum = 2
 integer, parameter :: vsmRSD = 3

! for tubes
 integer, parameter :: vsmMaxTubes = 100 ! for each MPI Process
 ! STRUCTURES
 !VISPARAM
  ! scalar data visualization
 type vismo_isodata
   real(kind=vdp) :: iso_level
   integer :: nscal
   real :: alpha
   real, dimension(3) :: amb, dif, spec
   real :: shin
   integer :: coloring
 end type vismo_isodata

 type vismo_slicedata
   real(kind=vdp), dimension(4) :: sc !a, b, c, d ! f(x,y,z) = ax+by+cz+d 
   integer :: nscal
   real :: alpha, amb, dif, spec, shin
 end type vismo_slicedata

 type vismo_volrendata
   integer :: nscal
   real :: amb, dif, spec, shin
 end type vismo_volrendata

 ! vector data visualization
 type vismo_strlinedata
   real(kind=vdp), dimension(3, vsmMaxSeed) :: seed
   integer :: num_seed
   integer, dimension(vsmMaxSeed) :: steps_remained
   integer :: nvec
   real, dimension(3) :: amb, dif, spec ! for monocolor lines
   real :: shin
   real(kind=vdp) :: radius
   logical :: monocolor
 end type vismo_strlinedata

 type vismo_arrowdata
   integer :: nvec
   real(kind=vdp), dimension(4) :: sc !a, b, c, d ! f(x,y,z) = ax+by+cz+d 
   real(kind=vdp), dimension(3, vsmARMaxSeed) :: seed
   integer :: num_seed
   real(kind=vdp) :: density
   real(kind=vdp) :: radius, length
   integer :: mode ! 0: eq, 1: all region, 2:seeds
   logical :: variable
 end type vismo_arrowdata

 ! particle data
 type vismo_ptcldata
  integer :: nptcl
  real(kind=vdp) ::radius
  real, dimension(3) :: amb, dif, spec
  real :: shin
  logical :: monocolor
 end type vismo_ptcldata

 ! Those are the common variables for any vis_kinds
 type vismo_interval
   real(kind=vdp) :: ray_interval ! for Ray-Casting
   real(kind=vdp) :: dh           ! for StreamLines' Runge-Kutta integration
 end type vismo_interval

 ! Those are the common variables for mpi
 type vismo_mpi
   integer :: comm   ! communicator
   integer :: myrank ! rank
   integer :: psize
   integer :: omp_tn ! OpenMP threads
 end type vismo_mpi

 ! visualization parameters 
 type vismo_visparameters
   ! isosurface
   type(vismo_isodata), dimension(vsmMaxIsosurf) :: isoparam
   integer :: iso_num
   logical :: iso_solid, iso_trans

   ! color slice
   type(vismo_slicedata), dimension(vsmMaxSlice) :: sliceparam
   integer :: slice_num
   logical :: slice_solid, slice_trans

   ! volume rendering
   type(vismo_volrendata), dimension(vsmMaxScalar) :: volparam
   integer :: volren_num

   ! stream line
   type(vismo_strlinedata), dimension(vsmMaxVector) :: slineparam
   integer :: sline_num

   ! arrows
   type(vismo_arrowdata), dimension(vsmMaxArrows) :: arrowparam
   integer :: arrow_num

   ! particle
   type(vismo_ptcldata), dimension(vsmMaxPtcl) :: ptclparam
   integer :: ptcl_num

   !lights
   integer :: light_num
   logical, dimension(vsmMaxLight) :: light_used

   !depthcue
   logical :: depthq
 end type vismo_visparameters

! Light
 type vismo_light
  real, dimension(3) :: pos
  real, dimension(3) :: direc
  real, dimension(3) :: amb, dif, spec
  logical :: para !parallel light or point light
 end type vismo_light

 !Image
  type vismo_image
   integer :: width, height
   real, dimension(:), allocatable :: image
   real, dimension(:), allocatable :: image_r
   real(kind=vdp), dimension(3) :: topleft
   real(kind=vdp) :: dp
   real, dimension(3) :: bgcolor
   integer :: pxlsize, rimgsize
   integer, dimension(:), allocatable :: rsdata    ! pixn, in, out, color
   integer, dimension(:), allocatable :: rsdata_r
   character, dimension(:), allocatable :: imageUC
   logical :: BGC
  end type vismo_image

!Depth Cueing
 type vismo_depthparam
  integer :: depth_kind ! 1:liner, 2:exp(-x), 3:exp(-x**2)
  real(kind=vdp), dimension(2) :: dist
 end type vismo_depthparam
 
 !Data
  type vismo_scaldata
#ifndef FLOATDATA
  real(kind=vdp), dimension(:,:,:), pointer :: scal
#else
  real, dimension(:,:,:), pointer :: scal
#endif
  real(kind=vdp) :: minvalue, maxvalue
 end type vismo_scaldata

 type vismo_vectdata
#ifndef FLOATDATA
  real(kind=vdp), dimension(:,:,:), pointer :: vect_x
  real(kind=vdp), dimension(:,:,:), pointer :: vect_y
  real(kind=vdp), dimension(:,:,:), pointer :: vect_z
#else
  real, dimension(:,:,:), pointer :: vect_x
  real, dimension(:,:,:), pointer :: vect_y
  real, dimension(:,:,:), pointer :: vect_z
#endif
  real(kind=vdp) :: minvalue, maxvalue
 end type vismo_vectdata

 type vismo_ptdata
#ifndef FLOATDATA
  real(kind=vdp), dimension(:), pointer :: x
  real(kind=vdp), dimension(:), pointer :: y
  real(kind=vdp), dimension(:), pointer :: z
  real(kind=vdp), dimension(:), pointer :: scal
#else
  real, dimension(:), pointer :: x
  real, dimension(:), pointer :: y
  real, dimension(:), pointer :: z
  real, dimension(:), pointer :: scal
#endif
  integer, dimension(:), pointer :: q ! q(m) == 0 : partcile(m) is drawn, q(m) /= 0: not drawn
  integer :: pNum                  
end type vismo_ptdata

 type vismo_datainfo
   integer :: NumOfScal
   integer :: NumOfVect
   integer :: NumOfPtcl
 end type vismo_datainfo

!Coordinates
 !whole data
 type vismo_coord
  integer, dimension(3) :: nx ! number of total points
  real(kind=vdp), dimension(3) :: origin  !origin
  real(kind=vdp), dimension(3) :: cend    !max
  real(kind=vdp), dimension(3) :: dx
  real(kind=vdp), dimension(:), allocatable :: coordx, coordy, coordz
  logical :: uniform
  real(kind=vdp), dimension(3, 8) :: wdvertex
  integer, dimension(2, 12) :: lvnum
 end type vismo_coord

 !local data (distributed data)
 type vismo_localcoord
  integer, dimension(3) :: nx  ! local number of total points (with overlapped points)
  integer, dimension(3) :: lx0 ! local real origin's number
  integer, dimension(3) :: lnx ! local real number of points (without overlapped points)
  real(kind=vdp), dimension(3) :: lorigin  !local origin (without overlapped points)
  real(kind=vdp), dimension(3) :: lcend    !local max (without overlapped points)
  real(kind=vdp), dimension(3) :: origin  !local origin
  real(kind=vdp), dimension(3) :: cend    !local max
  real(kind=vdp), dimension(3) :: dx
  real(kind=vdp), dimension(:), allocatable :: coordx, coordy, coordz
  logical :: uniform
 end type vismo_localcoord

! Color Map
 type vismo_scolormap
  real(kind=vdp), dimension(vsmMaxColmap) :: value
  real(kind=vdp) :: dv
  real, dimension(4, vsmMaxColmap) :: color
  integer :: colnum
  logical :: set, eqspace
 end type vismo_scolormap

 type vismo_vcolormap
  real(kind=vdp), dimension(vsmMaxColmap) :: value
  real(kind=vdp) :: dv
  real, dimension(3, vsmMaxColmap) :: color
  integer :: colnum
  logical :: set, eqspace
 end type vismo_vcolormap

!CAMERA
 ! pos, front, up are required
 type vismo_camera
  real(kind=vdp), dimension(3) :: pos
  real(kind=vdp), dimension(3) :: front, up, right
  real(kind=vdp) :: fovy
  real(kind=vdp) :: near, far
  real(kind=vdp) :: width
  real(kind=vdp) :: dp
  logical :: pers
 end type vismo_camera

!Binary Swap
 type vismo_bswap
   logical :: possible, bswap
   integer :: numstage
   integer :: myrank
   integer :: psize
   integer, dimension(3) :: myindex
   integer, dimension(3) :: ndv
   integer, dimension(:,:,:), allocatable :: index2rank
   integer, dimension(:), allocatable :: partner
   logical, dimension(:), allocatable :: above
 end type vismo_bswap

! tube
 type vismo_tubedata
#ifndef FLOATDATA
  real(kind=vdp), dimension(:, :), allocatable:: vertex
#else
  real, dimension(:, :), allocatable :: vertex
#endif
  integer, dimension(:), allocatable :: connect
  logical, dimension(vsmMaxVisKind) :: vison
  real(kind=vdp) :: radius
  integer :: nvertex, nconnect
  logical :: tubeon
end type vismo_tubedata


 end module vsmConst

