module vsmCoordinates
 use vsmConst
 use vsmFunction
 use vsmParam

 implicit none
 private

 type(vismo_coord), save, public :: vsmCoord
 type(vismo_localcoord), save, public :: vsmLocalCoord

 public vsmRayDist, vsmCheckLCoord, vsmIsInsidelWO
 public vsmIsInsidel
 public vsmRayDistInside
 public vsmFinalizeCoord, vsmFindCoord, vsmInitCoord, vsmPrepareCoord
 public vsmSetlrect, vsmSetLUniform, vsmSetRect, vsmSetUniform
 public vsmIsInsided
 contains

 subroutine vsmInitCoord(uni)
  logical, intent(in) :: uni
  integer :: i

  vsmCoord%uniform = uni
  vsmLocalCoord%uniform = uni
  do i=1, 3
    vsmCoord%nx(i)     = 0
    vsmCoord%origin(i) = 0.0_vdp
    vsmCoord%cend(i)   = 0.0_vdp
    vsmCoord%dx(i)     = 0.0_vdp

    vsmLocalCoord%nx(i)      = 0
    vsmLocalCoord%lx0(i)     = 0
    vsmLocalCoord%lnx(i)     = 0
    vsmLocalCoord%lorigin(i) = 0.0_vdp
    vsmLocalCoord%lcend(i)   = 0.0_vdp
    vsmLocalCoord%origin(i)  = 0.0_vdp
    vsmLocalCoord%cend(i)    = 0.0_vdp
    vsmLocalCoord%dx(i)      = 0.0_vdp
  end do
  vsmCoord%wdvertex = vsmFarAway
  vsmCoord%lvnum = -1
 end subroutine vsmInitCoord

 subroutine vsmPrepareCoord
! nothing
 end subroutine vsmPrepareCoord

 subroutine vsmSetUniform(nx, origin, dx)
  integer, dimension(3), intent(in) :: nx
  real(kind=vdp), dimension(3), intent(in) :: origin
  real(kind=vdp), dimension(3), intent(in) :: dx
  integer :: i

  do i=1, 3
    vsmCoord%nx(i) = nx(i)
    vsmCoord%origin(i) = origin(i)
    vsmCoord%dx(i) = dx(i)
    vsmCoord%cend(i) = origin(i)+(nx(i)-1)*dx(i)
  end do

  if(vsmDbg) then
    call vsmLogPrint("origin:", vsmCoord%origin)
    call vsmLogPrint("dx:", vsmCoord%dx)
    call vsmLogPrint("end:", vsmCoord%cend)
    call vsmLogPrint("nx:", vsmCoord%nx)
  end if

  call vsmCalcIntvalU(dx)
 end subroutine vsmSetUniform

 subroutine vsmSetLUniform(nx, lx0, lnx, origin, dx)
  integer, dimension(3), intent(in) :: nx, lx0, lnx
  real(kind=vdp), dimension(3), intent(in) :: origin
  real(kind=vdp), dimension(3), intent(in) :: dx
  integer :: i

  do i=1, 3
    vsmLocalCoord%nx(i)  = nx(i)
    vsmLocalCoord%lx0(i) = lx0(i)
    vsmLocalCoord%lnx(i) = lnx(i)
    vsmLocalCoord%dx(i)  = dx(i)
    vsmLocalCoord%lorigin(i) = origin(i)+(lx0(i)-1)*dx(i) !without overlap
    vsmLocalCoord%lcend(i)   = vsmLocalCoord%lorigin(i)+(lnx(i)-1)*dx(i)!without overlap
    vsmLocalCoord%origin(i)  = origin(i) !with overlap
    vsmLocalCoord%cend(i)    = vsmLocalCoord%origin(i)+(nx(i)-1)*dx(i) !with overlap
  end do

  if(vsmDbg) then
    call vsmLogPrint("local origin:", vsmLocalCoord%lorigin)
    call vsmLogPrint("local end:", vsmLocalCoord%lcend)
    call vsmLogPrint("local origin WO:", vsmLocalCoord%origin)
    call vsmLogPrint("local end WO:", vsmLocalCoord%cend)
  end if
 end subroutine vsmSetLUniform

 subroutine vsmSetRect(nx, coordx, coordy, coordz)
  integer, dimension(3), intent(in) :: nx
  real(kind=vdp), dimension(nx(1)), intent(in) :: coordx
  real(kind=vdp), dimension(nx(2)), intent(in) :: coordy
  real(kind=vdp), dimension(nx(3)), intent(in) :: coordz
  integer :: allocstate

  vsmCoord%nx(1:3) = nx(1:3)
  vsmCoord%origin(1) = coordx(1)
  vsmCoord%origin(2) = coordy(1)
  vsmCoord%origin(3) = coordz(1)
  vsmCoord%cend(1) = coordx(nx(1))
  vsmCoord%cend(2) = coordy(nx(2))
  vsmCoord%cend(3) = coordz(nx(3))

  allocate(vsmCoord%coordx(nx(1)), vsmCoord%coordy(nx(2)), &
             vsmCoord%coordz(nx(3)), stat=allocstate)
  if(allocstate /= 0) call vsmError('cant allocate coordinates')

  vsmCoord%coordx(1:nx(1)) = coordx(1:nx(1))
  vsmCoord%coordy(1:nx(2)) = coordy(1:nx(2))
  vsmCoord%coordz(1:nx(3)) = coordz(1:nx(3))
  call vsmCalcIntvalR(coordx, coordy, coordz, nx)
 end subroutine vsmSetRect

 subroutine vsmSetLRect(nx, lx0, lnx, coordx, coordy, coordz)
  integer, dimension(3), intent(in) :: nx, lx0, lnx
  real(kind=vdp), dimension(nx(1)), intent(in) :: coordx
  real(kind=vdp), dimension(nx(2)), intent(in) :: coordy
  real(kind=vdp), dimension(nx(3)), intent(in) :: coordz
  integer :: allocstate, i

  do i=1, 3
    vsmLocalCoord%nx(i) = nx(i)
    vsmLocalCoord%lx0(i) = lx0(i)
    vsmLocalCoord%lnx(i) = lnx(i)
  end do

  vsmLocalCoord%lorigin(1) = coordx(lx0(1))
  vsmLocalCoord%lorigin(2) = coordy(lx0(2))
  vsmLocalCoord%lorigin(3) = coordz(lx0(3))
  vsmLocalCoord%lcend(1)   = coordx(lx0(1)+lnx(1)-1)
  vsmLocalCoord%lcend(2)   = coordy(lx0(2)+lnx(2)-1)
  vsmLocalCoord%lcend(3)   = coordz(lx0(3)+lnx(3)-1)
  vsmLocalCoord%origin(1)  = coordx(1) !with overlap
  vsmLocalCoord%origin(2)  = coordy(1) !with overlap
  vsmLocalCoord%origin(3)  = coordz(1) !with overlap
  vsmLocalCoord%cend(1)    = coordx(nx(1)) !with overlap
  vsmLocalCoord%cend(2)    = coordy(nx(2)) !with overlap
  vsmLocalCoord%cend(3)    = coordz(nx(3)) !with overlap

  allocate(vsmLocalCoord%coordx(nx(1)), vsmLocalCoord%coordy(nx(2)), &
             vsmLocalCoord%coordz(nx(3)), stat=allocstate)
  if(allocstate /= 0) call vsmError('cant allocate coordinates')

  vsmLocalCoord%coordx(1:nx(1)) = coordx(1:nx(1))
  vsmLocalCoord%coordy(1:nx(2)) = coordy(1:nx(2))
  vsmLocalCoord%coordz(1:nx(3)) = coordz(1:nx(3))
 end subroutine vsmSetLRect

 subroutine vsmFindCoord(x, ratio, coordpos, loopn)
  real(kind=vdp), dimension(3), intent(in) :: x
  real(kind=vdp), dimension(3), intent(inout) :: ratio
  integer, dimension(3), intent(inout) :: coordpos
  integer, optional :: loopn

  if(vsmCoord%uniform) then
    if(present(loopn)) then
       call vsmFindUnicoord(x, ratio, coordpos, loopn)
    else 
       call vsmFindUnicoord(x, ratio, coordpos)
    end if
  else
    call vsmFindReccoord(x, ratio, coordpos)
  end if
! print *,"",x,ratio,coordpos
 end subroutine vsmFindCoord

 subroutine vsmFindUnicoord(x, ratio, coordpos, lnum)
  real(kind=vdp), dimension(3), intent(in) :: x
  real(kind=vdp), dimension(3), intent(inout) :: ratio
  integer, dimension(3), intent(inout) :: coordpos
  integer, optional :: lnum
  real(kind=vdp), dimension(3) :: rx

  if(x(1) < vsmLocalCoord%origin(1) .or. x(1) > vsmLocalCoord%cend(1)) then
    print *,"x origin end ", x(1),vsmLocalCoord%origin(1),vsmLocalCoord%cend(1)
    print *,"x lorigin end ", x(1),vsmLocalCoord%lorigin(1),vsmLocalCoord%lcend(1)
    if(present(lnum)) print *,"loop=", lnum
    call vsmError('out of local data x')
  end if

  if(x(2) < vsmLocalCoord%origin(2) .or. x(2) > vsmLocalCoord%cend(2)) then
    print *,"y origin end ", x(2),vsmLocalCoord%origin(2),vsmLocalCoord%cend(2)
    print *,"y lorigin end ", x(2),vsmLocalCoord%lorigin(2),vsmLocalCoord%lcend(2)
    if(present(lnum)) print *,"loop=", lnum
    call vsmError('out of local data y')
  end if

  if(x(3) < vsmLocalCoord%origin(3) .or. x(3) > vsmLocalCoord%cend(3)) then
    print *,"z origin end ", x(3),vsmLocalCoord%origin(3),vsmLocalCoord%cend(3)
    print *,"z lorigin end ", x(3),vsmLocalCoord%lorigin(3),vsmLocalCoord%lcend(3)
    if(present(lnum)) print *,"loop=", lnum
    call vsmError('out of local data z')
  end if

  rx(1:3) = (x(1:3) - vsmLocalCoord%origin(1:3))/vsmLocalCoord%dx(1:3)
  coordpos(1:3) = aint(rx(1:3)) + 1

  if(.not.vsmCheckLCoord(coordpos(1),coordpos(2),coordpos(3))) then 
      call vsmLogPrint("local coord:", x)
      print *,x
      call vsmLogPrint("local coord int:", coordpos)
      print *,coordpos
      call vsmError("Error :: local coord is strange :: vsmFindUnicoord")
  end if

  if(coordpos(1) == vsmLocalCoord%nx(1)) then
    coordpos(1) = vsmLocalCoord%nx(1)-1
    ratio(1) = 1.0_vdp
  else
    ratio(1) = rx(1) - aint(rx(1))
  end if

  if(coordpos(2) == vsmLocalCoord%nx(2)) then
    coordpos(2) = vsmLocalCoord%nx(2)-1
    ratio(2) = 1.0_vdp
  else
    ratio(2) = rx(2) - aint(rx(2))
  end if

  if(coordpos(3) == vsmLocalCoord%nx(3)) then
    coordpos(3) = vsmLocalCoord%nx(3)-1
    ratio(3) = 1.0_vdp
  else
    ratio(3) = rx(3) - aint(rx(3))
  end if

!  if(coordpos(1) == vsmLocalCoord%nx(1) .or. coordpos(2) == vsmLocalCoord%nx(2) &
!  	.or. coordpos(3) == vsmLocalCoord%nx(3)) then
!    call vsmLogPrint("x:",x(1:3))
!    call vsmLogPrint("ratio:",ratio(1:3))
!    call vsmLogPrint("local coord int:", coordpos)
!    call vsmError("Error :: local coord is strange :: vsmFindUnicoord 2")
!  end if
 end subroutine vsmFindUnicoord

 subroutine vsmFindReccoord(x, ratio, coordpos)
  !binary search
  real(kind=vdp), dimension(3), intent(in) :: x
  real(kind=vdp), dimension(3), intent(inout) :: ratio
  integer, dimension(3), intent(inout) :: coordpos
  integer, dimension(3) :: nx
  integer :: lower, upper, middle
  logical :: found

  nx(1:3) = vsmLocalCoord%nx(1:3)
  if(x(1) < vsmLocalCoord%coordx(1) .or. x(1) > vsmLocalCoord%coordx(nx(1))) then
    call vsmError('out of local data x')
  end if
  found = .false.

  if(x(1)==vsmLocalCoord%coordx(nx(1))) then
    coordpos(1) = nx(1)-1
    found = .true.
    ratio(1) = 1.0_vdp
  else
    lower = 1
    upper = nx(1)
    do
      if (upper- lower < 0) exit
      middle = (lower+upper)/2
      if(x(1) < vsmLocalCoord%coordx(middle)) then
        upper = middle
      else if(x(1) >= vsmLocalCoord%coordx(middle+1)) then
        lower = middle
      else
        found = .true.
        coordpos(1) = middle
        exit
      end if
    end do
    ratio(1) = (x(1) - vsmLocalCoord%coordx(coordpos(1)))/(vsmLocalCoord%coordx(coordpos(1)+1) &
      - vsmLocalCoord%coordx(coordpos(1)))
  end if

  if(.not.found) call vsmError('cant find coord x')

  if(x(2) < vsmLocalCoord%coordy(1) .or. x(2) > vsmLocalCoord%coordy(nx(2))) then
    call vsmError('out of local data y')
  end if
  found = .false.
  if(x(2)==vsmLocalCoord%coordy(nx(2))) then
    coordpos(2) = nx(2)-1
    found = .true.
    ratio(2) = 1.0_vdp
  else
    lower = 1
    upper = nx(2)
    do
      if (upper- lower < 0) exit
      middle = (lower+upper)/2
      if(x(2) < vsmLocalCoord%coordy(middle)) then
        upper = middle
      else if(x(2) >= vsmLocalCoord%coordy(middle+1)) then
        lower = middle
      else
        found = .true.
        coordpos(2) = middle
        exit
      end if
    end do
    ratio(2) = (x(2) - vsmLocalCoord%coordy(coordpos(2)))/(vsmLocalCoord%coordy(coordpos(2)+1) &
      - vsmLocalCoord%coordy(coordpos(2)))
  end if

  if(.not.found) call vsmError('cant find coord y')

  if(x(3) < vsmLocalCoord%coordz(1) .or. x(3) > vsmLocalCoord%coordz(nx(3))) then
    print *,"x(3)", x(3), vsmLocalCoord%coordz(1), vsmLocalCoord%coordz(nx(3))
    call vsmError('out of local data z')
  end if
  found = .false.
  if(x(3)==vsmLocalCoord%coordz(nx(3))) then
    coordpos(3) = nx(3) - 1
    found = .true.
    ratio(3) = 1.0_vdp
  else
    lower = 1
    upper = nx(3)
    do
      if (upper- lower < 0) exit
      middle = (lower+upper)/2
      if(x(3) < vsmLocalCoord%coordz(middle)) then
        upper = middle
      else if(x(3) >= vsmLocalCoord%coordz(middle+1)) then
        lower = middle
      else
        found = .true.
        coordpos(3) = middle
        exit
      end if
!    print *,"middle = ",x(3),nx(3),middle,vsmLocalCoord%coordz(middle),vsmLocalCoord%coordz(middle+1)
    end do
    ratio(3) = (x(3) - vsmLocalCoord%coordz(coordpos(3)))/(vsmLocalCoord%coordz(coordpos(3)+1) &
       - vsmLocalCoord%coordz(coordpos(3)))
  end if
  if(.not.found) call vsmError('cant find coord z')

  if(.not.vsmCheckLCoord(coordpos(1),coordpos(2),coordpos(3))) then
    call vsmError("Error :: localcoord is strange:vsmFindReccoord")
  end if

 end subroutine vsmFindReccoord

 function vsmIsInside(pos, origin, cend)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp), dimension(3), intent(in) :: origin, cend
   logical :: vsmIsInside

   if(pos(1) >= origin(1) .and. pos(1) < cend(1) .and. &
      pos(2) >= origin(2) .and. pos(2) < cend(2) .and. &
      pos(3) >= origin(3) .and. pos(3) < cend(3)) then
     vsmIsInside = .true.
   else
     vsmIsInside = .false.
   end if
 end function vsmIsInside

 function vsmIsInsided(pos)
   real(kind=vdp), dimension(3), intent(in) :: pos
   logical :: vsmIsInsided

   if(pos(1) >= vsmCoord%origin(1) .and. pos(1) < vsmCoord%cend(1) .and. &
      pos(2) >= vsmCoord%origin(2) .and. pos(2) < vsmCoord%cend(2) .and. &
      pos(3) >= vsmCoord%origin(3) .and. pos(3) < vsmCoord%cend(3)) then
     vsmIsInsided = .true.
   else
     vsmIsInsided = .false.
   end if
 end function vsmIsInsided

 function vsmIsInsidel(pos)
   real(kind=vdp), dimension(3), intent(in) :: pos
   logical :: vsmIsInsidel

   if(pos(1) >= vsmLocalCoord%lorigin(1) .and. pos(1) < vsmLocalCoord%lcend(1) .and. &
      pos(2) >= vsmLocalCoord%lorigin(2) .and. pos(2) < vsmLocalCoord%lcend(2) .and. &
      pos(3) >= vsmLocalCoord%lorigin(3) .and. pos(3) < vsmLocalCoord%lcend(3)) then
     vsmIsInsidel = .true.
   else
     vsmIsInsidel = .false.
   end if
 end function vsmIsInsidel

 function vsmIsInsidelWO(pos) ! With Overlap
   real(kind=vdp), dimension(3), intent(in) :: pos
   logical :: vsmIsInsidelWO

   if(pos(1) >= vsmLocalCoord%origin(1) .and. pos(1) < vsmLocalCoord%cend(1) .and. &
      pos(2) >= vsmLocalCoord%origin(2) .and. pos(2) < vsmLocalCoord%cend(2) .and. &
      pos(3) >= vsmLocalCoord%origin(3) .and. pos(3) < vsmLocalCoord%cend(3)) then
     vsmIsInsidelWO = .true.
   else
     vsmIsInsidelWO = .false.
   end if
 end function vsmIsInsidelWO

 function vsmIsInsidelWOE(pos) ! With Overlap and Equal
   real(kind=vdp), dimension(3), intent(in) :: pos
   logical :: vsmIsInsidelWOE

   if(pos(1) >= vsmLocalCoord%origin(1) .and. pos(1) <= vsmLocalCoord%cend(1) .and. &
      pos(2) >= vsmLocalCoord%origin(2) .and. pos(2) <= vsmLocalCoord%cend(2) .and. &
      pos(3) >= vsmLocalCoord%origin(3) .and. pos(3) <= vsmLocalCoord%cend(3)) then
     vsmIsInsidelWOE = .true.
   else
     vsmIsInsidelWOE = .false.
   end if
 end function vsmIsInsidelWOE

 subroutine vsmFinalizeCoord
  integer :: allocstate
  if(vsmCoord%uniform) return
  deallocate(vsmCoord%coordx, vsmCoord%coordy, vsmCoord%coordz, stat=allocstate)
  deallocate(vsmLocalCoord%coordx, vsmLocalCoord%coordy, vsmLocalCoord%coordz, stat=allocstate)
 end subroutine vsmFinalizeCoord

 function vsmCheckLCoord(i, j, k)
   integer, intent(in) :: i, j, k
   logical :: vsmCheckLCoord

   if(i<1 .or. i>vsmLocalCoord%nx(1)) then
     vsmCheckLCoord = .false.
     call vsmLogPrint("local x:", i)
     return
   end if

   if(j<1 .or. j>vsmLocalCoord%nx(2)) then
     vsmCheckLCoord = .false.
     call vsmLogPrint("local y:", j)
     return
   end if

   if(k<1 .or. k>vsmLocalCoord%nx(3)) then
     vsmCheckLCoord = .false.
     call vsmLogPrint("local z:", k)
     return
   end if

   vsmCheckLCoord = .true.
 end function vsmCheckLCoord

 function vsmIsXz(posx, posz, origin, cend)
   real(kind=vdp), intent(in) :: posx, posz
   real(kind=vdp), dimension(3), intent(in) :: origin, cend
   logical :: vsmIsXz

   if(posx >= origin(1) .and. posx <= cend(1) .and. &
      posz >= origin(3) .and. posz <= cend(3)) then
     vsmIsXz = .true.
   else
     vsmIsXz = .false.
   end if
 end function vsmIsXz

 function vsmIsXy(posx, posy, origin, cend)
   real(kind=vdp), intent(in) :: posx, posy
   real(kind=vdp), dimension(3), intent(in) :: origin, cend
   logical :: vsmIsXy

   if(posx >= origin(1) .and. posx <= cend(1) .and. &
      posy >= origin(2) .and. posy <= cend(2)) then
     vsmIsXy = .true.
   else
     vsmIsXy = .false.
   end if
 end function vsmIsXy

 function vsmIsYz(posy, posz, origin, cend)
   real(kind=vdp), intent(in) :: posy, posz
   real(kind=vdp), dimension(3), intent(in) :: origin, cend
   logical :: vsmIsYz

   if(posy >= origin(2) .and. posy <= cend(2) .and. &
      posz >= origin(3) .and. posz <= cend(3)) then
     vsmIsYz = .true.
   else
     vsmIsYz = .false.
   end if
 end function vsmIsYz

 function vsmRayDist(pos, norm, origin, cend)
   real(kind=vdp), dimension(3), intent(in) :: pos, norm, origin, cend
   real(kind=vdp), dimension(6) :: ray
   real(kind=vdp) :: vsmRayDist, p1, p2

   if(vsmIsInside (pos, origin, cend)) then
      vsmRayDist = 0.0_vdp
   else

      if(abs(norm(1)) /= 0.0_vdp) then
        ray(1) = (-pos(1)+origin(1))/norm(1)
        ray(2) = (-pos(1)+cend(1))/norm(1)
        p1 = pos(2)+ray(1)*norm(2); p2 = pos(3)+ray(1)*norm(3)
        if(.not.vsmIsYz(p1, p2, origin, cend)) ray(1) = vsmFaraway
        p1 = pos(2)+ray(2)*norm(2); p2 = pos(3)+ray(2)*norm(3) 
        if(.not.vsmIsYz(p1, p2, origin, cend)) ray(2) = vsmFaraway
      else
        ray(1:2) = vsmFaraway
      end if

      if(abs(norm(2)) /= 0.0_vdp) then
        ray(3) = (-pos(2)+origin(2))/norm(2)
        ray(4) = (-pos(2)+cend(2))/norm(2)
        p1 = pos(1)+ray(3)*norm(1); p2 = pos(3)+ray(3)*norm(3)
        if(.not.vsmIsXz(p1, p2, origin, cend)) ray(3) = vsmFaraway
        p1 = pos(1)+ray(4)*norm(1); p2 = pos(3)+ray(4)*norm(3)
        if(.not.vsmIsXz(p1, p2, origin, cend)) ray(4) = vsmFaraway
      else
        ray(3:4) = vsmFaraway
      end if

      if(abs(norm(3)) /= 0.0_vdp) then
        ray(5) = (-pos(3)+origin(3))/norm(3)
        ray(6) = (-pos(3)+cend(3))/norm(3)
        p1 = pos(1)+ray(5)*norm(1); p2 = pos(2)+ray(5)*norm(2)
        if(.not.vsmIsXy(p1, p2, origin, cend)) ray(5) = vsmFaraway
        p1 = pos(1)+ray(6)*norm(1) ; p2 = pos(2)+ray(6)*norm(2)
        if(.not.vsmIsXy(p1, p2, origin, cend)) ray(6) = vsmFaraway
      else
        ray(5:6) = vsmFaraway
      end if

      where(ray < 0.0_vdp)
        ray = vsmFaraway
      end where

      vsmRayDist = minval(ray)
      if(vsmRayDist < 0.0_vdp) call vsmLogPrint("ray_dist is negative!")

   end if

 end function vsmRayDist

 function vsmRayDistInside(pos, norm, origin, cend)
   real(kind=vdp), dimension(3), intent(in) :: pos, norm, origin, cend
   real(kind=vdp), dimension(6) :: ray
   real(kind=vdp) :: vsmRayDistInside, p1, p2

   if(.not.vsmIsInside (pos, origin, cend)) then
      call vsmLogPrint("pos ", pos)
      call vsmLogPrint("origin ", origin)
      call vsmLogPrint("cend ", cend)
      call vsmError("pos is not inside data")
   end if

   if(abs(norm(1)) /= 0.0_vdp) then
     ray(1) = (-pos(1)+origin(1))/norm(1)
     ray(2) = (-pos(1)+cend(1))/norm(1)
     p1 = pos(2)+ray(1)*norm(2); p2 = pos(3)+ray(1)*norm(3)
     if(.not.vsmIsYz(p1, p2, origin, cend)) ray(1) = vsmFaraway
     p1 = pos(2)+ray(2)*norm(2); p2 = pos(3)+ray(2)*norm(3) 
     if(.not.vsmIsYz(p1, p2, origin, cend)) ray(2) = vsmFaraway
   else
     ray(1:2) = vsmFaraway
   end if

   if(abs(norm(2)) /= 0.0_vdp) then
     ray(3) = (-pos(2)+origin(2))/norm(2)
     ray(4) = (-pos(2)+cend(2))/norm(2)
     p1 = pos(1)+ray(3)*norm(1); p2 = pos(3)+ray(3)*norm(3)
     if(.not.vsmIsXz(p1, p2, origin, cend)) ray(3) = vsmFaraway
     p1 = pos(1)+ray(4)*norm(1); p2 = pos(3)+ray(4)*norm(3)
     if(.not.vsmIsXz(p1, p2, origin, cend)) ray(4) = vsmFaraway
   else
     ray(3:4) = vsmFaraway
   end if

   if(abs(norm(3)) /= 0.0_vdp) then
     ray(5) = (-pos(3)+origin(3))/norm(3)
     ray(6) = (-pos(3)+cend(3))/norm(3)
     p1 = pos(1)+ray(5)*norm(1); p2 = pos(2)+ray(5)*norm(2)
     if(.not.vsmIsXy(p1, p2, origin, cend)) ray(5) = vsmFaraway
     p1 = pos(1)+ray(6)*norm(1) ; p2 = pos(2)+ray(6)*norm(2)
     if(.not.vsmIsXy(p1, p2, origin, cend)) ray(6) = vsmFaraway
   else
     ray(5:6) = vsmFaraway
   end if

   where(ray < 0.0_vdp)
     ray = vsmFaraway
   end where

   vsmRayDistInside = minval(ray)
   if(vsmRayDistInside < 0.0_vdp) call vsmLogPrint("ray_dist(inside) is negative!")

 end function vsmRayDistInside

end module vsmCoordinates

