module vsmData
 use vsmConst
 use vsmFunction
 use vsmColormap
 implicit none
 private

 type(vismo_scaldata), dimension(vsmMaxScalar), save, public :: vsmScalData
 type(vismo_vectdata), dimension(vsmMaxVector), save, public :: vsmVectData
 type(vismo_ptdata), dimension(vsmMaxPtcl), save, public :: vsmPtcData
 type(vismo_datainfo), save, public :: vsmDataInfo

 public vsmAddPtclData, vsmAddScalData, vsmAddVectData, vsmInitData, vsmSetScalMinMax, vsmSetVectMinMax
 public vsmAddPTCLDataCol, vsmSetDataMinMax

 contains

 subroutine vsmInitData
  vsmDataInfo%NumOfScal = 0
  vsmDataInfo%NumOfVect = 0
  vsmDataInfo%NumOfPtcl = 0
 end subroutine vsmInitData

 subroutine vsmSetDataMinMax
   integer :: i
   if(vsmDataInfo%NumOfScal > 0) then
     do i=1, vsmDataInfo%NumOfScal
       if(vsmSColormap(i)%set) then
         vsmScalData%minvalue = vsmSColormap(i)%value(1)
         vsmScalData%maxvalue = vsmSColormap(i)%value(vsmSColormap(i)%colnum)
       end if
     end do
   end if

   if(vsmDataInfo%NumOfVect > 0) then
     do i=1, vsmDataInfo%NumOfVect
       if(vsmVColormap(i)%set) then
         vsmVectData%minvalue = vsmVColormap(i)%value(1)
         vsmVectData%maxvalue = vsmVColormap(i)%value(vsmVColormap(i)%colnum)
       end if
     end do
   end if
 end subroutine vsmSetDataMinMax

 subroutine vsmAddScalData(scalar, nx, ny, nz)
   integer, intent(in) :: nx, ny, nz
#ifndef FLOATDATA
   real(kind=vdp), dimension(nx,ny,nz), target, intent(in) :: scalar
#else
   real, dimension(nx,ny,nz), target, intent(in) :: scalar
#endif
   integer :: NumOfScal

   NumOfScal = vsmDataInfo%NumOfScal + 1
   if(NumOfScal > vsmMaxScalar) call vsmError('too many scalar data')
   vsmScalData(NumOfScal)%scal => scalar
   vsmDataInfo%NumOfScal = NumOfScal
 end subroutine vsmAddScalData

 subroutine vsmAddVectData(vect_x, vect_y, vect_z, nx, ny, nz)
   integer, intent(in) :: nx, ny, nz
#ifndef FLOATDATA
   real(kind=vdp), dimension(nx,ny,nz), target, intent(in) :: vect_x, vect_y, vect_z
#else
   real, dimension(nx,ny,nz), target, intent(in) :: vect_x, vect_y, vect_z
#endif
   integer :: NumOfVect

   NumOfVect = vsmDataInfo%NumOfVect + 1
   if(NumOfVect > vsmMaxVector) call vsmError('too many vector data')
   vsmVectData(NumOfVect)%vect_x => vect_x
   vsmVectData(NumOfVect)%vect_y => vect_y
   vsmVectData(NumOfVect)%vect_z => vect_z
   vsmDataInfo%NumOfVect = NumOfVect
 end subroutine vsmAddVectData

 subroutine vsmAddPtclData(p_x, p_y, p_z, q, n)
   integer, intent(in) :: n
#ifndef FLOATDATA
   real(kind=vdp), dimension(n), target, intent(in) :: p_x, p_y, p_z
#else
   real, dimension(n), target, intent(in) :: p_x, p_y, p_z
#endif
   integer, dimension(n), target, intent(in) :: q
   integer :: NumOfPtcl

   if(n > vsmPCParticleC) call vsmError("too much particles")
   NumOfPtcl = vsmDataInfo%NumOfPtcl + 1
   if(NumOfPtcl > vsmMaxPtcl) call vsmError('too many particle data')
   vsmPtcData(NumOfPtcl)%x => p_x
   vsmPtcData(NumOfPtcl)%y => p_y
   vsmPtcData(NumOfPtcl)%z => p_z
   vsmPtcData(NumOfPtcl)%q => q
   vsmPtcData(NumOfPtcl)%pNum = n
   vsmDataInfo%NumOfPtcl = NumOfPtcl
 end subroutine vsmAddPtclData

 subroutine vsmAddPtclDataCol(p_x, p_y, p_z, q, v, n)
   integer, intent(in) :: n
#ifndef FLOATDATA
   real(kind=vdp), dimension(n), target, intent(in) :: p_x, p_y, p_z, v
#else
   real, dimension(n), target, intent(in) :: p_x, p_y, p_z, v
#endif
   integer, dimension(n), target, intent(in) :: q
   integer :: NumOfPtcl

   if(n > vsmPCParticleC) call vsmError("too much particles")
   NumOfPtcl = vsmDataInfo%NumOfPtcl + 1
   if(NumOfPtcl > vsmMaxPtcl) call vsmError('too many particle data')
   vsmPtcData(NumOfPtcl)%x => p_x
   vsmPtcData(NumOfPtcl)%y => p_y
   vsmPtcData(NumOfPtcl)%z => p_z
   vsmPtcData(NumOfPtcl)%scal => v
   vsmPtcData(NumOfPtcl)%q => q
   vsmPtcData(NumOfPtcl)%pNum = n
   vsmDataInfo%NumOfPtcl = NumOfPtcl
 end subroutine vsmAddPtclDataCol

 subroutine vsmSetScalMinMax(nscal, minvalue, maxvalue)
   integer, intent(in) :: nscal
   real(kind=vdp), intent(in) :: minvalue, maxvalue

   if(nscal > vsmMaxScalar .or. nscal < 1) call vsmError('set scalminmax ')
   vsmScalData(nscal)%minvalue = minvalue
   vsmScalData(nscal)%maxvalue = maxvalue
 end subroutine vsmSetScalMinMax

 subroutine vsmSetVectMinMax(nvec, minvalue, maxvalue)
   integer, intent(in) :: nvec
   real(kind=vdp), intent(in) :: minvalue, maxvalue

   if(nvec > vsmMaxVector .or. nvec < 1) call vsmError('set vectminmax')
   vsmVectData(nvec)%minvalue = minvalue
   vsmVectData(nvec)%maxvalue = maxvalue
 end subroutine vsmSetVectMinMax

end module vsmData

