module vsmFrame
 use vsmConst
 use vsmFunction
 use vsmCameraData
 use vsmColor
 use vsmCoordinates
! use vsmDepthQ
 use vsmSolidDrawer
 use vsmImage
 use vsmParam

 implicit none
 private

 logical, save :: bon ! blend on
 integer, save :: framewidth
 public vsmDrawFrame, vsmInitFrame, vsmSetFrameWidth, vsmGetFrameWidth
 contains

 subroutine vsmInitFrame
   framewidth = 1
 end subroutine vsmInitFrame
 
 subroutine vsmGetFrameWidth(width)
   integer, intent(out) :: width
   width = framewidth
 end subroutine vsmGetFrameWidth
 
 subroutine vsmSetFrameWidth(width)
   integer, intent(in) :: width
   if(width <= 0) then
      call vsmLogPrint("frame width should be positive ", width)
      return
   end if
   framewidth = width
 end subroutine vsmSetFrameWidth

 function vsmCheckFPix(pix)
   integer, dimension(2), intent(in) :: pix
   logical :: vsmCheckFPix

   if(pix(1) < 1 .or. pix(1) > vsmImageData%width .or. &
        pix(2) < 1 .or. pix(2) > vsmImageData%height) then
     vsmCheckFPix = .false.
   else
     vsmCheckFPix = .true.
   end if
 end function vsmCheckFPix

 subroutine vsmDrawFrame(vis_num, cam_num, blend)
   integer, intent(in) :: vis_num, cam_num
   logical, intent(in) :: blend
   real(kind=vdp), dimension(3, 8) :: pos
   real(kind=vdp) :: dx
   integer :: n

   bon = blend
   dx = vsmIntval%ray_interval
   pos(1:3, 1) = vsmCoord%origin(1:3) - dx
   pos(1:3, 8) = vsmCoord%cend(1:3) + dx

   pos(1,2) = vsmCoord%cend(1) + dx; pos(2,2) = pos(2, 1); pos(3,2) = pos(3, 1)
   pos(1,3) = pos(1, 1); pos(2,3) = vsmCoord%cend(2)+dx; pos(3,3) = pos(3, 1)
   pos(1,4) = vsmCoord%cend(1) + dx; pos(2,4) = vsmCoord%cend(2) + dx; pos(3,4) = pos(3, 1)
   pos(1:2, 5) = pos(1:2, 1); pos(3, 5) = vsmCoord%cend(3)+dx
   pos(1,6) = vsmCoord%cend(1)+dx; pos(2,6) = pos(2, 1); pos(3,6) = vsmCoord%cend(3)+dx
   pos(1,7) = pos(1, 1); pos(2,7) = vsmCoord%cend(2)+dx; pos(3,7) = vsmCoord%cend(3)+dx

   ! pos1 -> pos2
   n=1
   call vsmDrawLine(pos(:, 1), pos(:, 2), cam_num, vis_num, n)

   ! pos1 -> pos3
   n=2
   call vsmDrawLine(pos(:, 1), pos(:, 3), cam_num, vis_num, n)

   ! pos2 -> pos4
   n=3
   call vsmDrawLine(pos(:, 2), pos(:, 4), cam_num, vis_num, n)

   ! pos3 -> pos4
   n=4
   call vsmDrawLine(pos(:, 3), pos(:, 4), cam_num, vis_num, n)

   ! pos5 -> pos6
   n=5
   call vsmDrawLine(pos(:, 5), pos(:, 6), cam_num, vis_num, n)

   ! pos5 -> pos7
   n=6
   call vsmDrawLine(pos(:, 5), pos(:, 7), cam_num, vis_num, n)

   ! pos6 -> pos8
   n=7
   call vsmDrawLine(pos(:, 6), pos(:, 8), cam_num, vis_num, n)

   ! pos7 -> pos8
   n=8
   call vsmDrawLine(pos(:, 7), pos(:, 8), cam_num, vis_num, n)

   ! pos1 -> pos5
   n=9
   call vsmDrawLine(pos(:, 1), pos(:, 5), cam_num, vis_num, n)

   ! pos2 -> pos6
   n=10
   call vsmDrawLine(pos(:, 2), pos(:, 6), cam_num, vis_num, n)

   ! pos3 -> pos7
   n=11
   call vsmDrawLine(pos(:, 3), pos(:, 7), cam_num, vis_num, n)

   ! pos4 -> pos8
   n=12
   call vsmDrawLine(pos(:, 4), pos(:, 8), cam_num, vis_num, n)

 end subroutine vsmDrawFrame

 subroutine vsmDrawLine(pos1, pos2, cam_num, vis_num, lineNum)
   real(kind=vdp), dimension(3), intent(in) :: pos1, pos2
   integer, intent(in) :: cam_num, vis_num, lineNum
   logical :: cross1, cross2
   integer, dimension(2) :: pix1, pix2

   ! pos1 -> pos2
   call vsmPProjection4f(pos1, cam_num, cross1, pix1)
   call vsmPProjection4f(pos2, cam_num, cross2, pix2)
   if(cross1 .and. cross2) then
     call vsmConnectSolid(pix1, pix2, lineNum, pos1, cam_num, vis_num)
   else if(.not.cross1 .and. cross2) then
     call vsmLProjection4f(pos1, pos2, cam_num, cross1, pix1)
     if(.not.cross1) call vsmError("frame drawing 1")
     call vsmConnectSolid(pix1, pix2, lineNum, pos2, cam_num, vis_num)
   else if(cross1 .and. .not.cross2) then
     call vsmLProjection4f(pos2, pos1, cam_num, cross2, pix2)
     if(.not.cross2) call vsmError("frame drawing 2")
     call vsmConnectSolid(pix1, pix2, lineNum, pos1, cam_num, vis_num)
   end if

 end subroutine vsmDrawLine

 subroutine vsmConnectSolid(pix1, pix2, lineNum, pos, cam_num, vis_num)
   integer, dimension(2), intent(in) :: pix1, pix2
   real(kind=vdp), dimension(3), intent(in) :: pos
   real, dimension(3) :: color
   integer, intent(in) :: lineNum, cam_num, vis_num
   integer :: xs, xb, ys, yb ! small, big
   integer :: x, y, fs, bpx
   integer, dimension(2) :: pix, ppx
   integer :: i
   integer :: comm, myrank, psize, actpsize
   integer :: nl, st, ed, chnk
   real(kind=vdp) :: A
   logical :: draw
   real(kind=vdp) :: distance
   
   call vsmGetMPIVars(comm, myrank, psize)
!   print *,"frame ", comm, myrank, psize

   fs = framewidth; bpx = (fs-1)/2
   color = vsmFrameColor

   if(fs == 1) then
   if(pix1(1) == pix2(1)) then
     ys = min(pix1(2), pix2(2)); yb = max(pix1(2), pix2(2))
     pix(1) = pix1(1)

     nl = yb-ys+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl
!$omp parallel do private(i,distance, draw) firstprivate(pix)
     do i=st, ed
!     do i=1, yb-ys+1
       pix(2) = ys+i-1
       if(.not.vsmCheckFPix(pix)) cycle
       call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)

       if(draw) then
         call vsmPaintPixel3(pix, color, distance)
       end if
     end do

   else if(pix1(2) == pix2(2)) then

     xs = min(pix1(1), pix2(1)); xb = max(pix1(1), pix2(1))
     pix(2) = pix1(2)

     nl = xb-xs+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl
!$omp parallel do private(i, distance, draw) firstprivate(pix)
     do i=st, ed
!     do i=1, xb-xs+1
       pix(1) = xs+i-1
       if(.not.vsmCheckFPix(pix)) cycle
       call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)
       if(draw) then
         call vsmPaintPixel3(pix, color, distance)
       end if
     end do

   else 

     A = dble((pix2(2)-pix1(2)))/dble((pix2(1)-pix1(1)))

     if(abs(A) <= 1.0_vdp) then

       if(pix1(1) < pix2(1)) then
         xs = pix1(1); ys = pix1(2)
         xb = pix2(1); yb = pix2(2)
       else
         xs = pix2(1); ys = pix2(2)
         xb = pix1(1); yb = pix1(2)
       end if

     nl = xb-xs+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl
!$omp parallel do private(i, pix, distance, draw)
       do i=st, ed
!       do i=1, xb-xs+1
         pix(1) = xs+i-1
         pix(2) = nint(A*(pix(1)-xs) + ys)
         if(.not.vsmCheckFPix(pix)) cycle
         call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)
         if(draw) then
         call vsmPaintPixel3(pix, color, distance)
         end if
       end do

     else

       if(pix1(2) < pix2(2)) then
         xs = pix1(1); ys = pix1(2)
         xb = pix2(1); yb = pix2(2)
       else
         xs = pix2(1); ys = pix2(2)
         xb = pix1(1); yb = pix1(2)
       end if

     nl = yb-ys+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl
!$omp parallel do private(i, pix, distance, draw)
       do i=st, ed
!       do i=1, yb-ys+1
         pix(2) = ys+i-1
         pix(1) = nint((pix(2)-ys)/A + xs)
         if(.not.vsmCheckFPix(pix)) cycle
         call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)
         if(draw) then
         call vsmPaintPixel3(pix, color, distance)
         end if
       end do

     end if
   end if

   else ! Frame Size
   fs = fs - 1
   if(pix1(1) == pix2(1)) then
     ys = min(pix1(2), pix2(2)); yb = max(pix1(2), pix2(2))
     pix(1) = pix1(1)

     nl = yb-ys+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl

     do i=st, ed
!     do i=1, yb-ys+1
       pix(2) = ys+i-1
       if(.not.vsmCheckFPix(pix)) cycle
       call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)

       if(draw) then
         do y=pix(2) - bpx, pix(2)-bpx + fs
           do x = pix(1) - bpx, pix(1) - bpx + fs
             ppx(1) = x; ppx(2) = y
             if(.not.vsmCheckFPix(ppx)) cycle
             call vsmPaintPixel3(ppx, color, distance)
           end do
         end do
       end if
     end do

   else if(pix1(2) == pix2(2)) then

     xs = min(pix1(1), pix2(1)); xb = max(pix1(1), pix2(1))
     pix(2) = pix1(2)

     nl = xb-xs+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl

     do i=st, ed
!     do i=1, xb-xs+1
       pix(1) = xs+i-1
       if(.not.vsmCheckFPix(pix)) cycle
       call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)
       if(draw) then
         do y=pix(2) - bpx, pix(2)-bpx + fs
           do x = pix(1) - bpx, pix(1) - bpx + fs
             ppx(1) = x; ppx(2) = y
             if(.not.vsmCheckFPix(ppx)) cycle
             call vsmPaintPixel3(ppx, color, distance)
           end do
         end do
       end if
     end do

   else 

     A = dble((pix2(2)-pix1(2)))/dble((pix2(1)-pix1(1)))

     if(abs(A) <= 1.0_vdp) then

       if(pix1(1) < pix2(1)) then
         xs = pix1(1); ys = pix1(2)
         xb = pix2(1); yb = pix2(2)
       else
         xs = pix2(1); ys = pix2(2)
         xb = pix1(1); yb = pix1(2)
       end if

     nl = xb-xs+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl

       do i=st, ed
!       do i=1, xb-xs+1
         pix(1) = xs+i-1
         pix(2) = nint(A*(pix(1)-xs) + ys)
         if(.not.vsmCheckFPix(pix)) cycle
         call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)
         if(draw) then
           do y=pix(2) - bpx, pix(2)-bpx + fs
             do x = pix(1) - bpx, pix(1) - bpx + fs
               ppx(1) = x; ppx(2) = y
               if(.not.vsmCheckFPix(ppx)) cycle
               call vsmPaintPixel3(ppx, color, distance)
             end do
           end do
         end if
       end do

     else

       if(pix1(2) < pix2(2)) then
         xs = pix1(1); ys = pix1(2)
         xb = pix2(1); yb = pix2(2)
       else
         xs = pix2(1); ys = pix2(2)
         xb = pix1(1); yb = pix1(2)
       end if

     nl = yb-ys+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl

       do i=st, ed
!       do i=1, yb-ys+1
         pix(2) = ys+i-1
         pix(1) = nint((pix(2)-ys)/A + xs)
         if(.not.vsmCheckFPix(pix)) cycle
         call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)
         if(draw) then
           do y=pix(2) - bpx, pix(2)-bpx + fs
             do x = pix(1) - bpx, pix(1) - bpx + fs
               ppx(1) = x; ppx(2) = y
               if(.not.vsmCheckFPix(ppx)) cycle
               call vsmPaintPixel3(ppx, color, distance)
             end do
           end do
         end if
       end do

     end if
   end if

   end if

 end subroutine vsmConnectSolid

 subroutine vsmConnect(pix1, pix2, lineNum, pos, cam_num, vis_num)
   integer, dimension(2), intent(in) :: pix1, pix2
   real(kind=vdp), dimension(3), intent(in) :: pos
   real, dimension(3) :: color
   integer, intent(in) :: lineNum, cam_num, vis_num
   integer :: xs, xb, ys, yb ! small, big
   integer :: x, y, fs, bpx
   integer, dimension(2) :: pix, ppx
   integer :: i
   integer :: comm, myrank, psize, actpsize
   integer :: nl, st, ed, chnk
   real(kind=vdp) :: A
   logical :: draw
   real(kind=vdp) :: distance
   
   call vsmGetMPIVars(comm, myrank, psize)
!   print *,"frame ", comm, myrank, psize

   fs = framewidth; bpx = (fs-1)/2
   color = vsmFrameColor

   if(fs == 1) then
   if(pix1(1) == pix2(1)) then
     ys = min(pix1(2), pix2(2)); yb = max(pix1(2), pix2(2))
     pix(1) = pix1(1)

     nl = yb-ys+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl
!$omp parallel do private(i,distance, draw) firstprivate(pix)
     do i=st, ed
!     do i=1, yb-ys+1
       pix(2) = ys+i-1
       if(.not.vsmCheckFPix(pix)) cycle
       call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)

       if(draw) then
         if(bon) then
           call vsmPaintPixelBlend(pix, color, distance)
         else
           call vsmPaintPixel3(pix, color, distance)
         end if
       end if
     end do

   else if(pix1(2) == pix2(2)) then

     xs = min(pix1(1), pix2(1)); xb = max(pix1(1), pix2(1))
     pix(2) = pix1(2)

     nl = xb-xs+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl
!$omp parallel do private(i,distance, draw) firstprivate(pix)
     do i=st, ed
!     do i=1, xb-xs+1
       pix(1) = xs+i-1
       if(.not.vsmCheckFPix(pix)) cycle
       call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)
       if(draw) then
         if(bon) then
           call vsmPaintPixelBlend(pix, color, distance)
         else
           call vsmPaintPixel3(pix, color, distance)
         end if
       end if
     end do

   else 

     A = dble((pix2(2)-pix1(2)))/dble((pix2(1)-pix1(1)))

     if(abs(A) <= 1.0_vdp) then

       if(pix1(1) < pix2(1)) then
         xs = pix1(1); ys = pix1(2)
         xb = pix2(1); yb = pix2(2)
       else
         xs = pix2(1); ys = pix2(2)
         xb = pix1(1); yb = pix1(2)
       end if

     nl = xb-xs+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl
!$omp parallel do private(i,pix, distance, draw)
       do i=st, ed
!       do i=1, xb-xs+1
         pix(1) = xs+i-1
         pix(2) = nint(A*(pix(1)-xs) + ys)
         if(.not.vsmCheckFPix(pix)) cycle
         call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)
         if(draw) then
         if(bon) then
           call vsmPaintPixelBlend(pix, color, distance)
         else
           call vsmPaintPixel3(pix, color, distance)
         end if
         end if
       end do

     else

       if(pix1(2) < pix2(2)) then
         xs = pix1(1); ys = pix1(2)
         xb = pix2(1); yb = pix2(2)
       else
         xs = pix2(1); ys = pix2(2)
         xb = pix1(1); yb = pix1(2)
       end if

     nl = yb-ys+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl
!$omp parallel do private(i,pix, distance, draw)
       do i=st, ed
!       do i=1, yb-ys+1
         pix(2) = ys+i-1
         pix(1) = nint((pix(2)-ys)/A + xs)
         if(.not.vsmCheckFPix(pix)) cycle
         call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)
         if(draw) then
         if(bon) then
           call vsmPaintPixelBlend(pix, color, distance)
         else
           call vsmPaintPixel3(pix, color, distance)
         end if
         end if
       end do

     end if
   end if

   else ! Frame Size
   fs = fs - 1
   if(pix1(1) == pix2(1)) then
     ys = min(pix1(2), pix2(2)); yb = max(pix1(2), pix2(2))
     pix(1) = pix1(1)

     nl = yb-ys+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl

     do i=st, ed
!     do i=1, yb-ys+1
       pix(2) = ys+i-1
       if(.not.vsmCheckFPix(pix)) cycle
       call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)

       if(draw) then
         if(bon) then
           do y=pix(2) - bpx, pix(2)-bpx + fs
             do x = pix(1) - bpx, pix(1) - bpx + fs
               ppx(1) = x; ppx(2) = y
               if(.not.vsmCheckFPix(ppx)) cycle
               call vsmPaintPixelBlend(ppx, color, distance)
             end do
           end do
         else
           do y=pix(2) - bpx, pix(2)-bpx + fs
             do x = pix(1) - bpx, pix(1) - bpx + fs
               ppx(1) = x; ppx(2) = y
               if(.not.vsmCheckFPix(ppx)) cycle
               call vsmPaintPixel3(ppx, color, distance)
             end do
           end do
         end if
       end if
     end do

   else if(pix1(2) == pix2(2)) then

     xs = min(pix1(1), pix2(1)); xb = max(pix1(1), pix2(1))
     pix(2) = pix1(2)

     nl = xb-xs+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl

     do i=st, ed
!     do i=1, xb-xs+1
       pix(1) = xs+i-1
       if(.not.vsmCheckFPix(pix)) cycle
       call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)
       if(draw) then
         if(bon) then
           do y=pix(2) - bpx, pix(2)-bpx + fs
             do x = pix(1) - bpx, pix(1) - bpx + fs
               ppx(1) = x; ppx(2) = y
               if(.not.vsmCheckFPix(ppx)) cycle
               call vsmPaintPixelBlend(ppx, color, distance)
             end do
           end do
         else
           do y=pix(2) - bpx, pix(2)-bpx + fs
             do x = pix(1) - bpx, pix(1) - bpx + fs
               ppx(1) = x; ppx(2) = y
               if(.not.vsmCheckFPix(ppx)) cycle
               call vsmPaintPixel3(ppx, color, distance)
             end do
           end do
         end if
       end if
     end do

   else 

     A = dble((pix2(2)-pix1(2)))/dble((pix2(1)-pix1(1)))

     if(abs(A) <= 1.0_vdp) then

       if(pix1(1) < pix2(1)) then
         xs = pix1(1); ys = pix1(2)
         xb = pix2(1); yb = pix2(2)
       else
         xs = pix2(1); ys = pix2(2)
         xb = pix1(1); yb = pix1(2)
       end if

     nl = xb-xs+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl

       do i=st, ed
!       do i=1, xb-xs+1
         pix(1) = xs+i-1
         pix(2) = nint(A*(pix(1)-xs) + ys)
         if(.not.vsmCheckFPix(pix)) cycle
         call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)
         if(draw) then
           if(bon) then
           do y=pix(2) - bpx, pix(2)-bpx + fs
             do x = pix(1) - bpx, pix(1) - bpx + fs
               ppx(1) = x; ppx(2) = y
               if(.not.vsmCheckFPix(ppx)) cycle
               call vsmPaintPixelBlend(ppx, color, distance)
             end do
           end do
           else
           do y=pix(2) - bpx, pix(2)-bpx + fs
             do x = pix(1) - bpx, pix(1) - bpx + fs
               ppx(1) = x; ppx(2) = y
               if(.not.vsmCheckFPix(ppx)) cycle
               call vsmPaintPixel3(ppx, color, distance)
             end do
           end do
           end if
         end if
       end do

     else

       if(pix1(2) < pix2(2)) then
         xs = pix1(1); ys = pix1(2)
         xb = pix2(1); yb = pix2(2)
       else
         xs = pix2(1); ys = pix2(2)
         xb = pix1(1); yb = pix1(2)
       end if

     nl = yb-ys+1
     if(nl<psize) then
       actpsize = nl
     else
       actpsize = psize
     end if
     chnk = nl/actpsize
     if(myrank < actpsize) then
       st = chnk*myrank + 1
       ed = chnk*(myrank+1)
     else
       st = 1
       ed = 0
     end if
     if(myrank == actpsize-1) ed = nl

       do i=st, ed
!       do i=1, yb-ys+1
         pix(2) = ys+i-1
         pix(1) = nint((pix(2)-ys)/A + xs)
         if(.not.vsmCheckFPix(pix)) cycle
         call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)
         if(draw) then
           if(bon) then
           do y=pix(2) - bpx, pix(2)-bpx + fs
             do x = pix(1) - bpx, pix(1) - bpx + fs
               ppx(1) = x; ppx(2) = y
               if(.not.vsmCheckFPix(ppx)) cycle
               call vsmPaintPixelBlend(ppx, color, distance)
             end do
           end do
           else
           do y=pix(2) - bpx, pix(2)-bpx + fs
             do x = pix(1) - bpx, pix(1) - bpx + fs
               ppx(1) = x; ppx(2) = y
               if(.not.vsmCheckFPix(ppx)) cycle
               call vsmPaintPixel3(ppx, color, distance)
             end do
           end do
           end if
         end if
       end do

     end if
   end if

   end if

 end subroutine vsmConnect

 subroutine vsmMeasureDist(pix, num, pos, cam_num, dist, draw)
   real(kind=vdp), dimension(3), intent(in) :: pos
   integer, dimension(2), intent(in) :: pix
   integer, intent(in) :: num, cam_num
   real(kind=vdp), intent(out) :: dist
   logical, intent(out) :: draw
   real(kind=vdp), dimension(3) :: epos, ppos, direc

   dist = vsmFarAway
   draw = .false.

   if(num < 1 .or. num > 12) call vsmError("frame num is bad")

   ppos(1:3) = vsmImageData%topleft(1:3) + vsmImageData%dp*((pix(1)-1)*vsmCamera(cam_num)%right(1:3) &
                    - (pix(2)-1)*vsmCamera(cam_num)%up(1:3))
   if(vsmCamera(cam_num)%pers) then
   !perspective
     epos(1:3) = vsmCamera(cam_num)%pos(1:3)

     if(num==1 .or. num==5 .or. num==4 .or. num==8) then
       direc(1:3) = ppos(1:3) - epos(1:3)
       call vsmNormVec(direc)
       if(direc(2) /= 0.0_vdp) then
         dist = (pos(2) - epos(2))/ direc(2)
         draw = .true.
       else if(direc(3) /= 0.0_vdp) then
         dist = (pos(3) - epos(3))/ direc(3)
         draw = .true.
       else
         draw = .false.
       end if
       return
     end if

     if(num==2 .or. num==3 .or. num==6 .or. num==7) then
       direc(1:3) = ppos(1:3) - epos(1:3)
       call vsmNormVec(direc)
       if(direc(1) /= 0.0_vdp) then
         dist = (pos(1) - epos(1))/ direc(1)
         draw = .true.
       else if(direc(3) /= 0.0_vdp) then
         dist = (pos(3) - epos(3))/ direc(3)
         draw = .true.
       else
         draw = .false.
       end if
       return
     end if

     if(num==9 .or. num==10 .or. num==11 .or. num==12) then
       direc(1:3) = ppos(1:3) - epos(1:3)
       call vsmNormVec(direc)
       if(direc(1) /= 0.0_vdp) then
         dist = (pos(1) - epos(1))/ direc(1)
         draw = .true.
       else if(direc(2) /= 0.0_vdp) then
         dist = (pos(2) - epos(2))/ direc(2)
         draw = .true.
       else
         draw = .false.
       end if
       return
     end if
   else
   ! parallel
     direc(1:3) = vsmCamera(cam_num)%front(1:3)
   
     if(num==1 .or. num==5 .or. num==4 .or. num==8) then
       if(direc(2) /= 0.0_vdp .or. direc(3) /= 0.0_vdp) then
         dist = dot_product(direc, pos-ppos) + vsmCamera(cam_num)%near
         if(dist >= 0.0_vdp) then
            draw = .true.
         else
            draw = .false.
         end if
       else
         draw = .false.
       end if
       return
     end if

     if(num==2 .or. num==3 .or. num==6 .or. num==7) then
       if(direc(1) /= 0.0_vdp .or. direc(3) /= 0.0_vdp) then
         dist = dot_product(direc, pos-ppos) + vsmCamera(cam_num)%near
         if(dist >= 0.0_vdp) then
            draw = .true.
         else
            draw = .false.
         end if
       else
         draw = .false.
       end if
       return
     end if

     if(num==9 .or. num==10 .or. num==11 .or. num==12) then
       if(direc(1) /= 0.0_vdp .or. direc(2) /= 0.0_vdp) then
         dist = dot_product(direc, pos-ppos) + vsmCamera(cam_num)%near
         if(dist >= 0.0_vdp) then
            draw = .true.
         else
            draw = .false.
         end if
       else
         draw = .false.
       end if
       return
     end if
   end if

 end subroutine vsmMeasureDist

 subroutine vsmPProjection4f(pos, cam_num, cross, pix)
   real(kind=vdp), dimension(3), intent(in) :: pos
   integer, intent(in) :: cam_num
   logical, intent(out) :: cross
   integer, dimension(2), intent(out) :: pix

   real(kind=vdp), dimension(3) :: eyepos
   real(kind=vdp), dimension(3) :: norm, plane_c
   real(kind=vdp), dimension(3) :: direc
   real(kind=vdp) :: A, C, B, ip1, ip2
   real(kind=vdp), dimension(3) :: cross_pos
   real(kind=vdp), dimension(3) :: img_topleft, img_right, img_down

   norm(1:3) = vsmCamera(cam_num)%front(1:3)
   eyepos(1:3) = vsmCamera(cam_num)%pos(1:3)

   plane_c(1:3) = eyepos(1:3) + vsmCamera(cam_num)%near * norm(1:3)

   if(vsmCamera(cam_num)%pers) then
     !Perspective Projection
     direc(1:3) = pos(1:3) - eyepos(1:3)
     call vsmNormVec(direc)

     C = dot_product(norm, plane_c - eyepos)
     B = dot_product(norm, direc)

     if(B == 0.0_vdp) then
       cross = .false.
       return
     end if

     A = C/B

     if(A < 0.0_vdp) then
       cross = .false.
       return
     end if

     cross_pos(1:3) = A*direc(1:3) + eyepos(1:3)

   else
     !Parallel Projection
     direc(1:3) = -norm(1:3)

     C = dot_product(norm, plane_c - pos)
     !B = -1.0_vdp

     A = -C

     if(A < 0.0_vdp) then
       cross = .false.
       return
     end if

     cross_pos(1:3) = A*direc(1:3) + pos(1:3)
   end if

   img_topleft(1:3) = vsmImageData%topleft(1:3)
   img_right(1:3)   = vsmCamera(cam_num)%right(1:3)
   img_down(1:3)    = -vsmCamera(cam_num)%up(1:3)

   direc(1:3) = cross_pos(1:3) - img_topleft(1:3)

   ip1 = dot_product(direc, img_right)
   ip2 = dot_product(direc, img_down)

   pix(1) = nint(ip1/vsmImageData%dp)
   pix(2) = nint(ip2/vsmImageData%dp)

   cross = .true.
 end subroutine vsmPProjection4f

 subroutine vsmLProjection4f(pos1, pos2, cam_num, cross, pix)
   ! pos1 is below the plane and pos2 is above the plane
   real(kind=vdp), dimension(3), intent(in) :: pos1, pos2
   integer, intent(in) :: cam_num
   logical, intent(out) :: cross
   integer, dimension(2), intent(out) :: pix

   real(kind=vdp), dimension(3) :: eyepos
   real(kind=vdp), dimension(3) :: norm, plane_c
   real(kind=vdp), dimension(3) :: direc
   real(kind=vdp) :: A, C, B, ip1, ip2
   real(kind=vdp), dimension(3) :: cross_pos
   real(kind=vdp), dimension(3) :: img_topleft, img_right, img_down

   norm(1:3) = vsmCamera(cam_num)%front(1:3)
   eyepos(1:3) = vsmCamera(cam_num)%pos(1:3)

   plane_c(1:3) = eyepos(1:3) + vsmCamera(cam_num)%near * norm(1:3)

   !Perspective Projection & Orthogonal
   direc(1:3) = pos2(1:3) - pos1(1:3)
   call vsmNormVec(direc)

   C = dot_product(norm, plane_c - pos1)
   B = dot_product(norm, direc)

   if(B == 0.0_vdp) then
     cross = .false.
     return
   end if

   A = C/B

   if(A < 0.0_vdp) then
     cross = .false.
     return
   end if

   cross_pos(1:3) = A*direc(1:3) + pos1(1:3)

   img_topleft(1:3) = vsmImageData%topleft(1:3)
   img_right(1:3)   = vsmCamera(cam_num)%right(1:3)
   img_down(1:3)    = -vsmCamera(cam_num)%up(1:3)

   direc(1:3) = cross_pos(1:3) - img_topleft(1:3)

   ip1 = dot_product(direc, img_right)
   ip2 = dot_product(direc, img_down)

   pix(1) = nint(ip1/vsmImageData%dp)
   pix(2) = nint(ip2/vsmImageData%dp)

   cross = .true.
 end subroutine vsmLProjection4f


 end module vsmFrame

