module vsmIntpol
 use vsmConst
 use vsmFunction
 use vsmData
 use vsmCoordinates
 implicit none
 private

 public vsmGetScalVal, vsmGetIntpolVectd, vsmGetIntpolVect, vsmGetVectVal, vsmSetFrac, vsmSetScalGridVal, vsmSetVectGridVal
 contains

 subroutine vsmSetFrac(dx, vx)
   real(kind=vdp), dimension(3), intent(in) :: dx
   real(kind=vdp), dimension(8), intent(out) :: vx
   real(kind=vdp), dimension(3) :: one_minus_dx

   one_minus_dx(1:3) = 1.0_vdp - dx(1:3)

   vx(1) = one_minus_dx(1) * one_minus_dx(2) * one_minus_dx(3)
   vx(2) =           dx(1) * one_minus_dx(2) * one_minus_dx(3)
   vx(3) = one_minus_dx(1) *           dx(2) * one_minus_dx(3)
   vx(4) =           dx(1) *           dx(2) * one_minus_dx(3)
   vx(5) = one_minus_dx(1) * one_minus_dx(2) *           dx(3)
   vx(6) =           dx(1) * one_minus_dx(2) *           dx(3)
   vx(7) = one_minus_dx(1) *           dx(2) *           dx(3)
   vx(8) =           dx(1) *           dx(2) *           dx(3)

 end subroutine vsmSetFrac

 function vsmGetScalVal(vx, grid_val)
   real(kind=vdp) :: vsmGetScalVal
   real(kind=vdp), dimension(8), intent(in) :: vx
   real(kind=vdp), dimension(8), intent(in) :: grid_val

   vsmGetScalVal = dot_product(vx, grid_val)

 end function vsmGetScalVal

 subroutine vsmGetVectVal(vx, grid_val, vectval, normalize)
   real(kind=vdp), dimension(8), intent(in) :: vx
   real(kind=vdp), dimension(3, 8), intent(in) :: grid_val
   real(kind=vdp), dimension(3), intent(out) :: vectval
   logical, optional, intent(in) :: normalize

   integer :: i

   do i=1, 3
     vectval(i) = dot_product(vx, grid_val(i, 1:8))
   end do

   if(present(normalize)) then
     call vsmNormVec(vectval)
   end if
 end subroutine vsmGetVectVal

 subroutine vsmSetScalGridVal(nscal, ci, gridval)
   integer, intent(in) :: nscal
   integer, dimension(3), intent(in) :: ci
   real(kind=vdp), dimension(8), intent(out) :: gridval
   integer :: c1,c1p1,c2,c2p1,c3,c3p1

   c1=ci(1); c1p1=c1+1; c2=ci(2);c2p1=c2+1; c3=ci(3);c3p1=c3+1

   if(.not.vsmCheckLCoord(ci(1), ci(2), ci(3))) call vsmError("Error: vsmSetScalGridVal ci")
   if(.not.vsmCheckLCoord(c1p1, c2p1, c3p1)) call vsmError("Error: vsmSetScalGridVal cxp1")

   gridval(1) = vsmScalData(nscal)%scal(c1,   c2,   c3)
   gridval(2) = vsmScalData(nscal)%scal(c1p1, c2,   c3)
   gridval(3) = vsmScalData(nscal)%scal(c1,   c2p1, c3)
   gridval(4) = vsmScalData(nscal)%scal(c1p1, c2p1, c3)
   gridval(5) = vsmScalData(nscal)%scal(c1,   c2,   c3p1)
   gridval(6) = vsmScalData(nscal)%scal(c1p1, c2,   c3p1)
   gridval(7) = vsmScalData(nscal)%scal(c1,   c2p1, c3p1)
   gridval(8) = vsmScalData(nscal)%scal(c1p1, c2p1, c3p1)

 end subroutine vsmSetScalGridVal

 subroutine vsmSetVectGridVal(nvec, ci, gridval)
   integer, intent(in) :: nvec
   integer, dimension(3), intent(in) :: ci
   real(kind=vdp), dimension(3, 8), intent(out) :: gridval
   integer :: c1,c1p1,c2,c2p1,c3,c3p1

   c1=ci(1); c1p1=c1+1; c2=ci(2);c2p1=c2+1; c3=ci(3);c3p1=c3+1

   if(.not.vsmCheckLCoord(ci(1), ci(2), ci(3))) call vsmError("Error: vsmSetVectGridVal ci")
   if(.not.vsmCheckLCoord(c1p1, c2p1, c3p1)) call vsmError("Error: vsmSetVectGridVal p1")

   gridval(1, 1) = vsmVectData(nvec)%vect_x(c1,   c2,   c3)
   gridval(1, 2) = vsmVectData(nvec)%vect_x(c1p1, c2,   c3)
   gridval(1, 3) = vsmVectData(nvec)%vect_x(c1,   c2p1, c3)
   gridval(1, 4) = vsmVectData(nvec)%vect_x(c1p1, c2p1, c3)
   gridval(1, 5) = vsmVectData(nvec)%vect_x(c1,   c2,   c3p1)
   gridval(1, 6) = vsmVectData(nvec)%vect_x(c1p1, c2,   c3p1)
   gridval(1, 7) = vsmVectData(nvec)%vect_x(c1,   c2p1, c3p1)
   gridval(1, 8) = vsmVectData(nvec)%vect_x(c1p1, c2p1, c3p1)

   gridval(2, 1) = vsmVectData(nvec)%vect_y(c1,   c2,   c3)
   gridval(2, 2) = vsmVectData(nvec)%vect_y(c1p1, c2,   c3)
   gridval(2, 3) = vsmVectData(nvec)%vect_y(c1,   c2p1, c3)
   gridval(2, 4) = vsmVectData(nvec)%vect_y(c1p1, c2p1, c3)
   gridval(2, 5) = vsmVectData(nvec)%vect_y(c1,   c2,   c3p1)
   gridval(2, 6) = vsmVectData(nvec)%vect_y(c1p1, c2,   c3p1)
   gridval(2, 7) = vsmVectData(nvec)%vect_y(c1,   c2p1, c3p1)
   gridval(2, 8) = vsmVectData(nvec)%vect_y(c1p1, c2p1, c3p1)

   gridval(3, 1) = vsmVectData(nvec)%vect_z(c1,   c2,   c3)
   gridval(3, 2) = vsmVectData(nvec)%vect_z(c1p1, c2,   c3)
   gridval(3, 3) = vsmVectData(nvec)%vect_z(c1,   c2p1, c3)
   gridval(3, 4) = vsmVectData(nvec)%vect_z(c1p1, c2p1, c3)
   gridval(3, 5) = vsmVectData(nvec)%vect_z(c1,   c2,   c3p1)
   gridval(3, 6) = vsmVectData(nvec)%vect_z(c1p1, c2,   c3p1)
   gridval(3, 7) = vsmVectData(nvec)%vect_z(c1,   c2p1, c3p1)
   gridval(3, 8) = vsmVectData(nvec)%vect_z(c1p1, c2p1, c3p1)

 end subroutine vsmSetVectGridVal

 subroutine vsmGetIntpolVectd(norm, vx, intpol_norm)
   real(kind=vdp), dimension(3, 8), intent(in) :: norm
   real(kind=vdp), dimension(8), intent(in) :: vx
   real(kind=vdp), dimension(3), intent(out) :: intpol_norm
   integer :: i

   do i=1, 3
     intpol_norm(i) = dot_product(vx(1:8), norm(i, 1:8))
   end do
 end subroutine vsmGetIntpolVectd

 subroutine vsmGetIntpolVect(norm, vx, intpol_norm)
   real(kind=vdp), dimension(3, 8), intent(in) :: norm
   real(kind=vdp), dimension(8), intent(in) :: vx
   real, dimension(3), intent(out) :: intpol_norm
   integer :: i

   do i=1, 3
     intpol_norm(i) = real(dot_product(vx, norm(i, 1:8)))
   end do
 end subroutine vsmGetIntpolVect

end module vsmIntpol

