module vsmLightData
 use vsmConst
 use vsmFunction

 implicit none
 private

 type(vismo_light), dimension(vsmMaxLight), save, public :: vsmLight
 integer, save, public :: vsmLightNum
 
 public vsmAddLight, vsmInitLight, vsmCalcColor, vsmCalcColorOrtho, vsmCalcColorv
 public vsmCalcColorvOrtho

 contains

 subroutine vsmInitLight
  integer :: i
  vsmLight(1:vsmMaxLight)%para = .false.
  do i=1, vsmMaxLight
    vsmLight(i)%direc(1:3) = 0.0
    vsmLight(i)%pos(1:3) = vsmFarAway
    vsmLight(i)%amb(1:3)   = 0.0
    vsmLight(i)%dif(1:3)   = 0.0
    vsmLight(i)%spec(1:3)  = 0.0
  end do
  vsmLightNum = 0
 end subroutine vsmInitLight

 ! this subroutine needs to be called before setting a light
 subroutine vsmAddLight(p, pos, amb, dif, spec)
  logical, intent(in) :: p
  real, dimension(3), intent(in) :: pos
  real, dimension(3), intent(in) :: amb, dif, spec

  vsmLightNum = vsmLightNum + 1
  if(vsmLightNum > vsmMaxLight) call vsmError('set light: exceed max light')
  vsmLight(vsmLightNum)%para = p
  vsmLight(vsmLightNum)%pos(1:3) = pos(1:3)
  if(vsmLight(vsmLightNum)%para) then
    call vsmSetLightDirec()
  end if

  call vsmCheckFA01(amb); call vsmCheckFA01(dif); call vsmCheckFA01(spec); 
  vsmLight(vsmLightNum)%amb(1:3) = amb(1:3)
  vsmLight(vsmLightNum)%dif(1:3) = dif(1:3)
  vsmLight(vsmLightNum)%spec(1:3) = spec(1:3)
 end subroutine vsmAddLight

 subroutine vsmSetLightDirec()
  vsmLight(vsmLightNum)%direc(1:3) = vsmLight(vsmLightNum)%pos(1:3)
  call vsmNormVecF(vsmLight(vsmLightNum)%direc)
 end subroutine vsmSetLightDirec

 subroutine vsmCalcColorv(mat_amb, mat_dif, mat_spec, mat_shin, eye, pos, norm, used, color, transdat)
  real, dimension(3), intent(in) :: mat_amb, mat_dif, mat_spec
  real, intent(in) :: mat_shin
  real(kind=vdp), dimension(3), intent(in) :: eye ! camera pos
  real(kind=vdp), dimension(3), intent(in) :: pos
  real, dimension(3), intent(in) :: norm
  logical, dimension(vsmMaxLight), intent(in) :: used
  real, dimension(3), intent(out) :: color
  type(vismo_transdata), optional, intent(in) :: transdat

  real, dimension(3) :: eye_direc, light_direc, el_direc
  real, dimension(vsmMaxLight) :: ln, sn
  integer :: i

  !call vsmLogPrint("vsmCalcColor")
  color(1:3) = 0.0
  eye_direc(1:3) = real(eye(1:3) - pos(1:3))
  call vsmNormVecF(eye_direc)

  do i=1, vsmLightNum
    if(.not.used(i)) cycle
    !call vsmLogPrint("color calc light num =", i)
    if(vsmLight(i)%para) then
      if(.not.present(transdat)) then
        light_direc(1:3) = vsmLight(i)%direc(1:3)
      else
        call vsmMatrixCrossVectf(transdat%matrix, vsmLight(i)%direc, light_direc)
      end if
    else
      if(.not.present(transdat)) then
        light_direc(1:3) = vsmLight(i)%pos(1:3) - pos(1:3)
      else
        call vsmMatrixCrossVectf(transdat%matrix, vsmLight(i)%pos(1:3) - real(pos(1:3)), light_direc)
      end if
      call vsmNormVecF(light_direc)
    end if
    !call vsmLogPrint("light direc", light_direc)
    ln(i) = dot_product(norm, light_direc)
    !call vsmLogPrint("ln", ln(i))
    if(ln(i) <= 0.0) then
      ln(i) = 0.0
      sn(i) = 0.0
    else
      el_direc(1:3) = light_direc(1:3) + eye_direc(1:3)
      call vsmNormVecF(el_direc)
      sn(i) = dot_product(norm, el_direc)
      sn(i) = max(sn(i), 0.0)
    end if
  end do

  do i=1, vsmLightNum
    if(.not.used(i)) cycle ! 2013/8/30
    if(sn(i)>0.0) then
      if(log10(sn(i))*mat_shin<-5.0) then
        sn(i) = 0.0
      else
        sn(i) = sn(i)**mat_shin
      end if
    end if
    color(1:3) = color(1:3) + &
                 mat_amb(1:3)*vsmLight(i)%amb(1:3) + &
                 mat_dif(1:3)*vsmLight(i)%dif(1:3)*ln(i) + &
                 mat_spec(1:3)*vsmLight(i)%spec(1:3)*sn(i)
  end do
  !call vsmLogPrint("vsmCalcColor ends")
 end subroutine vsmCalcColorv

 subroutine vsmCalcColorvOrtho(mat_amb, mat_dif, mat_spec, mat_shin, dir, pos, norm, used, color, transdat)
  real, dimension(3), intent(in) :: mat_amb, mat_dif, mat_spec
  real, intent(in) :: mat_shin
  real(kind=vdp), dimension(3), intent(in) :: dir
  real(kind=vdp), dimension(3), intent(in) :: pos
  real, dimension(3), intent(in) :: norm
  logical, dimension(vsmMaxLight), intent(in) :: used
  real, dimension(3), intent(out) :: color
  type(vismo_transdata), optional, intent(in) :: transdat

  real, dimension(3) :: eye_direc, light_direc, el_direc, direc
  real, dimension(vsmMaxLight) :: ln, sn
  integer :: i

  color(1:3) = 0.0
  eye_direc(1:3) = -dir(1:3)
  call vsmNormVecF(eye_direc)

  do i=1, vsmLightNum
    if(.not.used(i)) cycle
    if(vsmLight(i)%para) then
      if(.not.present(transdat)) then
        light_direc(1:3) = vsmLight(i)%direc(1:3)
      else
        call vsmMatrixCrossVectf(transdat%matrix, vsmLight(i)%direc, direc)
        light_direc(1:3) = direc(1:3)
      end if
    else
      if(.not.present(transdat)) then
        light_direc(1:3) = vsmLight(i)%pos(1:3) - pos(1:3)
      else
        direc(1:3) = vsmLight(i)%pos(1:3) - real(pos(1:3))
        call vsmMatrixCrossVectf(transdat%matrix, direc, light_direc)
      end if
      call vsmNormVecF(light_direc)
    end if
    ln(i) = dot_product(norm, light_direc)
    if(ln(i) <= 0.0) then
      ln(i) = 0.0
      sn(i) = 0.0
    else
      el_direc(1:3) = light_direc(1:3) + eye_direc(1:3)
      call vsmNormVecF(el_direc)
      sn(i) = dot_product(norm, el_direc)
      sn(i) = max(sn(i), 0.0)
    end if
  end do

  do i=1, vsmLightNum
    if(.not.used(i)) cycle !2013/8/30
    if(sn(i)>0.0) then
      if(log10(sn(i))*mat_shin<-5.0) then
        sn(i) = 0.0
      else
        sn(i) = sn(i)**mat_shin
      end if
    end if
    color(1:3) = color(1:3) + mat_amb(1:3)*vsmLight(i)%amb(1:3) + &
                 mat_dif(1:3)*vsmLight(i)%dif(1:3)*ln(i) + &
                 mat_spec(1:3)*vsmLight(i)%spec(1:3)*sn(i)
  end do
 end subroutine vsmCalcColorvOrtho

 subroutine vsmCalcColor(mat_amb, mat_dif, mat_spec, mat_shin, eye, pos, norm, used, color, transdat)
  real, intent(in) :: mat_amb, mat_dif, mat_spec, mat_shin
  real(kind=vdp), dimension(3), intent(in) :: eye ! camera pos
  real(kind=vdp), dimension(3), intent(in) :: pos
  real, dimension(3), intent(in) :: norm
  logical, dimension(vsmMaxLight), intent(in) :: used
  real, dimension(3), intent(out) :: color
  type(vismo_transdata), optional, intent(in) :: transdat

  real, dimension(3) :: eye_direc, light_direc, el_direc, direc
  real, dimension(vsmMaxLight) :: ln, sn
  integer :: i

  color(1:3) = 0.0
  eye_direc(1:3) = real(eye(1:3) - pos(1:3))
  call vsmNormVecF(eye_direc)

  do i=1, vsmLightNum
    if(.not.used(i)) cycle
    if(vsmLight(i)%para) then
      if(.not.present(transdat)) then
        light_direc(1:3) = vsmLight(i)%direc(1:3)
      else
        call vsmMatrixCrossVectf(transdat%matrix, vsmLight(i)%direc, direc)
        light_direc(1:3) = direc(1:3)
      end if
    else
      if(.not.present(transdat)) then
        light_direc(1:3) = vsmLight(i)%pos(1:3) - pos(1:3)
      else
        direc(1:3) = vsmLight(i)%pos(1:3) - real(pos(1:3))
        call vsmMatrixCrossVectf(transdat%matrix, direc, light_direc)
      end if
      call vsmNormVecF(light_direc)
    end if
    ln(i) = dot_product(norm, light_direc)
    if(ln(i) <= 0.0) then
      ln(i) = 0.0
      sn(i) = 0.0
    else
      el_direc(1:3) = light_direc(1:3) + eye_direc(1:3)
      call vsmNormVecF(el_direc)
      sn(i) = dot_product(norm, el_direc)
      sn(i) = max(sn(i), 0.0)
    end if
  end do

  do i=1, vsmLightNum
    if(.not.used(i)) cycle !2013/8/30
    if(sn(i)>0.0) then
      if(log10(sn(i))*mat_shin<-5.0) then
        sn(i) = 0.0
      else
        sn(i) = sn(i)**mat_shin
      end if
    end if
    color(1:3) = color(1:3) + mat_amb*vsmLight(i)%amb(1:3) + &
                 mat_dif*vsmLight(i)%dif(1:3)*ln(i) + &
                 mat_spec*vsmLight(i)%spec(1:3)*sn(i)
  end do
 end subroutine vsmCalcColor

 subroutine vsmCalcColorOrtho(mat_amb, mat_dif, mat_spec, mat_shin, dir, pos, norm, used, color, transdat)
  real, intent(in) :: mat_amb, mat_dif, mat_spec, mat_shin
  real(kind=vdp), dimension(3), intent(in) :: dir
  real(kind=vdp), dimension(3), intent(in) :: pos
  real, dimension(3), intent(in) :: norm
  logical, dimension(vsmMaxLight), intent(in) :: used
  real, dimension(3), intent(out) :: color
  type(vismo_transdata), optional, intent(in) :: transdat

  real, dimension(3) :: eye_direc, light_direc, el_direc, direc
  real, dimension(vsmMaxLight) :: ln, sn
  integer :: i

  color(1:3) = 0.0
  eye_direc(1:3) = -dir(1:3)
  call vsmNormVecF(eye_direc)

  do i=1, vsmLightNum
    if(.not.used(i)) cycle
    if(vsmLight(i)%para) then
      if(.not.present(transdat)) then
        light_direc(1:3) = vsmLight(i)%direc(1:3)
      else
        call vsmMatrixCrossVectf(transdat%matrix, vsmLight(i)%direc, direc)
        light_direc(1:3) = direc(1:3)
      end if
    else
      if(.not.present(transdat)) then
        light_direc(1:3) = vsmLight(i)%pos(1:3) - pos(1:3)
      else
        direc(1:3) = vsmLight(i)%pos(1:3) - real(pos(1:3))
        call vsmMatrixCrossVectf(transdat%matrix, direc, light_direc)
      end if
      call vsmNormVecF(light_direc)
    end if
    ln(i) = dot_product(norm, light_direc)
    if(ln(i) <= 0.0) then
      ln(i) = 0.0
      sn(i) = 0.0
    else
      el_direc(1:3) = light_direc(1:3) + eye_direc(1:3)
      call vsmNormVecF(el_direc)
      sn(i) = dot_product(norm, el_direc)
      sn(i) = max(sn(i), 0.0)
    end if
  end do

  do i=1, vsmLightNum
    if(.not.used(i)) cycle !2013/8/30
    if(sn(i)>0.0) then
      if(log10(sn(i))*mat_shin<-5.0) then
        sn(i) = 0.0
      else
        sn(i) = sn(i)**mat_shin
      end if
    end if
    color(1:3) = color(1:3) + mat_amb*vsmLight(i)%amb(1:3) + &
                 mat_dif*vsmLight(i)%dif(1:3)*ln(i) + &
                 mat_spec*vsmLight(i)%spec(1:3)*sn(i)
  end do
 end subroutine vsmCalcColorOrtho

end module vsmLightData

