module vsmOctree
  use vsmConst
  use vsmFunction
  use vsmCoordinates
  use vsmColormap
  use vsmData
  use vsmParam
!$  use omp_lib
  implicit none
  private

type vismo_OctNode
  integer, dimension(3) :: vertex
  integer, dimension(3) :: OctDatSize
  !morton number  
  integer :: parent
  integer :: moi
  integer, dimension(0:7) :: children
  integer :: MyNumber ! morton number

  integer :: level ! level of octree
  logical :: skip
  real(kind=vdp) :: txmin, tymin, tzmin, txmax, tymax, tzmax
  integer, dimension(3) :: SkipVertexMin, SkipVertexMax
  real(kind=vdp), dimension(3) ::  SkipMin, SkipMax
 end type vismo_OctNode

 type vismo_OctCommon
  logical :: uniform
  integer :: maxLevel
  integer :: vis_num, Row
  real(kind=vdp) :: xmin, ymin, zmin
  real(kind=vdp), dimension(:), allocatable :: coordx, coordy, coordz
  logical, dimension(3) :: vis_on ! 1:isosurface, 2:volume rendering, 3:slices
 end type vismo_OctCommon

 type(vismo_OctNode), dimension(0:vsmMaxOctNum), save :: vsmOctant
 type(vismo_OctCommon), save :: vsmOctComDat

 public vsmInitOctree, vsmFinalizeOctree, vsmOctDivide, vsmOctPrepare, vsmRayJump, vsmRayJumpI
 public vsmSetOctComDat, vsmUpdateOctree
 public vsmRaySkipVoxel

 contains

 subroutine vsmInitOctree
  integer :: i

  do i=0, vsmMaxOctNum-1
    vsmOctant(i)%children(0:7) = -1

    vsmOctant(i)%parent = -1
    vsmOctant(i)%MyNumber = -1
    vsmOctant(i)%skip = .false.

    vsmOctant(i)%vertex(1:3) = -1
    vsmOctant(i)%SkipVertexMin(1:3) = -1
    vsmOctant(i)%SkipVertexMax(1:3) = -1
    vsmOctant(i)%SkipMin(1:3) = -1.0d+10
    vsmOctant(i)%SkipMax(1:3) = -1.0d+10

    vsmOctant(i)%txmin = 1.0d+10;  vsmOctant(i)%txmax = -1.0d+10
    vsmOctant(i)%tymin = 1.0d+10;  vsmOctant(i)%tymax = -1.0d+10
    vsmOctant(i)%tzmin = 1.0d+10;  vsmOctant(i)%tzmax = -1.0d+10
  end do

  vsmOctComDat%vis_num      = 0
  vsmOctComDat%uniform      = .true.
  vsmOctComDat%maxLevel     = vsmMaxOctLevel - 1
  vsmOctComDat%vis_on(1:3)  = .false.

  vsmOctComDat%xmin = 1.0d+10
  vsmOctComDat%ymin = 1.0d+10
  vsmOctComDat%zmin = 1.0d+10

  vsmOctComDat%Row  = 0

  call vsmLogPrint("Octree: vsmOctComDat%maxLevel", vsmOctComDat%maxLevel)
  call vsmLogPrint("Octree Initialized")
 end subroutine vsmInitOctree

 subroutine vsmSetOctComDat(vn)
   integer, intent(in) :: vn

   vsmOctComDat%vis_num = vn

   if(vsmVisParam(vn)%iso_num > 0 .and. vsmVisParam(vn)%iso_trans) then
     vsmOctComDat%vis_on(1) = .true.
   else
     vsmOctComDat%vis_on(1) = .false.
   end if

   if(vsmVisParam(vn)%volren_num > 0 ) then
     vsmOctComDat%vis_on(2) = .true.
   else
     vsmOctComDat%vis_on(2) = .false.
   end if

   if(vsmVisParam(vn)%slice_num > 0  .and. vsmVisParam(vn)%slice_trans) then
     vsmOctComDat%vis_on(3) = .true.
   else
     vsmOctComDat%vis_on(3) = .false.
   end if

 end subroutine vsmSetOctComDat

 subroutine vsmFinalizeOctree
   if(allocated(vsmOctComDat%coordx)) deallocate(vsmOctComDat%coordx)
   if(allocated(vsmOctComDat%coordy)) deallocate(vsmOctComDat%coordy)
   if(allocated(vsmOctComDat%coordz)) deallocate(vsmOctComDat%coordz)
 end subroutine vsmFinalizeOctree

 subroutine vsmRayJump(pos, dirc, ray_, r_skip)
  real(kind=vdp), dimension(3), intent(in) :: pos, dirc
  real(kind=vdp), intent(out) :: ray_
  logical, intent(out) :: r_skip
  integer :: i, nd
  integer :: div_num, integ_m, count_
  integer :: intx, inty, intz
  integer :: mnumray

  !call vsmLogPrint("Octree's vsmRayJump is called")
  intx = -1; inty = -1; intz = -1
  r_skip = .false.

  if(.not.vsmIsInsidelWO(pos)) then
    call vsmLogPrint("RayJump: out of data:", pos)
    call vsmError("RayJump: input pos is strange")
  end if

  div_num = vsmOctComDat%Row
  !div_num = ishft(div_num, vsmOctComDat%maxLevel)

!  integ_m = 1; count_ = 1
!  do i=1, vsmOctComDat%maxLevel-1
!    count_ = count_ * 8
!    integ_m = integ_m + count_
!  end do

  if(vsmOctComDat%coordx(div_num) == pos(1)) then
    intx = div_num-1 
  else
    do i=div_num-1, 0, -1
      if(vsmOctComDat%coordx(i) <= pos(1)) then
        intx = i
        exit
      end if
    end do
  end if

  if(vsmOctComDat%coordy(div_num) == pos(2)) then
    inty = div_num-1 
  else
    do i=div_num-1, 0, -1
      if(vsmOctComDat%coordy(i) <= pos(2)) then
        inty = i
        exit
      end if
    end do
  end if

  if(vsmOctComDat%coordz(div_num) == pos(3)) then
    intz = div_num-1 
  else
    do i=div_num-1, 0, -1
      if(vsmOctComDat%coordz(i) <= pos(3)) then
        intz = i
        exit
      end if
    end do
  end if

  if(intx == -1 .or. inty == -1 .or. intz == -1) then
    if(vsmDbg) then
      do i=0, div_num
        call vsmLogPrint("Octree:i=", i)
        call vsmLogPrint("Octree:coordx ", vsmOctComDat%coordx(i))
        call vsmLogPrint("Octree:coordy ", vsmOctComDat%coordy(i))
        call vsmLogPrint("Octree:coordz ", vsmOctComDat%coordz(i))
      end do
    end if
    call vsmError("RayJump: intx,y or z is negative")
  end if

  if(vsmOctComDat%maxLevel > 0) then
    integ_m = vsmOctNum(vsmOctComDat%maxLevel-1)
  else
    integ_m = 0
  end if
  mnumray =  vsmGetMortonNum(intx, inty, intz)
  nd = mnumray+integ_m

  if(vsmMaxOctNum <= nd ) call vsmError("OctNum: vsmRayJump")
  call vsmRaySkip(pos, dirc, nd, ray_, r_skip)

  !call vsmLogPrint("Octree's vsmRayJump ended")
 end subroutine vsmRayJump

 subroutine vsmRayJumpI(pos, dirc, jn, r_skip)
  real(kind=vdp), dimension(3), intent(in) :: pos, dirc
  integer, intent(inout) :: jn
  logical, intent(inout) :: r_skip
  integer :: i, nd
  integer :: div_num, integ_m!, count_
  integer :: intx, inty, intz
  integer :: mnumray

  !call vsmLogPrint("Octree's vsmRayJump is called")
  intx = -1; inty = -1; intz = -1
  r_skip = .false.

  if(.not.vsmIsInsidelWO(pos)) then
    call vsmLogPrint("RayJump: out of data:", pos)
    call vsmError("RayJump: input pos is strange")
  end if

  div_num = vsmOctComDat%Row

  if(vsmOctComDat%coordx(div_num) == pos(1)) then
    intx = div_num-1 
  else
    do i=div_num-1, 0, -1
      if(vsmOctComDat%coordx(i) <= pos(1)) then
        intx = i
        exit
      end if
    end do
  end if

  if(vsmOctComDat%coordy(div_num) == pos(2)) then
    inty = div_num-1 
  else
    do i=div_num-1, 0, -1
      if(vsmOctComDat%coordy(i) <= pos(2)) then
        inty = i
        exit
      end if
    end do
  end if

  if(vsmOctComDat%coordz(div_num) == pos(3)) then
    intz = div_num-1 
  else
    do i=div_num-1, 0, -1
      if(vsmOctComDat%coordz(i) <= pos(3)) then
        intz = i
        exit
      end if
    end do
  end if

  if(intx == -1 .or. inty == -1 .or. intz == -1) then
    if(vsmDbg) then
      do i=0, div_num
        call vsmLogPrint("Octree:i=", i)
        call vsmLogPrint("Octree:coordx ", vsmOctComDat%coordx(i))
        call vsmLogPrint("Octree:coordy ", vsmOctComDat%coordy(i))
        call vsmLogPrint("Octree:coordz ", vsmOctComDat%coordz(i))
      end do
    end if
    call vsmError("RayJump: intx,y or z is negative")
  end if

  if(vsmOctComDat%maxLevel > 0) then
    integ_m = vsmOctNum(vsmOctComDat%maxLevel-1)
  else
    integ_m = 0
  end if
  mnumray =  vsmGetMortonNum(intx, inty, intz)
  nd = mnumray+integ_m

  if(vsmMaxOctNum <= nd ) call vsmError("OctNum: vsmRayJump")
  call vsmRaySkipI(pos, dirc, nd, jn, r_skip)

  !call vsmLogPrint("Octree's vsmRayJump ended")
 end subroutine vsmRayJumpI

 function vsmIsInsideOctant(pos, on)
  real(kind=vdp), dimension(3), intent(in) :: pos
  integer, intent(in) :: on
  logical :: vsmIsInsideOctant

  if(vsmOctant(on)%txmin <= pos(1) .and. vsmOctant(on)%txmax > pos(1) .and. &
    vsmOctant(on)%tymin <= pos(2) .and. vsmOctant(on)%tymax > pos(2) .and. &
    vsmOctant(on)%tzmin <= pos(3) .and. vsmOctant(on)%tzmax > pos(3)) then
    vsmIsInsideOctant = .true.
  else
    vsmIsInsideOctant = .false.
  end if

 end function vsmIsInsideOctant

 subroutine vsmRaySkipVoxel(i0, pos, dirc, ray_, r_skip)
  integer, dimension(3), intent(in) :: i0
  real(kind=vdp), dimension(3), intent(in) :: pos
  real(kind=vdp), dimension(3), intent(in) :: dirc
  real(kind=vdp), intent(out) :: ray_
  logical, intent(out) :: r_skip

  integer :: i, j, k, ip1,jp1,kp1,im1,jm1,km1
  real(kind=vdp), dimension(6) :: a
  real(kind=vdp), dimension(3) :: cv_min, cv_max
  integer :: vis_num
  integer :: n, nscal, l
  real(kind=vdp) :: iso_level
  real(kind=vdp) :: max_val, min_val, b
  real(kind=vdp), dimension(8) :: dat
  real :: maxalpha
  real(kind=vdp), dimension(3, 8) :: tpos

  vis_num = vsmOctComDat%vis_num
  i=i0(1); j = i0(2); k=i0(3)
  r_skip = .false.

  ip1 = i+1; jp1 = j+1; kp1 = k+1
  im1=i-1; jm1=j-1; km1=k-1

  if(.not.vsmCheckLCoord(i, j, k)) call vsmError("Error:: vsmRaySkipVoxel")
  if(.not.vsmCheckLCoord(ip1, jp1, kp1)) call vsmError("Error:: vsmRaySkipVoxel2")

!  if(i == node%OctDatSize(1)) i = node%OctDatSize(1) - 1
!  if(j == node%OctDatSize(2)) j = node%OctDatSize(2) - 1
!  if(k == node%OctDatSize(3)) k = node%OctDatSize(3) - 1

!  if(i < 1 .or. i > node%OctDatSize(1) - 1) call vsmError("RaySkip voxel : i is strange")
!  if(j < 1 .or. j > node%OctDatSize(2) - 1) call vsmError("RaySkip voxel : j is strange")
!  if(k < 1 .or. k > node%OctDatSize(3) - 1) call vsmError("RaySkip voxel : k is strange")

  if(vsmOctComDat%vis_on(1)) then
    do n=1, vsmVisParam(vis_num)%iso_num
      nscal = vsmVisParam(vis_num)%isoparam(n)%nscal
      iso_level = vsmVisParam(vis_num)%isoparam(n)%iso_level

      dat(1) = vsmScalData(nscal)%scal(i,   j,   k)
      dat(2) = vsmScalData(nscal)%scal(ip1, j,   k)
      dat(3) = vsmScalData(nscal)%scal(i,   jp1, k)
      dat(4) = vsmScalData(nscal)%scal(ip1, jp1, k)
      dat(5) = vsmScalData(nscal)%scal(i,   j,   kp1)
      dat(6) = vsmScalData(nscal)%scal(ip1, j,   kp1)
      dat(7) = vsmScalData(nscal)%scal(i,   jp1, kp1)
      dat(8) = vsmScalData(nscal)%scal(ip1, jp1, kp1)

      max_val = maxval(dat); min_val = minval(dat)

      if(min_val <= iso_level .and. max_val >= iso_level) then
        r_skip = .false. 
        return
      end if
    end do
  end if

 if(vsmOctComDat%vis_on(2)) then
   do n=1, vsmVisParam(vis_num)%volren_num
     nscal = vsmVisParam(vis_num)%volparam(n)%nscal

      dat(1) = vsmScalData(nscal)%scal(i,   j,   k)
      dat(2) = vsmScalData(nscal)%scal(ip1, j,   k)
      dat(3) = vsmScalData(nscal)%scal(i,   jp1, k)
      dat(4) = vsmScalData(nscal)%scal(ip1, jp1, k)
      dat(5) = vsmScalData(nscal)%scal(i,   j,   kp1)
      dat(6) = vsmScalData(nscal)%scal(ip1, j,   kp1)
      dat(7) = vsmScalData(nscal)%scal(i,   jp1, kp1)
      dat(8) = vsmScalData(nscal)%scal(ip1, jp1, kp1)
      maxalpha = vsmGetMaxAlpha(dat, nscal)
 
     if(maxalpha > vsmSkipAlpha) then
        r_skip = .false. 
        return
     end if
   end do
 end if

 if(vsmOctComDat%vis_on(3)) then
   do n=1, vsmVisParam(vis_num)%slice_num
     if(vsmOctComDat%uniform) then
       tpos(1,1) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(im1)
       tpos(2,1) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(jm1)
       tpos(3,1) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(km1)

       tpos(1,2) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(i  )
       tpos(2,2) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(jm1)
       tpos(3,2) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(km1)

       tpos(1,3) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(im1)
       tpos(2,3) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(j  )
       tpos(3,3) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(km1)

       tpos(1,4) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(i  )
       tpos(2,4) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(j  )
       tpos(3,4) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(km1)

       tpos(1,5) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(im1)
       tpos(2,5) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(jm1)
       tpos(3,5) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(k)

       tpos(1,6) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(i  )
       tpos(2,6) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(jm1)
       tpos(3,6) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(k)

       tpos(1,7) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(im1)
       tpos(2,7) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(j  )
       tpos(3,7) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(k)

       tpos(1,8) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(i  )
       tpos(2,8) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(j  )
       tpos(3,8) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(k)
     else
       tpos(1,1) = vsmLocalCoord%coordx(i)
       tpos(2,1) = vsmLocalCoord%coordy(j)
       tpos(3,1) = vsmLocalCoord%coordz(k)

       tpos(1,2) = vsmLocalCoord%coordx(ip1)
       tpos(2,2) = vsmLocalCoord%coordy(j)
       tpos(3,2) = vsmLocalCoord%coordz(k)

       tpos(1,3) = vsmLocalCoord%coordx(i)
       tpos(2,3) = vsmLocalCoord%coordy(jp1)
       tpos(3,3) = vsmLocalCoord%coordz(k)

       tpos(1,4) = vsmLocalCoord%coordx(ip1)
       tpos(2,4) = vsmLocalCoord%coordy(jp1)
       tpos(3,4) = vsmLocalCoord%coordz(k)

       tpos(1,5) = vsmLocalCoord%coordx(i)
       tpos(2,5) = vsmLocalCoord%coordy(j)
       tpos(3,5) = vsmLocalCoord%coordz(kp1)

       tpos(1,6) = vsmLocalCoord%coordx(ip1)
       tpos(2,6) = vsmLocalCoord%coordy(j)
       tpos(3,6) = vsmLocalCoord%coordz(kp1)

       tpos(1,7) = vsmLocalCoord%coordx(i)
       tpos(2,7) = vsmLocalCoord%coordy(jp1)
       tpos(3,7) = vsmLocalCoord%coordz(kp1)

       tpos(1,8) = vsmLocalCoord%coordx(ip1)
       tpos(2,8) = vsmLocalCoord%coordy(jp1)
       tpos(3,8) = vsmLocalCoord%coordz(kp1)
     end if

     do l=1, 8
       dat(l) = vsmIsAboveSlice(vis_num, n, tpos(1:3,l))
     end do

     min_val =  minval(dat)
     max_val =  maxval(dat)

     if(min_val*max_val <= 0.0_vdp) then
       r_skip = .false. 
       return
     end if

   end do
 end if

 if(vsmOctComDat%uniform) then
   cv_min(1) = vsmLocalCoord%origin(1) + im1*vsmLocalCoord%dx(1)
   cv_max(1) = vsmLocalCoord%origin(1) + i*vsmLocalCoord%dx(1)

   cv_min(2) = vsmLocalCoord%origin(2) + jm1*vsmLocalCoord%dx(2)
   cv_max(2) = vsmLocalCoord%origin(2) + j*vsmLocalCoord%dx(2)

   cv_min(3) = vsmLocalCoord%origin(3) + km1*vsmLocalCoord%dx(3)
   cv_max(3) = vsmLocalCoord%origin(3) + k*vsmLocalCoord%dx(3)
 else
   cv_min(1) = vsmLocalCoord%coordx(i)
   cv_max(1) = vsmLocalCoord%coordx(ip1)

   cv_min(2) = vsmLocalCoord%coordy(j)
   cv_max(2) = vsmLocalCoord%coordy(jp1)

   cv_min(3) = vsmLocalCoord%coordz(k)
   cv_max(3) = vsmLocalCoord%coordz(kp1)
 end if

 if( pos(1) > cv_max(1) .or. pos(1) < cv_min(1) .or. & 
     pos(2) > cv_max(2) .or. pos(2) < cv_min(2) .or. &
     pos(3) > cv_max(3) .or. pos(3) < cv_min(3)) then
    call vsmLogPrint("Octree: Ray fail cvmin: ",cv_min)
    call vsmLogPrint("Octree: Ray fail cvmax: ",cv_max)
    call vsmLogPrint("Octree: Ray fail pos: ",pos)
    r_skip = .false.
    return
 end if

 a = -1.0_vdp

 if(dirc(1) /= 0.0_vdp) then
   a(1) = (cv_min(1)-pos(1))/dirc(1)
   a(2) = (cv_max(1)-pos(1))/dirc(1)
 end if
  
 if(dirc(2) /= 0.0_vdp) then
   a(3) = (cv_min(2)-pos(2))/dirc(2)
   a(4) = (cv_max(2)-pos(2))/dirc(2)
 end if

 if(dirc(3) /= 0.0_vdp) then
   a(5) = (cv_min(3)-pos(3))/dirc(3)
   a(6) = (cv_max(3)-pos(3))/dirc(3)
 end if

  where(a < 0.0_vdp)
    a = vsmFaraway
  end where

  b = minval(a)
  ray_ = vsmIntval%ray_interval * int(b/vsmIntval%ray_interval)

  if(int(b/vsmIntval%ray_interval) > 0 .and. ray_ == vsmIntval%ray_interval * b/vsmIntval%ray_interval) then
     ray_ = vsmIntval%ray_interval * (int(b/vsmIntval%ray_interval)-1)
   end if

  r_skip = .true.
 end subroutine vsmRaySkipVoxel

 recursive subroutine vsmUpdateOctree(node)
   integer, intent(in) :: node
   logical :: r_skip
   integer :: i

!   call vsmLogPrint("Octree: Update Octree is called")
   vsmOctant(node)%skip = .false.

   if(vsmOctant(vsmOctant(node)%parent)%skip) then
     !the parent of root is root itself. skip is always false
     vsmOctant(node)%skip = .true.
     call vsmGetSkipVertex(node, r_skip)

       if(vsmOctant(node)%moi == 0) then
         call vsmLogPrint("Octree: in node%parent%skip root ", vsmOctant(node)%skip)
       end if
   else if(SurveyData(node) == 0) then

     vsmOctant(node)%skip = .true.
     vsmOctant(node)%SkipVertexMin(1:3) = vsmOctant(node)%vertex(1:3)
     vsmOctant(node)%SkipVertexMax(1:3) = vsmOctant(node)%vertex(1:3) + vsmOctant(node)%OctDatSize(1:3)-1

     if(vsmOctant(node)%moi == 0) then
       call vsmLogPrint("Octree: in SurveyData root ", vsmOctant(node)%skip)
     end if

   end if

   if(vsmOctant(node)%moi == 0) call vsmLogPrint("Octree: root ", vsmOctant(node)%skip)

   if(vsmOctant(node)%skip) then

     if(vsmOctComDat%uniform) then
       vsmOctant(node)%SkipMin(1) = vsmOctComDat%xmin + vsmLocalCoord%dx(1)*(vsmOctant(node)%SkipVertexMin(1)-1)
       vsmOctant(node)%SkipMax(1) = vsmOctComDat%xmin + vsmLocalCoord%dx(1)*(vsmOctant(node)%SkipVertexMax(1)-1)
       vsmOctant(node)%SkipMin(2) = vsmOctComDat%ymin + vsmLocalCoord%dx(2)*(vsmOctant(node)%SkipVertexMin(2)-1)
       vsmOctant(node)%SkipMax(2) = vsmOctComDat%ymin + vsmLocalCoord%dx(2)*(vsmOctant(node)%SkipVertexMax(2)-1)
       vsmOctant(node)%SkipMin(3) = vsmOctComDat%zmin + vsmLocalCoord%dx(3)*(vsmOctant(node)%SkipVertexMin(3)-1)
       vsmOctant(node)%SkipMax(3) = vsmOctComDat%zmin + vsmLocalCoord%dx(3)*(vsmOctant(node)%SkipVertexMax(3)-1)
     else
       vsmOctant(node)%SkipMin(1) = vsmLocalCoord%coordx(vsmOctant(node)%SkipVertexMin(1))
       vsmOctant(node)%SkipMax(1) = vsmLocalCoord%coordx(vsmOctant(node)%SkipVertexMax(1))
       vsmOctant(node)%SkipMin(2) = vsmLocalCoord%coordy(vsmOctant(node)%SkipVertexMin(2))
       vsmOctant(node)%SkipMax(2) = vsmLocalCoord%coordy(vsmOctant(node)%SkipVertexMax(2))
       vsmOctant(node)%SkipMin(3) = vsmLocalCoord%coordz(vsmOctant(node)%SkipVertexMin(3))
       vsmOctant(node)%SkipMax(3) = vsmLocalCoord%coordz(vsmOctant(node)%SkipVertexMax(3))
     end if

  else
    vsmOctant(node)%SkipMin(1:3) = -1.0d+10
    vsmOctant(node)%SkipMax(1:3) = -1.0d+10
  end if 
 
  if(vsmOctant(node)%level == vsmOctComDat%maxLevel) then
    !call vsmLogPrint("Level == maxLevel: vsmOctant(node)%level", vsmOctant(node)%level)
    !call vsmLogPrint("Level == maxLevel: vsmOctComDat%maxLevel", vsmOctComDat%maxLevel)
    !call vsmLogPrint("Octree: Update Octree ends", vsmOctant(node)%moi)
    return
  end if

  do i=0, 7
    call vsmUpdateOctree(vsmOctant(node)%children(i))
  end do

  !call vsmLogPrint("Octree: Update Octree ends", node%moi)
 end subroutine vsmUpdateOctree

 subroutine vsmOctPrepare
  integer, dimension(3) :: vtx

  call vsmSetGrid
  vtx(1:3) = 1
  call vsmSetVertex(vtx, vsmLocalCoord%nx, 0)
 end subroutine vsmOctPrepare

 recursive subroutine vsmOctDivide(node)
  integer, intent(in) :: node
  logical :: root

  integer :: n, nx, ny, nz, b
  integer :: i, allocstat
  logical :: r_skip
  integer, dimension(3) :: hsize, hsizep1
  integer, dimension(3, 0:7) :: child_vertex, child_size
  integer :: next_level
  integer :: Mnum
  integer :: integ_num, tinteg_num

  if(node == 0) then
    root = .true.
  else
    root = .false.
  end if

  if(root) then
    call vsmLogPrint("vsmOctDivide is called")
   end if

  call vsmCalcMinMax(node)

  LEVEL0 : if(root) then
    vsmOctant(node)%level  = 0
    vsmOctant(node)%parent = 0
    nx = vsmOctant(node)%OctDatSize(1) - 1 ! num of control volume
    ny = vsmOctant(node)%OctDatSize(2) - 1
    nz = vsmOctant(node)%OctDatSize(3) - 1
    n  = vsmMaxOctLevel

    do i=1, vsmMaxOctLevel
      nx = nx/2; ny = ny/2; nz = nz/2
   
      if(vsmDbg) then
        call vsmLogPrint("nx = ",nx)
        call vsmLogPrint("ny = ",ny)
        call vsmLogPrint("nz = ",nz)
      end if

      if(nx < 1 .or. ny < 1 .or. nz < 1) exit
   
      n=n-1

      if(nx == 1 .or. ny == 1 .or. nz == 1) exit

    end do

    vsmOctComDat%maxLevel = vsmMaxOctLevel - n
    call vsmLogPrint("Octree: MaxLevel = ",vsmOctComDat%maxLevel)

    b = 1
    b = ishft(b, vsmOctComDat%maxLevel)
  
    call vsmLogPrint("Octree: Row = ",b)
    vsmOctComDat%Row = b
    allocate(vsmOctComDat%coordx(0:b), vsmOctComDat%coordy(0:b), vsmOctComDat%coordz(0:b), stat = allocstat)
    if(allocstat /= 0) call vsmError('cant allocate octree coord')

    vsmOctant(node)%moi = 0
    vsmOctant(node)%MyNumber = 0

  end if LEVEL0
 
  if(vsmOctant(node)%level == vsmOctComDat%maxLevel) then
    ! call vsmLogPrint("vsmOctant(node)level==vsmOctComDat%maxLevel:", vsmOctant(node)%level)
    return
  end if

  hsize(1:3) = vsmOctant(node)%OctDatSize(1:3)/2
  hsizep1(1:3) = hsize(1:3) + 1;
  
! child 0
  child_vertex(1:3, 0) = vsmOctant(node)%Vertex(1:3)
  child_size(1:3, 0) = hsizep1(1:3)

! child 1
  child_vertex(1,1) = vsmOctant(node)%Vertex(1) + hsize(1)
  child_vertex(2,1) = vsmOctant(node)%Vertex(2)
  child_vertex(3,1) = vsmOctant(node)%Vertex(3)
  child_size(1,1) = vsmOctant(node)%OctDatSize(1) - hsize(1)
  child_size(2,1) = hsizep1(2)
  child_size(3,1) = hsizep1(3)

! child 2
  child_vertex(1,2) = vsmOctant(node)%Vertex(1)
  child_vertex(2,2) = vsmOctant(node)%Vertex(2) + hsize(2)
  child_vertex(3,2) = vsmOctant(node)%Vertex(3)
  child_size(1,2) = hsizep1(1)
  child_size(2,2) = vsmOctant(node)%OctDatSize(2) - hsize(2)
  child_size(3,2) = hsizep1(3)

! child 3
  child_vertex(1,3) = vsmOctant(node)%Vertex(1) + hsize(1)
  child_vertex(2,3) = vsmOctant(node)%Vertex(2) + hsize(2)
  child_vertex(3,3) = vsmOctant(node)%Vertex(3)
  child_size(1,3) = vsmOctant(node)%OctDatSize(1) - hsize(1)
  child_size(2,3) = vsmOctant(node)%OctDatSize(2) - hsize(2)
  child_size(3,3) = hsizep1(3)

! child 4
  child_vertex(1, 4) = vsmOctant(node)%Vertex(1)
  child_vertex(2, 4) = vsmOctant(node)%Vertex(2)
  child_vertex(3, 4) = vsmOctant(node)%Vertex(3) + hsize(3)
  child_size(1, 4) = hsizep1(1)
  child_size(2, 4) = hsizep1(2)
  child_size(3, 4) = vsmOctant(node)%OctDatSize(3) - hsize(3)

! child 5
  child_vertex(1, 5) = vsmOctant(node)%Vertex(1) + hsize(1)
  child_vertex(2, 5) = vsmOctant(node)%Vertex(2)
  child_vertex(3, 5) = vsmOctant(node)%Vertex(3) + hsize(3)
  child_size(1, 5) = vsmOctant(node)%OctDatSize(1) - hsize(1)
  child_size(2, 5) = hsizep1(2)
  child_size(3, 5) = vsmOctant(node)%OctDatSize(3) - hsize(3)

! child 6
  child_vertex(1, 6) = vsmOctant(node)%Vertex(1)
  child_vertex(2, 6) = vsmOctant(node)%Vertex(2) + hsize(2)
  child_vertex(3, 6) = vsmOctant(node)%Vertex(3) + hsize(3)
  child_size(1, 6) = hsizep1(1)
  child_size(2, 6) = vsmOctant(node)%OctDatSize(2) - hsize(2)
  child_size(3, 6) = vsmOctant(node)%OctDatSize(3) - hsize(3)

! child 7
  child_vertex(1, 7) = vsmOctant(node)%Vertex(1) + hsize(1)
  child_vertex(2, 7) = vsmOctant(node)%Vertex(2) + hsize(2)
  child_vertex(3, 7) = vsmOctant(node)%Vertex(3) + hsize(3)
  child_size(1, 7) = vsmOctant(node)%OctDatSize(1) - hsize(1)
  child_size(2, 7) = vsmOctant(node)%OctDatSize(2) - hsize(2)
  child_size(3, 7) = vsmOctant(node)%OctDatSize(3) - hsize(3)

  next_level = vsmOctant(node)%level + 1
  Mnum = vsmOctant(node)%MyNumber*8

!  integ_num = 1; tinteg_num = 1
!  do i=1, vsmOctant(node)%level
!    tinteg_num = tinteg_num * 8
!    integ_num = integ_num + tinteg_num
!  end do

  integ_num = vsmOctNum(vsmOctant(node)%level)

  do i=0, 7
    vsmOctant(node)%children(i) = integ_num + Mnum + i
    vsmOctant(vsmOctant(node)%children(i))%level = next_level
    vsmOctant(vsmOctant(node)%children(i))%MyNumber = Mnum+i
  end do

  !call vsmLogPrint("moi : ", vsmOctant(node)%moi)
  !call vsmLogPrint("children's numbers: ", vsmOctant(node)%children)
 
  do i=0, 7
    if(vsmOctant(node)%children(i) >= vsmMaxOctNum) then
      call vsmLogPrint("vsmOctant(node)%children(i) >= vsmMaxOctNum:", vsmOctant(node)%children(i))
      call vsmError("vsmOctant(node) number is strange")
    end if
    vsmOctant(vsmOctant(node)%children(i))%moi = vsmOctant(node)%children(i)
    vsmOctant(vsmOctant(node)%children(i))%parent = vsmOctant(node)%moi
    call vsmSetVertex(child_vertex(1:3, i), child_size(1:3, i), vsmOctant(node)%children(i))
    !call vsmLogPrint("call vsmOctDivide: ", vsmOctant(node)%children(i))
    call vsmOctDivide(vsmOctant(node)%children(i))
  end do

  if(vsmOctant(node)%level == 0) then
    call vsmLogPrint("call vsmCalcCoord")
    call vsmCalcCoord
  end if

  !call vsmLogPrint("vsmOctDivide ends:", vsmOctant(node)%moi)
 end subroutine vsmOctDivide

 subroutine vsmSetGrid
  vsmOctComDat%uniform = vsmLocalCoord%uniform
  vsmOctComDat%xmin = vsmLocalCoord%origin(1)
  vsmOctComDat%ymin = vsmLocalCoord%origin(2)
  vsmOctComDat%zmin = vsmLocalCoord%origin(3)
 end subroutine vsmSetGrid 

 subroutine vsmSetVertex(vx, sz, node)
   integer, dimension(3), intent(in) :: vx, sz
   integer, intent(in) :: node

   vsmOctant(node)%vertex(1:3) = vx(1:3)
   vsmOctant(node)%OctDatSize(1:3) = sz(1:3)
   !call vsmLogPrint("Octree:sz", sz)
 end subroutine vsmSetVertex

 function SurveyData(node)
  integer, intent(in) :: node
  integer ::  SurveyData
  integer :: i, j, k, nscal, n, vis_num, ip1,jp1,kp1
  real(kind=vdp) :: iso_level, max_val, min_val
  real(kind=vdp), dimension(8) :: dat
  real(kind=vdp), dimension(3) :: pos
  real :: alpha
  integer :: above, below

  !call vsmLogPrint("Octree: survey data is called, level = ",node%level)
  vis_num = vsmOctComDat%vis_num

  !call vsmLogPrint("vsnum",vis_num)
  if(vis_num < 1) call vsmError("survayData visnum")

    if(vsmOctant(node)%Vertex(1)<1 .or. vsmOctant(node)%Vertex(2)<1 .or. &
       vsmOctant(node)%Vertex(3)<1) call vsmError("i or j or k<1")

    if(vsmOctant(node)%Vertex(1)+vsmOctant(node)%OctDatSize(1)-1 > vsmLocalCoord%nx(1) &
      .or. vsmOctant(node)%Vertex(2)+vsmOctant(node)%OctDatSize(2)-1 > vsmLocalCoord%nx(2) &
      .or. vsmOctant(node)%Vertex(3)+vsmOctant(node)%OctDatSize(3)-1 > vsmLocalCoord%nx(3)) then
       call vsmError("i or j or k too big")
    end if

  SurveyData = 0
  if(vsmOctComDat%vis_on(1)) then ! isosurface

    do n=1, vsmVisParam(vis_num)%iso_num
      if(vsmVisParam(vis_num)%isoparam(n)%alpha >= vsmErtValue) cycle
      nscal = vsmVisParam(vis_num)%isoparam(n)%nscal
      iso_level = vsmVisParam(vis_num)%isoparam(n)%iso_level
      do k=vsmOctant(node)%Vertex(3), vsmOctant(node)%Vertex(3)+vsmOctant(node)%OctDatSize(3)-2
!$omp parallel do private(i,j,ip1,jp1,kp1,dat,max_val,min_val)
        do j=vsmOctant(node)%Vertex(2), vsmOctant(node)%Vertex(2)+vsmOctant(node)%OctDatSize(2)-2
          do i=vsmOctant(node)%Vertex(1), vsmOctant(node)%Vertex(1)+vsmOctant(node)%OctDatSize(1)-2
            if(SurveyData == 0) then
               ip1 = i+1; jp1 = j+1; kp1 = k+1
               dat(1) = vsmScalData(nscal)%scal(i,   j,   k)
               dat(2) = vsmScalData(nscal)%scal(ip1, j,   k)
               dat(3) = vsmScalData(nscal)%scal(i,   jp1, k)
               dat(4) = vsmScalData(nscal)%scal(ip1, jp1, k)
               dat(5) = vsmScalData(nscal)%scal(i,   j,   kp1)
               dat(6) = vsmScalData(nscal)%scal(ip1, j,   kp1)
               dat(7) = vsmScalData(nscal)%scal(i,   jp1, kp1)
               dat(8) = vsmScalData(nscal)%scal(ip1, jp1, kp1)

               max_val = maxval(dat)
               min_val = minval(dat)

               if(min_val <= iso_level .and. max_val >= iso_level) then
                  SurveyData = 1
                  call vsmLogPrint("SurveyData: Isosurface")
               end if
            end if

          end do !i
        end do !j
      if(SurveyData > 0) exit
      end do !k
      if(SurveyData > 0) exit
    end do
   !call vsmLogPrint("survey data for isosurface ends")
  end if

  if(SurveyData > 0) return

  if(vsmOctComDat%vis_on(2)) then !volume rendering

     do n=1, vsmVisParam(vis_num)%volren_num
       nscal = vsmVisParam(vis_num)%volparam(n)%nscal
       do k=vsmOctant(node)%Vertex(3), vsmOctant(node)%Vertex(3)+vsmOctant(node)%OctDatSize(3)-1

!$omp parallel do private(i,j,ip1,jp1,kp1,dat,alpha)
         do j=vsmOctant(node)%Vertex(2), vsmOctant(node)%Vertex(2)+vsmOctant(node)%OctDatSize(2)-1
           do i=vsmOctant(node)%Vertex(1), vsmOctant(node)%Vertex(1)+vsmOctant(node)%OctDatSize(1)-1
             if(SurveyData == 0) then
               ip1 = i+1; jp1 = j+1; kp1 = k+1
               dat(1) = vsmScalData(nscal)%scal(i,   j,   k)
               dat(2) = vsmScalData(nscal)%scal(ip1, j,   k)
               dat(3) = vsmScalData(nscal)%scal(i,   jp1, k)
               dat(4) = vsmScalData(nscal)%scal(ip1, jp1, k)
               dat(5) = vsmScalData(nscal)%scal(i,   j,   kp1)
               dat(6) = vsmScalData(nscal)%scal(ip1, j,   kp1)
               dat(7) = vsmScalData(nscal)%scal(i,   jp1, kp1)
               dat(8) = vsmScalData(nscal)%scal(ip1, jp1, kp1)

               alpha = vsmGetMaxAlpha(dat, nscal)

                if(alpha > vsmSkipAlpha) then
                  SurveyData = 1
                  call vsmLogPrint("SurveyData: VolumeRendering")
                end if
             end if

           end do !i
         end do !j
         if(SurveyData > 0) exit
       end do !k
     if(SurveyData > 0) exit
     end do
    !call vsmLogPrint("survey data for volumerendering ends")
  end if

  if(SurveyData > 0) return

   if(vsmVisParam(vis_num)%slice_solid) return
   if(vsmOctComDat%vis_on(3) .and. vsmPCmode /= 2) then ! slicing

    do n=1, vsmVisParam(vis_num)%slice_num
       if(vsmVisParam(vis_num)%sliceparam(n)%alpha >= vsmErtValue) cycle
       above = 0; below = 0
!$omp parallel do private(i,j,k,pos) firstprivate(above, below)
       do k=vsmOctant(node)%Vertex(3), vsmOctant(node)%Vertex(3)+vsmOctant(node)%OctDatSize(3)-1
         do j=vsmOctant(node)%Vertex(2), vsmOctant(node)%Vertex(2)+vsmOctant(node)%OctDatSize(2)-1
           do i=vsmOctant(node)%Vertex(1), vsmOctant(node)%Vertex(1)+vsmOctant(node)%OctDatSize(1)-1

             if(SurveyData == 0) then
                if(vsmOctComDat%uniform) then
                  pos(1) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(i-1)
                  pos(2) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(j-1)
                  pos(3) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(k-1)
                else
                  pos(1) = vsmLocalCoord%coordx(i)
                  pos(2) = vsmLocalCoord%coordy(j)
                  pos(3) = vsmLocalCoord%coordz(k)
                end if

                if(vsmIsAboveSlice(vis_num, n, pos) > 0.0_vdp) then
                  above = above + 1
                else
                  below = below + 1
                end if

                if(above * below > 0) then
                  SurveyData = 1
                  call vsmLogPrint("SurveyData: Slicing")
                end if
            end if
          end do
        end do
      end do
      if(SurveyData > 0) exit
    end do
     !call vsmLogPrint("Octree: survey data for slicing a",above)
     !call vsmLogPrint("Octree: survey data for slicing b",below)
   end if

  if(SurveyData > 0) return

  !call vsmLogPrint("Octree: survey data ends")
 end function SurveyData

 subroutine vsmGetGMinMax(gmin, gmax, node)
  real(kind=vdp), dimension(3), intent(out) :: gmin, gmax
  integer, intent(in) :: node

  gmin(1) = vsmOctant(node)%txmin; gmin(2) = vsmOctant(node)%tymin; gmin(3) = vsmOctant(node)%tzmin
  gmax(1) = vsmOctant(node)%txmax; gmax(2) = vsmOctant(node)%tymax; gmax(3) = vsmOctant(node)%tzmax

 end subroutine vsmGetGMinMax

 subroutine vsmCalcMinMax(node)
   integer, intent(in) :: node
   real(kind=vdp), dimension(3) :: temp

   if(vsmOctComDat%uniform) then
      vsmOctant(node)%txmin = vsmOctComDat%xmin + vsmLocalCoord%dx(1)*(vsmOctant(node)%Vertex(1)-1)
      vsmOctant(node)%txmax = vsmOctant(node)%txmin + vsmLocalCoord%dx(1)*(vsmOctant(node)%OctDatSize(1)-1)

      vsmOctant(node)%tymin = vsmOctComDat%ymin + vsmLocalCoord%dx(2)*(vsmOctant(node)%Vertex(2)-1)
      vsmOctant(node)%tymax = vsmOctant(node)%tymin + vsmLocalCoord%dx(2)*(vsmOctant(node)%OctDatSize(2)-1)

      vsmOctant(node)%tzmin = vsmOctComDat%zmin + vsmLocalCoord%dx(3)*(vsmOctant(node)%Vertex(3)-1)
      vsmOctant(node)%tzmax = vsmOctant(node)%tzmin + vsmLocalCoord%dx(3)*(vsmOctant(node)%OctDatSize(3)-1)

      temp(1) = vsmOctant(node)%txmin; temp(2) = vsmOctant(node)%tymin; temp(3) = vsmOctant(node)%tzmin;
      if(.not.vsmisinsidelwo(temp)) then
         call vsmLogPrint("calc min max1:", temp)
      end if

      temp(1) = vsmOctant(node)%txmax; temp(2) = vsmOctant(node)%tymax; temp(3) = vsmOctant(node)%tzmax;
      if(.not.vsmisinsidelwo(temp)) then
         call vsmLogPrint("calc min max2:", temp)
      end if
    else
      vsmOctant(node)%txmin = vsmLocalCoord%coordx(vsmOctant(node)%Vertex(1))
      vsmOctant(node)%txmax = vsmLocalCoord%coordx(vsmOctant(node)%Vertex(1)+vsmOctant(node)%OctDatSize(1)-1)

      vsmOctant(node)%tymin = vsmLocalCoord%coordy(vsmOctant(node)%Vertex(2))
      vsmOctant(node)%tymax = vsmLocalCoord%coordy(vsmOctant(node)%Vertex(2)+vsmOctant(node)%OctDatSize(2)-1)

      vsmOctant(node)%tzmin = vsmLocalCoord%coordz(vsmOctant(node)%Vertex(3))
      vsmOctant(node)%tzmax = vsmLocalCoord%coordz(vsmOctant(node)%Vertex(3)+vsmOctant(node)%OctDatSize(3)-1)
   end if
 end subroutine vsmCalcMinMax

 function vsmGetMortonNum(x, y, z)
   integer, intent(in) :: x, y, z
   integer :: vsmGetMortonNum
   integer ::  n, ax, ay, az, bx, by, bz, b, i, di
   !call vsmLogPrint("maxlevel",vsmOctComDat%maxLevel)
   n=1 
   bx=0; by=0; bz=0
   do i=0, vsmOctComDat%maxLevel-1
     ax = iand(x, n)
     ay = iand(y, n)
     az = iand(z, n)

     !ax = ishft(ax, 3*i  )
     !ay = ishft(ay, 3*i+1)
     !az = ishft(az, 3*i+2)

     di = 2*i

     ax = ishft(ax, di  )
     ay = ishft(ay, di+1)
     az = ishft(az, di+2)

     bx = ior(bx, ax)
     by = ior(by, ay)
     bz = ior(bz, az)

     n = ishft(n, 1)
   end do

   b = 0
   b = ior(b, bx)
   b = ior(b, by)
   vsmGetMortonNum = ior(b, bz)
  
!   vsmGetMortonNum = b
   if(vsmGetMortonNum < 0 .or. 8**vsmOctComDat%maxLevel <= vsmGetMortonNum) then
     call vsmLogPrint("Morton num is strange:", vsmGetMortonNum)
     call vsmError("ends")
   end if
 end function vsmGetMortonNum

 subroutine vsmGetSkipVertex(node, r_skip)
   !integer, dimension(3), intent(out) :: minvtx, maxvtx
   !type(vismo_octnode), intent(in) :: node
   integer, intent(in) :: node
   logical, intent(out) :: r_skip
   integer :: parent

   if(vsmOctant(node)%skip) then
     parent = vsmOctant(node)%parent
     vsmOctant(node)%SkipVertexMin(1:3) = vsmOctant(parent)%SkipVertexMin(1:3)
     vsmOctant(node)%SkipVertexMax(1:3) = vsmOctant(parent)%SkipVertexMax(1:3)
     r_skip = .true.
   else
     r_skip = .false.
   end if
 end subroutine vsmGetSkipVertex

 subroutine vsmRaySkip(pos, dirc, node, ray_, r_skip)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp), dimension(3), intent(in) :: dirc
   integer, intent(in) :: node
   real(kind=vdp), intent(out) :: ray_
   logical, intent(inout) :: r_skip

   real(kind=vdp), dimension(6) :: a
   real(kind=vdp) :: b

   ray_ = 0.0_vdp; 
   r_skip = .false.;
   a=-1.0_vdp; 
   b=0.0_vdp

   if(.not. vsmOctant(node)%skip) then
     r_skip = .false.
     return
   end if

   if(.not. vsmIsInsideNode(pos, node)) then
     r_skip = .false.
     return
   end if

   if(dirc(1) /= 0.0_vdp) then
     a(1) = (vsmOctant(node)%SkipMin(1)-pos(1))/dirc(1)
     a(2) = (vsmOctant(node)%SkipMax(1)-pos(1))/dirc(1)
   end if
  
   if(dirc(2) /= 0.0_vdp) then
     a(3) = (vsmOctant(node)%SkipMin(2)-pos(2))/dirc(2)
     a(4) = (vsmOctant(node)%SkipMax(2)-pos(2))/dirc(2)
   end if

   if(dirc(3) /= 0.0_vdp) then
     a(5) = (vsmOctant(node)%SkipMin(3)-pos(3))/dirc(3)
     a(6) = (vsmOctant(node)%SkipMax(3)-pos(3))/dirc(3)
   end if

   where(a < 0.0_vdp)
     a = vsmFaraway
   end where

   b = minval(a)
   ray_ = vsmIntval%ray_interval * int(b/vsmIntval%ray_interval)
   if(int(b/vsmIntval%ray_interval) > 0 .and. int(b/vsmIntval%ray_interval) == b/vsmIntval%ray_interval) then
!   if(int(b/vsmIntval%ray_interval) > 0) then
     ray_ = vsmIntval%ray_interval * (int(b/vsmIntval%ray_interval)-1)
   end if
   r_skip = .true. 
   if(ray_ < 0.0_vdp) then
     if(vsmDbg) then
       call vsmLogPrint("a=",a)
       call vsmLogPrint("minval(a)=",minval(a))
      !call vsmLogPrint("b=",b)
      !call vsmLogPrint("intval=",vsmIntval%ray_interval)
      !call vsmLogPrint("real=",b/vsmIntval%ray_interval)
      !call vsmLogPrint("int=",int(b/vsmIntval%ray_interval))
       call vsmLogPrint("skipmin",vsmOctant(node)%skipmin)
       call vsmLogPrint("skipmax",vsmOctant(node)%skipmax)
     end if
     call vsmError("ray_ is negative")
   end if
 end subroutine vsmRaySkip

 subroutine vsmRaySkipI(pos, dirc, node, jn, r_skip)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp), dimension(3), intent(in) :: dirc
   integer, intent(in) :: node
   integer, intent(inout) :: jn
   logical, intent(inout) :: r_skip

   real(kind=vdp), dimension(6) :: a
   real(kind=vdp) :: b, db

   r_skip = .false.
   a=-1.0_vdp; 
   b=0.0_vdp

   if(.not. vsmOctant(node)%skip) then
     r_skip = .false.
     return
   end if

   if(.not. vsmIsInsideNode(pos, node)) then
     r_skip = .false.
     return
   end if

   if(dirc(1) /= 0.0_vdp) then
     a(1) = (vsmOctant(node)%SkipMin(1)-pos(1))/dirc(1)
     a(2) = (vsmOctant(node)%SkipMax(1)-pos(1))/dirc(1)
   end if
  
   if(dirc(2) /= 0.0_vdp) then
     a(3) = (vsmOctant(node)%SkipMin(2)-pos(2))/dirc(2)
     a(4) = (vsmOctant(node)%SkipMax(2)-pos(2))/dirc(2)
   end if

   if(dirc(3) /= 0.0_vdp) then
     a(5) = (vsmOctant(node)%SkipMin(3)-pos(3))/dirc(3)
     a(6) = (vsmOctant(node)%SkipMax(3)-pos(3))/dirc(3)
   end if

   where(a < 0.0_vdp)
     a = vsmFaraway
   end where

   b = minval(a)
   jn = int(b/vsmIntval%ray_interval)
   db = b/vsmIntval%ray_interval - jn
   if(jn > 0 .and. db == 0.0_vdp) then
     jn = jn - 1
   end if
   r_skip = .true. 
   if(jn < 0) then
     if(vsmDbg) then
 !      ray_ = vsmIntval%ray_interval * jn
       call vsmLogPrint("a=",a)
       call vsmLogPrint("minval(a)=",minval(a))
      !call vsmLogPrint("b=",b)
      !call vsmLogPrint("intval=",vsmIntval%ray_interval)
      !call vsmLogPrint("real=",b/vsmIntval%ray_interval)
      !call vsmLogPrint("int=",int(b/vsmIntval%ray_interval))
       call vsmLogPrint("skipmin",vsmOctant(node)%skipmin)
       call vsmLogPrint("skipmax",vsmOctant(node)%skipmax)
     end if
     call vsmError("ray_ is negative")
   end if
 end subroutine vsmRaySkipI

 function vsmIsInsideNode(pos, node) !is_rayin
  logical :: vsmIsInsideNode
  integer, intent(in) :: node
  real (kind=vdp), dimension(3), intent(in) :: pos

  if(pos(1) >= vsmOctant(node)%txmax .or. pos(1) < vsmOctant(node)%txmin .or. pos(2) >= vsmOctant(node)%tymax .or. &
  pos(2) < vsmOctant(node)%tymin .or. pos(3) >= vsmOctant(node)%tzmax .or. pos(3) < vsmOctant(node)%tzmin) then
    vsmIsInsideNode = .false.
  else 
    vsmIsInsideNode = .true.
  end if
 end function vsmIsInsideNode

 subroutine vsmCalcCoord
  integer :: numRow, num_sv, tsv
  integer :: i, nd
  integer :: bx, by, bz
  real(kind=vdp), dimension(3) :: gmin, gmax

  call vsmLogPrint("vsmCalcCoord is called")
  if(vsmOctComDat%maxLevel == 0) then 
    vsmOctComDat%coordx(0) = vsmOctant(0)%txmin
    vsmOctComDat%coordy(0) = vsmOctant(0)%tymin
    vsmOctComDat%coordz(0) = vsmOctant(0)%tzmin

    vsmOctComDat%coordx(1) = vsmOctant(0)%txmax
    vsmOctComDat%coordy(1) = vsmOctant(0)%tymax
    vsmOctComDat%coordz(1) = vsmOctant(0)%tzmax
    return
  end if
  
!  num_sv = 1; tsv = 1
!  do i=1, vsmOctComDat%maxLevel-1
!   tsv = ishft(tsv, 3)
!   num_sv = num_sv + tsv
!  end do

  if(vsmOctComDat%maxLevel > 0) then
    num_sv = vsmOctNum(vsmOctComDat%maxLevel-1)
  else
    num_sv = 0
  end if

  call vsmLogPrint("numsv: ",num_sv)
!  num_sv = num_sv - 1

  numRow = vsmOctComDat%Row !ishft(a, vsmOctComDat%maxLevel)
  call vsmLogPrint("numRow: ",numRow)


  vsmOctComDat%coordx(0) = vsmOctant(0)%txmin
  vsmOctComDat%coordy(0) = vsmOctant(0)%tymin
  vsmOctComDat%coordz(0) = vsmOctant(0)%tzmin

  do i=0, numRow - 1

   bx = vsmGetMortonNum(i, 0, 0)
   call vsmLogPrint("mortonum i,0,0 ",bx)
   nd = num_sv+bx
   call vsmGetGMinMax(gmin, gmax, nd)
   call vsmLogPrint("minmax i,0,0 ")
   vsmOctComDat%coordx(i+1) = gmax(1)

   by = vsmGetMortonNum(0, i, 0)
   call vsmLogPrint("mortonum 0,i,0 ",by)
   nd = num_sv+by
   call vsmGetGMinMax(gmin, gmax, nd)
   call vsmLogPrint("minmax 0,i,0 ")
   vsmOctComDat%coordy(i+1) = gmax(2)
   
   bz = vsmGetMortonNum(0, 0, i)
   call vsmLogPrint("mortonum 0,0,i ",bz)
   nd = num_sv+bz
   call vsmGetGMinMax(gmin, gmax, nd)
   call vsmLogPrint("minmax 0,0,i ")
   vsmOctComDat%coordz(i+1) = gmax(3)

   if(vsmDbg) then
     call vsmLogPrint("Row ", i)
     call vsmLogPrint("bx:", bx)
     call vsmLogPrint("gmax(1):", gmax(1))
     call vsmLogPrint("by:", by)
     call vsmLogPrint("gmax(2):", gmax(2))
     call vsmLogPrint("bz:", bz)
     call vsmLogPrint("gmax(3):", gmax(3))
   end if
  end do

  call vsmLogPrint("vsmCalcCoord ends")
 end subroutine vsmCalcCoord
 
end module

