module vsmOctree4Iso
  use vsmConst
  use vsmFunction
  use vsmColormap
  use vsmData
  use vsmCoordinates
  use vsmParam

  implicit none
  private

type vismo_OctNode4Iso
  integer, dimension(3) :: vertex
  integer, dimension(3) :: OctDatSize
  !morton number  
  integer :: parent
  integer :: moi
  integer, dimension(0:7) :: children
  integer :: MyNumber ! morton number

  integer :: level ! level of octree
  logical :: skip
  real(kind=vdp) :: txmin, tymin, tzmin, txmax, tymax, tzmax
  integer, dimension(3) :: SkipVertexMin, SkipVertexMax
  real(kind=vdp), dimension(3) ::  SkipMin, SkipMax
 end type vismo_OctNode4Iso

 type vismo_OctCommon4Iso
  logical :: uniform
  integer :: maxLevel
  integer :: vis_num, Row
  real(kind=vdp) :: xmin, ymin, zmin
  real(kind=vdp), dimension(:), allocatable :: coordx, coordy, coordz
  logical, dimension(3) :: vis_on ! 1:isosurface, 2:volume rendering, 3:slices
 end type vismo_OctCommon4Iso

 type(vismo_OctNode4Iso), dimension(0:vsmMaxOctNum), save :: vsmOctant4Iso
 type(vismo_OctCommon4Iso), save :: vsmOctComDat4Iso

 public vsmInitOctree4iso, vsmFinalizeOctree4iso, vsmOctDivide4iso, vsmOctPrepare4iso, vsmRayJump4iso
 public vsmSetOctComDat4iso, vsmUpdateOctree4iso

 contains

 subroutine vsmInitOctree4Iso
  integer :: i

  do i=0, vsmMaxOctNum-1
    vsmOctant4Iso(i)%children(0:7) = -1

    vsmOctant4Iso(i)%parent = -1
    vsmOctant4Iso(i)%MyNumber = -1
    vsmOctant4Iso(i)%skip = .false.

    vsmOctant4Iso(i)%vertex(1:3) = -1
    vsmOctant4Iso(i)%SkipVertexMin(1:3) = -1
    vsmOctant4Iso(i)%SkipVertexMax(1:3) = -1
    vsmOctant4Iso(i)%SkipMin(1:3) = -1.0d+10
    vsmOctant4Iso(i)%SkipMax(1:3) = -1.0d+10

    vsmOctant4Iso(i)%txmin = 1.0d+10;  vsmOctant4Iso(i)%txmax = -1.0d+10
    vsmOctant4Iso(i)%tymin = 1.0d+10;  vsmOctant4Iso(i)%tymax = -1.0d+10
    vsmOctant4Iso(i)%tzmin = 1.0d+10;  vsmOctant4Iso(i)%tzmax = -1.0d+10
  end do

  vsmOctComDat4Iso%vis_num      = 0
  vsmOctComDat4Iso%uniform      = .true.
  vsmOctComDat4Iso%maxLevel     = vsmMaxOctLevel - 1
  vsmOctComDat4Iso%vis_on(1:3)  = .false.

  vsmOctComDat4Iso%xmin = 1.0d+10
  vsmOctComDat4Iso%ymin = 1.0d+10
  vsmOctComDat4Iso%zmin = 1.0d+10

  vsmOctComDat4Iso%Row  = 0

  call vsmLogPrint("Octree: vsmOctComDat4Iso%maxLevel", vsmOctComDat4Iso%maxLevel)
  call vsmLogPrint("Octree Initialized")
 end subroutine vsmInitOctree4Iso

 subroutine vsmSetOctComDat4Iso(vn)
   integer, intent(in) :: vn

   vsmOctComDat4Iso%vis_num = vn

   if(vsmVisParam(vn)%iso_num > 0 .and. vsmVisParam(vn)%iso_solid) then
     vsmOctComDat4Iso%vis_on(1) = .true.
   else
     vsmOctComDat4Iso%vis_on(1) = .false.
   end if

   if(vsmVisParam(vn)%volren_num > 0 ) then
     vsmOctComDat4Iso%vis_on(2) = .true.
   else
     vsmOctComDat4Iso%vis_on(2) = .false.
   end if

   if(vsmVisParam(vn)%slice_num > 0 ) then
     vsmOctComDat4Iso%vis_on(3) = .true.
   else
     vsmOctComDat4Iso%vis_on(3) = .false.
   end if

 end subroutine vsmSetOctComDat4Iso

 subroutine vsmFinalizeOctree4Iso
   if(allocated(vsmOctComDat4Iso%coordx)) deallocate(vsmOctComDat4Iso%coordx)
   if(allocated(vsmOctComDat4Iso%coordy)) deallocate(vsmOctComDat4Iso%coordy)
   if(allocated(vsmOctComDat4Iso%coordz)) deallocate(vsmOctComDat4Iso%coordz)
 end subroutine vsmFinalizeOctree4Iso

 subroutine vsmRayJump4Iso(pos, dirc, ray_, r_skip)
  real(kind=vdp), dimension(3), intent(in) :: pos, dirc
  real(kind=vdp), intent(out) :: ray_
  logical, intent(out) :: r_skip
  integer :: i, nd
  integer :: div_num, integ_m, count_
  integer :: intx, inty, intz
  integer :: mnumray

  !call vsmLogPrint("Octree's vsmRayJump is called")
  intx = -1; inty = -1; intz = -1
  r_skip = .false.

  if(.not.vsmIsInsidelWO(pos)) then
    call vsmLogPrint("RayJump: out of data:", pos)
    call vsmError("RayJump: input pos is strange")
  end if

  div_num = vsmOctComDat4Iso%Row
  !div_num = ishft(div_num, vsmOctComDat4Iso%maxLevel)

!  integ_m = 1; count_ = 1
!  do i=1, vsmOctComDat4Iso%maxLevel-1
!    count_ = count_ * 8
!    integ_m = integ_m + count_
!  end do

  if(vsmOctComDat4Iso%coordx(div_num) == pos(1)) then
    intx = div_num-1 
  else
    do i=div_num-1, 0, -1
      if(vsmOctComDat4Iso%coordx(i) <= pos(1)) then
        intx = i
        exit
      end if
    end do
  end if

  if(vsmOctComDat4Iso%coordy(div_num) == pos(2)) then
    inty = div_num-1 
  else
    do i=div_num-1, 0, -1
      if(vsmOctComDat4Iso%coordy(i) <= pos(2)) then
        inty = i
        exit
      end if
    end do
  end if

  if(vsmOctComDat4Iso%coordz(div_num) == pos(3)) then
    intz = div_num-1 
  else
    do i=div_num-1, 0, -1
      if(vsmOctComDat4Iso%coordz(i) <= pos(3)) then
        intz = i
        exit
      end if
    end do
  end if

  if(intx == -1 .or. inty == -1 .or. intz == -1) then
    if(vsmDbg) then
      do i=0, div_num
        call vsmLogPrint("Octree:i=", i)
        call vsmLogPrint("Octree:coordx ", vsmOctComDat4Iso%coordx(i))
        call vsmLogPrint("Octree:coordy ", vsmOctComDat4Iso%coordy(i))
        call vsmLogPrint("Octree:coordz ", vsmOctComDat4Iso%coordz(i))
      end do
    end if
    call vsmError("RayJump: intx,y or z is negative")
  end if

  if(vsmOctComDat4Iso%maxLevel > 0) then
    integ_m = vsmOctNum(vsmOctComDat4Iso%maxLevel-1)
  else
    integ_m = 0
  end if
  mnumray =  vsmGetMortonNum4Iso(intx, inty, intz)
  nd = mnumray+integ_m

  if(vsmMaxOctNum <= nd ) call vsmError("OctNum: vsmRayJump")
  call vsmRaySkip4Iso(pos, dirc, nd, ray_, r_skip)

  !call vsmLogPrint("Octree's vsmRayJump ended")
 end subroutine vsmRayJump4Iso

 function vsmIsInsideOctant4iso(pos, on)
  real(kind=vdp), dimension(3), intent(in) :: pos
  integer, intent(in) :: on
  logical :: vsmIsInsideOctant4iso

  if(vsmOctant4iso(on)%txmin <= pos(1) .and. vsmOctant4iso(on)%txmax > pos(1) .and. &
    vsmOctant4iso(on)%tymin <= pos(2) .and. vsmOctant4iso(on)%tymax > pos(2) .and. &
    vsmOctant4iso(on)%tzmin <= pos(3) .and. vsmOctant4iso(on)%tzmax > pos(3)) then
    vsmIsInsideOctant4iso = .true.
  else
    vsmIsInsideOctant4iso = .false.
  end if

 end function vsmIsInsideOctant4iso

 recursive subroutine vsmUpdateOctree4Iso(node)
   integer, intent(in) :: node
   logical :: r_skip
   integer :: i

!   call vsmLogPrint("Octree: Update Octree is called")
   vsmOctant4Iso(node)%skip = .false.

   if(vsmOctant4Iso(vsmOctant4Iso(node)%parent)%skip) then
     !the parent of root is root itself. skip is always false
     vsmOctant4Iso(node)%skip = .true.
     call vsmGetSkipVertex4Iso(node, r_skip)

       if(vsmOctant4Iso(node)%moi == 0) then
         call vsmLogPrint("Octree: in node%parent%skip root ", vsmOctant4Iso(node)%skip)
       end if
   else if(SurveyData4Iso(node) == 0) then

     vsmOctant4Iso(node)%skip = .true.
     vsmOctant4Iso(node)%SkipVertexMin(1:3) = vsmOctant4Iso(node)%vertex(1:3)
     vsmOctant4Iso(node)%SkipVertexMax(1:3) = vsmOctant4Iso(node)%vertex(1:3) + vsmOctant4Iso(node)%OctDatSize(1:3)-1

     if(vsmOctant4Iso(node)%moi == 0) then
       call vsmLogPrint("Octree: in SurveyData root ", vsmOctant4Iso(node)%skip)
     end if

   end if

   if(vsmOctant4Iso(node)%moi == 0) call vsmLogPrint("Octree: root ", vsmOctant4Iso(node)%skip)

   if(vsmOctant4Iso(node)%skip) then

     if(vsmOctComDat4Iso%uniform) then
       vsmOctant4Iso(node)%SkipMin(1) = vsmOctComDat4Iso%xmin + vsmLocalCoord%dx(1)*(vsmOctant4Iso(node)%SkipVertexMin(1)-1)
       vsmOctant4Iso(node)%SkipMax(1) = vsmOctComDat4Iso%xmin + vsmLocalCoord%dx(1)*(vsmOctant4Iso(node)%SkipVertexMax(1)-1)
       vsmOctant4Iso(node)%SkipMin(2) = vsmOctComDat4Iso%ymin + vsmLocalCoord%dx(2)*(vsmOctant4Iso(node)%SkipVertexMin(2)-1)
       vsmOctant4Iso(node)%SkipMax(2) = vsmOctComDat4Iso%ymin + vsmLocalCoord%dx(2)*(vsmOctant4Iso(node)%SkipVertexMax(2)-1)
       vsmOctant4Iso(node)%SkipMin(3) = vsmOctComDat4Iso%zmin + vsmLocalCoord%dx(3)*(vsmOctant4Iso(node)%SkipVertexMin(3)-1)
       vsmOctant4Iso(node)%SkipMax(3) = vsmOctComDat4Iso%zmin + vsmLocalCoord%dx(3)*(vsmOctant4Iso(node)%SkipVertexMax(3)-1)
     else
       vsmOctant4Iso(node)%SkipMin(1) = vsmLocalCoord%coordx(vsmOctant4Iso(node)%SkipVertexMin(1))
       vsmOctant4Iso(node)%SkipMax(1) = vsmLocalCoord%coordx(vsmOctant4Iso(node)%SkipVertexMax(1))
       vsmOctant4Iso(node)%SkipMin(2) = vsmLocalCoord%coordy(vsmOctant4Iso(node)%SkipVertexMin(2))
       vsmOctant4Iso(node)%SkipMax(2) = vsmLocalCoord%coordy(vsmOctant4Iso(node)%SkipVertexMax(2))
       vsmOctant4Iso(node)%SkipMin(3) = vsmLocalCoord%coordz(vsmOctant4Iso(node)%SkipVertexMin(3))
       vsmOctant4Iso(node)%SkipMax(3) = vsmLocalCoord%coordz(vsmOctant4Iso(node)%SkipVertexMax(3))
     end if

  else
    vsmOctant4Iso(node)%SkipMin(1:3) = -1.0d+10
    vsmOctant4Iso(node)%SkipMax(1:3) = -1.0d+10
  end if 
 
  if(vsmOctant4Iso(node)%level == vsmOctComDat4Iso%maxLevel) then
    !call vsmLogPrint("Level == maxLevel: vsmOctant4Iso(node)%level", vsmOctant4Iso(node)%level)
    !call vsmLogPrint("Level == maxLevel: vsmOctComDat4Iso%maxLevel", vsmOctComDat4Iso%maxLevel)
    !call vsmLogPrint("Octree: Update Octree ends", vsmOctant4Iso(node)%moi)
    return
  end if

  do i=0, 7
    call vsmUpdateOctree4Iso(vsmOctant4Iso(node)%children(i))
  end do

  !call vsmLogPrint("Octree: Update Octree ends", node%moi)
 end subroutine vsmUpdateOctree4Iso

 subroutine vsmOctPrepare4Iso
  integer, dimension(3) :: vtx

  call vsmSetGrid4Iso
  vtx(1:3) = 1
  call vsmSetVertex4Iso(vtx, vsmLocalCoord%nx, 0)
 end subroutine vsmOctPrepare4Iso

 recursive subroutine vsmOctDivide4Iso(node)
  integer, intent(in) :: node
  logical :: root

  integer :: n, nx, ny, nz, b
  integer :: i, allocstat
  logical :: r_skip
  integer, dimension(3) :: hsize, hsizep1
  integer, dimension(3, 0:7) :: child_vertex, child_size
  integer :: next_level
  integer :: Mnum
  integer :: integ_num, tinteg_num

  if(node == 0) then
    root = .true.
  else
    root = .false.
  end if

  if(root) then
    call vsmLogPrint("vsmOctDivide is called")
   end if

  call vsmCalcMinMax4Iso(node)

  LEVEL0 : if(root) then
    vsmOctant4Iso(node)%level  = 0
    vsmOctant4Iso(node)%parent = 0
    nx = vsmOctant4Iso(node)%OctDatSize(1) - 1 ! num of control volume
    ny = vsmOctant4Iso(node)%OctDatSize(2) - 1
    nz = vsmOctant4Iso(node)%OctDatSize(3) - 1
    n  = vsmMaxOctLevel

    do i=1, vsmMaxOctLevel
      nx = nx/2; ny = ny/2; nz = nz/2
   
      if(vsmDbg) then
        call vsmLogPrint("nx = ",nx)
        call vsmLogPrint("ny = ",ny)
        call vsmLogPrint("nz = ",nz)
      end if

      if(nx < 1 .or. ny < 1 .or. nz < 1) exit
   
      n=n-1

      if(nx == 1 .or. ny == 1 .or. nz == 1) exit

    end do

    vsmOctComDat4Iso%maxLevel = vsmMaxOctLevel - n
    call vsmLogPrint("Octree: MaxLevel = ",vsmOctComDat4Iso%maxLevel)

    b = 1
    b = ishft(b, vsmOctComDat4Iso%maxLevel)
  
    call vsmLogPrint("Octree: Row = ",b)
    vsmOctComDat4Iso%Row = b
    allocate(vsmOctComDat4Iso%coordx(0:b), vsmOctComDat4Iso%coordy(0:b), vsmOctComDat4Iso%coordz(0:b), stat = allocstat)
    if(allocstat /= 0) call vsmError('cant allocate octree coord')

    vsmOctant4Iso(node)%moi = 0
    vsmOctant4Iso(node)%MyNumber = 0

  end if LEVEL0
 
  if(vsmOctant4Iso(node)%level == vsmOctComDat4Iso%maxLevel) then
    ! call vsmLogPrint("vsmOctant4Iso(node)level==vsmOctComDat4Iso%maxLevel:", vsmOctant4Iso(node)%level)
    return
  end if

  hsize(1:3) = vsmOctant4Iso(node)%OctDatSize(1:3)/2
  hsizep1(1:3) = hsize(1:3) + 1;
  
! child 0
  child_vertex(1:3, 0) = vsmOctant4Iso(node)%Vertex(1:3)
  child_size(1:3, 0) = hsizep1(1:3)

! child 1
  child_vertex(1,1) = vsmOctant4Iso(node)%Vertex(1) + hsize(1)
  child_vertex(2,1) = vsmOctant4Iso(node)%Vertex(2)
  child_vertex(3,1) = vsmOctant4Iso(node)%Vertex(3)
  child_size(1,1) = vsmOctant4Iso(node)%OctDatSize(1) - hsize(1)
  child_size(2,1) = hsizep1(2)
  child_size(3,1) = hsizep1(3)

! child 2
  child_vertex(1,2) = vsmOctant4Iso(node)%Vertex(1)
  child_vertex(2,2) = vsmOctant4Iso(node)%Vertex(2) + hsize(2)
  child_vertex(3,2) = vsmOctant4Iso(node)%Vertex(3)
  child_size(1,2) = hsizep1(1)
  child_size(2,2) = vsmOctant4Iso(node)%OctDatSize(2) - hsize(2)
  child_size(3,2) = hsizep1(3)

! child 3
  child_vertex(1,3) = vsmOctant4Iso(node)%Vertex(1) + hsize(1)
  child_vertex(2,3) = vsmOctant4Iso(node)%Vertex(2) + hsize(2)
  child_vertex(3,3) = vsmOctant4Iso(node)%Vertex(3)
  child_size(1,3) = vsmOctant4Iso(node)%OctDatSize(1) - hsize(1)
  child_size(2,3) = vsmOctant4Iso(node)%OctDatSize(2) - hsize(2)
  child_size(3,3) = hsizep1(3)

! child 4
  child_vertex(1, 4) = vsmOctant4Iso(node)%Vertex(1)
  child_vertex(2, 4) = vsmOctant4Iso(node)%Vertex(2)
  child_vertex(3, 4) = vsmOctant4Iso(node)%Vertex(3) + hsize(3)
  child_size(1, 4) = hsizep1(1)
  child_size(2, 4) = hsizep1(2)
  child_size(3, 4) = vsmOctant4Iso(node)%OctDatSize(3) - hsize(3)

! child 5
  child_vertex(1, 5) = vsmOctant4Iso(node)%Vertex(1) + hsize(1)
  child_vertex(2, 5) = vsmOctant4Iso(node)%Vertex(2)
  child_vertex(3, 5) = vsmOctant4Iso(node)%Vertex(3) + hsize(3)
  child_size(1, 5) = vsmOctant4Iso(node)%OctDatSize(1) - hsize(1)
  child_size(2, 5) = hsizep1(2)
  child_size(3, 5) = vsmOctant4Iso(node)%OctDatSize(3) - hsize(3)

! child 6
  child_vertex(1, 6) = vsmOctant4Iso(node)%Vertex(1)
  child_vertex(2, 6) = vsmOctant4Iso(node)%Vertex(2) + hsize(2)
  child_vertex(3, 6) = vsmOctant4Iso(node)%Vertex(3) + hsize(3)
  child_size(1, 6) = hsizep1(1)
  child_size(2, 6) = vsmOctant4Iso(node)%OctDatSize(2) - hsize(2)
  child_size(3, 6) = vsmOctant4Iso(node)%OctDatSize(3) - hsize(3)

! child 7
  child_vertex(1, 7) = vsmOctant4Iso(node)%Vertex(1) + hsize(1)
  child_vertex(2, 7) = vsmOctant4Iso(node)%Vertex(2) + hsize(2)
  child_vertex(3, 7) = vsmOctant4Iso(node)%Vertex(3) + hsize(3)
  child_size(1, 7) = vsmOctant4Iso(node)%OctDatSize(1) - hsize(1)
  child_size(2, 7) = vsmOctant4Iso(node)%OctDatSize(2) - hsize(2)
  child_size(3, 7) = vsmOctant4Iso(node)%OctDatSize(3) - hsize(3)

  next_level = vsmOctant4Iso(node)%level + 1
  Mnum = vsmOctant4Iso(node)%MyNumber*8

!  integ_num = 1; tinteg_num = 1
!  do i=1, vsmOctant4Iso(node)%level
!    tinteg_num = tinteg_num * 8
!    integ_num = integ_num + tinteg_num
!  end do

  integ_num = vsmOctNum(vsmOctant4Iso(node)%level)

  do i=0, 7
    vsmOctant4Iso(node)%children(i) = integ_num + Mnum + i
    vsmOctant4Iso(vsmOctant4Iso(node)%children(i))%level = next_level
    vsmOctant4Iso(vsmOctant4Iso(node)%children(i))%MyNumber = Mnum+i
  end do

  !call vsmLogPrint("moi : ", vsmOctant4Iso(node)%moi)
  !call vsmLogPrint("children's numbers: ", vsmOctant4Iso(node)%children)
 
  do i=0, 7
    if(vsmOctant4Iso(node)%children(i) >= vsmMaxOctNum) then
      call vsmLogPrint("vsmOctant4Iso(node)%children(i) >= vsmMaxOctNum:", vsmOctant4Iso(node)%children(i))
      call vsmError("vsmOctant4Iso(node) number is strange")
    end if
    vsmOctant4Iso(vsmOctant4Iso(node)%children(i))%moi = vsmOctant4Iso(node)%children(i)
    vsmOctant4Iso(vsmOctant4Iso(node)%children(i))%parent = vsmOctant4Iso(node)%moi
    call vsmSetVertex4Iso(child_vertex(1:3, i), child_size(1:3, i), vsmOctant4Iso(node)%children(i))
    !call vsmLogPrint("call vsmOctDivide: ", vsmOctant4Iso(node)%children(i))
    call vsmOctDivide4Iso(vsmOctant4Iso(node)%children(i))
  end do

  if(vsmOctant4Iso(node)%level == 0) then
    call vsmLogPrint("call vsmCalcCoord")
    call vsmCalcCoord4Iso
  end if

  !call vsmLogPrint("vsmOctDivide ends:", vsmOctant4Iso(node)%moi)
 end subroutine vsmOctDivide4Iso

 subroutine vsmSetGrid4Iso
  vsmOctComDat4Iso%uniform = vsmLocalCoord%uniform
  vsmOctComDat4Iso%xmin = vsmLocalCoord%origin(1)
  vsmOctComDat4Iso%ymin = vsmLocalCoord%origin(2)
  vsmOctComDat4Iso%zmin = vsmLocalCoord%origin(3)
 end subroutine vsmSetGrid4Iso 

 subroutine vsmSetVertex4Iso(vx, sz, node)
   integer, dimension(3), intent(in) :: vx, sz
   integer, intent(in) :: node

   vsmOctant4Iso(node)%vertex(1:3) = vx(1:3)
   vsmOctant4Iso(node)%OctDatSize(1:3) = sz(1:3)
   !call vsmLogPrint("Octree:sz", sz)
 end subroutine vsmSetVertex4Iso

 function SurveyData4Iso(node)
  integer, intent(in) :: node
  integer ::  SurveyData4Iso
  integer :: i, j, k, nscal, n, vis_num, ip1,jp1,kp1
  real(kind=vdp) :: iso_level, max_val, min_val
  real(kind=vdp), dimension(8) :: dat

  !call vsmLogPrint("Octree: survey data is called, level = ",node%level)
  vis_num = vsmOctComDat4Iso%vis_num

  !call vsmLogPrint("vsnum",vis_num)
  if(vis_num < 1) call vsmError("survayData visnum")

  SurveyData4Iso = 0
  if(vsmOctComDat4Iso%vis_on(1)) then ! isosurface
    if(vsmOctant4Iso(node)%Vertex(1)<1 .or. vsmOctant4Iso(node)%Vertex(2)<1 .or. &
       vsmOctant4Iso(node)%Vertex(3)<1) call vsmError("i or j or k<1")

    if(vsmOctant4Iso(node)%Vertex(1)+vsmOctant4Iso(node)%OctDatSize(1)-1 > vsmLocalCoord%nx(1) &
      .or. vsmOctant4Iso(node)%Vertex(2)+vsmOctant4Iso(node)%OctDatSize(2)-1 > vsmLocalCoord%nx(2) &
      .or. vsmOctant4Iso(node)%Vertex(3)+vsmOctant4Iso(node)%OctDatSize(3)-1 > vsmLocalCoord%nx(3)) then
       call vsmError("i or j or k too big")
    end if

    do n=1, vsmVisParam(vis_num)%iso_num
      if(vsmVisParam(vis_num)%isoparam(n)%alpha < vsmErtValue) cycle
      nscal = vsmVisParam(vis_num)%isoparam(n)%nscal
      iso_level = vsmVisParam(vis_num)%isoparam(n)%iso_level
      do k=vsmOctant4Iso(node)%Vertex(3), vsmOctant4Iso(node)%Vertex(3)+vsmOctant4Iso(node)%OctDatSize(3)-2
!$omp parallel do private(i,j,ip1,jp1,kp1,dat,max_val,min_val)
        do j=vsmOctant4Iso(node)%Vertex(2), vsmOctant4Iso(node)%Vertex(2)+vsmOctant4Iso(node)%OctDatSize(2)-2
          do i=vsmOctant4Iso(node)%Vertex(1), vsmOctant4Iso(node)%Vertex(1)+vsmOctant4Iso(node)%OctDatSize(1)-2
            if(SurveyData4Iso == 0) then
               ip1 = i+1; jp1 = j+1; kp1 = k+1
               dat(1) = vsmScalData(nscal)%scal(i,   j,   k)
               dat(2) = vsmScalData(nscal)%scal(ip1, j,   k)
               dat(3) = vsmScalData(nscal)%scal(i,   jp1, k)
               dat(4) = vsmScalData(nscal)%scal(ip1, jp1, k)
               dat(5) = vsmScalData(nscal)%scal(i,   j,   kp1)
               dat(6) = vsmScalData(nscal)%scal(ip1, j,   kp1)
               dat(7) = vsmScalData(nscal)%scal(i,   jp1, kp1)
               dat(8) = vsmScalData(nscal)%scal(ip1, jp1, kp1)

               max_val = maxval(dat)
               min_val = minval(dat)

               if(min_val <= iso_level .and. max_val >= iso_level) then
                  SurveyData4Iso = 1
                  call vsmLogPrint("SurveyData: Isosurface")
               end if
            end if

          end do !i
        end do !j
      if(SurveyData4Iso > 0) exit
      end do !k
      if(SurveyData4Iso > 0) exit
    end do
   !call vsmLogPrint("survey data for isosurface ends")
  end if

  !call vsmLogPrint("Octree: survey data ends")
 end function SurveyData4Iso

 subroutine vsmGetGMinMax4Iso(gmin, gmax, node)
  real(kind=vdp), dimension(3), intent(out) :: gmin, gmax
  integer, intent(in) :: node

  gmin(1) = vsmOctant4Iso(node)%txmin; gmin(2) = vsmOctant4Iso(node)%tymin; gmin(3) = vsmOctant4Iso(node)%tzmin
  gmax(1) = vsmOctant4Iso(node)%txmax; gmax(2) = vsmOctant4Iso(node)%tymax; gmax(3) = vsmOctant4Iso(node)%tzmax

 end subroutine vsmGetGMinMax4Iso

 subroutine vsmCalcMinMax4Iso(node)
   integer, intent(in) :: node
   real(kind=vdp), dimension(3) :: temp

   if(vsmOctComDat4Iso%uniform) then
      vsmOctant4Iso(node)%txmin = vsmOctComDat4Iso%xmin + vsmLocalCoord%dx(1)*(vsmOctant4Iso(node)%Vertex(1)-1)
      vsmOctant4Iso(node)%txmax = vsmOctant4Iso(node)%txmin + vsmLocalCoord%dx(1)*(vsmOctant4Iso(node)%OctDatSize(1)-1)

      vsmOctant4Iso(node)%tymin = vsmOctComDat4Iso%ymin + vsmLocalCoord%dx(2)*(vsmOctant4Iso(node)%Vertex(2)-1)
      vsmOctant4Iso(node)%tymax = vsmOctant4Iso(node)%tymin + vsmLocalCoord%dx(2)*(vsmOctant4Iso(node)%OctDatSize(2)-1)

      vsmOctant4Iso(node)%tzmin = vsmOctComDat4Iso%zmin + vsmLocalCoord%dx(3)*(vsmOctant4Iso(node)%Vertex(3)-1)
      vsmOctant4Iso(node)%tzmax = vsmOctant4Iso(node)%tzmin + vsmLocalCoord%dx(3)*(vsmOctant4Iso(node)%OctDatSize(3)-1)

      temp(1) = vsmOctant4Iso(node)%txmin; temp(2) = vsmOctant4Iso(node)%tymin; temp(3) = vsmOctant4Iso(node)%tzmin;
      if(.not.vsmisinsidelwo(temp)) then
         call vsmLogPrint("calc min max1:", temp)
      end if

      temp(1) = vsmOctant4Iso(node)%txmax; temp(2) = vsmOctant4Iso(node)%tymax; temp(3) = vsmOctant4Iso(node)%tzmax;
      if(.not.vsmisinsidelwo(temp)) then
         call vsmLogPrint("calc min max2:", temp)
      end if
    else
      vsmOctant4Iso(node)%txmin = vsmLocalCoord%coordx(vsmOctant4Iso(node)%Vertex(1))
      vsmOctant4Iso(node)%txmax = vsmLocalCoord%coordx(vsmOctant4Iso(node)%Vertex(1)+vsmOctant4Iso(node)%OctDatSize(1)-1)

      vsmOctant4Iso(node)%tymin = vsmLocalCoord%coordy(vsmOctant4Iso(node)%Vertex(2))
      vsmOctant4Iso(node)%tymax = vsmLocalCoord%coordy(vsmOctant4Iso(node)%Vertex(2)+vsmOctant4Iso(node)%OctDatSize(2)-1)

      vsmOctant4Iso(node)%tzmin = vsmLocalCoord%coordz(vsmOctant4Iso(node)%Vertex(3))
      vsmOctant4Iso(node)%tzmax = vsmLocalCoord%coordz(vsmOctant4Iso(node)%Vertex(3)+vsmOctant4Iso(node)%OctDatSize(3)-1)
   end if
 end subroutine vsmCalcMinMax4Iso

 function vsmGetMortonNum4Iso(x, y, z)
   integer, intent(in) :: x, y, z
   integer :: vsmGetMortonNum4Iso
   integer ::  n, ax, ay, az, bx, by, bz, b, i, di
   !call vsmLogPrint("maxlevel",vsmOctComDat4Iso%maxLevel)
   n=1 
   bx=0; by=0; bz=0
   do i=0, vsmOctComDat4Iso%maxLevel-1
     ax = iand(x, n)
     ay = iand(y, n)
     az = iand(z, n)

     di = 2*i

     ax = ishft(ax, di  )
     ay = ishft(ay, di+1)
     az = ishft(az, di+2)

     bx = ior(bx, ax)
     by = ior(by, ay)
     bz = ior(bz, az)

     n = ishft(n, 1)
   end do

   b = 0
   b = ior(b, bx)
   b = ior(b, by)
   vsmGetMortonNum4Iso = ior(b, bz)
  
!   vsmGetMortonNum4Iso = b
   if(vsmGetMortonNum4Iso < 0 .or. 8**vsmOctComDat4Iso%maxLevel <= vsmGetMortonNum4Iso) then
     call vsmLogPrint("Morton num is strange:", vsmGetMortonNum4Iso)
     call vsmError("ends")
   end if
 end function vsmGetMortonNum4Iso

 subroutine vsmGetSkipVertex4Iso(node, r_skip)
   !integer, dimension(3), intent(out) :: minvtx, maxvtx
   !type(vismo_octnode), intent(in) :: node
   integer, intent(in) :: node
   logical, intent(out) :: r_skip
   integer :: parent

   if(vsmOctant4Iso(node)%skip) then
     parent = vsmOctant4Iso(node)%parent
     vsmOctant4Iso(node)%SkipVertexMin(1:3) = vsmOctant4Iso(parent)%SkipVertexMin(1:3)
     vsmOctant4Iso(node)%SkipVertexMax(1:3) = vsmOctant4Iso(parent)%SkipVertexMax(1:3)
     r_skip = .true.
   else
     r_skip = .false.
   end if
 end subroutine vsmGetSkipVertex4Iso

 subroutine vsmRaySkip4Iso(pos, dirc, node, ray_, r_skip)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp), dimension(3), intent(in) :: dirc
   integer, intent(in) :: node
   real(kind=vdp), intent(out) :: ray_
   logical, intent(inout) :: r_skip
   integer :: rayi
   real(kind=vdp), dimension(6) :: a
   real(kind=vdp) :: b

   ray_ = 0.0_vdp; 
   r_skip = .false.;
   a=-1.0_vdp; 
   b=0.0_vdp

   if(.not. vsmOctant4Iso(node)%skip) then
     r_skip = .false.
     return
   end if

   if(.not. vsmIsInsideNode4Iso(pos, node)) then
     r_skip = .false.
     return
   end if

   if(dirc(1) /= 0.0_vdp) then
     a(1) = (vsmOctant4Iso(node)%SkipMin(1)-pos(1))/dirc(1)
     a(2) = (vsmOctant4Iso(node)%SkipMax(1)-pos(1))/dirc(1)
   end if
  
   if(dirc(2) /= 0.0_vdp) then
     a(3) = (vsmOctant4Iso(node)%SkipMin(2)-pos(2))/dirc(2)
     a(4) = (vsmOctant4Iso(node)%SkipMax(2)-pos(2))/dirc(2)
   end if

   if(dirc(3) /= 0.0_vdp) then
     a(5) = (vsmOctant4Iso(node)%SkipMin(3)-pos(3))/dirc(3)
     a(6) = (vsmOctant4Iso(node)%SkipMax(3)-pos(3))/dirc(3)
   end if

   where(a < 0.0_vdp)
     a = vsmFaraway
   end where

   b = minval(a)
   rayi = int(b/vsmIntval%ray_interval)
   if(rayi > 0 .and. b/vsmIntval%ray_interval - rayi == 0.0_vdp) then
     ray_ = vsmIntval%ray_interval * (rayi- 1)
     r_skip = .true. 
   else if(rayi == 0) then
     r_skip = .false.
   end if
   if(ray_ < 0.0_vdp) then
     if(vsmDbg) then
       call vsmLogPrint("a=",a)
       call vsmLogPrint("minval(a)=",minval(a))
      !call vsmLogPrint("b=",b)
      !call vsmLogPrint("intval=",vsmIntval%ray_interval)
      !call vsmLogPrint("real=",b/vsmIntval%ray_interval)
      !call vsmLogPrint("int=",int(b/vsmIntval%ray_interval))
       call vsmLogPrint("skipmin",vsmOctant4Iso(node)%skipmin)
       call vsmLogPrint("skipmax",vsmOctant4Iso(node)%skipmax)
     end if
     call vsmError("ray_ is negative")
   end if
 end subroutine vsmRaySkip4Iso

 function vsmIsInsideNode4Iso(pos, node) !is_rayin
  logical :: vsmIsInsideNode4Iso
  integer, intent(in) :: node
  real (kind=vdp), dimension(3), intent(in) :: pos

  if(pos(1) >= vsmOctant4Iso(node)%txmax .or. pos(1) < vsmOctant4Iso(node)%txmin .or. pos(2) >= vsmOctant4Iso(node)%tymax .or. &
  pos(2) < vsmOctant4Iso(node)%tymin .or. pos(3) >= vsmOctant4Iso(node)%tzmax .or. pos(3) < vsmOctant4Iso(node)%tzmin) then
    vsmIsInsideNode4Iso = .false.
  else 
    vsmIsInsideNode4Iso = .true.
  end if
 end function vsmIsInsideNode4Iso

 subroutine vsmCalcCoord4Iso
  integer :: numRow, num_sv, tsv
  integer :: i, nd
  integer :: bx, by, bz
  real(kind=vdp), dimension(3) :: gmin, gmax

  call vsmLogPrint("vsmCalcCoord is called")
  if(vsmOctComDat4Iso%maxLevel == 0) then 
    vsmOctComDat4Iso%coordx(0) = vsmOctant4Iso(0)%txmin
    vsmOctComDat4Iso%coordy(0) = vsmOctant4Iso(0)%tymin
    vsmOctComDat4Iso%coordz(0) = vsmOctant4Iso(0)%tzmin

    vsmOctComDat4Iso%coordx(1) = vsmOctant4Iso(0)%txmax
    vsmOctComDat4Iso%coordy(1) = vsmOctant4Iso(0)%tymax
    vsmOctComDat4Iso%coordz(1) = vsmOctant4Iso(0)%tzmax
    return
  end if
  
!  num_sv = 1; tsv = 1
!  do i=1, vsmOctComDat4Iso%maxLevel-1
!   tsv = ishft(tsv, 3)
!   num_sv = num_sv + tsv
!  end do

  if(vsmOctComDat4Iso%maxLevel > 0) then
    num_sv = vsmOctNum(vsmOctComDat4Iso%maxLevel-1)
  else
    num_sv = 0
  end if

  call vsmLogPrint("numsv: ",num_sv)
!  num_sv = num_sv - 1

  numRow = vsmOctComDat4Iso%Row !ishft(a, vsmOctComDat4Iso%maxLevel)
  call vsmLogPrint("numRow: ",numRow)


  vsmOctComDat4Iso%coordx(0) = vsmOctant4Iso(0)%txmin
  vsmOctComDat4Iso%coordy(0) = vsmOctant4Iso(0)%tymin
  vsmOctComDat4Iso%coordz(0) = vsmOctant4Iso(0)%tzmin

  do i=0, numRow - 1

   bx = vsmGetMortonNum4Iso(i, 0, 0)
   nd = num_sv+bx
   call vsmGetGMinMax4Iso(gmin, gmax, nd)
   vsmOctComDat4Iso%coordx(i+1) = gmax(1)

   by = vsmGetMortonNum4Iso(0, i, 0)
   nd = num_sv+by
   call vsmGetGMinMax4Iso(gmin, gmax, nd)
   vsmOctComDat4Iso%coordy(i+1) = gmax(2)
   
   bz = vsmGetMortonNum4Iso(0, 0, i)
   nd = num_sv+bz
   call vsmGetGMinMax4Iso(gmin, gmax, nd)
   vsmOctComDat4Iso%coordz(i+1) = gmax(3)

   if(vsmDbg) then
     call vsmLogPrint("Row ", i)
     call vsmLogPrint("bx:", bx)
     call vsmLogPrint("gmax(1):", gmax(1))
     call vsmLogPrint("by:", by)
     call vsmLogPrint("gmax(2):", gmax(2))
     call vsmLogPrint("bz:", bz)
     call vsmLogPrint("gmax(3):", gmax(3))
   end if
  end do

  call vsmLogPrint("vsmCalcCoord ends")
 end subroutine vsmCalcCoord4Iso
 
end module

