module vsmParamBroadcast
 use vsmConst
 use vsmFunction
 use vsmParam
 use vsmImage
 use vsmFrame
 use vsmLightData
 use vsmCameraData
 use vsmDepthQ
 use vsmColormap
#ifndef CLANG
 use mpi
#endif
 implicit none
 private

 ! for bcast
 public vsmParamBCast

 contains

 subroutine vsmParamBcast
   integer :: comm, myrank
   call vsmGetMPIVars(comm, myrank)
   call vsmLogPrint('**broadcasting parameters rank:0->others')
   call vsmBcast1(comm, myrank)
   call vsmBcastIso(comm, myrank)
   call vsmBcastSlice(comm, myrank)
   call vsmBcastVol(comm, myrank)
   call vsmBcastStl(comm, myrank)
   call vsmBcastArw(comm, myrank)
   call vsmBcastPtcl(comm, myrank)
   call vsmBcastLight(comm, myrank)
   call vsmBcastCamera(comm, myrank)
   call vsmBcastDepth(comm, myrank)
   call vsmBcastCmap(comm, myrank)
   call vsmLogPrint('**broadcasting parameters ends')
 end subroutine vsmParamBcast

! type vismo_scolormap
!  real(kind=vdp), dimension(vsmMaxColmap) :: value
!  real, dimension(4, vsmMaxColmap) :: color
!  integer :: colnum
! end type vismo_scolormap
 subroutine vsmBcastSCmap(comm, myrank)
   integer, intent(in) :: comm, myrank
   integer, dimension(vsmMaxScalar) :: intdata
   real, dimension(4*vsmMaxScalar*vsmMaxColmap) :: floatdata
   real(kind=vdp), dimension(vsmMaxScalar*vsmMaxColmap) :: doubledata
   integer :: i, j, tmp, ierr, num,src

   call vsmLogPrint('scalar color map')
   if(myrank == 0) then
     do j=1, vsmSCMapNum
       intdata(j) = vsmSColormap(j)%colnum
       call vsmLogPrint('colnum = ', vsmSColormap(j)%colnum) 
       do i=1, vsmMaxColmap
         doubledata(i + vsmMaxColmap*(j-1)) = vsmSColormap(j)%value(i)
         tmp = 4*((i-1) + vsmMaxColmap*(j-1))
         floatdata(1+tmp:4+tmp) = vsmSColormap(j)%color(1:4, i)
       end do
     end do
   end if
   src = 0
#ifndef CLANG
   call mpi_bcast(intdata, vsmSCMapNum, mpi_integer, src, comm, ierr)
   num = vsmSCMapNum*vsmMaxColmap
   call mpi_bcast(doubledata, num, mpi_double_precision, src, comm, ierr)
   num = num*4
   call mpi_bcast(floatdata, num, mpi_real, src, comm, ierr)
#else
   call vsmmpibcastint(intdata, vsmSCMapNum, src)
   num = vsmSCMapNum*vsmMaxColmap
   call vsmmpibcastdble(doubledata, num, src)
   num = num*4
   call vsmmpibcastfloat(floatdata, num, src)
#endif
   if(myrank /= 0) then
     do j=1, vsmSCMapNum
       vsmSColormap(j)%colnum = intdata(j)
       if(vsmSColormap(j)%colnum > 0) then
          vsmSColormap(j)%set = .true.
       end if
       call vsmLogPrint('colnum = ', vsmSColormap(j)%colnum) 
       do i=1, vsmMaxColmap
         vsmSColormap(j)%value(i) = doubledata(i + vsmMaxColmap*(j-1))
         tmp = 4*((i-1) + vsmMaxColmap*(j-1))
         vsmSColormap(j)%color(1:4, i) = floatdata(1+tmp:4+tmp)
       end do
     end do
   end if
 end subroutine vsmBcastSCmap

! type vismo_vcolormap
!  real(kind=vdp), dimension(vsmMaxColmap) :: value
!  real, dimension(3, vsmMaxColmap) :: color
!  integer :: colnum
! end type vismo_vcolormap
 subroutine vsmBcastVCmap(comm, myrank)
   integer, intent(in) :: comm, myrank
   integer, dimension(vsmMaxVector) :: intdata
   real, dimension(3*vsmMaxVector*vsmMaxColmap) :: floatdata
   real(kind=vdp), dimension(vsmMaxVector*vsmMaxColmap) :: doubledata
   integer :: i, j, tmp, ierr, num, src

   call vsmLogPrint('vector color map')
   if(myrank == 0) then
     do j=1, vsmVCMapNum
       intdata(j) = vsmVColormap(j)%colnum
       call vsmLogPrint('colnum = ', vsmVColormap(j)%colnum) 
       do i=1, vsmMaxColmap
         doubledata(i + vsmMaxColmap*(j-1)) = vsmVColormap(j)%value(i)
         tmp = 3*((i-1) + vsmMaxColmap*(j-1))
         floatdata(1+tmp:3+tmp) = vsmVColormap(j)%color(1:3, i)
       end do
     end do
   end if
   src=0
#ifndef CLANG
   call mpi_bcast(intdata, vsmVCMapNum, mpi_integer, src, comm, ierr)
   num = vsmVCMapNum*vsmMaxColmap
   call mpi_bcast(doubledata, num, mpi_double_precision, src, comm, ierr)
   num = num*3
   call mpi_bcast(floatdata, num, mpi_real, src, comm, ierr)
#else
   call vsmmpibcastint(intdata, vsmVCMapNum, src)
   num = vsmVCMapNum*vsmMaxColmap
   call vsmmpibcastdble(doubledata, num, src)
   num = num*3
   call vsmmpibcastfloat(floatdata, num, src)
#endif
   if(myrank /= 0) then
     do j=1, vsmVCMapNum
       vsmVColormap(j)%colnum = intdata(j)
       if(vsmVColormap(j)%colnum > 0) then
          vsmVColormap(j)%set = .true.
       end if
       call vsmLogPrint('colnum = ', vsmVColormap(j)%colnum) 
       do i=1, vsmMaxColmap
         vsmVColormap(j)%value(i) = doubledata(i + vsmMaxColmap*(j-1))
         tmp = 3*((i-1) + vsmMaxColmap*(j-1))
         vsmVColormap(j)%color(1:3, i) = floatdata(1+tmp:3+tmp)
       end do
     end do
   end if
 end subroutine vsmBcastVCmap


! type vismo_scolormap
!  real(kind=vdp), dimension(vsmMaxColmap) :: value
!  real, dimension(4, vsmMaxColmap) :: color
!  integer :: colnum
! end type vismo_scolormap
 subroutine vsmBcastPCmap(comm, myrank)
   integer, intent(in) :: comm, myrank
   integer, dimension(vsmMaxPtcl) :: intdata
   real, dimension(4*vsmMaxPtcl*vsmMaxColmap) :: floatdata
   real(kind=vdp), dimension(vsmMaxPtcl*vsmMaxColmap) :: doubledata
   integer :: i, j, tmp, ierr, num, src

   call vsmLogPrint('scalar color map')
   src = 0
   if(myrank == 0) then
     do j=1, vsmPCMapNum
       intdata(j) = vsmPColormap(j)%colnum
       call vsmLogPrint('colnum = ', vsmPColormap(j)%colnum) 
       do i=1, vsmMaxColmap
         doubledata(i + vsmMaxColmap*(j-1)) = vsmPColormap(j)%value(i)
         tmp = 4*((i-1) + vsmMaxColmap*(j-1))
         floatdata(1+tmp:4+tmp) = vsmPColormap(j)%color(1:4, i)
       end do
     end do
   end if
#ifndef CLANG
   call mpi_bcast(intdata, vsmPCMapNum, mpi_integer, src, comm, ierr)
   num = vsmPCMapNum*vsmMaxColmap
   call mpi_bcast(doubledata, num, mpi_double_precision, src, comm, ierr)
   num = num*4
   call mpi_bcast(floatdata, num, mpi_real, src, comm, ierr)
#else
   call vsmmpibcastint(intdata, vsmPCMapNum, src)
   num = vsmPCMapNum*vsmMaxColmap
   call vsmmpibcastdble(doubledata, num, src)
   num = num*4
   call vsmmpibcastfloat(floatdata, num, src)
#endif
   if(myrank /= 0) then
     do j=1, vsmPCMapNum
       vsmPColormap(j)%colnum = intdata(j)
       if(vsmPColormap(j)%colnum > 0) then
          vsmPColormap(j)%set = .true.
       end if
       call vsmLogPrint('colnum = ', vsmPColormap(j)%colnum) 
       do i=1, vsmMaxColmap
         vsmPColormap(j)%value(i) = doubledata(i + vsmMaxColmap*(j-1))
         tmp = 4*((i-1) + vsmMaxColmap*(j-1))
         vsmPColormap(j)%color(1:4, i) = floatdata(1+tmp:4+tmp)
       end do
     end do
   end if
 end subroutine vsmBcastPCmap

 subroutine vsmBcastCmap(comm, myrank)
   integer, intent(in) :: comm, myrank

   call vsmBcastSCmap(comm, myrank)
   call vsmBcastVCmap(comm, myrank)
   call vsmBcastPCmap(comm, myrank)
 end subroutine vsmBcastCmap

! type vismo_depthparam
!  integer :: depth_kind ! 1:liner, 2:exp(-x), 3:exp(-x**2)
!  real(kind=vdp), dimension(2) :: dist
! end type vismo_depthparam
 subroutine vsmBcastDepth(comm, myrank)
   integer, intent(in) :: comm, myrank
   integer :: intdata
   real(kind=vdp), dimension(2) :: doubledata
   integer :: ierr, src, count

   call vsmLogPrint('Depth')
   if(myrank == 0) then
     intdata = vsmDepthParam%depth_kind
     doubledata(1:2) = vsmDepthParam%dist(1:2)
     call vsmLogPrint('kind = ', vsmDepthParam%depth_kind)
     call vsmLogPrint('dist = ', vsmDepthParam%dist)
   end if

   src = 0
#ifndef CLANG
   count = 1
   call mpi_bcast(intdata, count, mpi_integer, src, comm, ierr)
   count = 2
   call mpi_bcast(doubledata, count, mpi_double_precision, src, comm, ierr)
#else
   count = 1
   call vsmmpibcastint(intdata, count, src)
   count = 2
   call vsmmpibcastdble(doubledata, count, src)
#endif
   if(myrank /= 0) then
     vsmDepthParam%depth_kind = intdata 
     vsmDepthParam%dist(1:2) = doubledata(1:2)
     call vsmLogPrint('kind = ', vsmDepthParam%depth_kind)
     call vsmLogPrint('dist = ', vsmDepthParam%dist)
   end if

 end subroutine vsmBcastDepth

! type vismo_camera
!  real(kind=vdp), dimension(3) :: pos
!  real(kind=vdp), dimension(3) :: front
!  real(kind=vdp), dimension(3) :: up
!  real(kind=vdp), dimension(3) :: right
!  real(kind=vdp) :: fovy
!  real(kind=vdp) :: near, far
!  real(kind=vdp) :: width
!  logical :: pers
! end type vismo_camera

 subroutine vsmBcastCamera(comm, myrank)
   integer, intent(in) :: comm, myrank
   integer, dimension(vsmMaxCamera) :: intdata
   real(kind=vdp), dimension(16*vsmMaxCamera) :: doubledata
   integer :: i, k, tmp, ierr, src, count

   call vsmLogPrint('Camera')
   if(myrank == 0) then
     call vsmLogPrint('vsmCameranum = ', vsmCameraNum)
     do i=1, vsmCameraNum
       if(vsmCamera(i)%pers) then 
         intdata(i) = 1
       else
         intdata(i) = 0
       end if

       do k=1, 3
         tmp = k+16*(i-1)
         doubledata(   tmp) = vsmCamera(i)%pos(k)
         doubledata(3 +tmp) = vsmCamera(i)%front(k)
         doubledata(6 +tmp) = vsmCamera(i)%up(k)
         doubledata(9 +tmp) = vsmCamera(i)%right(k)
       end do
       tmp = 16*(i-1)
       doubledata(13+tmp) = vsmCamera(i)%fovy
       doubledata(14+tmp) = vsmCamera(i)%near
       doubledata(15+tmp) = vsmCamera(i)%far
       doubledata(16+tmp) = vsmCamera(i)%width

       if(vsmDbg) then
         call vsmLogPrint('pers = ', vsmCamera(i)%pers)
         call vsmLogPrint('pos = ', vsmCamera(i)%pos)
         call vsmLogPrint('front = ', vsmCamera(i)%front)
         call vsmLogPrint('up = ', vsmCamera(i)%up)
         call vsmLogPrint('right = ', vsmCamera(i)%right)
         call vsmLogPrint('fovy = ', vsmCamera(i)%fovy)
         call vsmLogPrint('near = ', vsmCamera(i)%near)
         call vsmLogPrint('far = ', vsmCamera(i)%far)
         call vsmLogPrint('width = ', vsmCamera(i)%width)
       end if
     end do
   end if
   src=0
#ifndef CLANG
   call mpi_bcast(intdata, vsmCameraNum, mpi_integer, src, comm, ierr)
   count = 16*vsmCameraNum
   call mpi_bcast(doubledata, count, mpi_double_precision, src, comm, ierr)
#else
   call vsmmpibcastint(intdata, vsmCameraNum, src)
   count = 16*vsmCameraNum
   call vsmmpibcastdble(doubledata, count, src)
#endif
   if(myrank /= 0) then
     call vsmLogPrint('vsmCameranum = ', vsmCameraNum)
     do i=1, vsmCameraNum
       if(intdata(i) == 1) then 
         vsmCamera(i)%pers = .true.
       else
         vsmCamera(i)%pers = .false.
       end if

       do k=1, 3
         tmp = k+16*(i-1)
         vsmCamera(i)%pos(k)   = doubledata(   tmp) 
         vsmCamera(i)%front(k) = doubledata(3 +tmp) 
         vsmCamera(i)%up(k)    = doubledata(6 +tmp) 
         vsmCamera(i)%right(k) = doubledata(9 +tmp) 
       end do
       tmp = 16*(i-1)
       vsmCamera(i)%fovy = doubledata(13+tmp) 
       vsmCamera(i)%near = doubledata(14+tmp) 
       vsmCamera(i)%far  = doubledata(15+tmp) 
       vsmCamera(i)%width  = doubledata(16+tmp) 

       if(vsmDbg) then
         call vsmLogPrint('para = ', vsmCamera(i)%pers)
         call vsmLogPrint('pos = ', vsmCamera(i)%pos)
         call vsmLogPrint('front = ', vsmCamera(i)%front)
         call vsmLogPrint('up = ', vsmCamera(i)%up)
         call vsmLogPrint('right = ', vsmCamera(i)%right)
         call vsmLogPrint('fovy = ', vsmCamera(i)%fovy)
         call vsmLogPrint('near = ', vsmCamera(i)%near)
         call vsmLogPrint('far = ', vsmCamera(i)%far)
         call vsmLogPrint('width = ', vsmCamera(i)%width)
       end if
     end do
   end if

 end subroutine vsmBcastCamera

! type vismo_light
!  real, dimension(3) :: pos
!  real, dimension(3) :: direc
!  real, dimension(3) :: amb, dif, spec
!  logical :: para !parallel light or point light
! end type vismo_light
 subroutine vsmBcastLight(comm, myrank)
   integer, intent(in) :: comm, myrank
   integer, dimension(vsmMaxLight) :: intdata
   real, dimension((3*5)*vsmMaxLight) :: floatdata
   integer :: i, k, tmp, ierr, src, count

   call vsmLogPrint('Light')
   if(myrank == 0) then
     call vsmLogPrint('lightnum = ', vsmLightNum)
     do i=1, vsmLightNum
       if(vsmLight(i)%para) then 
         intdata(i) = 1
       else
         intdata(i) = 0
       end if

       do k=1, 3
         tmp = k+15*(i-1)
         floatdata(   tmp) = vsmLight(i)%pos(k)
         floatdata(3 +tmp) = vsmLight(i)%direc(k)
         floatdata(6 +tmp) = vsmLight(i)%amb(k)
         floatdata(9 +tmp) = vsmLight(i)%dif(k)
         floatdata(12+tmp) = vsmLight(i)%spec(k)
       end do

       if(vsmDbg) then
         call vsmLogPrint('para = ', vsmLight(i)%para)
         call vsmLogPrint('pos = ', vsmLight(i)%pos)
         call vsmLogPrint('direc = ', vsmLight(i)%direc)
         call vsmLogPrint('amb = ', vsmLight(i)%amb)
         call vsmLogPrint('dif = ', vsmLight(i)%dif)
         call vsmLogPrint('spec = ', vsmLight(i)%spec)
       end if
     end do
   end if
   src = 0
#ifndef CLANG
   call mpi_bcast(intdata, vsmLightNum, mpi_integer, src, comm, ierr)
   count = 15*vsmLightNum
   call mpi_bcast(floatdata, count, mpi_real, src, comm, ierr)
#else
   call vsmmpibcastint(intdata, vsmLightNum, src)
   count = 15*vsmLightNum
   call vsmmpibcastfloat(floatdata, count, src)
#endif
   if(myrank /= 0) then
     call vsmLogPrint('lightnum = ', vsmLightNum)
     do i=1, vsmLightNum
       if(intdata(i) == 1) then 
         vsmLight(i)%para = .true. 
       else
         vsmLight(i)%para = .false. 
       end if

       do k=1, 3
         tmp = k+15*(i-1)
         vsmLight(i)%pos(k)   = floatdata(   tmp) 
         vsmLight(i)%direc(k) = floatdata(3 +tmp)
         vsmLight(i)%amb(k)   = floatdata(6 +tmp) 
         vsmLight(i)%dif(k)   = floatdata(9 +tmp)
         vsmLight(i)%spec(k)  = floatdata(12+tmp)
       end do

       if(vsmDbg) then
         call vsmLogPrint('para = ', vsmLight(i)%para)
         call vsmLogPrint('pos = ', vsmLight(i)%pos)
         call vsmLogPrint('direc = ', vsmLight(i)%direc)
         call vsmLogPrint('amb = ', vsmLight(i)%amb)
         call vsmLogPrint('dif = ', vsmLight(i)%dif)
         call vsmLogPrint('spec = ', vsmLight(i)%spec)
       end if
     end do
   end if

 end subroutine vsmBcastLight

! type vismo_ptcldata
!  integer :: nptcl
!  real(kind=vdp) ::radius
!  real, dimension(3) :: amb, dif, spec
!  real :: shin
!  logical :: mono
! end type vismo_ptcldata
 subroutine vsmBcastPtcl(comm, myrank)
   integer, intent(in) :: comm, myrank
   integer, dimension(2*vsmMaxPtcl*vsmMaxVisKind) :: intdata
   real, dimension((3*3+1)*vsmMaxPtcl*vsmMaxVisKind) :: floatdata
   real(kind=vdp), dimension(vsmMaxPtcl*vsmMaxVisKind) :: doubledata
   integer :: i, j, k, tmp, ierr, num, src

   intdata = 0
   floatdata = 0.0
   doubledata = 0.0_vdp
   call vsmLogPrint('Particle')
   if(myrank == 0) then
     do j=1, vsmNumViskind
       do i=1, vsmMaxPtcl
         intdata(1+(i-1)*2+2*(j-1)*vsmMaxPtcl) = vsmVisParam(j)%ptclparam(i)%nptcl
         if(vsmVisParam(j)%ptclparam(i)%monocolor) then
           intdata(2+(i-1)*2+2*(j-1)*vsmMaxPtcl) = 0
         else
           intdata(2+(i-1)*2+2*(j-1)*vsmMaxPtcl) = 1
         end if
         doubledata(i+(j-1)*vsmMaxPtcl) = vsmVisParam(j)%ptclparam(i)%radius
         do k=1, 3
           tmp = k+10*(i-1)+10*vsmMaxPtcl*(j-1)
           floatdata(tmp)   = vsmVisParam(j)%ptclparam(i)%amb(k)
           floatdata(3+tmp) = vsmVisParam(j)%ptclparam(i)%dif(k)
           floatdata(6+tmp) = vsmVisParam(j)%ptclparam(i)%spec(k)
         end do
         floatdata(10+10*(i-1)+10*vsmMaxPtcl*(j-1)) = vsmVisParam(j)%ptclparam(i)%shin

         if(vsmDbg) then
           if(i > vsmVisParam(j)%ptcl_num) cycle
           call vsmLogPrint('viskind = ', j)
           call vsmLogPrint('nptcl = ', vsmVisParam(j)%ptclparam(i)%nptcl)
           call vsmLogPrint('rad = ', vsmVisParam(j)%ptclparam(i)%radius)
           call vsmLogPrint('amb = ', vsmVisParam(j)%ptclparam(i)%amb)
           call vsmLogPrint('dif = ', vsmVisParam(j)%ptclparam(i)%dif)
           call vsmLogPrint('spec = ', vsmVisParam(j)%ptclparam(i)%spec)
           call vsmLogPrint('shin = ', vsmVisParam(j)%ptclparam(i)%shin)
           call vsmLogPrint('mono = ', vsmVisParam(j)%ptclparam(i)%monocolor)
         end if
 !          print *, 'mono = ', vsmVisParam(j)%ptclparam(i)%monocolor, myrank,intdata(2+(i-1)*2+2*(j-1)*vsmMaxPtcl)
       end do
     end do
!           print *, 'intdata = ',intdata
 !           print *, 'intdata = ',intdata(1:2*vsmMaxPTCL), myrank
   end if
   src = 0
#ifndef CLANG
   num = vsmMaxPtcl*vsmNumViskind*2
   call mpi_bcast(intdata, num, mpi_integer, src, comm, ierr)
   num = num / 2
   call mpi_bcast(doubledata, num, mpi_double_precision, src, comm, ierr)
   num = 10*num
   call mpi_bcast(floatdata, num, mpi_real, src, comm, ierr)
#else
   num = vsmMaxPtcl*vsmNumViskind*2
   call vsmmpibcastint(intdata, num, src)
   num=num/2
   call vsmmpibcastdble(doubledata, num, src)
   num = 10*num
   call vsmmpibcastfloat(floatdata, num, src)
#endif
   if(myrank /= 0) then
     do j=1, vsmNumViskind
       do i=1, vsmMaxPtcl
         vsmVisParam(j)%ptclparam(i)%nptcl  = intdata(1+(i-1)*2+2*(j-1)*vsmMaxPtcl)
         if(intdata(2+(i-1)*2+2*(j-1)*vsmMaxPtcl) == 1) then
           vsmVisParam(j)%ptclparam(i)%monocolor  = .false.
         else
           vsmVisParam(j)%ptclparam(i)%monocolor  = .true.
         end if
         vsmVisParam(j)%ptclparam(i)%radius = doubledata(i+(j-1)*vsmMaxPtcl)
         do k=1, 3
           tmp = k+10*(i-1)+10*vsmMaxPtcl*(j-1)
           vsmVisParam(j)%ptclparam(i)%amb(k) = floatdata(tmp)
           vsmVisParam(j)%ptclparam(i)%dif(k) = floatdata(3+tmp)
           vsmVisParam(j)%ptclparam(i)%spec(k) = floatdata(6+tmp)
         end do
         vsmVisParam(j)%ptclparam(i)%shin = floatdata(10+10*(i-1)+10*vsmMaxPtcl*(j-1))
         if(vsmDbg) then
           if(i > vsmVisParam(j)%ptcl_num) cycle
           call vsmLogPrint('viskind = ', j)
           call vsmLogPrint('nptcl = ', vsmVisParam(j)%ptclparam(i)%nptcl)
           call vsmLogPrint('rad = ', vsmVisParam(j)%ptclparam(i)%radius)
           call vsmLogPrint('amb = ', vsmVisParam(j)%ptclparam(i)%amb)
           call vsmLogPrint('dif = ', vsmVisParam(j)%ptclparam(i)%dif)
           call vsmLogPrint('spec = ', vsmVisParam(j)%ptclparam(i)%spec)
           call vsmLogPrint('shin = ', vsmVisParam(j)%ptclparam(i)%shin)
           call vsmLogPrint('mono = ', vsmVisParam(j)%ptclparam(i)%monocolor)
         end if
 !          print *, 'mono = ', vsmVisParam(j)%ptclparam(i)%monocolor,myrank,intdata(2+(i-1)*2+2*(j-1)*vsmMaxPtcl)
       end do
     end do
!            print *, 'intdata = ',intdata(1:2*vsmMaxPTCL), myrank
  end if

 end subroutine vsmBcastPtcl

! type vismo_arrowdata
!   integer :: nvec
!   real(kind=vdp), dimension(4) :: sc !a, b, c, d ! f(x,y,z) = ax+by+cz+d 
!   real(kind=vdp) :: density
!   real(kind=vdp) :: radius, length
! end type vismo_arrowdata
 subroutine vsmBcastArw(comm, myrank)
   integer, intent(in) :: comm, myrank
   integer, dimension(vsmMaxArrows*vsmMaxVisKind*3) :: intdata
   real(kind=vdp), dimension((4+1+2+vsmArMaxSeed*3)*vsmMaxArrows*vsmMaxVisKind) :: doubledata
   integer :: i, j, k, tmp, ierr, num, offset,src

   call vsmLogPrint('Arrows')
   offset = (4+1+2)*vsmMaxArrows*vsmMaxVisKind
   if(myrank == 0) then
     do j=1, vsmNumViskind
       do i=1, vsmMaxArrows
         intdata(3*(i-1)+1+3*(j-1)*vsmMaxArrows) = vsmVisParam(j)%arrowparam(i)%nvec
         intdata(3*(i-1)+2+3*(j-1)*vsmMaxArrows) = vsmVisParam(j)%arrowparam(i)%mode
         if(vsmVisParam(j)%arrowparam(i)%variable) then
            intdata(3*(i-1)+3+3*(j-1)*vsmMaxArrows) = 1
         else
            intdata(3*(i-1)+3+3*(j-1)*vsmMaxArrows) = 0
         end if

         tmp = 7*((i-1)+vsmMaxArrows*(j-1))
         doubledata(1+tmp:4+tmp) = vsmVisParam(j)%arrowparam(i)%sc(1:4)
         doubledata(5+tmp) = vsmVisParam(j)%arrowparam(i)%density
         doubledata(6+tmp) = vsmVisParam(j)%arrowparam(i)%radius
         doubledata(7+tmp) = vsmVisParam(j)%arrowparam(i)%length
         if(vsmDbg) then
           if(i > vsmVisParam(j)%arrow_num) cycle
           call vsmLogPrint('viskind = ', j)
           call vsmLogPrint('nvec = ', vsmVisParam(j)%arrowparam(i)%nvec)
           call vsmLogPrint('sc = ', vsmVisParam(j)%arrowparam(i)%sc)
           call vsmLogPrint('dens = ', vsmVisParam(j)%arrowparam(i)%density)
           call vsmLogPrint('rad = ', vsmVisParam(j)%arrowparam(i)%radius)
           call vsmLogPrint('len = ', vsmVisParam(j)%arrowparam(i)%length)
         end if
       end do
     end do

     do j=1, vsmNumViskind
       do i=1, vsmMaxArrows
         do k=1, vsmArMaxSeed
            tmp = 3*(k-1)+(j-1)*vsmMaxArrows*vsmArMaxSeed*3+(i-1)*vsmArMaxSeed*3+offset
            doubledata(tmp+1:tmp+3) = vsmVisParam(j)%arrowparam(i)%seed(1:3, k)
         end do
       end do
     end do
   end if
   src = 0
#ifndef CLANG
   num = vsmMaxArrows*vsmNumViskind*3
   call mpi_bcast(intdata, num, mpi_integer, src, comm, ierr)
   num = (4+1+2+vsmArMaxSeed*3)*num/3
   call mpi_bcast(doubledata, num, mpi_double_precision, src, comm, ierr)
#else
   num = vsmMaxArrows*vsmNumViskind*3
   call vsmmpibcastint(intdata, num, src)
   num = (4+1+2+vsmArMaxSeed*3)*num/3
   call vsmmpibcastdble(doubledata, num, src)
#endif
   if(myrank /= 0) then
     do j=1, vsmNumViskind
       do i=1, vsmMaxArrows
         vsmVisParam(j)%arrowparam(i)%nvec = intdata(3*(i-1)+1+3*(j-1)*vsmMaxArrows)
         vsmVisParam(j)%arrowparam(i)%mode = intdata(3*(i-1)+2+3*(j-1)*vsmMaxArrows) 
         if(intdata(3*(i-1)+3+3*(j-1)*vsmMaxArrows) == 1) then
            vsmVisParam(j)%arrowparam(i)%variable = .true.
         else
            vsmVisParam(j)%arrowparam(i)%variable = .false.
         end if

         tmp = 7*((i-1)+vsmMaxArrows*(j-1))
           vsmVisParam(j)%arrowparam(i)%sc(1:4) = doubledata(1+tmp:4+tmp)
           vsmVisParam(j)%arrowparam(i)%density = doubledata(5+tmp)
           vsmVisParam(j)%arrowparam(i)%radius  = doubledata(6+tmp)
           vsmVisParam(j)%arrowparam(i)%length  = doubledata(7+tmp)
         if(vsmDbg) then
           if(i > vsmVisParam(j)%arrow_num) cycle
           call vsmLogPrint('viskind = ', j)
           call vsmLogPrint('nvec = ', vsmVisParam(j)%arrowparam(i)%nvec)
           call vsmLogPrint('sc = ', vsmVisParam(j)%arrowparam(i)%sc)
           call vsmLogPrint('dens = ', vsmVisParam(j)%arrowparam(i)%density)
           call vsmLogPrint('rad = ', vsmVisParam(j)%arrowparam(i)%radius)
           call vsmLogPrint('len = ', vsmVisParam(j)%arrowparam(i)%length)
         end if
       end do
     end do

     do j=1, vsmNumViskind
       do i=1, vsmMaxArrows
         do k=1, vsmArMaxSeed
            tmp = 3*(k-1)+(j-1)*vsmMaxArrows*vsmArMaxSeed*3+(i-1)*vsmArMaxSeed*3+offset
            vsmVisParam(j)%arrowparam(i)%seed(1:3, k) = doubledata(tmp+1:tmp+3)
         end do
       end do
     end do

   end if

 end subroutine vsmBcastArw

! type vismo_strlinedata
!   real(kind=vdp), dimension(3, vsmMaxSeed) :: seed
!   integer :: num_seed
!   integer, dimension(vsmMaxSeed) :: steps_remained
!   integer :: nvec
!   real, dimension(3) :: amb, dif, spec ! for monocolor lines
!   real :: shin
!   real(kind=vdp) :: radius
!   logical :: monocolor
! end type vismo_strlinedata
 subroutine vsmBcastStl(comm, myrank)
   integer, intent(in) :: comm, myrank
   integer, dimension(3*vsmMaxVector*vsmMaxVisKind) :: intdata !steps_remaineds are initial values
   real, dimension(10*vsmMaxVector*vsmMaxVisKind) :: floatdata
   real(kind=vdp), dimension((1+vsmMaxSeed*3)*vsmMaxVector*vsmMaxVisKind) :: doubledata
   integer :: i, j, k, tmp, ierr, num, src, count

   call vsmLogPrint('StreamLines')
   if(myrank == 0) then
     do j=1, vsmNumViskind
       do i=1, vsmMaxVector
         tmp = 3*((i-1)+(j-1)*vsmMaxVector)
         intdata(1+tmp) = vsmVisParam(j)%slineparam(i)%num_seed
         intdata(2+tmp) = vsmVisParam(j)%slineparam(i)%nvec
         if(vsmVisParam(j)%slineparam(i)%monocolor) then
           intdata(3+tmp) = 1
         else 
           intdata(3+tmp) = 0
         end if
         do k=1, 3
           tmp = k+10*(i-1) + 10*vsmMaxVector*(j-1)
           floatdata(  tmp) = vsmVisParam(j)%slineparam(i)%amb(k)
           floatdata(3+tmp) = vsmVisParam(j)%slineparam(i)%dif(k)
           floatdata(6+tmp) = vsmVisParam(j)%slineparam(i)%spec(k)
         end do
         floatdata(10+ 10*(i-1) + 10*vsmMaxVector*(j-1)) = vsmVisParam(j)%slineparam(i)%shin

         doubledata(1+(1+vsmMaxSeed*3)*(i-1)+(1+vsmMaxSeed*3)*vsmMaxVector*(j-1)) = &
                      vsmVisParam(j)%slineparam(i)%radius
         do k=1, vsmMaxSeed
           tmp = 1 + 3*(k-1)+(1+vsmMaxSeed*3)*(i-1)+(1+vsmMaxSeed*3)*vsmMaxVector*(j-1)
           doubledata(1+tmp:3+tmp) = vsmVisParam(j)%slineparam(i)%seed(1:3, k)
         end do
         if(vsmDbg) then
           if(i > vsmVisParam(j)%sline_num) cycle
           call vsmLogPrint('viskind = ', j)
           call vsmLogPrint('nvec = ', vsmVisParam(j)%slineparam(i)%nvec)
           call vsmLogPrint('numseed = ', vsmVisParam(j)%slineparam(i)%num_seed)
           call vsmLogPrint('mono = ', vsmVisParam(j)%slineparam(i)%monocolor)
           call vsmLogPrint('amb = ', vsmVisParam(j)%slineparam(i)%amb)
           call vsmLogPrint('dif = ', vsmVisParam(j)%slineparam(i)%dif)
           call vsmLogPrint('spec = ', vsmVisParam(j)%slineparam(i)%spec)
           call vsmLogPrint('shin = ', vsmVisParam(j)%slineparam(i)%shin)
           do k=1, vsmVisParam(j)%slineparam(i)%num_seed
             call vsmLogPrint("seed = ", vsmVisParam(j)%slineparam(i)%seed(1:3, k))
           end do
         end if
       end do
     end do
   end if
   src = 0
   num = vsmMaxVector*vsmNumViskind
#ifndef CLANG
   count = num*3
   call mpi_bcast(intdata, count, mpi_integer, src, comm, ierr)
   count = num*10
   call mpi_bcast(floatdata, count, mpi_real, src, comm, ierr)
   num = (1+vsmMaxSeed*3)*num
   call mpi_bcast(doubledata, num, mpi_double_precision, src, comm, ierr)
#else
   count = num*3
   call vsmmpibcastint(intdata, count, src)
   count = num*10
   call vsmmpibcastfloat(floatdata, count, src)
   num = (1+vsmMaxSeed*3)*num
   call vsmmpibcastdble(doubledata, num, src)
#endif
   if(myrank /= 0) then
     do j=1, vsmNumViskind
       do i=1, vsmMaxVector
         tmp = 3*((i-1)+(j-1)*vsmMaxVector)
         vsmVisParam(j)%slineparam(i)%num_seed = intdata(1+tmp)
         vsmVisParam(j)%slineparam(i)%nvec     = intdata(2+tmp) 
         if(intdata(3+tmp) == 1) then
           vsmVisParam(j)%slineparam(i)%monocolor = .true.
         else 
           vsmVisParam(j)%slineparam(i)%monocolor = .false.
         end if
         do k=1, 3
           tmp = k+10*(i-1) + 10*vsmMaxVector*(j-1)
           vsmVisParam(j)%slineparam(i)%amb(k) = floatdata(tmp)
           vsmVisParam(j)%slineparam(i)%dif(k) = floatdata(3+tmp)
           vsmVisParam(j)%slineparam(i)%spec(k) = floatdata(6+tmp)
         end do
         vsmVisParam(j)%slineparam(i)%shin = floatdata(10+ 10*(i-1) + 10*vsmMaxVector*(j-1)) 

         vsmVisParam(j)%slineparam(i)%radius = &
              doubledata(1+(1+vsmMaxSeed*3)*(i-1)+(1+vsmMaxSeed*3)*vsmMaxVector*(j-1))
         do k=1, vsmMaxSeed
           tmp = 3*(k-1)+(1+vsmMaxSeed*3)*(i-1)+(1+vsmMaxSeed*3)*vsmMaxVector*(j-1) + 1
           vsmVisParam(j)%slineparam(i)%seed(1, k) = doubledata(1+tmp)
           vsmVisParam(j)%slineparam(i)%seed(2, k) = doubledata(2+tmp)
           vsmVisParam(j)%slineparam(i)%seed(3, k) = doubledata(3+tmp)
         end do
         if(vsmDbg) then
           if(i > vsmVisParam(j)%sline_num) cycle
           call vsmLogPrint('viskind = ', j)
           call vsmLogPrint('nvec = ', vsmVisParam(j)%slineparam(i)%nvec)
           call vsmLogPrint('numseed = ', vsmVisParam(j)%slineparam(i)%num_seed)
           call vsmLogPrint('mono = ', vsmVisParam(j)%slineparam(i)%monocolor)
           call vsmLogPrint('amb = ', vsmVisParam(j)%slineparam(i)%amb)
           call vsmLogPrint('dif = ', vsmVisParam(j)%slineparam(i)%dif)
           call vsmLogPrint('spec = ', vsmVisParam(j)%slineparam(i)%spec)
           call vsmLogPrint('shin = ', vsmVisParam(j)%slineparam(i)%shin)
           do k=1, vsmVisParam(j)%slineparam(i)%num_seed
             call vsmLogPrint("seed = ", vsmVisParam(j)%slineparam(i)%seed(1:3, k))
           end do
         end if
       end do
     end do
   end if

 end subroutine vsmBcastStl

! type vismo_volrendata
!   integer :: nscal
!   real :: amb, dif, spec, shin
! end type vismo_volrendata
 subroutine vsmBcastVol(comm, myrank)
   integer, intent(in) :: comm, myrank
   integer, dimension(vsmMaxScalar*vsmMaxVisKind) :: intdata
   real, dimension(4*vsmMaxScalar*vsmMaxVisKind) :: floatdata
   integer :: i, j, tmp, ierr, num, src

   call vsmLogPrint('Volume Rendering')
   if(myrank == 0) then
     do j=1, vsmNumViskind
       do i=1, vsmMaxScalar
         intdata(i+(j-1)*vsmMaxScalar) = vsmVisParam(j)%volparam(i)%nscal

         tmp = 4*((i-1)+vsmMaxScalar*(j-1))
         floatdata(1+tmp) = vsmVisParam(j)%volparam(i)%amb
         floatdata(2+tmp) = vsmVisParam(j)%volparam(i)%dif
         floatdata(3+tmp) = vsmVisParam(j)%volparam(i)%spec
         floatdata(4+tmp) = vsmVisParam(j)%volparam(i)%shin
         if(vsmDbg) then
           if(i > vsmVisParam(j)%volren_num) cycle
           call vsmLogPrint('viskind = ', j)
           call vsmLogPrint('nscal = ', vsmVisParam(j)%volparam(i)%nscal)
           call vsmLogPrint('amb = ', vsmVisParam(j)%volparam(i)%amb)
           call vsmLogPrint('dif = ', vsmVisParam(j)%volparam(i)%dif)
           call vsmLogPrint('spec = ', vsmVisParam(j)%volparam(i)%spec)
           call vsmLogPrint('shin = ', vsmVisParam(j)%volparam(i)%shin)
         end if
       end do
     end do
   end if
   src = 0
#ifndef CLANG
   num = vsmMaxScalar*vsmNumViskind
   call mpi_bcast(intdata, num, mpi_integer, src, comm, ierr)
   num = 4*num
   call mpi_bcast(floatdata, num, mpi_real, src, comm, ierr)
#else
   num = vsmMaxScalar*vsmNumViskind
   call vsmmpibcastint(intdata, num, src)
   num = 4*num
   call vsmmpibcastfloat(floatdata, num, src)
#endif
   if(myrank /= 0) then
     do j=1, vsmNumViskind
       do i=1, vsmMaxScalar
         vsmVisParam(j)%volparam(i)%nscal = intdata(i+(j-1)*vsmMaxScalar)

         tmp = 4*((i-1)+vsmMaxScalar*(j-1))
         vsmVisParam(j)%volparam(i)%amb = floatdata(1+tmp)
         vsmVisParam(j)%volparam(i)%dif = floatdata(2+tmp) 
         vsmVisParam(j)%volparam(i)%spec = floatdata(3+tmp)
         vsmVisParam(j)%volparam(i)%shin = floatdata(4+tmp)
         if(vsmDbg) then
           if(i > vsmVisParam(j)%volren_num) cycle
           call vsmLogPrint('viskind = ', j)
           call vsmLogPrint('nscal = ', vsmVisParam(j)%volparam(i)%nscal)
           call vsmLogPrint('amb = ', vsmVisParam(j)%volparam(i)%amb)
           call vsmLogPrint('dif = ', vsmVisParam(j)%volparam(i)%dif)
           call vsmLogPrint('spec = ', vsmVisParam(j)%volparam(i)%spec)
           call vsmLogPrint('shin = ', vsmVisParam(j)%volparam(i)%shin)
         end if
       end do
     end do
   end if
 end subroutine vsmBcastVol

! type vismo_slicedata
!   real(kind=vdp), dimension(4) :: sc !a, b, c, d ! f(x,y,z) = ax+by+cz+d 
!   integer :: nscal
!   real :: alpha, amb, dif, spec, shin
! end type vismo_slicedata
 subroutine vsmBcastSlice(comm, myrank)
   integer, intent(in) :: comm, myrank
   integer, dimension(vsmMaxSlice*vsmMaxVisKind) :: intdata
   real, dimension(5*vsmMaxSlice*vsmMaxVisKind) :: floatdata
   real(kind=vdp), dimension(4*vsmMaxSlice*vsmMaxVisKind) :: doubledata
   integer :: i, j, tmp, ierr, num, src

   call vsmLogPrint('Slicing')
   if(myrank == 0) then
     do j=1, vsmNumViskind
       do i=1, vsmMaxSlice
         intdata(i+(j-1)*vsmMaxSlice) = vsmVisParam(j)%sliceparam(i)%nscal

         tmp = 5*((i-1)+vsmMaxSlice*(j-1))
         floatdata(1+tmp) = vsmVisParam(j)%sliceparam(i)%alpha
         floatdata(2+tmp) = vsmVisParam(j)%sliceparam(i)%amb
         floatdata(3+tmp) = vsmVisParam(j)%sliceparam(i)%dif
         floatdata(4+tmp) = vsmVisParam(j)%sliceparam(i)%spec
         floatdata(5+tmp) = vsmVisParam(j)%sliceparam(i)%shin

         tmp = 4*((i-1)+vsmMaxSlice*(j-1))
         doubledata(1+tmp:4+tmp) = vsmVisParam(j)%sliceparam(i)%sc(1:4)
         if(vsmDbg) then
           if(i > vsmVisParam(j)%slice_num) cycle
           call vsmLogPrint('viskind = ', j)
           call vsmLogPrint('nscal = ', vsmVisParam(j)%sliceparam(i)%nscal)
           call vsmLogPrint('alpha = ', vsmVisParam(j)%sliceparam(i)%alpha)
           call vsmLogPrint('amb = ', vsmVisParam(j)%sliceparam(i)%amb)
           call vsmLogPrint('dif = ', vsmVisParam(j)%sliceparam(i)%dif)
           call vsmLogPrint('spec = ', vsmVisParam(j)%sliceparam(i)%spec)
           call vsmLogPrint('shin = ', vsmVisParam(j)%sliceparam(i)%shin)
           call vsmLogPrint('eq = ', vsmVisParam(j)%sliceparam(i)%sc)
         end if
       end do
     end do
   end if
   src = 0
#ifndef CLANG
   num = vsmNumViskind*vsmMaxSlice
   call mpi_bcast(intdata, num, mpi_integer, src, comm, ierr)

   num = 5*num
   call mpi_bcast(floatdata, num, mpi_real, src, comm, ierr)

   num = 4*vsmMaxSlice*vsmNumViskind
   call mpi_bcast(doubledata, num, mpi_double_precision, src, comm, ierr)
#else
   num = vsmNumViskind*vsmMaxSlice
   call vsmmpibcastint(intdata, num, src)

   num = 5*num
   call vsmmpibcastfloat(floatdata, num, src)

   num = 4*vsmMaxSlice*vsmNumViskind
   call vsmmpibcastdble(doubledata, num, src)
#endif
   if(myrank /= 0) then
     do j=1, vsmNumViskind
       do i=1, vsmMaxSlice
         vsmVisParam(j)%sliceparam(i)%nscal = intdata(i+(j-1)*vsmMaxSlice)

         tmp = 5*((i-1)+vsmMaxSlice*(j-1))
         vsmVisParam(j)%sliceparam(i)%alpha = floatdata(1+tmp)
         vsmVisParam(j)%sliceparam(i)%amb   = floatdata(2+tmp)
         vsmVisParam(j)%sliceparam(i)%dif   = floatdata(3+tmp)
         vsmVisParam(j)%sliceparam(i)%spec  = floatdata(4+tmp) 
         vsmVisParam(j)%sliceparam(i)%shin  = floatdata(5+tmp) 

         tmp = 4*((i-1)+vsmMaxSlice*(j-1))
         vsmVisParam(j)%sliceparam(i)%sc(1:4) = doubledata(1+tmp:4+tmp)
         if(vsmDbg) then
           if(i > vsmVisParam(j)%slice_num) cycle
           call vsmLogPrint('viskind = ', j)
           call vsmLogPrint('nscal = ', vsmVisParam(j)%sliceparam(i)%nscal)
           call vsmLogPrint('alpha = ', vsmVisParam(j)%sliceparam(i)%alpha)
           call vsmLogPrint('amb = ', vsmVisParam(j)%sliceparam(i)%amb)
           call vsmLogPrint('dif = ', vsmVisParam(j)%sliceparam(i)%dif)
           call vsmLogPrint('spec = ', vsmVisParam(j)%sliceparam(i)%spec)
           call vsmLogPrint('shin = ', vsmVisParam(j)%sliceparam(i)%shin)
           call vsmLogPrint('eq = ', vsmVisParam(j)%sliceparam(i)%sc)
         end if
       end do
     end do
   end if

 end subroutine vsmBcastSlice

! type vismo_isodata
!   real(kind=vdp) :: iso_level
!   integer :: nscal
!   real :: alpha
!   real, dimension(3) :: amb, dif, spec
!   real :: shin
! end type vismo_isodata
 subroutine vsmBcastIso(comm, myrank)
   integer, intent(in) :: comm, myrank
   integer, dimension(vsmMaxIsosurf*vsmMaxVisKind*2) :: intdata
   real, dimension((2+3*3)*vsmMaxIsosurf*vsmMaxVisKind) :: floatdata
   real(kind=vdp), dimension(vsmMaxIsosurf*vsmMaxVisKind) :: doubledata
   integer :: i, j, k, tmp, ierr, num, src

   call vsmLogPrint('Isosurface')
   if(myrank == 0) then
     do j=1, vsmNumViskind
       do i=1, vsmMaxIsosurf
         intdata((i-1)*2+1+(j-1)*vsmMaxIsosurf*2) = vsmVisParam(j)%isoparam(i)%nscal
         intdata((i-1)*2+2+(j-1)*vsmMaxIsosurf*2) = vsmVisParam(j)%isoparam(i)%coloring

         floatdata(1+   11*((i-1)+(j-1)*vsmMaxIsosurf)) = vsmVisParam(j)%isoparam(i)%alpha
         do k=1, 3
           tmp = k+ ((i-1)+(j-1)*vsmMaxIsosurf)*11
           floatdata(1+tmp) = vsmVisParam(j)%isoparam(i)%amb(k)
           floatdata(4+tmp) = vsmVisParam(j)%isoparam(i)%dif(k)
           floatdata(7+tmp) = vsmVisParam(j)%isoparam(i)%spec(k)
         end do
         floatdata(11*(1 + (i-1)+(j-1)*vsmMaxIsosurf)) = vsmVisParam(j)%isoparam(i)%shin
         doubledata(i+(j-1)*vsmMaxIsosurf) = vsmVisParam(j)%isoparam(i)%iso_level

         if(vsmDbg) then
           if(i > vsmVisParam(j)%iso_num) cycle
           call vsmLogPrint('viskind = ', j)
           call vsmLogPrint('nscal = ', vsmVisParam(j)%isoparam(i)%nscal)
           call vsmLogPrint('alpha = ', vsmVisParam(j)%isoparam(i)%alpha)
           call vsmLogPrint('amb = ', vsmVisParam(j)%isoparam(i)%amb)
           call vsmLogPrint('dif = ', vsmVisParam(j)%isoparam(i)%dif)
           call vsmLogPrint('spec = ', vsmVisParam(j)%isoparam(i)%spec)
           call vsmLogPrint('shin = ', vsmVisParam(j)%isoparam(i)%shin)
           call vsmLogPrint('isolevel = ', vsmVisParam(j)%isoparam(i)%iso_level)
         end if

       end do
     end do
   end if
   src = 0
#ifndef CLANG
   num = vsmNumViskind*vsmMaxIsosurf*2
   call mpi_bcast(intdata, num, mpi_integer, src, comm, ierr)
   num = vsmNumViskind*vsmMaxIsosurf
   call mpi_bcast(doubledata, num, mpi_double_precision, src, comm, ierr)
   num = 11*num
   call mpi_bcast(floatdata, num, mpi_real, src, comm, ierr)
#else
   num = vsmNumViskind*vsmMaxIsosurf*2
   call vsmmpibcastint(intdata, num, src)
   num = vsmNumViskind*vsmMaxIsosurf
   call vsmmpibcastdble(doubledata, num, src)
   num = 11*num
   call vsmmpibcastfloat(floatdata, num, src)
#endif


   if(myrank /= 0) then
     do j=1, vsmNumViskind
       do i=1, vsmMaxIsosurf
         vsmVisParam(j)%isoparam(i)%nscal = intdata((i-1)*2+1+(j-1)*vsmMaxIsosurf*2)
         vsmVisParam(j)%isoparam(i)%coloring = intdata((i-1)*2+2+(j-1)*vsmMaxIsosurf*2)

         vsmVisParam(j)%isoparam(i)%alpha     = floatdata(1+  11*((i-1)+(j-1)*vsmMaxIsosurf)) 
         do k=1, 3
           tmp = k+ ((i-1)+(j-1)*vsmMaxIsosurf)*11
           vsmVisParam(j)%isoparam(i)%amb(k)  = floatdata(1+tmp) 
           vsmVisParam(j)%isoparam(i)%dif(k)  = floatdata(4+tmp) 
           vsmVisParam(j)%isoparam(i)%spec(k) = floatdata(7+tmp) 
         end do
         vsmVisParam(j)%isoparam(i)%shin      = floatdata(11*(1 +  (i-1)+(j-1)*vsmMaxIsosurf)) 
         vsmVisParam(j)%isoparam(i)%iso_level = doubledata(i+(j-1)*vsmMaxIsosurf) 

         if(vsmDbg) then
           if(i > vsmVisParam(j)%iso_num) cycle
           call vsmLogPrint('viskind = ', j)
           call vsmLogPrint('nscal = ', vsmVisParam(j)%isoparam(i)%nscal)
           call vsmLogPrint('alpha = ', vsmVisParam(j)%isoparam(i)%alpha)
           call vsmLogPrint('amb = ', vsmVisParam(j)%isoparam(i)%amb)
           call vsmLogPrint('dif = ', vsmVisParam(j)%isoparam(i)%dif)
           call vsmLogPrint('spec = ', vsmVisParam(j)%isoparam(i)%spec)
           call vsmLogPrint('shin = ', vsmVisParam(j)%isoparam(i)%shin)
           call vsmLogPrint('isolevel = ', vsmVisParam(j)%isoparam(i)%iso_level)
         end if
       end do
     end do
   end if

 end subroutine vsmBcastIso

 subroutine vsmBcast1(comm, myrank)
   integer, intent(in) :: comm, myrank
   integer :: i, j, k, num, tmp, ierr

   ! isonum, slicenum, volnum, slinenum, arrownum, numptcl, lightnum, depthcue(logical), vsmMaxLight x vsmMaxVisKind
   ! width, height(image), vsmCameraNum, vsmLightNum
   integer, parameter :: size1 = (8+vsmMaxLight)*vsmMaxVisKind + 4 
   integer, dimension(size1) :: intdata 
   integer :: size_, src
   real, dimension(4) :: rgb

   call vsmLogPrint('General')
   src = 0
   size_ = 1
#ifndef CLANG
   call mpi_bcast(vsmNumViskind, size_, mpi_integer, src, comm, ierr)
   call mpi_bcast(vsmPCmode, size_, mpi_integer, src, comm, ierr)
#else
   call vsmmpibcastint(vsmNumViskind, size_, src)
   call vsmmpibcastint(vsmPCmode, size_, src)
#endif

   if(myrank == 0) then
      rgb(1:3) = vsmImageData%bgcolor(1:3)
      call vsmGetFrameWidth(tmp)
      rgb(4) = real(tmp)
   end if
   size_=4
#ifndef CLANG
   call mpi_bcast(rgb, size_, mpi_real, src, comm, ierr)
#else
   call vsmmpibcastfloat(rgb, size_, src)
#endif
   if(myrank /= 0) then
      call vsmSetBGcolor(rgb(1:3))
      tmp = int(rgb(4))
      call vsmSetFrameWidth(tmp)
   end if

   if(myrank == 0) then
     call vsmLogPrint('send: numviskind = ', vsmNumViskind)
     call vsmLogPrint('send: vsmPCmode = ', vsmPCmode)
   else
     call vsmLogPrint('receive: numviskind = ', vsmNumViskind)
     call vsmLogPrint('receive: vsmPCmode = ', vsmPCmode)
   end if

   size_ = (8+vsmMaxLight)*vsmNumViskind + 4
   num = 8+vsmMaxLight ! sending receiving 8 int data
   if(myrank == 0) then
     j = 0
     do i=1, vsmNumViskind
       tmp = j*num
       intdata(1+tmp) = vsmVisParam(i)%iso_num
       intdata(2+tmp) = vsmVisParam(i)%slice_num
       intdata(3+tmp) = vsmVisParam(i)%volren_num
       intdata(4+tmp) = vsmVisParam(i)%sline_num
       intdata(5+tmp) = vsmVisParam(i)%arrow_num
       intdata(6+tmp) = vsmVisParam(i)%ptcl_num
       intdata(7+tmp) = vsmVisParam(i)%light_num
       if(vsmVisParam(i)%depthq) then
         intdata(8+tmp) = 1
       else
         intdata(8+tmp) = 0
       end if

       do k=1, vsmMaxLight
         if(vsmVisParam(i)%light_used(k)) then
           intdata(8+k+tmp) = 1
         else
           intdata(8+k+tmp) = 0
         end if
       end do

       if(vsmDbg) then
         call vsmLogPrint('--------------------')
         call vsmLogPrint('viskind = ', i)
         call vsmLogPrint('isonum = ', vsmVisParam(i)%iso_num)
         call vsmLogPrint('slicenum = ', vsmVisParam(i)%slice_num)
         call vsmLogPrint('volrennum = ', vsmVisParam(i)%volren_num)
         call vsmLogPrint('slinenum = ', vsmVisParam(i)%sline_num)
         call vsmLogPrint('arrownum = ', vsmVisParam(i)%arrow_num)
         call vsmLogPrint('num_ptcl = ', vsmVisParam(i)%ptcl_num)
         call vsmLogPrint('vsmLightNum = ', vsmVisParam(i)%light_num)
         call vsmLogPrint('depthq = ', vsmVisParam(i)%depthq)
         call vsmLogPrint('light_used = ', vsmVisParam(i)%light_used)
         call vsmLogPrint('--------------------')
       end if

     j = j + 1
     end do
     tmp = num*vsmNumViskind
     intdata(tmp+1) = vsmImageData%width
     intdata(tmp+2) = vsmImageData%height
     intdata(tmp+3) = vsmCameraNum
     intdata(tmp+4) = vsmLightNum
     if(vsmDbg) then
       call vsmLogPrint('vsmNumViskind = ', vsmNumViskind)
       call vsmLogPrint('width = ', vsmImageData%width)
       call vsmLogPrint('height = ', vsmImageData%height)
       call vsmLogPrint('vsmCameraNum = ', vsmCameraNum)
       call vsmLogPrint('vsmLightNum = ', vsmLightNum)
     end if
   end if

#ifndef CLANG
   call mpi_bcast(intdata, size_, mpi_integer, src, comm, ierr)
#else
   call vsmmpibcastint(intdata, size_, src)
#endif
   if(myrank /= 0) then 
     j = 0
     do i=1, vsmNumViskind
       tmp = j*num
       vsmVisParam(i)%iso_num    = intdata(1+tmp) 
       vsmVisParam(i)%slice_num  = intdata(2+tmp) 
       vsmVisParam(i)%volren_num = intdata(3+tmp) 
       vsmVisParam(i)%sline_num  = intdata(4+tmp) 
       vsmVisParam(i)%arrow_num  = intdata(5+tmp) 
       vsmVisParam(i)%ptcl_num   = intdata(6+tmp) 
       vsmVisParam(i)%light_num  = intdata(7+tmp) 
       if(intdata(8+tmp) == 1) then
         vsmVisParam(i)%depthq = .true.
       else
         vsmVisParam(i)%depthq = .false.
       end if

       do k=1, vsmMaxLight
         if(intdata(8+k+tmp) == 1) then
           vsmVisParam(i)%light_used(k) = .true.
         else
           vsmVisParam(i)%light_used(k) = .false.
         end if
       end do

       if(vsmDbg) then
         call vsmLogPrint('--------------------')
         call vsmLogPrint('viskind = ', i)
         call vsmLogPrint('isonum = ', vsmVisParam(i)%iso_num)
         call vsmLogPrint('slicenum = ', vsmVisParam(i)%slice_num)
         call vsmLogPrint('volrennum = ', vsmVisParam(i)%volren_num)
         call vsmLogPrint('slinenum = ', vsmVisParam(i)%sline_num)
         call vsmLogPrint('arrownum = ', vsmVisParam(i)%arrow_num)
         call vsmLogPrint('num_ptcl = ', vsmVisParam(i)%ptcl_num)
         call vsmLogPrint('light_num = ', vsmVisParam(i)%light_num)
         call vsmLogPrint('depthq = ', vsmVisParam(i)%depthq)
         call vsmLogPrint('light_used = ', vsmVisParam(i)%light_used)
         call vsmLogPrint('--------------------')
       end if
     j = j + 1
     end do
    tmp = num*vsmNumViskind
    vsmImageData%width  = intdata(tmp+1)
    vsmImageData%height = intdata(tmp+2)
    vsmCameraNum   = intdata(tmp+3) 
    vsmLightNum    = intdata(tmp+4)

    if(vsmDbg) then
      call vsmLogPrint('vsmNumViskind = ', vsmNumViskind)
      call vsmLogPrint('width = ', vsmImageData%width)
      call vsmLogPrint('height = ', vsmImageData%height)
      call vsmLogPrint('vsmCameraNum = ', vsmCameraNum)
      call vsmLogPrint('vsmLightNum = ', vsmLightNum)
    end if
   end if

 end subroutine

end module vsmParamBroadcast

