module vsmParamReader

 use vsmConst
 use vsmFunction
 use vsmParam
 use vsmImage
 use vsmLightData
 use vsmCameraData
 use vsmDepthQ
 use vsmColormap
 use vsmColor
 use vsmData
 use vsmFrame

 implicit none
 private

 ! variables used only in this file
 integer, parameter :: token_len = 64
 integer, parameter :: token_num = 20
 integer, parameter :: buf_size  = 512
! character(buf_size), private :: buf
 logical, save :: imagesize, lightparam, camparam, visualization
 integer,save :: backup = 0

 public vsmCheckParam, vsmInitParamReader, vsmInvestigateFiles, vsmReadParam
 contains

 subroutine vsmInitParamReader
   imagesize     = .false.
   lightparam    = .false.
   camparam      = .false.
   visualization = .false.
 end subroutine vsmInitParamReader

 subroutine vsmCheckParam
   integer :: i, j, tmp
   integer :: nscal, nvect, nptcl
   integer :: comm, rank, psize

   call vsmGetMPIVars(comm, rank, psize)

   if(rank /= 0) return

   if(.not.imagesize) then
      call vsmError("imagesize was not set")
   end if

   if(.not.lightparam) then
      call vsmError("lightparam was not set")
   end if

   if(.not. camparam) then
      call vsmError("cameraparam was not set")
   end if

   if(.not. visualization ) then
      call vsmError("visualization param was not specified")
   end if
   nscal = vsmDataInfo%NumOfScal
   nvect = vsmDataInfo%NumOfVect
   nptcl = vsmDataInfo%NumOfPtcl

   do j=1, vsmNumVisKind
     do i=1, vsmVisParam(j)%iso_num
       tmp = vsmVisParam(j)%isoparam(i)%nscal
       if(tmp < 1 .or. tmp > nscal) then
         call vsmLogPrint("iso-nscal", tmp)
         call vsmLogPrint("nscal", nscal)
         call vsmError("ScalNum of isoparam is bad.")
       end if
     end do

     do i=1, vsmVisParam(j)%volren_num
       tmp = vsmVisParam(j)%volparam(i)%nscal
       if(tmp < 1 .or. tmp > nscal) then
         call vsmLogPrint("volren-nscal", tmp)
         call vsmLogPrint("nscal", nscal)
         call vsmError("ScalNum of volparam is bad.")
       end if

       if(.not.vsmSColormap(tmp)%set) then
         call vsmError("colormap for scalar (volren) is not set.")
       end if

     end do

     do i=1, vsmVisParam(j)%slice_num
       tmp = vsmVisParam(j)%sliceparam(i)%nscal
       if(tmp < 1 .or. tmp > nscal) then
         call vsmLogPrint("slice-nscal", tmp)
         call vsmLogPrint("nscal", nscal)
         call vsmError("ScalNum of sliceparam is bad.")
       end if

       if(.not.vsmSColormap(tmp)%set) then
         call vsmError("colormap for scalar (slice) is not set.")
       end if
     end do

     do i=1, vsmVisParam(j)%arrow_num
       tmp = vsmVisParam(j)%arrowparam(i)%nvec
       if(tmp < 1 .or. tmp > nvect) then
         call vsmLogPrint("arrow-nvec", tmp)
         call vsmLogPrint("nvec", nvect)
         call vsmError("VectNum of arrowparam is bad.")
       end if
     end do

     do i=1, vsmVisParam(j)%sline_num
       tmp = vsmVisParam(j)%slineparam(i)%nvec
       if(tmp < 1 .or. tmp > nvect) then
         call vsmLogPrint("slice-nvec", tmp)
         call vsmLogPrint("nvec", nvect)
         call vsmError("VectNum of sline param is bad.")
       end if
       if(.not.vsmVisParam(j)%slineparam(i)%monocolor) then
         if(.not.vsmVColormap(tmp)%set) then
           call vsmError("colormap for vector (sline) is not set.")
         end if
       end if
     end do

     do i=1, vsmVisParam(j)%ptcl_num
       tmp = vsmVisParam(j)%ptclparam(i)%nptcl
       if(tmp < 1 .or. tmp > nptcl) then
         call vsmLogPrint("ptclnum", tmp)
         call vsmLogPrint("nptcl", nptcl)
         call vsmError("PtclNum of ptclparam is bad.")
       end if
     end do

  end do
 end subroutine vsmCheckParam

 recursive subroutine vsmInvestigateFiles(filename, result)
   character(*), intent(in) :: filename
   logical, intent(out) :: result
   integer :: UnNum
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: iostatus
   logical :: FileExist

   result = .true.
   UnNum = vsmSearchUnitNum("Investigate FIles")

     open(UnNum, file=trim(filename), status='old', action='read', iostat=iostatus)

     if(iostatus > 0) then
       call vsmLogPrint('cant open paramfile '//trim(filename))
       result = .false.
       return
     end if

     do
       call vsmClearTokens(tokens, num_tokens)
       read(UnNum, '(A)', iostat=iostatus) line
       if(iostatus < 0) exit
       if(iostatus > 0) then 
         call vsmError('file read error : investigating config files ')
       end if

       call vsmGetTokens(line, tokens, num_tokens)

       if(num_tokens /= 0) then
         if(tokens(1) .eq. 'INCLUDE') then
            call vsmInvestigateFiles(tokens(2), FileExist)
            if(FIleExist .eqv. .false.) then
              close(UnNum)
              result = .false.
              return
            end if
         end if
       end if
     end do
   close(UnNum)

 end subroutine vsmInvestigateFiles

 recursive subroutine vsmReadParam(filename, root)
   character(*), intent(in) :: filename
   logical, intent(in), optional :: root
   !integer, intent(in) :: UnNum
   integer :: UnNum, UnNumb
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: ival1, ival2
   real :: fval1, fval2, fval3
   real(kind = vdp) :: dval1, dval2
   integer :: iostatus
   character(len=token_len) :: temp_
   character(len=4) :: BUpNum

   if(present(root) .and. vsmReInitMode) then
     backup = backup + 1
     UnNum = vsmSearchUnitNum("ReInit")

     if(backup > 2) then
     open(UnNum, file=trim(vsmConfigChangeFile), action='readwrite', &
         position='append', iostat=iostatus)
     else
     open(UnNum, file=trim(vsmConfigChangeFile), status='replace', &
         iostat=iostatus)
     endif

     if(iostatus > 0) then
       call vsmError("cant open config change file")
     end if

     write(UnNum, *) "ConfigFile = ", backup
     close(UnNum)
   end if

   UnNum = vsmSearchUnitNum("ReadParam")

   open(UnNum, file=trim(filename), status='old', action='read', iostat=iostatus)
   if(iostatus > 0) then
      call vsmError('cant open paramfile '//trim(filename))
   end if

     do
       call vsmClearTokens(tokens, num_tokens)
       read(UnNum, '(A)', iostat=iostatus) line
       if(iostatus < 0) exit
       if(iostatus > 0) then 
         call vsmError('file read error: reading visualization param ')
       end if

       call vsmGetTokens(line, tokens, num_tokens)
       !print *, num_tokens
       !do i=1, num_tokens
       !  write(*, '(A)', advance='no') trim(tokens(i))
       !end do
       if(num_tokens == 0) then
         cycle
       else
         ! skip comments
         temp_ = tokens(1)
         if(temp_(1:1) == '#') then
          call vsmLogPrint(trim(tokens(1))//" :skipped")
          cycle
         end if

         select case (tokens(1))
         ! other file
           case('INCLUDE')
             call vsmReadParam(tokens(2))

         ! image size
           case('IMAGESIZE')
             ival1 = vsmAtoI(tokens(2))
             ival2 = vsmAtoI(tokens(3))
             call vsmSetImageParam(ival1, ival2)

         ! background color
           case('BGCOLOR')
             fval1 = vsmAtoF(tokens(2))
             fval2 = vsmAtoF(tokens(3))
             fval3 = vsmAtoF(tokens(4))
             call vsmSetBGCParam(fval1, fval2, fval3)

         ! frame size
           case('FRAMEWIDTH')
             ival1 = vsmAtoI(tokens(2))
             call vsmSetFRameWidth(ival1)

         ! light
           case('LIGHT')
             if(num_tokens == 1) call vsmSetLightParam(UnNum)

         ! depth cue
           case('DEPTHCUE')
             ival1 = vsmAtoI(tokens(2))
             dval1 = vsmAtoD(tokens(3))
             dval2 = vsmAtoD(tokens(4))
             call vsmSetDqParam(ival1, dval1, dval2)

         ! camera
           case('CAMERA')
             call vsmSetCameraParam(UnNum)

         ! camera
           case('MULTICAMERAS')
             call vsmSetMultiCameraParam(UnNum)

         ! dataminmax
           case('SCAL_MINMAX')
             ival1 = vsmAtoI(tokens(2))
             dval1 = vsmAtoD(tokens(3))
             dval2 = vsmAtoD(tokens(4))
!             call vsmSetScalMinMax(ival1, dval1, dval2)

           case('VECT_MINMAX')
             ival1 = vsmAtoI(tokens(2))
             dval1 = vsmAtoD(tokens(3))
             dval2 = vsmAtoD(tokens(4))
!             call vsmSetVectMinMax(ival1, dval1, dval2)

         ! color map
           case('SCAL_COLORMAP')
             call vsmSetSCmap(UnNum)

           case('VECT_COLORMAP')
             call vsmSetVCmap(UnNum)

           case('PTCL_COLORMAP')
             call vsmSetPCmap(UnNum)

         ! visualization
           case ('VISUALIZATION')
             call vsmAddVisualization
             call vsmSetvisualization(UnNum)

         ! point cloud
!           case ('IMAGE_POINTCLOUD_ON')
!             vsmPCmode = 1

         ! point cloud
!           case ('ONLY_POINTCLOUD_ON')
!             vsmPCmode = 2

           case default
             call vsmError("Invalid Token : "//trim(line))
         end select
       end if
     end do
   close(UnNum)
 
   if(.not.vsmReInitMode) return
   ! copying config file
   call vsmInt2Asc4(backup, BUpNum)
   UnNum = vsmSearchUnitNum("ReadParam")
   open(UnNum, file=trim(filename), action='readwrite', iostat=iostatus)
   if(iostatus > 0) call vsmError('cant open paramfile')

   UnNumb = vsmSearchUnitNum("CopyParam")
   open(UnNumb, file=trim(trim(filename)//trim(adjustl('.bak'))//trim(adjustl(BUpNum))), &
        action='write', iostat=iostatus)
   if(iostatus > 0) call vsmError('cant open file for copying paramfile')

     do
       read(UnNum, '(A)', iostat=iostatus) line
       if(iostatus < 0) exit
       if(iostatus > 0) then 
         call vsmError('file read error : reading visualization param ')
       end if
       write(UnNumb, '(A)') line(1:len_trim(line))
     end do
     close(UnNumb)
   close(UnNum, status='delete')

 end subroutine vsmReadParam

 ! -----------  private subroutines ---------------!

 subroutine vsmSetImageParam(w, h)
   integer, intent(in) :: w, h

   call vsmLogPrint("vsmSetImageParam")
   if(imagesize) return ! already set
   call vsmSetImageSize(w, h)
   imagesize = .true.
   if(vsmDbg) then
     call vsmLogPrint("width  = ", w)
     call vsmLogPrint("height = ", h)
     call vsmLogPrint("end image")
   end if
 end subroutine vsmSetImageParam

 subroutine vsmSetBGCParam(r,g,b)
   real, intent(in) :: r,g,b
   real, dimension(3) :: rgb

   call vsmLogPrint("vsmSetBGCParam")
   rgb(1) = r; rgb(2) = g; rgb(3) = b
   
   call vsmSetBGcolor(rgb)
   if(vsmDbg) then
     call vsmLogPrint("bgcolor = ", rgb)
     call vsmLogPrint("end bgcolor")
   end if
 end subroutine vsmSetBGCParam

 subroutine vsmSetDqParam(dqkind, near, far)
   integer, intent(in) :: dqkind
   real(kind=vdp), intent(in) :: near, far
   real(kind=vdp), dimension(2) :: dist

   call vsmLogPrint("vsmSetDqParam")
   dist(1) = near;  dist(2) = far
   call vsmSetDepthParam(dqkind, dist)
   if(vsmDbg) then
     call vsmLogPrint("depthcue  kind:", dqkind)
     call vsmLogPrint("dist:", dist)
     call vsmLogPrint("end dq")
   end if
 end subroutine vsmSetDqParam

 subroutine vsmSetLightParam(UnNum)
   integer, intent(in) :: UnNum
   real, dimension(3) :: pos, amb, dif, spec
   logical :: para
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: iostatus, lnpo
   character(len=token_len) :: temp_

   if(vsmLightNum+1 > vsmMaxLight) call vsmError("too many light")
   lnpo = vsmLightNum+1
   pos(1:3) = vsmLightDefPos(1:3, lnpo)
   amb(1:3) = vsmLightDefAmb(1:3, lnpo)
   dif(1:3) = vsmLightDefDif(1:3, lnpo)
   spec(1:3) = vsmLightSpec(1:3, lnpo)
   para = .true.
   call vsmLogPrint("vsmSetLightParam")
   do
     call vsmClearTokens(tokens, num_tokens)
     read(UnNum, '(A)', iostat=iostatus) line
     if(iostatus /= 0) then
       call vsmLogPrint("iostat = ", iostatus)
       call vsmError('file read error : rading light param ')
     end if

     call vsmGetTokens(line, tokens, num_tokens)

    if(num_tokens == 0) cycle
     temp_ = tokens(1)
     if(temp_(1:1) == '#') then
       call vsmLogPrint(trim(tokens(1))//" :skipped")
       cycle
     end if

     select case (tokens(1))
       case('PARALLEL')
         para = .true.

       case('POINT')
         para = .false.

       case('POSITION')
         pos(1) = vsmAtoF(tokens(2))
         pos(2) = vsmAtoF(tokens(3))
         pos(3) = vsmAtoF(tokens(4))

       case('AMB')
         amb(1) = vsmAtoF(tokens(2))
         amb(2) = vsmAtoF(tokens(3))
         amb(3) = vsmAtoF(tokens(4))

       case('DIF')
         dif(1) = vsmAtoF(tokens(2))
         dif(2) = vsmAtoF(tokens(3))
         dif(3) = vsmAtoF(tokens(4))

       case('SPEC')
         spec(1) = vsmAtoF(tokens(2))
         spec(2) = vsmAtoF(tokens(3))
         spec(3) = vsmAtoF(tokens(4))

       case('END_LIGHT')
         exit

       case default
        call vsmError("Invalid Token in LIGHT : "//trim(line))

     end select

   end do

   call vsmAddLight(para, pos, amb, dif, spec)
   lightparam = .true.
   if(vsmDbg) then
     call vsmLogPrint("para = ", para)
     call vsmLogPrint("pos = ", pos)
     call vsmLogPrint("amb = ", amb)
     call vsmLogPrint("dif = ", dif)
     call vsmLogPrint("spec = ", spec)
     call vsmLogPrint("end light ")
   end if
 end subroutine vsmSetLightParam

 subroutine vsmSetCameraParam(UnNum)
   integer, intent(in) :: UnNum
   real(kind=vdp), dimension(3) :: pos, front, up
   real(kind=vdp) :: fovy, width
   real(kind=vdp), dimension(2) :: nearfar
   logical :: pers
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: iostatus
   character(len=token_len) :: temp_

   call vsmLogPrint("vsmSetCameraParam")
   fovy = 0.0_vdp; width=0.0_vdp
   pers = .true.
   do
     call vsmClearTokens(tokens, num_tokens)
     read(UnNum, '(A)', iostat=iostatus) line
     if(iostatus /= 0) then
       call vsmLogPrint("iostat = ", iostatus)
       call vsmError('file read error : rading camera param ')
     end if

     call vsmGetTokens(line, tokens, num_tokens)

     if(num_tokens == 0) cycle

     temp_ = tokens(1)
     if(temp_(1:1) == '#') then
       call vsmLogPrint(trim(tokens(1))//" :skipped")
       cycle
     end if

     select case (tokens(1))
       case('ORTHOGONAL')
         pers = .false.

       case('WIDTH')
         width = vsmAtoD(tokens(2))

       case('POSITION')
         pos(1) = vsmAtoD(tokens(2))
         pos(2) = vsmAtoD(tokens(3))
         pos(3) = vsmAtoD(tokens(4))

       case('FRONT')
         front(1) = vsmAtoD(tokens(2))
         front(2) = vsmAtoD(tokens(3))
         front(3) = vsmAtoD(tokens(4))

       case('UP')
         up(1) = vsmAtoD(tokens(2))
         up(2) = vsmAtoD(tokens(3))
         up(3) = vsmAtoD(tokens(4))

       case('NEARFAR')
         nearfar(1) = vsmAtoD(tokens(2))
         nearfar(2) = vsmAtoD(tokens(3))

       case('FOVY')
         fovy = vsmAtoF(tokens(2))

       case('END_CAMERA')
         exit

       case default
        call vsmError("Invalid Token in CAMERA : "//trim(line))

     end select

   end do

   call vsmSetCamera(nearfar(1), nearfar(2), pos, up, front, pers, fovy, width)

   camparam = .true.
   if(vsmDbg) then
     call vsmLogPrint("pers = ", pers)
     call vsmLogPrint("pos = ", pos)
     call vsmLogPrint("front = ", front)
     call vsmLogPrint("up = ", up)
     call vsmLogPrint("nearfar = ", nearfar)
     call vsmLogPrint("fovy = ", fovy)
     call vsmLogPrint("end camera")
   end if
 end subroutine vsmSetCameraParam

 subroutine vsmSetMultiCameraParam(UnNum)
   integer, intent(in) :: UnNum
   real(kind=vdp), dimension(3, vsmMaxCamera) :: pos, front
   real(kind=vdp) :: fovy, width
   real(kind=vdp), dimension(2) :: nearfar
   logical :: pers
   real(kind=vdp), dimension(3) :: up
   real(kind=vdp), dimension(3) :: axis, center, x, y, base
   real(kind=vdp) :: radius, dw
   integer :: i, NumCamera
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: iostatus
   character(len=token_len) :: temp_

   call vsmLogPrint("vsmSetMultiCameraParam")
   fovy = 0.0_vdp; width=0.0_vdp
   pers = .true.
   NumCamera = 0
   do
     call vsmClearTokens(tokens, num_tokens)
     read(UnNum, '(A)', iostat=iostatus) line
     if(iostatus /= 0) then
       call vsmLogPrint("iostat = ", iostatus)
       call vsmError('file read error : rading camera param ')
     end if

     call vsmGetTokens(line, tokens, num_tokens)

     if(num_tokens == 0) cycle

     temp_ = tokens(1)
     if(temp_(1:1) == '#') then
       call vsmLogPrint(trim(tokens(1))//" :skipped")
       cycle
     end if

     select case (tokens(1))
       case('ORTHOGONAL')
         pers = .false.

       case('WIDTH')
         width = vsmAtoD(tokens(2))

       case('AXIS')
         axis(1) = vsmAtoD(tokens(2))
         axis(2) = vsmAtoD(tokens(3))
         axis(3) = vsmAtoD(tokens(4))

       case('RADIUS')
         radius = vsmAtoD(tokens(2))

       case('NCAMERA')
         NumCamera = vsmAtoI(tokens(2))

       case('CENTER')
         center(1) = vsmAtoD(tokens(2))
         center(2) = vsmAtoD(tokens(3))
         center(3) = vsmAtoD(tokens(4))

       case('NEARFAR')
         nearfar(1) = vsmAtoD(tokens(2))
         nearfar(2) = vsmAtoD(tokens(3))

       case('FOVY')
         fovy = vsmAtoF(tokens(2))

       case('END_MULTICAMERAS')
         exit

       case default
        call vsmError("Invalid Token in CAMERA : "//trim(line))

     end select

   end do
   !now
   if(NumCamera == 0) call vsmError(" camera num = 0")
   call vsmNormVec(axis)
   up = axis; 
   if(axis(1) == 1.0_vdp .and. axis(2) == 0.0_vdp .and. axis(3) == 0.0_vdp) then
     base(1) = 0.0_vdp; base(2) = 1.0_vdp; base(3) = 0.0_vdp
   else
     base(1) = 1.0_vdp; base(2) = 0.0_vdp; base(3) = 0.0_vdp
   end if
   call vsmCrossProd(x, up, base)
   call vsmNormVec(x)
   call vsmCrossProd(y, up, x)
   dw = 2.0_vdp*pi/NumCamera
   
   do i=1, NumCamera
     pos(1:3, i) = cos(dw*(i-1))*x(1:3)+sin(dw*(i-1))*y(1:3)
     front(1:3, i) = - pos(1:3, i)
     pos(1:3, i) = radius*pos(1:3, i) + center(1:3)
   end do

   do i=1, NumCamera
     call vsmSetCamera(nearfar(1), nearfar(2), pos(1:3, i), up, front(1:3,i), pers, fovy, width)
   end do

   camparam = .true.
   vsmMultipleCameras = .true.
   if(vsmDbg) then
     call vsmLogPrint("pers = ", pers)
     call vsmLogPrint("aixs = ", axis)
     call vsmLogPrint("center = ", center)
     call vsmLogPrint("radius = ", radius)
     call vsmLogPrint("num = ", NumCamera)
     call vsmLogPrint("nearfar = ", nearfar)
     call vsmLogPrint("fovy = ", fovy)
     call vsmLogPrint("end camera")
   end if
 end subroutine vsmSetMultiCameraParam

 subroutine vsmSetisosurface(UnNum)
   integer, intent(in) :: UnNum
   real, dimension(3) :: amb, dif, spec
   real :: shin, alpha
   real(kind=vdp) :: level
   integer :: nscal, isonum, coloring
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: iostatus
   character(len=token_len) :: temp_

   call vsmLogPrint("vsmSetisosurface")
   ! defalt color
   isonum = vsmGetCurrentIsonum() + 1
   if(isonum > vsmMaxIsosurf) call vsmError("too many isosurf")
   amb(1:3) = vsmIsoDefAmb(1:3, isonum)
   dif(1:3) = vsmIsoDefDif(1:3, isonum)
   spec(1:3) = vsmIsoDefSpec(1:3, isonum)
   shin = vsmIsoDefShin(isonum)
   alpha = iso_defa_alpha
   coloring = -1
   do
     call vsmClearTokens(tokens, num_tokens)
     read(UnNum, '(A)', iostat=iostatus) line
     if(iostatus /= 0) then
       call vsmLogPrint("iostat = ", iostatus)
       call vsmError('file read error : reading visualization param : isosurface')
     end if

     call vsmGetTokens(line, tokens, num_tokens)

     if(num_tokens==0) cycle
     temp_ = tokens(1)
     if(temp_(1:1) == '#') then
       call vsmLogPrint(trim(tokens(1))//" :skipped")
       cycle
     end if

     select case (tokens(1))
       case('SCALNUM')
         nscal = vsmAtoI(tokens(2))

       case('LEVEL')
         level = vsmAtoD(tokens(2))

       case('AMB')
         amb(1) = vsmAtoF(tokens(2))
         amb(2) = vsmAtoF(tokens(3))
         amb(3) = vsmAtoF(tokens(4))

       case('DIF')
         dif(1) = vsmAtoF(tokens(2))
         dif(2) = vsmAtoF(tokens(3))
         dif(3) = vsmAtoF(tokens(4))

       case('SPEC')
         spec(1) = vsmAtoF(tokens(2))
         spec(2) = vsmAtoF(tokens(3))
         spec(3) = vsmAtoF(tokens(4))

       case('SHIN')
         shin = vsmAtoF(tokens(2))

       case('ALPHA')
         alpha = vsmAtoF(tokens(2))

       case('END_ISOSURFACE')
         exit

       case default
        call vsmError("Invalid Token in ISOSURFACE : "//trim(line))

     end select
   end do
   call vsmSetIsoParam(level, nscal, alpha, amb, dif, spec, shin, coloring)

   if(vsmDbg) then
     call vsmLogPrint("level = ", level)
     call vsmLogPrint("scalnum = ", nscal)
     call vsmLogPrint("amb = ", amb)
     call vsmLogPrint("dif = ", dif)
     call vsmLogPrint("spec = ", spec)
     call vsmLogPrint("shin = ", shin)
     call vsmLogPrint("alpha = ", alpha)
     call vsmLogPrint("end isosurface")
   end if
 end subroutine vsmSetisosurface

 subroutine vsmSetSlice(UnNum)
   integer, intent(in) :: UnNum
   real :: amb, dif, spec, shin, alpha
   real(kind=vdp), dimension(4) :: eq
   real(kind=vdp) :: dptemp
   integer :: nscal
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens, i
   integer :: iostatus
   character(len=token_len) :: temp_

   call vsmLogPrint("vsmSetSlice")
   ! defalt color
   amb = vsmSliceDefAmb
   dif = vsmSliceDefDif
   spec = vsmSliceDefSpec
   shin = vsmSliceDefShin
   alpha = vsmSliceDefAlpha
   eq = vsmDefPlaneEq
   do
     call vsmClearTokens(tokens, num_tokens)
     read(UnNum, '(A)', iostat=iostatus) line
     if(iostatus /= 0) then
       call vsmLogPrint("iostat = ", iostatus)
       call vsmError('file read error : reading visualization param : slice')
     end if

     call vsmGetTokens(line, tokens, num_tokens)

     if(num_tokens == 0) cycle
     temp_ = tokens(1)
     if(temp_(1:1) == '#') then
       call vsmLogPrint(trim(tokens(1))//" :skipped")
       cycle
     end if

     select case (tokens(1))
       case('SCALNUM')
         nscal = vsmAtoI(tokens(2))

       case('AMB')
         amb = vsmAtoF(tokens(2))

       case('DIF')
         dif = vsmAtoF(tokens(2))

       case('SPEC')
         spec = vsmAtoF(tokens(2))

       case('SHIN')
         shin = vsmAtoF(tokens(2))

       case('ALPHA')
         alpha = vsmAtoF(tokens(2))

       case('EQ')
         if(num_tokens < 4) call vsmError("invalid Token for  Slice equation")
         eq(1) = vsmAtoD(tokens(2))
         eq(2) = vsmAtoD(tokens(3))
         eq(3) = vsmAtoD(tokens(4))
         eq(4) = vsmAtoD(tokens(5))

       case('END_SLICE')
         exit

        case default
          call vsmError("Invalid Token in SLICE : "//trim(line))

     end select
   end do
   dptemp = sqrt(eq(1)*eq(1)+eq(2)*eq(2)+eq(3)*eq(3))
   if(dptemp <= 0.0_vdp) call vsmError("slice eq is strange ...")
   do i=1, 4
     eq(i) = eq(i)/dptemp
   end do
   call vsmSetSliceParam(eq, nscal, alpha, amb, dif, spec, shin)
   if(vsmDbg) then
     call vsmLogPrint("scalnum = ", nscal)
     call vsmLogPrint("f(x) = ax+by+cz+d : ", eq)
     call vsmLogPrint("amb = ", amb)
     call vsmLogPrint("dif = ", dif)
     call vsmLogPrint("spec = ", spec)
     call vsmLogPrint("shin = ", shin)
     call vsmLogPrint("alpha = ", alpha)
     call vsmLogPrint("end slice")
   end if
 end subroutine vsmSetSlice

 subroutine vsmSetvolumerend(UnNum)
   integer, intent(in) :: UnNum
   real :: amb, dif, spec, shin
   integer :: nscal
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: iostatus
   character(len=token_len) :: temp_

   call vsmLogPrint("vsmSetvolumerendering")
   amb = vsmVolDefAmb
   dif = vsmVolDefDif
   spec = vsmVolDefSpec
   shin = vsmVolDefShin
   do
     call vsmClearTokens(tokens, num_tokens)
     read(UnNum, '(A)', iostat=iostatus) line
     if(iostatus /= 0) then
       call vsmLogPrint("iostat = ", iostatus)
       call vsmError('file read error : reading visualization param : volumerendering')
     end if

     call vsmGetTokens(line, tokens, num_tokens)

     if(num_tokens == 0) cycle
     temp_ = tokens(1)
     if(temp_(1:1) == '#') then
       call vsmLogPrint(trim(tokens(1))//" :skipped")
       cycle
     end if

     select case (tokens(1))
       case('SCALNUM')
         nscal = vsmAtoI(tokens(2))

       case('AMB')
         amb = vsmAtoF(tokens(2))

       case('DIF')
         dif = vsmAtoF(tokens(2))

       case('SPEC')
         spec = vsmAtoF(tokens(2))

       case('SHIN')
         shin = vsmAtoF(tokens(2))

       case('END_VOLUME_RENDERING')
         exit

       case default
        call vsmError("Invalid Token in VOLUME RENDERING : "//trim(line))

     end select
   end do
   call vsmSetVolren(nscal, amb, dif, spec, shin)
   if(vsmDbg) then
     call vsmLogPrint("scalnum = ", nscal)
     call vsmLogPrint("amb = ", amb)
     call vsmLogPrint("dif = ", dif)
     call vsmLogPrint("spec = ", spec)
     call vsmLogPrint("shin = ", shin)
   end if
 end subroutine vsmSetvolumerend

 subroutine vsmSetseed(UnNum)
   integer, intent(in) :: UnNum
   real(kind=vdp), dimension(3) :: x
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: iostatus
   character(len=token_len) :: temp_

   call vsmLogPrint("vsmSetseed")
   do
     call vsmClearTokens(tokens, num_tokens)
     read(UnNum, '(A)', iostat=iostatus) line
     if(iostatus /= 0) then
       call vsmLogPrint("iostat = ", iostatus)
       call vsmError('file read error : reading visualization param : streamline seed')
     end if

     call vsmGetTokens(line, tokens, num_tokens)

     if(num_tokens == 0) cycle
     temp_ = tokens(1)
     if(temp_(1:1) == '#') then
       call vsmLogPrint(trim(tokens(1))//" :skipped")
       cycle
     end if

     select case (tokens(1))
       case('END_SEED')
         exit

       case default
         if(num_tokens < 3) call vsmError("invalid postion of seed : "//trim(line))
         x(1) = vsmAtoD(tokens(1))
         x(2) = vsmAtoD(tokens(2))
         x(3) = vsmAtoD(tokens(3))
         call vsmSetSeedParam(x)
         call vsmLogPrint("seed = ", x)

     end select
   end do
   call vsmLogPrint("vsmSetseed Ends")
 end subroutine vsmSetseed

 subroutine vsmSetSline(UnNum)
   integer, intent(in) :: UnNum
   real, dimension(3) :: amb, dif, spec
   real :: shin
   real(kind=vdp) :: rad
   integer :: nvec
   logical :: mono
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: iostatus
   character(len=token_len) :: temp_

   call vsmLogPrint("vsmSetSline")
   mono = .false.
   amb(1:3) = 0
   dif(1:3) = 0.0
   spec(1:3)=0.0
   do
     call vsmClearTokens(tokens, num_tokens)
     read(UnNum, '(A)', iostat=iostatus) line
     if(iostatus /= 0) then
       call vsmLogPrint("iostat = ", iostatus)
       call vsmError('file read error : reading visualization param : streamline')
     end if

     call vsmGetTokens(line, tokens, num_tokens)

     if(num_tokens == 0) cycle
     temp_ = tokens(1)
     if(temp_(1:1) == '#') then
       call vsmLogPrint(trim(tokens(1))//" :skipped")
       cycle
     end if

     select case (tokens(1))
       case('VECTNUM')
         nvec = vsmAtoI(tokens(2))

       case('AMB')
         amb(1) = vsmAtoF(tokens(2))
         if(num_tokens == 4) then
           amb(2) = vsmAtoF(tokens(3))
           amb(3) = vsmAtoF(tokens(4))
         end if

       case('DIF')
         dif(1) = vsmAtoF(tokens(2))
         if(num_tokens == 4) then
           dif(2) = vsmAtoF(tokens(3))
           dif(3) = vsmAtoF(tokens(4))
         end if

       case('SPEC')
         spec(1) = vsmAtoF(tokens(2))
         if(num_tokens == 4) then
           spec(2) = vsmAtoF(tokens(3))
           spec(3) = vsmAtoF(tokens(4))
         end if

       case('SHIN')
         shin = vsmAtoF(tokens(2))

       case('RADIUS')
         rad = vsmAtoD(tokens(2))

       case('MONO')
         mono = .true.

       case('COLOR')
         mono = .false.

       case('SEED')
         call vsmSetseed(UnNum)

       case('END_STREAM_LINES')
         exit

       case default
        call vsmError("Invalid Token in STREAMLINE : "//trim(line))

     end select
   end do
   call vsmLogPrint("vsmSetSlineParam")
   call vsmSetSlineParam(nvec, mono, amb, dif, spec, shin, rad)
   if(vsmDbg) then
     call vsmLogPrint("vectnum = ", nvec)
     call vsmLogPrint("amb = ", amb)
     call vsmLogPrint("dif = ", dif)
     call vsmLogPrint("spec = ", spec)
     call vsmLogPrint("shin = ", shin)
     call vsmLogPrint("mono = ", mono)
     call vsmLogPrint("end sline")
   end if
 end subroutine vsmSetSline

 subroutine vsmSetArseed(UnNum)
   integer, intent(in) :: UnNum
   real(kind=vdp), dimension(3) :: x
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: iostatus
   character(len=token_len) :: temp_

   call vsmLogPrint("vsmSetArseed")
   do
     call vsmClearTokens(tokens, num_tokens)
     read(UnNum, '(A)', iostat=iostatus) line
     if(iostatus /= 0) then
       call vsmLogPrint("iostat = ", iostatus)
       call vsmError('file read error : reading visualization param : streamline seed')
     end if

     call vsmGetTokens(line, tokens, num_tokens)

     if(num_tokens == 0) cycle
     temp_ = tokens(1)
     if(temp_(1:1) == '#') then
       call vsmLogPrint(trim(tokens(1))//" :skipped")
       cycle
     end if

     select case (tokens(1))
       case('END_SEED')
         exit

       case default
         if(num_tokens < 3) call vsmError("invalid postion of seed : "//trim(line))
         x(1) = vsmAtoD(tokens(1))
         x(2) = vsmAtoD(tokens(2))
         x(3) = vsmAtoD(tokens(3))
         call vsmSetArSeedParam(x)
         call vsmLogPrint("seed = ", x)

     end select
   end do
   call vsmLogPrint("vsmSetseed Ends")
 end subroutine vsmSetArseed

 subroutine vsmSetArrow(UnNum)
   integer, intent(in) :: UnNum
   integer :: nvec, mode
   real(kind=vdp), dimension(4) :: eq
   real(kind=vdp) :: dens, rad, length
   logical :: vari
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: iostatus
   character(len=token_len) :: temp_

   call vsmLogPrint("vsmSetArrows")

   eq(1:4) = vsmDefPlaneEq(1:4)
   rad = vsmDefArrowRad; length = vsmDefArrowLen
   mode = 0; vari = .false.
   do
     call vsmClearTokens(tokens, num_tokens)
     read(UnNum, '(A)', iostat=iostatus) line
     if(iostatus /= 0) then
       call vsmLogPrint("iostat = ", iostatus)
       call vsmError('file read error : reading visualization param : arrows')
     end if

     call vsmGetTokens(line, tokens, num_tokens)

     if(num_tokens == 0) cycle
     temp_ = tokens(1)
     if(temp_(1:1) == '#') then
       call vsmLogPrint(trim(tokens(1))//" :skipped")
       cycle
     end if

     select case (tokens(1))
       case('VECTNUM')
         nvec = vsmAtoI(tokens(2))

       case('DENSITY')
         dens = vsmAtoD(tokens(2))

       case('RADIUS')
         rad = vsmAtoD(tokens(2))

       case('LENGTH')
         length = vsmAtoD(tokens(2))

       case('EQ')
        if(num_tokens < 5) call vsmError("invalid eq in Arrows"//trim(line))
         eq(1) = vsmAtoF(tokens(2))
         eq(2) = vsmAtoF(tokens(3))
         eq(3) = vsmAtoF(tokens(4))
         eq(4) = vsmAtoF(tokens(5))
         mode = 0

       case('END_ARROWS')
         exit

       case default
        call vsmError("Invalid Token in ARROWS : "//trim(line))

     end select
   end do
   call vsmSetArrowParam(eq, nvec, dens, rad, length, mode, vari)
   if(vsmDbg) then
     call vsmLogPrint("vectnum = ", nvec)
     call vsmLogPrint("mode = ", mode)
     if(mode == 0) call vsmLogPrint("f(x) = ax+by+cz+d : ", eq)
     call vsmLogPrint("dens = ", dens)
     call vsmLogPrint("rad = ", rad)
     call vsmLogPrint("length = ", length)
     call vsmLogPrint("variable = ", vari)
     call vsmLogPrint("end arrows")
   end if
 end subroutine vsmSetArrow

 subroutine vsmSetParticle(UnNum)
   integer, intent(in) :: UnNum
   real, dimension(3) :: amb, dif, spec
   real :: shin
   real(kind=vdp) :: radius
   integer :: nptcl
   logical :: mono
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: iostatus
   character(len=token_len) :: temp_

   call vsmLogPrint("vsmSetParticle")
   radius = vsmDefPtclRad
   mono = .true.
   do
     call vsmClearTokens(tokens, num_tokens)
     read(UnNum, '(A)', iostat=iostatus) line
     if(iostatus /= 0) then
       call vsmLogPrint("iostat = ", iostatus)
       call vsmError('file read error : reading visualization param : isosurface')
     end if

     call vsmGetTokens(line, tokens, num_tokens)

     if(num_tokens == 0) cycle
     temp_ = tokens(1)
     if(temp_(1:1) == '#') then
       call vsmLogPrint(trim(tokens(1))//" :skipped")
       cycle
     end if

     select case (tokens(1))
       case('PTCLNUM')
         nptcl = vsmAtoI(tokens(2))
         if(nptcl < 1 .or. nptcl > vsmMaxPtcl) call vsmError("ptcl num is strange")
         amb(1:3) = vsmPtclDefAmb(1:3, nptcl)
         dif(1:3) = vsmPtclDefDif(1:3, nptcl)
         spec(1:3) = vsmPtclDefSpec(1:3, nptcl)
         shin = vsmPtclDefShin(nptcl)

       case('RADIUS')
         radius = vsmAtoD(tokens(2))

       case('AMB')
         amb(1) = vsmAtoF(tokens(2))
         if(num_tokens == 4) then
           amb(2) = vsmAtoF(tokens(3))
           amb(3) = vsmAtoF(tokens(4))
         end if

       case('DIF')
         dif(1) = vsmAtoF(tokens(2))
         if(num_tokens == 4) then
           dif(2) = vsmAtoF(tokens(3))
           dif(3) = vsmAtoF(tokens(4))
         end if

       case('SPEC')
         spec(1) = vsmAtoF(tokens(2))
         if(num_tokens == 4) then
           spec(2) = vsmAtoF(tokens(3))
           spec(3) = vsmAtoF(tokens(4))
         end if

       case('SHIN')
         shin = vsmAtoF(tokens(2))

       case('COLOR')
         mono = .false.

       case('END_PARTICLE')
         exit

       case default
        call vsmError("Invalid Token in PARTICLE : "//trim(line))

     end select
   end do
   call vsmSetPtclParam(nptcl, mono, radius, amb, dif, spec, shin)
   if(vsmDbg) then
     call vsmLogPrint("radius = ", radius)
     call vsmLogPrint("ptclnum = ", nptcl)
     call vsmLogPrint("amb = ", amb)
     call vsmLogPrint("dif = ", dif)
     call vsmLogPrint("spec = ", spec)
     call vsmLogPrint("shin = ", shin)
     call vsmLogPrint("mono = ", mono)
     call vsmLogPrint("end particle")
   end if
 end subroutine vsmSetParticle

 subroutine vsmSetSCmap(UnNum)
   integer, intent(in) :: UnNum
   real(kind=vdp), dimension(vsmMaxColmap) :: val
   real, dimension(4, vsmMaxColmap) :: color
   integer :: nscal
   integer :: n
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: iostatus
   logical :: es
   character(len=token_len) :: temp_

   es = .false.
   call vsmLogPrint("vsmSetscalcolormap")
   n = 0
   do
     call vsmClearTokens(tokens, num_tokens)
     read(UnNum, '(A)', iostat=iostatus) line
     if(iostatus /= 0) then
       call vsmLogPrint("iostat = ", iostatus)
       call vsmError('file read error : reading visualization param : vsmSColormap')
     end if

     call vsmGetTokens(line, tokens, num_tokens)

     if(num_tokens == 0) cycle
     temp_ = tokens(1)
     if(temp_(1:1) == '#') then
       call vsmLogPrint(trim(tokens(1))//" :skipped")
       cycle
     end if

     select case(tokens(1))
       case('SCALNUM')
         nscal = vsmAtoI(tokens(2))

       case('EQUALLY_SPACED')
         es = .true.

       case('END_SCAL_COLORMAP')
         exit

       case default
         n = n + 1
         if(n > vsmMaxColmap) call vsmError("SCALCOLMAP : number of color exceeded the limit")
         if(num_tokens < 5) call vsmError("SCAL COLORMAP invalid color")
         val(n) = vsmAtoD(tokens(1))
         color(1, n) = vsmAtoF(tokens(2))
         color(2, n) = vsmAtoF(tokens(3))
         color(3, n) = vsmAtoF(tokens(4))
         color(4, n) = vsmAtoF(tokens(5))
         if(n > 1) then
           if(val(n-1) > val(n)) call vsmError("SCAL COLORMAP invalid: values are decreasing")
         end if
         if(vsmDbg) then
            call vsmLogPrint('val = ', val(n))
            call vsmLogPrint('col = ', color(1:4,n))
         end if
     end select
   end do
   call vsmSetSColorArray(nscal, n, val, color, es)
   if(vsmDbg) then
     call vsmLogPrint("scalnum = ", nscal)
     call vsmLogPrint("colnum = ", n)
     call vsmLogPrint("eq space = ", es)
     call vsmLogPrint("end scalcolormap")
   end if

 end subroutine vsmSetSCmap

 subroutine vsmSetVCmap(UnNum)
   integer, intent(in) :: UnNum
   real(kind=vdp), dimension(vsmMaxColmap) :: val
   real, dimension(3, vsmMaxColmap) :: color
   integer :: nvec
   integer :: n
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: iostatus
   logical :: es
   character(len=token_len) :: temp_

   es = .false.
   call vsmLogPrint("vsmSetvcalcolormap")
   n = 0
   do
     call vsmClearTokens(tokens, num_tokens)
     read(UnNum, '(A)', iostat=iostatus) line
     if(iostatus /= 0) then
       call vsmLogPrint("iostat = ", iostatus)
       call vsmError('file read error : reading visualization param: vsmVColormap')
     end if

     call vsmGetTokens(line, tokens, num_tokens)

     if(num_tokens == 0) cycle
     temp_ = tokens(1)
     if(temp_(1:1) == '#') then
       call vsmLogPrint(trim(tokens(1))//" :skipped")
       cycle
     end if

     select case(tokens(1))
       case('VECTNUM')
         nvec = vsmAtoI(tokens(2))

       case('EQUALLY_SPACED')
         es = .true.

       case('END_VECT_COLORMAP')
         exit

       case default
         n = n + 1
         if(n > vsmMaxColmap) call vsmError("VECTCOLMAP : number of color exceeded the limit")
         if(num_tokens < 4) call vsmError("VECT COLORMAP invalid color")
         val(n) = vsmAtoD(tokens(1))
         color(1, n) = vsmAtoF(tokens(2))
         color(2, n) = vsmAtoF(tokens(3))
         color(3, n) = vsmAtoF(tokens(4))
         if(n > 1) then
           if(val(n-1) > val(n)) call vsmError("VECT COLORMAP invalid: values are decreasing")
         end if

         if(vsmDbg) then
            call vsmLogPrint('val = ', val(n))
            call vsmLogPrint('col = ', color(1:3,n))
        end if
     end select
   end do
   call vsmSetVColorArray(nvec, n, val, color, es)
   if(vsmDbg) then
     call vsmLogPrint("vectnum = ", nvec)
     call vsmLogPrint("colnum = ", n)
     call vsmLogPrint("eq space = ", es)
     call vsmLogPrint("end vcalcolormap")
   end if

 end subroutine vsmSetVCmap


 subroutine vsmSetPCmap(UnNum)
   integer, intent(in) :: UnNum
   real(kind=vdp), dimension(vsmMaxColmap) :: val
   real, dimension(4, vsmMaxColmap) :: color
   integer :: nptcl
   integer :: n
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: iostatus
   logical :: es
   character(len=token_len) :: temp_

   es = .false.
   call vsmLogPrint("vsmSetptclcolormap")
   n = 0
   do
     call vsmClearTokens(tokens, num_tokens)
     read(UnNum, '(A)', iostat=iostatus) line
     if(iostatus /= 0) then
       call vsmLogPrint("iostat = ", iostatus)
       call vsmError('file read error : reading visualization param : vsmPColormap')
     end if

     call vsmGetTokens(line, tokens, num_tokens)

     if(num_tokens == 0) cycle
     temp_ = tokens(1)
     if(temp_(1:1) == '#') then
       call vsmLogPrint(trim(tokens(1))//" :skipped")
       cycle
     end if

     select case(tokens(1))
       case('PTCLNUM')
         nptcl = vsmAtoI(tokens(2))

       case('EQUALLY_SPACED')
         es = .true.

       case('END_PTCL_COLORMAP')
         exit

       case default
         if(n > vsmMaxColmap) call vsmError("PTCLCOLMAP : number of color exceeded the limit")
         if(num_tokens < 5) call vsmError("PTCL COLORMAP invalid color")
         n = n + 1
         val(n) = vsmAtoD(tokens(1))
         color(1, n) = vsmAtoF(tokens(2))
         color(2, n) = vsmAtoF(tokens(3))
         color(3, n) = vsmAtoF(tokens(4))
         color(4, n) = vsmAtoF(tokens(5))
         if(n > 1 .and. (val(n-1) > val(n))) call vsmError("PTCL COLORMAP invalid: values are decreasing")
         if(vsmDbg) then
            call vsmLogPrint('val = ', val(n))
            call vsmLogPrint('col = ', color(1:4,n))
         end if
     end select
   end do
   call vsmSetPColorArray(nptcl, n, val, color, es)
   if(vsmDbg) then
     call vsmLogPrint("ptclnum = ", nptcl)
     call vsmLogPrint("colnum = ", n)
     call vsmLogPrint("eq space = ", es)
     call vsmLogPrint("end ptclcolormap")
   end if

 end subroutine vsmSetPCmap

 subroutine vsmSetvisualization(UnNum)
   integer, intent(in) :: UnNum
   integer :: ival
   character(token_len), dimension(token_num) :: tokens
   character(buf_size) :: line
   integer :: num_tokens
   integer :: iostatus
   character(len=token_len) :: temp_

   call vsmLogPrint("vsmSetvisualization")

   do
     call vsmClearTokens(tokens, num_tokens)
     read(UnNum, '(A)', iostat=iostatus) line
     if(iostatus /= 0) then
       call vsmLogPrint("iostat = ", iostatus)
       call vsmError('file read error : reading visualization param ')
     end if

     call vsmGetTokens(line, tokens, num_tokens)

     if(num_tokens == 0) cycle
     temp_ = tokens(1)
     if(temp_(1:1) == '#') then
       call vsmLogPrint(trim(tokens(1))//" :skipped")
       cycle
     end if

     select case (tokens(1))
       case('ISOSURFACE')
         call vsmSetisosurface(UnNum)

       case('SLICE')
         call vsmSetSlice(UnNum)

       case('VOLUME_RENDERING')
         call vsmSetvolumerend(UnNum)

       case('STREAM_LINES')
         call vsmAddSline
         call vsmSetSline(UnNum)

       case('ARROWS')
         call vsmAddArrow
         call vsmSetArrow(UnNum)

       case('PARTICLE')
         call vsmSetParticle(UnNum)

       case('DEPTHCUE')
         call vsmSetDepthQ
         call vsmLogPrint('depthcue on')

       case('LIGHT')
         ival = vsmAtoI(tokens(2))
         call vsmLightOn(ival)

       case('END_VISUALIZATION')
         exit
         
        case default
           call vsmError("Invalid Token in VISUALIZATION : "//trim(line))
    end select

   end do
   visualization = .true.
 end subroutine vsmSetvisualization

 subroutine vsmGetTokens(line, tokens, num)
   character(buf_size), intent(in) :: line
   character(token_len), dimension(token_num), intent(out) :: tokens
   integer, intent(out) :: num
   character(buf_size) :: temp_char
   integer :: spos
   integer :: len_string
   character(buf_size) :: buf

   !buf(1:buf_size) = line(1:buf_size)
   buf = line
   num = 0
   do
     buf = trim(adjustl(buf))
     len_string = len_trim(buf)

     if(len_string == 0) exit

     spos = scan(buf, ' ')
     temp_char = buf

     num = num + 1

     tokens(num) = temp_char(1:spos)
     tokens(num) = trim(tokens(num))

     buf = temp_char(spos+1:)
   end do
 end subroutine vsmGetTokens

 subroutine vsmClearTokens(tkns, ntkns)
   character(token_len), dimension(token_num), intent(inout) :: tkns
   integer, intent(inout) :: ntkns
   integer :: i
   do i=1, token_num
     tkns(i) = ''
   end do
   ntkns = 0
 end subroutine vsmClearTokens

end module vsmParamReader

