module vsmRayCaster
    use vsmConst
    use vsmFunction
    use vsmData
    use vsmCoordinates
    use vsmColormap
    use vsmDepthQ
    use vsmImage
    use vsmLightData
    use vsmCameraData
    use vsmParam
    use vsmIntpol
    use vsmOctree
    use vsmOctree4Iso
#ifndef SERIAL
    use vsmBSwap
    use vsmBTree
#endif
    implicit none
    private

    type vismo_cslice
        integer :: num
        real(kind=vdp) :: val
    end type vismo_cslice

   public vsmScalarVisualization

    contains

 recursive subroutine vsmQSort(cslidat, sbeg, send)
     type(vismo_cslice), dimension(:), intent(inout) :: cslidat
     integer, intent(in) :: sbeg, send
     integer :: smid, smidm, smidp

     if(sbeg < send) then
         call vsmDiv(cslidat, sbeg, send, smid)
         smidp = smid + 1; smidm = smid - 1
         call vsmQSort(cslidat, sbeg, smidm)
         call vsmQSort(cslidat, smidp, send)
     end if
 end subroutine vsmQSort

 subroutine vsmDiv(cslidat, sbeg, send, smid)
   type(vismo_cslice), dimension(:), intent(inout) :: cslidat
   integer, intent(in) :: sbeg, send
   integer, intent(out) :: smid
   integer :: lbeg, lend
   type(vismo_cslice) :: temp_sdat, piv

   lbeg = sbeg; lend = send;
   piv = cslidat(sbeg)

   do
     if(lbeg >= lend) exit
     do
        if(lbeg >= lend .or. cslidat(lend)%val < piv%val) exit
        lend = lend - 1
     end do

     do
        if(lbeg >= lend .or. cslidat(lbeg)%val > piv%val) exit
        lbeg = lbeg + 1
     end do

     if(lbeg < lend) then
        temp_sdat = cslidat(lbeg)
        cslidat(lbeg) = cslidat(lend)
        cslidat(lend) = temp_sdat
     end if
   end do

   cslidat(sbeg) = cslidat(lend)
   cslidat(lend) = piv
   smid = lend
 end subroutine vsmDiv

 subroutine vsmGradScal(i, j, k, nscal, vec)
   integer, intent(in) :: i, j, k, nscal
   real(kind=vdp), dimension(3), intent(out) :: vec
   real(kind=vdp) :: dis, x1, x2

   if(.not.vsmCheckLCoord(i,j,k)) call vsmError("coord is strange:: vsmGradScal")

   if (i == vsmLocalCoord%nx(1)) then
      x1 = vsmScalData(nscal)%scal(i, j, k);
      x2 = vsmScalData(nscal)%scal(i-1, j, k);

      if(vsmLocalCoord%uniform) then
        dis = vsmLocalCoord%dx(1)
      else
        dis = vsmLocalCoord%coordx(i) - vsmLocalCoord%coordx(i-1)
      end if
   else if(i == 1) then
      x1 = vsmScalData(nscal)%scal(2, j, k)
      x2 = vsmScalData(nscal)%scal(1, j, k)

      if (vsmLocalCoord%uniform) then
        dis = vsmLocalCoord%dx(1)
      else
        dis = vsmLocalCoord%coordx(2) - vsmLocalCoord%coordx(1)
      end if
   else
      x1 = vsmScalData(nscal)%scal(i+1, j, k)
      x2 = vsmScalData(nscal)%scal(i-1, j, k)

      if (vsmLocalCoord%uniform) then
        dis = 2.0_vdp*vsmLocalCoord%dx(1)
      else
        dis = vsmLocalCoord%coordx(i+1) - vsmLocalCoord%coordx(i-1)
      end if
   end if

   vec(1) = (x2 - x1)/ dis

   if(isnan(vec(1))) then
	print *,"vec1 in grad",vec(1)
	print *,"x2, x1 in grad ", x2,x1
	print *,"dis in grad", dis
	print *,"i, j, k in grad",i, j,k
	call vsmError("vec(1) is NAN")
   end if

   if (j == vsmLocalCoord%nx(2)) then
      x1 = vsmScalData(nscal)%scal(i, j, k)
      x2 = vsmScalData(nscal)%scal(i, j-1, k)

      if(vsmLocalCoord%uniform) then
        dis = vsmLocalCoord%dx(2)
      else
        dis = vsmLocalCoord%coordy(j) - vsmLocalCoord%coordy(j-1)
      end if
   else if(j == 1) then
      x1 = vsmScalData(nscal)%scal(i, 2, k)
      x2 = vsmScalData(nscal)%scal(i, 1, k)

      if (vsmLocalCoord%uniform) then
        dis = vsmLocalCoord%dx(2)
      else
        dis = vsmLocalCoord%coordy(2) - vsmLocalCoord%coordy(1)
      end if
   else
      x1 = vsmScalData(nscal)%scal(i, j+1, k)
      x2 = vsmScalData(nscal)%scal(i, j-1, k)

      if (vsmLocalCoord%uniform) then
        dis = 2.0_vdp*vsmLocalCoord%dx(2)
      else
        dis = vsmLocalCoord%coordy(j+1) - vsmLocalCoord%coordy(j-1)
      end if
   end if

   vec(2) = (x2 - x1) / dis

   if(isnan(vec(2))) then
	print *,"vec2 in grad",vec(2)
	print *,"vec2 in grad x2, x1", x2,x1
	print *,"dis in grad", dis
	print *,"i, j, k in grad",i, j,k
	call vsmError("vec(2) is NAN")
   end if

   if (k == vsmLocalCoord%nx(3)) then
      x1 = vsmScalData(nscal)%scal(i, j, k)
      x2 = vsmScalData(nscal)%scal(i, j, k-1)

      if(vsmLocalCoord%uniform) then
        dis = vsmLocalCoord%dx(3)
      else
        dis = vsmLocalCoord%coordz(k) - vsmLocalCoord%coordz(k-1)
      end if
   else if(k == 1) then
      x1 = vsmScalData(nscal)%scal(i, j, 2)
      x2 = vsmScalData(nscal)%scal(i, j, 1)

      if (vsmLocalCoord%uniform) then
        dis = vsmLocalCoord%dx(3)
      else
        dis = vsmLocalCoord%coordz(2) - vsmLocalCoord%coordz(1)
      end if
   else
      x1 = vsmScalData(nscal)%scal(i, j, k+1)
      x2 = vsmScalData(nscal)%scal(i, j, k-1)

      if (vsmLocalCoord%uniform) then
        dis = 2.0_vdp*vsmLocalCoord%dx(3)
      else
        dis = vsmLocalCoord%coordz(k+1) - vsmLocalCoord%coordz(k-1)
      end if
   end if

   vec(3) = (x2 - x1) / dis

   if(isnan(vec(3))) then
	print *,"vec3 in grad",vec(3)
	print *,"vec3 in grad x2, x1", x2,x1
	print *,"dis in grad", dis
	print *,"i, j, k in grad",i, j,k
	call vsmError("vec(3) is NAN")
   end if

   !call vsmLogPrint("vsmGradScal: vec = ", vec)
 end subroutine vsmGradScal

 subroutine vsmCopyNormvec (dx, dy, dz, norm_vect, temp_vect, calcd)
   integer, intent(in) :: dx, dy, dz
   real(kind=vdp), dimension(3, 8), intent(inout) :: norm_vect, temp_vect
   integer, dimension(8), intent(inout) :: calcd
   integer :: i

  select case (dx)
    case (-1) ! dx

      select case (dy)
    case (-1) ! dy

        select case (dz)
    case (-1)
      temp_vect(1:3, 8) = norm_vect(1:3, 1)
      calcd(8) = 0

    case (0)
      temp_vect(1:3, 4) = norm_vect(1:3, 1)
      calcd(4) = 0;
      temp_vect(1:3, 8) = norm_vect(1:3, 5)
      calcd(8) = 0

    case (1)
      temp_vect(1:3, 4) = norm_vect(1:3, 5)
      calcd(4) = 0;

    case default
      call vsmError ('failed to copy normal vectors')
          end select

   case (0) ! dy

   select case(dz)
     case (-1)
       temp_vect(1:3, 6) = norm_vect(1:3, 1)
       calcd(6) = 0
       temp_vect(1:3, 8) = norm_vect(1:3, 3)
       calcd(8) = 0

     case (0)
       temp_vect(1:3, 2) = norm_vect(1:3, 1)
       calcd(2) = 0
       temp_vect(1:3, 4) = norm_vect(1:3, 3)
       calcd(4) = 0
       temp_vect(1:3, 6) = norm_vect(1:3, 5)
       calcd(6) = 0
       temp_vect(1:3, 8) = norm_vect(1:3, 7)
       calcd(8) = 0

     case (1)
       temp_vect(1:3, 2) = norm_vect(1:3, 5)
       calcd(2) = 0
       temp_vect(1:3, 4) = norm_vect(1:3, 7)
       calcd(4) = 0

     case default
       call vsmError ('failed to copy normal vectors')
          end select

   case (1) ! dy

   select case (dz)
     case (-1)
       temp_vect(1:3, 6) = norm_vect(1:3, 3)
       calcd(6) = 0

     case (0)
       temp_vect(1:3, 2) = norm_vect(1:3, 3)
       calcd(2) = 0
       temp_vect(1:3, 6) = norm_vect(1:3, 7)
       calcd(6) = 0

     case (1)
       temp_vect(1:3, 2) = norm_vect(1:3, 7)
       calcd(2) = 0

     case default
       call vsmError ('failed to copy normal vectors')

     end select

      case default ! dy
         call vsmError ('failed to copy normal vectors')

    end select

    case (0) ! dx

      select case (dy)
    case (-1) ! dy

   select case(dz)
     case (-1)
       temp_vect(1:3, 7) = norm_vect(1:3, 1)
       calcd(7) = 0
       temp_vect(1:3, 8) = norm_vect(1:3, 2)
       calcd(8) = 0

     case (0)
       temp_vect(1:3, 3) = norm_vect(1:3, 1)
       calcd(3) = 0
       temp_vect(1:3, 4) = norm_vect(1:3, 2)
       calcd(4) = 0
       temp_vect(1:3, 7) = norm_vect(1:3, 5)
       calcd(7) = 0
       temp_vect(1:3, 8) = norm_vect(1:3, 6)
       calcd(8) = 0

     case (1)
       temp_vect(1:3, 3) = norm_vect(1:3, 5)
       calcd(3) = 0
       temp_vect(1:3, 4) = norm_vect(1:3, 6)
       calcd(4) = 0

     case default
       call vsmError ('failed to copy normal vectors')

   end select

 case (0) ! dy

   select case (dz)
     case (-1)
       temp_vect(1:3, 5) = norm_vect(1:3, 1)
       calcd(5) = 0;
       temp_vect(1:3, 6) = norm_vect(1:3, 2)
       calcd(6) = 0;
       temp_vect(1:3, 7) = norm_vect(1:3, 3)
       calcd(7) = 0;
       temp_vect(1:3, 8) = norm_vect(1:3, 4)
       calcd(8) = 0;

     case (1)
       temp_vect(1:3, 1) = norm_vect(1:3, 5)
       calcd(1) = 0;
       temp_vect(1:3, 2) = norm_vect(1:3, 6)
       calcd(2) = 0;
       temp_vect(1:3, 3) = norm_vect(1:3, 7)
       calcd(3) = 0;
       temp_vect(1:3, 4) = norm_vect(1:3, 8)
       calcd(4) = 0;

     case default
       call vsmError ('failed to copy normal vectors')

           end select

 case (1)! dy

   select case(dz)
     case (-1)
       temp_vect(1:3, 5) = norm_vect(1:3, 3)
       calcd(5) = 0;
       temp_vect(1:3, 6) = norm_vect(1:3, 4)
       calcd(6) = 0;

     case (0)
       temp_vect(1:3, 1) = norm_vect(1:3, 3)
       calcd(1) = 0;
       temp_vect(1:3, 2) = norm_vect(1:3, 4)
       calcd(2) = 0;
       temp_vect(1:3, 5) = norm_vect(1:3, 7)
       calcd(5) = 0;
       temp_vect(1:3, 6) = norm_vect(1:3, 8)
       calcd(6) = 0;

     case (1)
       temp_vect(1:3, 1) = norm_vect(1:3, 7)
       calcd(1) = 0;
       temp_vect(1:3, 2) = norm_vect(1:3, 8)
       calcd(2) = 0;

     case default
       call vsmError ('failed to copy normal vectors')

   end select

        case default
    call vsmError ('failed to copy normal vectors')

       end select


    case (1)
      select case(dy)
    case (-1) ! dy

   select case (dz)
     case (-1)
       temp_vect(1:3, 7) = norm_vect(1:3, 2)
       calcd(7) = 0

     case (0)
       temp_vect(1:3, 3) = norm_vect(1:3, 2)
       calcd(3) = 0
       temp_vect(1:3, 7) = norm_vect(1:3, 6)
       calcd(7) = 0

     case (1)
       temp_vect(1:3, 3) = norm_vect(1:3, 6)
       calcd(3) = 0

            case default
              call vsmError ('failed to copy normal vectors')

          end select

 case (0) ! dy

   select case (dz)
     case (-1)
       temp_vect(1:3, 5) = norm_vect(1:3, 2)
       calcd(5) = 0
       temp_vect(1:3, 7) = norm_vect(1:3, 4)
       calcd(7) = 0

     case (0)
       temp_vect(1:3, 1) = norm_vect(1:3, 2)
       calcd(1) = 0
       temp_vect(1:3, 3) = norm_vect(1:3, 4)
       calcd(3) = 0
       temp_vect(1:3, 5) = norm_vect(1:3, 6)
       calcd(5) = 0
       temp_vect(1:3, 7) = norm_vect(1:3, 8)
       calcd(7) = 0

     case (1)
       temp_vect(1:3, 1) = norm_vect(1:3, 6)
       calcd(1) = 0
       temp_vect(1:3, 3) = norm_vect(1:3, 8)
       calcd(3) = 0

            case default
              call vsmError ('failed to copy normal vectors')

         end select

 case (1)

   select case(dz)
     case (-1)
       temp_vect(1:3, 5) = norm_vect(1:3, 4)
       calcd(5) = 0

     case (0)
       temp_vect(1:3, 1) = norm_vect(1:3, 4)
       calcd(1) = 0
       temp_vect(1:3, 5) = norm_vect(1:3, 8)
       calcd(5) = 0

     case (1)
       temp_vect(1:3, 1) = norm_vect(1:3, 8)
       calcd(1) = 0

            case default
              call vsmError ('failed to copy normal vectors')

          end select

         case default
            call vsmError ('failed to copy normal vectors')

       end select

     case default
           call vsmError ('failed to copy normal vectors')

    end select
    do i=1, 8
      if(calcd(i) == 0) then
        if(isnan(temp_vect(1,i)) .or. isnan(temp_vect(1,i)) .or. isnan(temp_vect(3,i))) then
     	 print *,"vsmCopyNormvec ",temp_vect(1:3,i)
   	     print *,"dx dy dz i=",dx,dy,dz,i
         call vsmError("temp_vect is NAN")
        end if
      end if
   end do
 end subroutine vsmCopyNormvec

 subroutine vsmNormScalar(ci, nci, nscal, norm)
   integer, dimension(3), intent(in) :: ci
   integer, dimension(3), intent(inout) :: nci
   integer, intent(in) :: nscal
   real(kind=vdp), dimension(3, 8), intent(inout) :: norm
   integer :: l
   integer, dimension(3) :: dxyz
   integer, dimension(8) :: calcd
   real(kind=vdp), dimension(3, 8) :: temp_norm
   !call vsmLogPrint("normscalar ci=",ci)
   !call vsmLogPrint("normscalar nci=",nci)

   if (ci(1) /= nci(1) .or. ci(2) /= nci(2) .or. ci(3) /= nci(3)) then

     dxyz(1:3) = ci(1:3) - nci(1:3)
     calcd(1:8) = 1
     if (maxval(abs(dxyz)) < 2) then
       call vsmCopyNormvec (dxyz(1), dxyz(2), dxyz(3), norm, temp_norm, calcd)
     end if
!     call vsmLogPrint("dx dy dz = ",dxyz)
!     call vsmLogPrint("calcd = ",calcd)

     do l = 1, 8
       if (calcd(l) == 1) then
         select case (l)
           case (1)
             call vsmGradScal(ci(1),   ci(2),   ci(3), nscal, temp_norm(1:3,1))
           case (2)
             call vsmGradScal(ci(1)+1, ci(2),   ci(3), nscal, temp_norm(1:3,2))
           case (3)
             call vsmGradScal(ci(1),   ci(2)+1, ci(3), nscal, temp_norm(1:3,3))
           case (4)
             call vsmGradScal(ci(1)+1, ci(2)+1, ci(3), nscal, temp_norm(1:3,4))
           case (5)
             call vsmGradScal(ci(1),   ci(2),   ci(3)+1, nscal, temp_norm(1:3,5))
           case (6)
             call vsmGradScal(ci(1)+1, ci(2),   ci(3)+1, nscal, temp_norm(1:3,6))
           case (7)
             call vsmGradScal(ci(1),   ci(2)+1, ci(3)+1, nscal, temp_norm(1:3,7))
           case (8)
             call vsmGradScal(ci(1)+1, ci(2)+1, ci(3)+1, nscal, temp_norm(1:3,8))
         end select
       end if
     norm(1:3, l) = temp_norm(1:3, l)

     if(isnan(norm(1, l)) .or. isnan(norm(2, l)) .or. isnan(norm(3, l))) then
	   print *,"norm in normscalar",norm
	   print *,"ci in normscalar", ci
	   call vsmError("NAN exist in norm scalar")
     end if

     end do

    nci(1:3) = ci(1:3)
!    call vsmLogPrint("normscalar ci=",ci)
!    call vsmLogPrint("normscalar nci=",nci)
   end if
 end subroutine vsmNormScalar

 subroutine vsmSlicerIMT(pixel, norm, ray, dist2pxl, vis_num, cam_num, px)
   ! inputed variables
   real(kind=vdp), dimension(3), intent(in) :: pixel, norm
   real(kind=vdp), intent(in) :: ray, dist2pxl
   integer, intent(in) :: vis_num, cam_num
   integer, dimension(2), intent(in) :: px
   logical :: z_on

   ! variables
   type(vismo_cslice), dimension(vsmMaxSlice) :: cslidat
   integer :: i, sl_c, crs_points, snum, beg
   integer :: nscal
   integer, dimension(3) :: ci
   real, dimension(4) :: color ! RGBA
   real :: one_m_salpha
   real, dimension(4) :: color_
   real, dimension(3) :: lumi
   real :: amb, dif, spec, shin, alpha, lumi_a
   real(kind=vdp), dimension(3) :: ray_pos, tpos
   real(kind=vdp) :: ray_, inpro, zdis
   real(kind=vdp) :: v1, v2
   real(kind=vdp), dimension(3) :: dx
   real(kind=vdp), dimension(8) :: vx
   real(kind=vdp), dimension(8) :: grid_val
   real(kind=vdp) :: scal_val
   real, dimension(3) :: normal_vect
   real(kind=vdp), dimension(4) :: eq
   real :: dq_factor
   real(kind=vdp) :: firstin_ray, fdistance

   if(vsmPCmode == 2) call vsmError("vsmSlicerIMT is called thoght vsmPCmode is not 0 or 1")
   !initialization
   call vsmSetZeroF(color); call vsmSetZeroF(color_)
   zdis = vsmGetZbuf(px)
   tpos(1:3) = pixel(1:3) + norm(1:3) * (zdis-dist2pxl)
   z_on = vsmIsInsidel(tpos)

   crs_points = 0
   ray_ = ray
   ci(1:3) = -1
   inpro = dot_product(norm, vsmCamera(cam_num)%front)
   firstin_ray = ray
   color(1:4) = 0.0
   ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_

   do i = 1, vsmVisParam(vis_num)%slice_num
      if(vsmVisParam(vis_num)%sliceparam(i)%alpha >= vsmErtValue) cycle

      eq(1:4) = vsmVisParam(vis_num)%sliceparam(i)%sc(1:4)
      v1 = eq(1)* norm(1) + eq(2)* norm(2) + eq(3)* norm(3)

      if (abs(v1) < 1.0D-15) cycle

      v2 = -(eq(1)* pixel(1) + eq(2)* pixel(2) + eq(3)* pixel(3) + eq(4))
      ray_ = v2 / v1;
      if (ray_ <= 0.0_vdp) cycle !Near

      crs_points = crs_points + 1
      cslidat(crs_points)%num = i
      cslidat(crs_points)%val = ray_
   end do

   if(crs_points == 0) return

   beg = 1
   call vsmQSort(cslidat, beg, crs_points)

   do sl_c=1, crs_points
     snum = cslidat(sl_c)%num
     nscal = vsmVisParam(vis_num)%sliceparam(snum)%nscal

     ray_ = cslidat(sl_c)%val
     fdistance = ray_+dist2pxl

     if(fdistance*inpro >= vsmCamera(cam_num)%far) exit

     ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_

     if(.not.vsmIsInsidel(ray_pos)) cycle

     if(z_on .and. fdistance >= zdis) then
       ! in fact color_(4) is 1.0
       call vsmGetPixelColor4(px, color_)
       one_m_salpha = 1.0 - color(4)
       lumi_a = color_(4) * one_m_salpha
       color(1:3) = color(1:3) + color_(1:3)*lumi_a ! RGB
       color(4)   = color(4)   + lumi_a           ! A
       exit
     end if

     call vsmFindCoord(ray_pos, dx, ci)
     ! for interpolation
     call vsmSetFrac(dx, vx)
     call vsmSetScalGridVal(nscal, ci, grid_val)

     ! calculate each scalar value at current point
     scal_val = vsmGetScalVal(vx, grid_val)

     ! color slices
     call vsmGetSliceColor(vis_num, snum, alpha, amb, dif, spec, shin)
     call vsmGetSliceNormf(vis_num, snum, normal_vect)

     if(vsmCamera(cam_num)%pers) then
        call vsmCalcColor(amb, dif, spec, shin, vsmCamera(cam_num)%pos, &
                      ray_pos, normal_vect, vsmVisParam(vis_num)%light_used, lumi)
     else
             call vsmCalcColorOrtho(amb, dif, spec, shin, vsmCamera(cam_num)%front, &
                      ray_pos, normal_vect, vsmVisParam(vis_num)%light_used, lumi)
     end if

     call vsmSetZeroF(color_)
     call vsmGetSColor(scal_val, nscal, color_)
     
     lumi(1:3)=lumi(1:3)*color_(1:3)

     if(vsmVisParam(vis_num)%depthq) then
        call vsmDepthCueing(fdistance, lumi)
     end if

     one_m_salpha = 1.0 - color(4);
     lumi_a = alpha * one_m_salpha;

     color(1:3) = color(1:3) + lumi(1:3)*lumi_a ! RGB
     color(4)   = color(4)   + lumi_a           ! A

     !early ray termination
     if(color(4) >= vsmErtValue) exit

     ! end of color slices
   end do

   if(z_on .and. color(4) < vsmErtValue) then
      ! in fact color_(4) is 1.0
     call vsmGetPixelColor4(px, color_)
     one_m_salpha = 1.0 - color(4)
     lumi_a = color_(4) * one_m_salpha
     color(1:3) = color(1:3) + color_(1:3)*lumi_a ! RGB
     color(4)   = color(4)   + lumi_a           ! A
   end if

     call vsmRaySegment(firstin_ray + dist2pxl, fdistance-vsmIntval%ray_interval, color, px)
 end subroutine vsmSlicerIMT

! solid slicer
 subroutine vsmSlicerIMS(pixel, norm, ray, dist2pxl, vis_num, cam_num, px)
   ! inputed variables
   real(kind=vdp), dimension(3), intent(in) :: pixel, norm
   real(kind=vdp), intent(in) :: ray, dist2pxl
   integer, intent(in) :: vis_num, cam_num
   integer, dimension(2), intent(in) :: px

   ! variables
   integer :: i, snum, sslice
   integer :: nscal
   integer, dimension(3) :: ci
   real, dimension(4) :: color ! RGBA
   real :: one_m_salpha
   real, dimension(4) :: color_
   real, dimension(3) :: lumi
   real :: amb, dif, spec, shin, alpha, lumi_a
   real(kind=vdp), dimension(3) :: ray_pos, tpos
   real(kind=vdp) :: ray_, inpro, zdis
   real(kind=vdp) :: v1, v2
   real(kind=vdp), dimension(3) :: dx
   real(kind=vdp), dimension(8) :: vx
   real(kind=vdp), dimension(8) :: grid_val
   real(kind=vdp) :: scal_val
   real, dimension(3) :: normal_vect
   real(kind=vdp), dimension(4) :: eq
   real :: dq_factor
   real(kind=vdp) :: firstin_ray, fdistance

   if(vsmPCmode == 2) call vsmError("vsmSlicerIMS is called thoght vsmPCmode is not 0 or 1")
   !initialization
   call vsmSetZeroF(color); call vsmSetZeroF(color_)
   zdis = vsmGetZbuf(px)
   ray_ = ray
   ci(1:3) = -1
   inpro = dot_product(norm, vsmCamera(cam_num)%front)
   firstin_ray = ray
   color(1:4) = 0.0

   ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_
   sslice = 0
   do i = 1, vsmVisParam(vis_num)%slice_num
     if(vsmVisParam(vis_num)%sliceparam(i)%alpha >= vsmErtValue) then
       sslice = sslice + 1
     end if
   end do

   if(sslice == 0) return

   do i = 1, vsmVisParam(vis_num)%slice_num
      if(vsmVisParam(vis_num)%sliceparam(i)%alpha < vsmErtValue) cycle

      eq(1:4) = vsmVisParam(vis_num)%sliceparam(i)%sc(1:4)
      v1 = eq(1)* norm(1) + eq(2)* norm(2) + eq(3)* norm(3)

      if(abs(v1) < 1.0D-15) cycle

      v2 = -(eq(1)* pixel(1) + eq(2)* pixel(2) + eq(3)* pixel(3) + eq(4))
      ray_ = v2 / v1
      if (ray_ <= 0.0_vdp) cycle !Near

      snum = i
     nscal = vsmVisParam(vis_num)%sliceparam(snum)%nscal

     fdistance = ray_+dist2pxl

     if(fdistance*inpro >= vsmCamera(cam_num)%far) cycle

     ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_

     if(.not.vsmIsInsidel(ray_pos)) cycle

     if(fdistance >= zdis) then
       cycle
     end if

     call vsmFindCoord(ray_pos, dx, ci)
     ! for interpolation
     call vsmSetFrac(dx, vx)
     call vsmSetScalGridVal(nscal, ci, grid_val)

     ! calculate each scalar value at current point
     scal_val = vsmGetScalVal(vx, grid_val)

     ! color slices
     call vsmGetSliceColor(vis_num, snum, alpha, amb, dif, spec, shin)
     call vsmGetSliceNormf(vis_num, snum, normal_vect)

     if(vsmCamera(cam_num)%pers) then
        call vsmCalcColor(amb, dif, spec, shin, vsmCamera(cam_num)%pos, &
                      ray_pos, normal_vect, vsmVisParam(vis_num)%light_used, lumi)
     else
        call vsmCalcColorOrtho(amb, dif, spec, shin, vsmCamera(cam_num)%front, &
                      ray_pos, normal_vect, vsmVisParam(vis_num)%light_used, lumi)
     end if

     call vsmSetZeroF(color_)
     call vsmGetSColor(scal_val, nscal, color_)

     color(1:3) = lumi(1:3)*color_(1:3) ! RGB
     color(4)   = 1.0           ! A

     call vsmPaintPixel3(px, color, fdistance)
     ! end of color slices
   end do
 end subroutine vsmSlicerIMS

 subroutine vsmIsoRendererIMS(pixel, norm, ray, ray2ldat, dist2pxl, vis_method, vis_num, cam_num, px)
    ! inputed variables
   real(kind=vdp), dimension(3), intent(in) :: pixel, norm
   real(kind=vdp), intent(in) :: ray, ray2ldat, dist2pxl
   logical, dimension(3), intent(in) :: vis_method
   integer, intent(in) :: vis_num, cam_num
   integer, dimension(2), intent(in) :: px

   ! variables
   integer :: integ_loop, i, sl_c, iso_c, vol_c
   integer :: nscal, NumScal
   integer, dimension(3) :: ci, prev_ci
   integer, dimension(3, vsmMaxScalar) :: nci
   real, dimension(4) :: color ! RGBA
   real :: one_m_salpha
   real, dimension(4) :: color_, isocolor
   real, dimension(3) :: amb, dif, spec, lumi
   real :: shin, alpha, lumi_a
   real(kind=vdp), dimension(3) :: ray_pos, prev_ray_pos
   real(kind=vdp) :: rintval, ray_, skip_intval, inpro, zdis
   real(kind=vdp), dimension(3) :: dx
   real(kind=vdp), dimension(8) :: vx
   real(kind=vdp) :: judge1, judge2, rayBack

   integer :: iso_s, cison
   integer, dimension(3) :: prev_isoci, isoci
   real(kind=vdp), dimension(3) :: isodx, iso_pos, iso_pos_f, iso_pos_b, slice_pos
   real(kind=vdp) :: isolength, seg1,seg2,judgeI,judgeIf, judgeIb,isoval,isovalf,isovalb
   real(kind=vdp), dimension(8) :: isovx, iso_grid_val

   real(kind=vdp), dimension(3, 8, vsmMaxScalar) :: scalnorm

   real(kind=vdp), dimension(8) :: prev_vx
   real(kind=vdp), dimension(3, 8) :: prev_scalnorm
   real, dimension(3) :: prev_normal_vect
   real(kind=vdp) :: ray_ratio
   integer, dimension(3) :: dummy_nci

   real(kind=vdp), dimension(8, vsmMaxScalar) :: grid_val
   real(kind=vdp), dimension(vsmMaxScalar) :: scal_val, prev_scal_val
   real, dimension(3) :: normal_vect
   real :: dq_factor
   real(kind=vdp) :: fdistance
   !octree
   logical :: skip, skipcalc, endFlag

   integer :: comm, myrank, psize

   if(vsmPCmode /=0 ) call vsmError("vsmIsoRendererIMS is called though vsmPCmode /= 0")
   call vsmGetMPIVars(comm, myrank, psize)
    !initialization
   endFlag = .false.
   NumScal = vsmDataInfo%NumOfScal
   prev_ci(1:3) = -1
   nci(1:3, 1:vsmMaxScalar) = -2 ! different from prev_ci values
   call vsmSetZeroF(color)
   !insidelocal = .false.

   prev_vx=0.0_vdp
   prev_scalnorm = 0.0_vdp

   zdis = vsmGetZbuf(px)
   fdistance = ray+dist2pxl
   if(zdis < fdistance) then
     return
   end if

   rintval = vsmIntval%ray_interval
   ray_ = ray
   prev_ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_
   inpro = dot_product(norm, vsmCamera(cam_num)%front)

   ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_
   dx = 0.0_vdp
   if(.not.vsmIsInsidel(ray_pos)) then
     do i=1, 100
        ray_ = ray_ + rintval*0.01_vdp
        ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_
        if(vsmIsInsidel(ray_pos)) then
            exit
        end if
     end do

     if(.not.vsmIsInsidel(ray_pos)) then
         return
     end if
   end if

   prev_ray_pos(1:3) = ray_pos(1:3) - norm(1:3)*rintval
   if(.not.vsmIsInsidelWO(prev_ray_pos)) then
      rayBack = ray2ldat
      do i=1, 100
         prev_ray_pos(1:3) = pixel(1:3) + norm(1:3) * rayBack
         if(vsmIsInsidelWO(prev_ray_pos)) then
             !call vsmLogPrint("Modified(prev ray pos) ", prev_ray_pos)
             exit
         end if
         rayBack = rayBack + rintval*0.01_vdp
      end do

      if(.not.vsmIsInsidelWO(prev_ray_pos)) then
         prev_ray_pos(1:3) = ray_pos(1:3)
         !call vsmLogPrint("Modified2(prev ray pos) ", prev_ray_pos)
      end if
   end if

   call vsmFindCoord(prev_ray_pos, dx, prev_ci, 0)
   call vsmSetFrac(dx, prev_vx)
   do i=1, NumScal
     call vsmSetScalGridVal(i, prev_ci, grid_val(1:8, i))
   end do
   do i=1, NumScal
     prev_scal_val(i) = vsmGetScalVal(prev_vx, grid_val(1:8, i))
   end do

   do integ_loop = 1, vsmMaxRayCount

     fdistance = ray_ + dist2pxl
     if(fdistance*inpro >= vsmCamera(cam_num)%far) exit

     ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_

     if(.not.vsmIsInsidel(ray_pos)) then
       endFlag = .true.
       !call vsmLogPrint("prev_ray_pos ", prev_ray_pos)
       rayBack = vsmRayDistInside(prev_ray_pos, norm, vsmLocalCoord%lorigin, vsmLocalCoord%lcend)
       ray_pos(1:3) = prev_ray_pos(1:3) + norm(1:3) * rayBack

       if(.not.vsmIsInsidelWO(ray_pos)) then
         do i=1, 100
            rayBack = rayBack - rintval*0.01_vdp
            ray_pos(1:3) = prev_ray_pos(1:3) + norm(1:3) * rayBack
            if(vsmIsInsidelWO(ray_pos)) then
                exit
            end if
         end do
         if(.not.vsmIsInsidelWO(ray_pos)) then
            ray_pos(1:3) = prev_ray_pos(1:3)
           ! call vsmError("ray cannot enter the data : raycaster")
         end if
       end if
     end if

     if(.not.vsmIsInsidelWO(ray_pos)) then
       exit
     end if

     if(fdistance >= zdis) then
       call vsmGetPixelColor4(px, color_)
       one_m_salpha = 1.0 - color(4)
       lumi_a = color_(4) * one_m_salpha
       color(1:3) = color(1:3) + color_(1:3)*lumi_a ! RGB
       color(4)   = color(4)   + lumi_a           ! A
       ! solid drawing
       call vsmPaintPixel3(px, color, zdis)
       return
     end if

     ! for interpolation
     call vsmFindCoord(ray_pos, dx, ci, integ_loop)
     call vsmSetFrac(dx, vx)

     if(ci(1)/=prev_ci(1) .or. ci(2)/=prev_ci(2) .or. ci(3)/=prev_ci(3) .or. integ_loop==1) then
       do i=1, NumScal
         call vsmSetScalGridVal(i, ci, grid_val(1:8, i))
       end do
       skipcalc = .true.
     else
       skipcalc = .false.
     end if

     ! calculate each scalar value at current point
     do i=1, NumScal
       scal_val(i) = vsmGetScalVal(vx, grid_val(1:8, i))
     end do

     ! isosurface
       do iso_c = 1, vsmVisParam(vis_num)%iso_num
         if(vsmVisParam(vis_num)%isoparam(iso_c)%alpha < vsmErtValue) cycle
         nscal = vsmVisParam(vis_num)%isoparam(iso_c)%nscal
         judge1 = vsmIsAboveIsoL(vis_num, iso_c, scal_val(nscal))
         judge2 = vsmIsAboveIsoL(vis_num, iso_c, prev_scal_val(nscal))

         if(judge1* judge2 <= 0.0_vdp) then

           ray_ratio = abs(judge1)/abs(scal_val(nscal)-prev_scal_val(nscal)) ! always positive or 0.0
          ! for safty
           ray_ratio = vsmScaleD01(ray_ratio)

           iso_pos_b(1:3) = prev_ray_pos(1:3)
           judgeIb = judge2
           isovalb = prev_scal_val(nscal)

           iso_pos_f(1:3) = ray_pos(1:3)
           judgeIf = judge1
           isovalf = scal_val(nscal)

           iso_pos(1:3) = iso_pos_f(1:3) + (iso_pos_b(1:3) - iso_pos_f(1:3))*ray_ratio
           prev_isoci(1:3) = ci(1:3)
           iso_grid_val(1:8) = grid_val(1:8, nscal)

           do iso_s =1, 3
              call vsmFindCoord(iso_pos, isodx, isoci, -integ_loop)
              call vsmSetFrac(isodx, isovx)
              if(prev_isoci(1)/=isoci(1) .or. prev_isoci(2)/=isoci(2) .or. prev_isoci(3)/=isoci(3)) then
                  call vsmSetScalGridVal(nscal, isoci, iso_grid_val(1:8))
              end if

              isoval = vsmGetScalVal(isovx, iso_grid_val(1:8))
              judgeI = vsmIsAboveIsoL(vis_num, iso_c, isoval)

              if(judgeI*judgeIf <= 0.0_vdp) then
                 iso_pos_b(1:3) = iso_pos(1:3)
                 judgeIb = judgeI
                 isovalb = isoval
              else
                 iso_pos_f(1:3) = iso_pos(1:3)
                 judgeIf = judgeI
                 isovalf = isoval
              end if
              ray_ratio = abs(judgeIf)/abs(isovalf - isovalb) ! always positive or 0.0
              ! for safty
              ray_ratio = vsmScaleD01(ray_ratio)

              iso_pos(1:3) = iso_pos_f(1:3) + (iso_pos_b(1:3) - iso_pos_f(1:3))*ray_ratio
              prev_isoci(1:3) = isoci(1:3)
           end do

           if(dot_product(iso_pos-vsmCamera(cam_num)%pos, vsmCamera(cam_num)%front) &
                < vsmCamera(cam_num)%near) then
             cycle
           end if

           call vsmGetIsoColor(vis_num, iso_c, alpha, amb, dif, spec, shin)

           if(.not.vsmIsInsidel(iso_pos) .and. alpha < vsmErtValue) then
             cycle
           end if

           call vsmFindCoord(iso_pos, isodx, isoci, -integ_loop)
           call vsmSetFrac(isodx, isovx)
           if(prev_isoci(1)/=isoci(1) .or. prev_isoci(2)/=isoci(2) .or. prev_isoci(3)/=isoci(3)) then
               call vsmSetScalGridVal(nscal, isoci, iso_grid_val(1:8))
           end if

           rayBack = sqrt(dot_product(ray_pos - iso_pos, ray_pos - iso_pos))

           if(isoci(1) /= nci(1, nscal) .or. isoci(2) /= nci(2, nscal) .or. isoci(3) /= nci(3, nscal)) then
             call vsmNormScalar (isoci, nci(1:3, nscal), nscal, scalnorm(1:3, 1:8, nscal))
           end if

           call vsmGetIntpolVect(scalnorm(1:3, 1:8, nscal), isovx, normal_vect)

           call vsmNormVecF(normal_vect)

           call vsmSetZerof(lumi)

           if(vsmCamera(cam_num)%pers) then
              call vsmCalcColorv(amb, dif, spec, shin, vsmCamera(cam_num)%pos, &
                      ray_pos, normal_vect, vsmVisParam(vis_num)%light_used, lumi)
           else
             call vsmCalcColorvOrtho(amb, dif, spec, shin, vsmCamera(cam_num)%front, &
                      ray_pos, normal_vect, vsmVisParam(vis_num)%light_used, lumi)
           end if

           one_m_salpha = 1.0 - color(4);
           lumi_a = alpha * one_m_salpha;

           color(1:3) = color(1:3) + lumi(1:3)*lumi_a ! RGB
           color(4)   = color(4)   + lumi_a           ! A

           ! solid drawing
           call vsmPaintPixel3(px, color, fdistance-rayBack)
           return
         end if
       end do
     !early ray termination
     if(color(4) >= vsmErtValue) exit
     ! end of isosurface

     if(endFlag) exit

     prev_ci(1:3) = ci(1:3)

     ! ------------- skip some space ----------
     if(skipcalc) then
!       call vsmLogPrint("skipcalc in")
       skip = .false.
       ! skip a super voxel
       call vsmRayJump4iso(ray_pos, norm, skip_intval, skip)
       if(skip) then

         ray_pos(1:3) = ray_pos(1:3) + skip_intval * norm(1:3)
         ray_ = ray_ + skip_intval

         if(.not.vsmIsInsidelWO(ray_pos)) then
            call vsmLogPrint("isorendererims:by skip space, ray went out of local data", skip_intval)
            call vsmFindCoord(ray_pos, dx, ci)
            call vsmLogPrint("out pos", ray_pos)
            call vsmLogPrint("out ci", ci)
            call vsmError("by skip, ray went out of local data")
         end if

         if(.not.vsmIsInsidel(ray_pos)) exit

         ! for interpolation
         call vsmFindCoord(ray_pos, dx, ci, vsmMaxRayCount+1)
         call vsmSetFrac(dx, vx)

         do i=1, NumScal
           call vsmSetScalGridVal(i, ci, grid_val(1:8, i))
         end do
         prev_ci(1:3) = ci(1:3)

         ! calculate each scalar value at current point
         do i=1, NumScal
           scal_val(i) = vsmGetScalVal(vx, grid_val(1:8, i))
         end do

         prev_scal_val(1:NumScal) = scal_val(1:NumScal)
         ray_ = ray_ + rintval
         prev_ray_pos(1:3) = ray_pos(1:3)
         prev_vx = vx
         cycle
       end if
!       call vsmLogPrint("skipcalc out")
     end if
     !-----------------------------------------

     prev_ray_pos(1:3) = ray_pos(1:3)
     ray_ = ray_ + rintval
     prev_scal_val(1:NumScal) = scal_val(1:NumScal)
     prev_vx = vx
   end do
 end subroutine vsmIsoRendererIMS

 subroutine vsmRayCastIM(pixel, norm, ray, ray2ldat, dist2pxl, vis_method, vis_num, cam_num, px)
    ! inputed variables
   real(kind=vdp), dimension(3), intent(in) :: pixel, norm
   real(kind=vdp), intent(in) :: ray, ray2ldat, dist2pxl
   logical, dimension(3), intent(in) :: vis_method
   integer, intent(in) :: vis_num, cam_num
   integer, dimension(2), intent(in) :: px
   logical :: z_on

   ! variables
   integer :: integ_loop, i, sl_c, iso_c, vol_c
   integer :: nscal, NumScal
   integer, dimension(3) :: ci, prev_ci
   integer, dimension(3, vsmMaxScalar) :: nci
   real, dimension(4) :: color ! RGBA
   real :: one_m_salpha
   real, dimension(4) :: color_, isocolor
   real, dimension(3) :: amb, dif, spec, lumi
   real :: shin, alpha, lumi_a
   real(kind=vdp), dimension(3) :: ray_pos, prev_ray_pos
   real(kind=vdp) :: rintval, ray_, skip_intval, inpro, zdis
   real(kind=vdp), dimension(3) :: dx
   real(kind=vdp), dimension(8) :: vx
   real(kind=vdp) :: judge1, judge2, rayBack

   integer :: iso_s, cison
   integer, dimension(3) :: prev_isoci, isoci
   real(kind=vdp), dimension(3) :: isodx, iso_pos, iso_pos_f, iso_pos_b, slice_pos
   real(kind=vdp) :: isolength, seg1,seg2,judgeI,judgeIf, judgeIb,isoval,isovalf,isovalb
   real(kind=vdp), dimension(8) :: isovx, iso_grid_val

   real(kind=vdp), dimension(3, 8, vsmMaxScalar) :: scalnorm

   real(kind=vdp), dimension(8) :: prev_vx
   real(kind=vdp), dimension(3, 8) :: prev_scalnorm
   real, dimension(3) :: prev_normal_vect
   real(kind=vdp) :: ray_ratio
   integer, dimension(3) :: dummy_nci

   real(kind=vdp), dimension(8, vsmMaxScalar) :: grid_val
   real(kind=vdp), dimension(vsmMaxScalar) :: scal_val, prev_scal_val
   real, dimension(3) :: normal_vect
   real :: dq_factor
   logical :: first_in
   real(kind=vdp) :: firstin_ray, fdistance
   !octree
   !real(kind=vdp), dimension(3) :: npos
   logical :: skip, skipcalc, endFlag

   integer :: comm, myrank, psize

   if(vsmPCmode /=0 ) call vsmError("vsmRayCastIM is called though vsmPCmode /= 0")

   call vsmGetMPIVars(comm, myrank, psize)

   !initialization
   endFlag = .false.
   NumScal = vsmDataInfo%NumOfScal
   prev_ci(1:3) = -1
   nci(1:3, 1:vsmMaxScalar) = -2 ! different from prev_ci values
   call vsmSetZeroF(color)
   !insidelocal = .false.

   prev_vx=0.0_vdp
   prev_scalnorm = 0.0_vdp

   first_in = .false.
   zdis = vsmGetZbuf(px)

   ray_pos(1:3) = pixel(1:3)+ (zdis-dist2pxl)*norm(1:3)
   z_on =  vsmIsInsidel(ray_pos)
   fdistance = ray+dist2pxl
   rintval = vsmIntval%ray_interval
   if(.not. z_on .and. zdis < fdistance) then
     call vsmGetPixelColor4(px, color)
     call vsmRaySegment(zdis, fdistance-rintval, color, px)
     return
   end if

   ray_ = ray
   firstin_ray = ray
   prev_ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_
   inpro = dot_product(norm, vsmCamera(cam_num)%front)

   ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_
   dx = 0.0_vdp
   if(.not.vsmIsInsidel(ray_pos)) then
     ray_ = ray_ + rintval
     ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_

     if(.not.vsmIsInsidel(ray_pos)) then
         call vsmClearPixel(px)
         return
     else
         if(first_in .eqv. .false.) then
           first_in = .true.
           firstin_ray = ray_
         end if
     end if
   end if

   prev_ray_pos(1:3) = ray_pos(1:3) - norm(1:3)*rintval
   if(.not.vsmIsInsidelWO(prev_ray_pos)) then
      rayBack = ray2ldat
      do i=1, 100
         prev_ray_pos(1:3) = pixel(1:3) + norm(1:3) * rayBack
         if(vsmIsInsidelWO(prev_ray_pos)) then
             !call vsmLogPrint("Modified(prev ray pos) ", prev_ray_pos)
             exit
         end if
         rayBack = rayBack + rintval*0.01_vdp
      end do

      if(.not.vsmIsInsidelWO(prev_ray_pos)) then
         prev_ray_pos(1:3) = ray_pos(1:3)
         !call vsmLogPrint("Modified2(prev ray pos) ", prev_ray_pos)
      end if
   end if

   call vsmFindCoord(prev_ray_pos, dx, prev_ci, 0)
   call vsmSetFrac(dx, prev_vx)
   do i=1, NumScal
     call vsmSetScalGridVal(i, prev_ci, grid_val(1:8, i))
   end do
   do i=1, NumScal
     prev_scal_val(i) = vsmGetScalVal(prev_vx, grid_val(1:8, i))
   end do

   do integ_loop = 1, vsmMaxRayCount
     fdistance = ray_ + dist2pxl
     if(fdistance*inpro >= vsmCamera(cam_num)%far) exit

     ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_

     if(.not.vsmIsInsidel(ray_pos)) then
       endFlag = .true.
       !call vsmLogPrint("prev_ray_pos ", prev_ray_pos)
!       call vsmRaySegment(firstin_ray + dist2pxl, fdistance-rintval, color, px)
       rayBack = vsmRayDistInside(prev_ray_pos, norm, vsmLocalCoord%lorigin, vsmLocalCoord%lcend)
       ray_pos(1:3) = prev_ray_pos(1:3) + norm(1:3) * rayBack

       if(.not.vsmIsInsidelWO(ray_pos)) then
         do i=1, 100
            rayBack = rayBack - rintval*0.01_vdp
            ray_pos(1:3) = prev_ray_pos(1:3) + norm(1:3) * rayBack
            if(vsmIsInsidelWO(ray_pos)) then
                exit
            end if
         end do
         if(.not.vsmIsInsidelWO(ray_pos)) then
            ray_pos(1:3) = prev_ray_pos(1:3)
           ! call vsmError("ray cannot enter the data : raycaster")
         end if
       end if
     end if

     if(.not.vsmIsInsidelWO(ray_pos)) then
       exit
     end if

     if(z_on .and. fdistance >= zdis) then
       call vsmGetPixelColor4(px, color_)
       one_m_salpha = 1.0 - color(4)
       lumi_a = color_(4) * one_m_salpha
       color(1:3) = color(1:3) + color_(1:3)*lumi_a ! RGB
       color(4)   = color(4)   + lumi_a           ! A
       if(first_in .and. integ_loop == 1) then
          call vsmRaySegment(firstin_ray + dist2pxl, fdistance-rintval, color, px)
          return
       end if
       exit
     end if

     ! for interpolation
     call vsmFindCoord(ray_pos, dx, ci, integ_loop)
     call vsmSetFrac(dx, vx)

     if(ci(1)/=prev_ci(1) .or. ci(2)/=prev_ci(2) .or. ci(3)/=prev_ci(3) .or. integ_loop==1) then
       do i=1, NumScal
         call vsmSetScalGridVal(i, ci, grid_val(1:8, i))
       end do
       skipcalc = .true.
     else
       skipcalc = .false.
     end if

     ! calculate each scalar value at current point
     do i=1, NumScal
       scal_val(i) = vsmGetScalVal(vx, grid_val(1:8, i))
     end do

     ! color slices
     if (vis_method(1)) then
       do sl_c = 1, vsmVisParam(vis_num)%slice_num
         if(vsmVisParam(vis_num)%sliceparam(sl_c)%alpha >= vsmErtValue) cycle
         judge1 = vsmIsAboveSlice(vis_num, sl_c, ray_pos)
         judge2 = vsmIsAboveSlice(vis_num, sl_c, prev_ray_pos)
         if(judge1* judge2 <= 0.0_vdp) then
           ! ray has been crossed
           call vsmGetSliceColor(vis_num, sl_c, alpha, amb(1), dif(1), spec(1), shin)
           call vsmGetSliceNormf(vis_num, sl_c, normal_vect)

           if(vsmCamera(cam_num)%pers) then
              call vsmCalcColor(amb(1), dif(1), spec(1), shin, vsmCamera(cam_num)%pos, &
                      ray_pos, normal_vect, vsmVisParam(vis_num)%light_used, lumi)
           else
              call vsmCalcColorOrtho(amb(1), dif(1), spec(1), shin, vsmCamera(cam_num)%front, &
                      ray_pos, normal_vect, vsmVisParam(vis_num)%light_used, lumi)
           end if

           nscal = vsmVisParam(vis_num)%sliceparam(sl_c)%nscal
           call vsmSetZeroF(color_)
           call vsmGetSColor(scal_val(nscal), nscal, color_)

           ray_ratio = abs(judge1)/abs(judge1 - judge2)
           rayBack = ray_ratio*sqrt(dot_product(ray_pos-prev_ray_pos, ray_pos-prev_ray_pos))

           slice_pos(1:3) = vsmCamera(cam_num)%pos(1:3) + (fdistance-rayBack)*norm(1:3)

           if(dot_product(slice_pos-vsmCamera(cam_num)%pos, vsmCamera(cam_num)%front) &
                < vsmCamera(cam_num)%near) then
             cycle
           end if

           if(.not.vsmIsInsidel(slice_pos) .and. alpha < vsmErtValue) then
             cycle
           end if

           lumi(1:3) = lumi(1:3)*color_(1:3)
           if(vsmVisParam(vis_num)%depthq) then
             call vsmDepthCueing(fdistance-rayBack, lumi)
           end if

           one_m_salpha = 1.0 - color(4);
           lumi_a = alpha * one_m_salpha;

!           color(1:3) = color(1:3) + lumi(1:3)*color_(1:3)*lumi_a ! RGB
           color(1:3) = color(1:3) + lumi(1:3)*lumi_a ! RGB
           color(4)   = color(4)   + lumi_a           ! A
         end if
       end do
     !early ray termination
     if(color(4) >= vsmErtValue) exit
     end if
     ! end of color slices

     ! isosurface
      if(vis_method(2)) then

       do iso_c = 1, vsmVisParam(vis_num)%iso_num
         if(vsmVisParam(vis_num)%isoparam(iso_c)%alpha >= vsmErtValue) cycle
         nscal = vsmVisParam(vis_num)%isoparam(iso_c)%nscal
         judge1 = vsmIsAboveIsoL(vis_num, iso_c, scal_val(nscal))
         judge2 = vsmIsAboveIsoL(vis_num, iso_c, prev_scal_val(nscal))

         if(judge1* judge2 <= 0.0_vdp) then

           ray_ratio = abs(judge1)/abs(scal_val(nscal)-prev_scal_val(nscal)) ! always positive or 0.0
          ! for safty
           ray_ratio = vsmScaleD01(ray_ratio)

           iso_pos_b(1:3) = prev_ray_pos(1:3)
           judgeIb = judge2
           isovalb = prev_scal_val(nscal)

           iso_pos_f(1:3) = ray_pos(1:3)
           judgeIf = judge1
           isovalf = scal_val(nscal)

           iso_pos(1:3) = iso_pos_f(1:3) + (iso_pos_b(1:3) - iso_pos_f(1:3))*ray_ratio
           prev_isoci(1:3) = ci(1:3)
           iso_grid_val(1:8) = grid_val(1:8, nscal)

           do iso_s =1, 3
              call vsmFindCoord(iso_pos, isodx, isoci, -integ_loop)
              call vsmSetFrac(isodx, isovx)
              if(prev_isoci(1)/=isoci(1) .or. prev_isoci(2)/=isoci(2) .or. prev_isoci(3)/=isoci(3)) then
                  call vsmSetScalGridVal(nscal, isoci, iso_grid_val(1:8))
              end if

              isoval = vsmGetScalVal(isovx, iso_grid_val(1:8))
              judgeI = vsmIsAboveIsoL(vis_num, iso_c, isoval)

              if(judgeI*judgeIf <= 0.0_vdp) then
                 iso_pos_b(1:3) = iso_pos(1:3)
                 judgeIb = judgeI
                 isovalb = isoval
              else
                 iso_pos_f(1:3) = iso_pos(1:3)
                 judgeIf = judgeI
                 isovalf = isoval
              end if
              ray_ratio = abs(judgeIf)/abs(isovalf - isovalb) ! always positive or 0.0
              ! for safty
              ray_ratio = vsmScaleD01(ray_ratio)

              iso_pos(1:3) = iso_pos_f(1:3) + (iso_pos_b(1:3) - iso_pos_f(1:3))*ray_ratio
              prev_isoci(1:3) = isoci(1:3)
           end do

           if(dot_product(iso_pos-vsmCamera(cam_num)%pos, vsmCamera(cam_num)%front) &
                < vsmCamera(cam_num)%near) then
             cycle
           end if

           call vsmGetIsoColor(vis_num, iso_c, alpha, amb, dif, spec, shin)

           if(.not.vsmIsInsidel(iso_pos) .and. alpha < vsmErtValue) then
             cycle
           end if

           call vsmFindCoord(iso_pos, isodx, isoci, -integ_loop)
           call vsmSetFrac(isodx, isovx)
           if(prev_isoci(1)/=isoci(1) .or. prev_isoci(2)/=isoci(2) .or. prev_isoci(3)/=isoci(3)) then
               call vsmSetScalGridVal(nscal, isoci, iso_grid_val(1:8))
           end if

           rayBack = sqrt(dot_product(ray_pos - iso_pos, ray_pos - iso_pos))

           if(isoci(1) /= nci(1, nscal) .or. isoci(2) /= nci(2, nscal) .or. isoci(3) /= nci(3, nscal)) then
             call vsmNormScalar (isoci, nci(1:3, nscal), nscal, scalnorm(1:3, 1:8, nscal))
           end if

           call vsmGetIntpolVect(scalnorm(1:3, 1:8, nscal), isovx, normal_vect)

           call vsmNormVecF(normal_vect)

           call vsmSetZerof(lumi)

           if(vsmCamera(cam_num)%pers) then
              call vsmCalcColorv(amb, dif, spec, shin, vsmCamera(cam_num)%pos, &
                      ray_pos, normal_vect, vsmVisParam(vis_num)%light_used, lumi)
           else
             call vsmCalcColorvOrtho(amb, dif, spec, shin, vsmCamera(cam_num)%front, &
                      ray_pos, normal_vect, vsmVisParam(vis_num)%light_used, lumi)
           end if

           if(vsmVisParam(vis_num)%depthq) then
             call vsmDepthCueing(fdistance-rayBack, lumi)
           end if

           one_m_salpha = 1.0 - color(4);
           lumi_a = alpha * one_m_salpha;

           color(1:3) = color(1:3) + lumi(1:3)*lumi_a ! RGB
           color(4)   = color(4)   + lumi_a           ! A
         end if
       end do
     !early ray termination
     if(color(4) >= vsmErtValue) exit
     end if
     ! end of isosurface

     if(endFlag) exit
     ! volume rendering
     if(vis_method(3) .and. .not.endFlag .and. vsmIsInsidel(ray_pos) .and. color(4) < vsmErtValue) then !the last is for pc
       do vol_c=1, vsmVisParam(vis_num)%volren_num
         nscal = vsmVisParam(vis_num)%volparam(vol_c)%nscal
         call vsmGetVolColor(vis_num, vol_c, amb(1), dif(1), spec(1), shin)
         ! calculating normal vector
         if(ci(1) /= nci(1, nscal) .or. ci(2) /= nci(2, nscal) .or. ci(3) /= nci(3, nscal)) then
           call vsmNormScalar (ci, nci(1:3, nscal), nscal, scalnorm(1:3, 1:8, nscal))
         end if
         call vsmGetIntpolVect(scalnorm(1:3, 1:8, nscal), vx, normal_vect)
         call vsmNormVecF(normal_vect)

         if(vsmCamera(cam_num)%pers) then
            call vsmCalcColor(amb(1), dif(1), spec(1), shin, vsmCamera(cam_num)%pos, &
                      ray_pos, normal_vect, vsmVisParam(vis_num)%light_used, lumi)
         else
             call vsmCalcColorOrtho(amb(1), dif(1), spec(1), shin, vsmCamera(cam_num)%front, &
                      ray_pos, normal_vect, vsmVisParam(vis_num)%light_used, lumi)
         end if

         call vsmSetZeroF(color_)
         call vsmGetSColor(scal_val(nscal), nscal, color_)

         lumi(1:3) = lumi(1:3)*color_(1:3)
         if(vsmVisParam(vis_num)%depthq) then
             call vsmDepthCueing(fdistance, lumi)
         end if

         ! recalc alpha for short intval
         color_(4) = 1.0 - (1.0-color_(4))**vsmRayStride

         one_m_salpha = 1.0 - color(4);
         lumi_a = color_(4) * one_m_salpha;

!         color(1:3) = color(1:3) + lumi(1:3)*color_(1:3)*lumi_a ! RGB
         color(1:3) = color(1:3) + lumi(1:3)*lumi_a ! RGB
         color(4)   = color(4)   + lumi_a           ! A
       end do
     if(color(4) >= vsmErtValue) exit
     end if
     ! end of volume rendering

     prev_ci(1:3) = ci(1:3)

     ! ------------- skip some space ----------

     if(skipcalc) then
!       call vsmLogPrint("skipcalc in")
       skip = .false.
       ! skip a super voxel
       call vsmRayJump(ray_pos, norm, skip_intval, skip)
       if(skip) then
         ray_pos(1:3) = ray_pos(1:3) + skip_intval * norm(1:3)
         ray_ = ray_ + skip_intval
         if(.not.vsmIsInsidelWO(ray_pos)) then
            call vsmLogPrint("raycastim:by skip space, ray went out of local data", skip_intval)
            call vsmFindCoord(ray_pos, dx, ci)
            call vsmLogPrint("out pos", ray_pos)
            call vsmLogPrint("out ci", ci)
            call vsmError("by skip, ray went out of local data")
         end if

         if(.not.vsmIsInsidel(ray_pos)) then
            fdistance = ray_+ dist2pxl
            if(z_on .and. fdistance >= zdis) then
               call vsmGetPixelColor4(px, color_)
               one_m_salpha = 1.0 - color(4)
               lumi_a = color_(4) * one_m_salpha
               color(1:3) = color(1:3) + color_(1:3)*lumi_a ! RGB
               color(4)   = color(4)   + lumi_a           ! A
            end if
            exit
         end if
         ! for interpolation
         call vsmFindCoord(ray_pos, dx, ci)
         call vsmSetFrac(dx, vx)

         do i=1, NumScal
           call vsmSetScalGridVal(i, ci, grid_val(1:8, i))
         end do
         prev_ci(1:3) = ci(1:3)

         ! calculate each scalar value at current point
         do i=1, NumScal
           scal_val(i) = vsmGetScalVal(vx, grid_val(1:8, i))
         end do

         prev_scal_val(1:NumScal) = scal_val(1:NumScal)
         ray_ = ray_ + rintval
         prev_ray_pos(1:3) = ray_pos(1:3)
         prev_vx = vx
         cycle
       end if

       ! skip a voxel
       skip = .false.
       if(vis_method(3)) call vsmRaySkipVoxel(ci, ray_pos, norm, skip_intval, skip)
       if(skip) then
         ray_pos(1:3) = ray_pos(1:3) + skip_intval * norm(1:3)
         ray_ = ray_ + skip_intval

         ! for interpolation
         call vsmFindCoord(ray_pos, dx, ci)
         call vsmSetFrac(dx, vx)

         if(ci(1) /= prev_ci(1) .or. ci(2) /= prev_ci(2) .or. ci(3) /= prev_ci(3)) then
           do i=1, NumScal
             call vsmSetScalGridVal(i, ci, grid_val(1:8, i))
           end do
         end if
         prev_ci(1:3) = ci(1:3)

         ! calculate each scalar value at current point
         do i=1, NumScal
           scal_val(i) = vsmGetScalVal(vx, grid_val(1:8, i))
         end do

         prev_scal_val(1:NumScal) = scal_val(1:NumScal)

         ray_ = ray_ + rintval
         prev_ray_pos(1:3) = ray_pos(1:3)
         prev_vx = vx
         cycle
       end if
!       call vsmLogPrint("skipcalc out")
     end if
     !-----------------------------------------

     prev_ray_pos(1:3) = ray_pos(1:3)
     ray_ = ray_ + rintval
     prev_scal_val(1:NumScal) = scal_val(1:NumScal)
     prev_vx = vx
   end do

   call vsmRaySegment(firstin_ray + dist2pxl, fdistance-rintval, color, px)
 end subroutine vsmRayCastIM

 subroutine vsmScalarVisualization(vis_num, cam_num, time, blend, dp)
     integer, intent(in) :: vis_num, cam_num, time
     logical, intent(in) :: blend
     real(kind=vdp), intent(in) :: dp
     integer :: i, j, k
     real(kind=vdp), dimension(3) :: norm, pixel, topleft, pos

     real(kind=vdp), dimension(3, 4) :: tnorm
     real(kind=vdp), dimension(4) :: inpro
     real(kind=vdp) :: a, b

     real(kind=vdp) :: dist2pxl, ray2ldat, ray2dat, ray
     integer :: step
     integer, dimension(2) :: px
     logical, dimension(3) :: v_method, rmv_method
     real(kind=vdp), dimension(2, vsmImageData%width,vsmImageData%height) :: hit
     real(kind=vdp) :: min_ray2ldat = vsmFaraway
     real, dimension(4) :: color
     real(kind=vdp) :: zdis
     real :: alpha

!   integer :: comm, myrank, psize
!   call vsmGetMPIVars(comm, myrank, psize)

     if(vsmVisParam(vis_num)%slice_num > 0) then
         v_method(1) = .true.; rmv_method(1) = .true. 
     else
         v_method(1) = .false.; rmv_method(1) = .false.
     end if

     if(vsmVisParam(vis_num)%iso_num > 0) then
         v_method(2) = .true.; rmv_method(2) = .true.
     else
         v_method(2) = .false.; rmv_method(2) = .false.
     end if

     if(vsmVisParam(vis_num)%volren_num > 0) then
         v_method(3) = .true.; rmv_method(3) = .true.
     else
         v_method(3) = .false.; rmv_method(3) = .false.
     end if

!     z_on = .true. ! z is always on because of the frame

     call vsmLogPrint("entering scalar visualization")

     topleft(1:3) = vsmImageData%topleft(1:3) + vsmImageData%dp/2.0_vdp * (-vsmCamera(cam_num)%up(1:3) &
                        + vsmCamera(cam_num)%right(1:3))

!$omp parallel do private(i,j,pixel, norm, tnorm, dist2pxl, px, ray2ldat, ray2dat) schedule(dynamic, 1)
     do j=1, vsmImageData%height
         do i=1, vsmImageData%width
             ! calculate normal vactor
             pixel(1:3)  = topleft(1:3) + vsmImageData%dp*((i-1)*vsmCamera(cam_num)%right(1:3) &
                  - (j-1)*vsmCamera(cam_num)%up(1:3))

             if(vsmCamera(cam_num)%pers) then
             ! Perspective Projection
                 norm(1:3) = pixel(1:3) - vsmCamera(cam_num)%pos(1:3)
                 dist2pxl = sqrt(dot_product(norm, norm))
                 call vsmNormVec(norm)
             else
             ! Parallel Projection
                 norm(1:3) = vsmCamera(cam_num)%front(1:3)
                 dist2pxl = vsmCamera(cam_num)%near
             end if

             px(1) = i; px(2) = j

             ray2dat = vsmRayDist(pixel, norm, vsmCoord%origin, vsmCoord%cend)
             if(ray2dat > vsmFarAwayd10) then
                hit(1,i,j) = vsmFarAway
                hit(2,i,j) = vsmFarAway
                cycle
             end if
             hit(1,i,j) = ray2dat

             ray2ldat = vsmRayDist(pixel, norm, vsmLocalCoord%lorigin, vsmLocalCoord%lcend)
             if(ray2ldat > vsmFarAwayd10) then
             ! ray does not hit the local data
                hit(2,i,j) = vsmFarAway
                cycle
             end if
             hit(2,i,j) = ray2ldat
         end do
     end do

!
!   First, Opaque Rendering
!
     if(vsmVisParam(vis_num)%slice_solid .or. vsmVisParam(vis_num)%iso_solid) then
     ! Opaque Slices and Isosurface
!$omp parallel do private(i,j,pixel, norm, dist2pxl, px, ray2ldat, ray2dat, ray) schedule(dynamic, 1)
         do j=1, vsmImageData%height
             do i=1, vsmImageData%width
                 if(hit(1,i,j) > vsmFarAwayd10 .or. hit(2,i,j) > vsmFarAwayd10) cycle
                 ! calculate normal vactor
                 pixel(1:3)  = topleft(1:3) + vsmImageData%dp*((i-1)*vsmCamera(cam_num)%right(1:3) &
                      - (j-1)*vsmCamera(cam_num)%up(1:3))

                 if(vsmCamera(cam_num)%pers) then
                 ! Perspective Projection
                     norm(1:3) = pixel(1:3) - vsmCamera(cam_num)%pos(1:3)
                     dist2pxl = sqrt(dot_product(norm, norm))
                     call vsmNormVec(norm)
                 else
                 ! Parallel Projection
                     norm(1:3) = vsmCamera(cam_num)%front(1:3)
                     dist2pxl = vsmCamera(cam_num)%near
                 end if

                px(1) = i; px(2) = j
                ray2dat = hit(1,i,j);  ray2ldat = hit(2,i,j)

                if(vsmVisParam(vis_num)%slice_solid) then
                  select case (vsmPCmode)
                      case (0)
                          call vsmSlicerIMS(pixel, norm, ray2ldat, dist2pxl, vis_num, cam_num, px)
                      case (1)
                          call vsmSlicerIMS(pixel, norm, ray2ldat, dist2pxl, vis_num, cam_num, px)
                      case (2)

                      case default
                          call vsmError("pcmode is strange: slicerims")
                  end select
                end if

                if(vsmVisParam(vis_num)%iso_solid) then
                   ray = ray2ldat
                   select case (vsmPCmode)
                      case (0)
                         call vsmIsoRendererIMS(pixel, norm, ray, ray2ldat, dist2pxl, rmv_method, vis_num, cam_num, px)
                      case (1)
                      case (2)
                      case default
                         call vsmError("pcmode is strange: isorenderer")
                   end select
                 end if
            end do
        end do
          if(vsmVisParam(vis_num)%depthq) then
             call vsmDepthCueingImage
          end if
#ifndef SERIAL
           select case (vsmPCmode)
             case (0)
                call vsmBinaryTreeIM(cam_num, vis_num, time)

             case (1)
             case (2)

             case default
               call vsmError("pc mode is strange : vsmVisualization scalblend")

           end select
#endif
        end if

        if(vsmVisParam(vis_num)%slice_trans .or. vsmVisParam(vis_num)%iso_trans .or. rmv_method(3)) then
         call vsmClearRImage
!$omp parallel do private(i,j,pixel, norm, dist2pxl, px, ray2ldat, ray2dat, color, zdis, pos) schedule(dynamic, 1)
         do j=1, vsmImageData%height
             do i=1, vsmImageData%width
                 px(1) = i; px(2) = j
                 if(hit(1,i,j) > vsmFarAwayd10) then
                    zdis = vsmGetZbuf(px)
                    if(zdis < vsmFarAwayd10) then
                       call vsmGetPixelColor4(px, color)
                       call vsmPaintPixel4R(px, color, zdis)
                       call vsmClearPixel(px)
                    end if
                    cycle
                 end if
                 ! calculate normal vactor
                 pixel(1:3)  = topleft(1:3) + vsmImageData%dp*((i-1)*vsmCamera(cam_num)%right(1:3) &
                      - (j-1)*vsmCamera(cam_num)%up(1:3))

                 if(vsmCamera(cam_num)%pers) then
                 ! Perspective Projection
                     norm(1:3) = pixel(1:3) - vsmCamera(cam_num)%pos(1:3)
                     dist2pxl = sqrt(dot_product(norm, norm))
                     call vsmNormVec(norm)
                 else
                 ! Parallel Projection
                     norm(1:3) = vsmCamera(cam_num)%front(1:3)
                     dist2pxl = vsmCamera(cam_num)%near
                 end if

                 zdis = vsmGetZbuf(px)
                 if(zdis < vsmFarAwayd10) then
                    pos(1:3) = pixel(1:3) + norm(1:3)*(zdis - dist2pxl)
                    if(.not.vsmIsInsided(pos)) then
                      call vsmGetPixelColor4(px, color)
                      call vsmPaintPixel4R(px, color, zdis)
                      call vsmClearPixel(px)
                    end if
                 end if

            end do
        end do

        if(vsmVisParam(vis_num)%slice_trans .and. .not. vsmVisParam(vis_num)%iso_trans .and. .not.rmv_method(3)) then
!$omp parallel do private(i,j,pixel, norm, dist2pxl, px, ray2ldat, ray2dat, color, zdis) schedule(dynamic, 1)
         do j=1, vsmImageData%height
             do i=1, vsmImageData%width
                 px(1) = i; px(2) = j
                 if(hit(1,i,j) > vsmFarAwayd10 .or. hit(2,i,j) > vsmFarAwayd10) then
                    cycle
                 end if
                 ! calculate normal vactor
                 pixel(1:3)  = topleft(1:3) + vsmImageData%dp*((i-1)*vsmCamera(cam_num)%right(1:3) &
                      - (j-1)*vsmCamera(cam_num)%up(1:3))

                 if(vsmCamera(cam_num)%pers) then
                 ! Perspective Projection
                     norm(1:3) = pixel(1:3) - vsmCamera(cam_num)%pos(1:3)
                     dist2pxl = sqrt(dot_product(norm, norm))
                     call vsmNormVec(norm)
                 else
                 ! Parallel Projection
                     norm(1:3) = vsmCamera(cam_num)%front(1:3)
                     dist2pxl = vsmCamera(cam_num)%near
                 end if

                 ray2dat = hit(1,i,j);  ray2ldat = hit(2,i,j)

                select case (vsmPCmode)
                    case (0)
                        call vsmSlicerIMT(pixel, norm, ray2ldat, dist2pxl, vis_num, cam_num, px)
                    case (1)
                        call vsmSlicerIMT(pixel, norm, ray2ldat, dist2pxl, vis_num, cam_num, px)
                    case (2)

                    case default
                        call vsmError("pcmode is strange: slicerimt")
                end select
            end do
        end do

        else
        
!$omp parallel do private(i,j,pixel, norm, dist2pxl, px, ray2ldat, ray2dat, step, ray, color, zdis) schedule(dynamic, 1)
         do j=1, vsmImageData%height
             do i=1, vsmImageData%width
                 px(1) = i; px(2) = j
                 if(hit(1,i,j) > vsmFarAwayd10 .or. hit(2,i,j) > vsmFarAwayd10) then
                    cycle
                 end if
                 ! calculate normal vactor
                 pixel(1:3)  = topleft(1:3) + vsmImageData%dp*((i-1)*vsmCamera(cam_num)%right(1:3) &
                      - (j-1)*vsmCamera(cam_num)%up(1:3))

                 if(vsmCamera(cam_num)%pers) then
                     ! Perspective Projection
                     norm(1:3) = pixel(1:3) - vsmCamera(cam_num)%pos(1:3)
                     dist2pxl = sqrt(dot_product(norm, norm))
                     call vsmNormVec(norm)
                 else
                     ! Parallel Projection
                     norm(1:3) = vsmCamera(cam_num)%front(1:3)
                     dist2pxl = vsmCamera(cam_num)%near
                 end if

                 ray2dat = hit(1,i,j); ray2ldat = hit(2,i,j)

                 if(vsmPCmode /=2 .and. v_method(3)) then
                    step = int((ray2ldat - ray2dat)/vsmIntval%ray_interval)
                    if(step == 0) then !1.0D-10 : for secure
                        if(ray2dat == ray2ldat) then
                           ray = ray2dat
                        else
                           ray = ray2dat + vsmIntval%ray_interval
                        end if
                     else
                        if(step > 0) then
                            if(step*vsmIntval%ray_interval == ray2ldat - ray2dat) then
                                ray = ray2dat + vsmIntval%ray_interval*step
                            else
                                ray = ray2dat + vsmIntval%ray_interval*(step+1)                            
                          end if
                        else
                            ray = ray2ldat
                        end if
                     end if
                   else
                     ray = ray2ldat
                   end if

                   select case (vsmPCmode)
                      case (0)
                         call vsmRayCastIM(pixel, norm, ray, ray2ldat, dist2pxl, rmv_method, vis_num, cam_num, px)
                      case (1)
                         call vsmRayCastIM(pixel, norm, ray, ray2ldat, dist2pxl, rmv_method, vis_num, cam_num, px)
                      case (2)

                      case default
                         call vsmError("pcmode is strange: raycaster")
                   end select
              end do
            end do
        end if
        end if
     
        if(vsmPCmode == 0) return
   call vsmLogPrint("scalar visualization ends")

 end subroutine vsmScalarVisualization

end module vsmRayCaster

