module vsmParam
 use vsmConst
 use vsmFunction

 implicit none
 private


 !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 !---- declaration of variables ----
 !
 type(vismo_visparameters), dimension(vsmMaxVisKind), save, public :: vsmVisParam
 integer, save, public :: vsmNumViskind
 integer, save, public :: vsmPCmode ! 0:image, 1:image, point cloud, 2:point cloud
 type(vismo_interval), save, public :: vsmIntval
 type(vismo_mpi), save :: vsmMPIVars
 !
 !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 public vsmGetCurrentIsonum, vsmIsAboveSlice, vsmIsAboveIsoL
 public vsmAddSline, vsmAddArrow, vsmAddVisualization, vsmCalcIntValr, vsmCalcIntvalu, vsmGetIsoAlpha
 public vsmGetIsoColor, vsmGetMPIVars, vsmGetSliceAlpha, vsmGetSliceColor, vsmGetVolColor, vsmInitVisParam
 public vsmLightOn
 public vsmSetArrowParam, vsmSetDepthq, vsmSetIsoParam, vsmSetPtclParam, vsmSetSeedParam, vsmSetSliceParam
 public vsmSetSlineParam, vsmSetVolren
 public vsmSetARSeedParam, vsmGetSliceNormF

 contains

 subroutine vsmInitVisParam(comm, rank, psize, ompth)
   integer, intent(in) :: comm, rank, psize, ompth
   ! after mpi_init
   call vsmInitIsoparam
   call vsmInitSliceparam
   call vsmInitVolparam
   call vsmInitSlineparam
   call vsmInitArrowparam
   call vsmInitPtclparam
   call vsmInitLight
   call vsmInitDepthq
   vsmNumViskind = 0
   vsmPCmode = 0
   vsmMPIVars%comm   = comm
   vsmMPIVars%myrank = rank
   vsmMPIVars%psize  = psize
   vsmMPIVars%omp_tn = ompth
 end subroutine vsmInitVisParam

 subroutine vsmAddVisualization
   vsmNumViskind = vsmNumViskind + 1
   if(vsmNumViskind > vsmMaxVisKind) call vsmError('add visualization : exceed vsmMaxVisKind')
 end subroutine vsmAddVisualization

 function vsmGetCurrentIsonum()
   integer :: vsmGetCurrentIsonum
   vsmGetCurrentIsonum = vsmVisParam(vsmNumViskind)%iso_num
 end function vsmGetCurrentIsonum

 ! subroutine for lights
 subroutine vsmInitLight
  integer :: i
  vsmVisParam(1:vsmMaxVisKind)%light_num = 0
  do i=1, vsmMaxVisKind
    vsmVisParam(i)%light_used(1:vsmMaxLight) = .false.
  end do
 end subroutine vsmInitLight

 subroutine vsmLightOn(nlight)
   integer, intent(in) :: nlight
   integer :: light_num

   if(nlight > vsmMaxLight .or. nlight < 1) call vsmError('num of lights')
   light_num = vsmVisParam(vsmNumViskind)%light_num + 1
   vsmVisParam(vsmNumViskind)%light_num = light_num
   vsmVisParam(vsmNumViskind)%light_used(nlight) = .true.
 end subroutine vsmLightOn

 ! subroutines for depthcue
 subroutine vsmInitDepthq
   vsmVisParam(1:vsmMaxVisKind)%depthq = .false.
 end subroutine vsmInitDepthq

 subroutine vsmSetDepthQ
   vsmVisParam(vsmNumViskind)%depthq = .true.
 end subroutine vsmSetDepthQ

 ! subroutines for isosurface
 subroutine vsmInitIsoparam
   integer :: i, k

   do k=1, vsmMaxVisKind
     do i=1, vsmMaxIsosurf
       vsmVisParam(k)%isoparam(i)%iso_level = 0.0_vdp
       vsmVisParam(k)%isoparam(i)%alpha = 0.0
       vsmVisParam(k)%isoparam(i)%nscal = -1
       vsmVisParam(k)%isoparam(i)%coloring = -1
       vsmVisParam(k)%isoparam(i)%amb(1:3)  = 0.0
       vsmVisParam(k)%isoparam(i)%dif(1:3)  = 0.0
       vsmVisParam(k)%isoparam(i)%spec(1:3) = 0.0
       vsmVisParam(k)%isoparam(i)%shin = 0.0
     end do
     vsmVisParam(k)%iso_num = 0
     vsmVisParam(k)%iso_solid = .false.
     vsmVisParam(k)%iso_trans = .false.
   end do
 end subroutine vsmInitIsoparam

 function vsmCheckIso(vnum, niso)
   integer, intent(in) :: vnum, niso
   logical :: vsmCheckIso

   if(vnum > vsmMaxVisKind .or. vnum < 1) then
     vsmCheckIso = .false.
     return
   end if

   if(niso < 1 .or. niso > vsmVisParam(vnum)%iso_num) then
     vsmCheckIso = .false.
     return
   end if
   vsmCheckIso = .true.
 end function vsmCheckIso

 function vsmCheckIso2(vnum, niso, lpse)
   integer, intent(in) :: vnum
   integer, dimension(vsmVecLen+1), intent(in) :: niso
   integer, dimension(2), intent(in) :: lpse
   logical :: vsmCheckIso2
   integer :: i, err

   if(vnum > vsmMaxVisKind .or. vnum < 1) then
     vsmCheckIso2 = .false.
     return
   end if

   err = 0
   do i=lpse(1), lpse(2)
      if(niso(i) < 1 .or. niso(i) > vsmVisParam(vnum)%iso_num) then
         err = err + 1
      end if 
   end do

   if(err > 0) then
     vsmCheckIso2 = .false.
     return
   end if

   vsmCheckIso2 = .true.
 end function vsmCheckIso2

 subroutine vsmSetIsoParam(level, nscal, alpha, amb, dif, spec, shin, coloring)
   integer, intent(in) :: nscal, coloring
   real(kind=vdp), intent(in) :: level
   real, intent(in) :: alpha, shin
   real, dimension(3), intent(in) :: amb, dif, spec
   integer :: iso_num

   if(nscal > vsmMaxScalar .or. nscal < 1 .or. vsmNumViskind > vsmMaxVisKind) then
     call vsmError('vsmSetIsoParam')
   end if
   iso_num = vsmVisParam(vsmNumViskind)%iso_num + 1

   if(iso_num > vsmMaxIsosurf) call vsmError('set_isparam : exceed vsmMaxIsosurf')
   vsmVisParam(vsmNumViskind)%iso_num = iso_num
   vsmVisParam(vsmNumViskind)%isoparam(iso_num)%iso_level = level
   vsmVisParam(vsmNumViskind)%isoparam(iso_num)%alpha = alpha
   vsmVisParam(vsmNumViskind)%isoparam(iso_num)%nscal = nscal
   if(coloring < 0 .or. nscal == coloring) then
      vsmVisParam(vsmNumViskind)%isoparam(iso_num)%coloring = -1
   else
      vsmVisParam(vsmNumViskind)%isoparam(iso_num)%coloring = coloring
   end if

   call vsmCheckFA01(amb); call vsmCheckFA01(dif); call vsmCheckFA01(spec)
   call vsmCheckF01(alpha)
   vsmVisParam(vsmNumViskind)%isoparam(iso_num)%amb(1:3)  = amb(1:3)
   vsmVisParam(vsmNumViskind)%isoparam(iso_num)%dif(1:3)  = dif(1:3)
   vsmVisParam(vsmNumViskind)%isoparam(iso_num)%spec(1:3) = spec(1:3)

   vsmVisParam(vsmNumViskind)%isoparam(iso_num)%shin = shin
 end subroutine vsmSetIsoParam

subroutine vsmChangeIsoLevel(visnum, isonum, level)
   integer, intent(in) :: visnum, isonum
   real(kind=vdp), intent(in) :: level

   if(visnum > vsmNumViskind .or. visnum < 1) then
     call vsmError('vsmChangeIsoLevel :: visnum is strange')   
   end if

   if(isonum > vsmVisParam(visnum)%iso_num .or. isonum < 1) then
     call vsmError('vsmChangeIsoLevel :: isonum is strange')   
   end if

   vsmVisParam(visnum)%isoparam(isonum)%iso_level = level
   if(vsmDbg) then
     call vsmLogPrint("change iso level visnum =", visnum)
     call vsmLogPrint("change iso level isonum =", isonum)
     call vsmLogPrint("change iso level to ", level)
   end if
 end subroutine vsmChangeIsoLevel

 subroutine vsmGetIsoColor(visnum, niso, alpha, amb, dif, spec, shin)
   integer, intent(in) :: visnum, niso
   real, intent(out) :: alpha, shin
   real, dimension(3), intent(out) :: amb, dif, spec

   if(.not.vsmCheckIso(visnum, niso)) call vsmError("error: vsmGetIsoColor")
   alpha = vsmVisParam(visnum)%isoparam(niso)%alpha
   amb(1:3) = vsmVisParam(visnum)%isoparam(niso)%amb(1:3)
   dif(1:3) = vsmVisParam(visnum)%isoparam(niso)%dif(1:3)
   spec(1:3) = vsmVisParam(visnum)%isoparam(niso)%spec(1:3)
   shin = vsmVisParam(visnum)%isoparam(niso)%shin
 end subroutine vsmGetIsoColor

 function vsmIsAboveIsoL(visnum, niso, s)
   integer, intent(in) :: niso, visnum
   real(kind=vdp), intent(in) :: s
   real(kind=vdp) :: vsmIsAboveIsoL

   if(.not.vsmCheckIso(visnum, niso)) call vsmError("error: vsmIsAboveIsoL")
   vsmIsAboveIsoL = vsmVisParam(visnum)%isoparam(niso)%iso_level - s
 end function vsmIsAboveIsoL

 subroutine vsmGetIsoAlpha(visnum, niso, alpha)
   integer, intent(in) :: visnum, niso
   real, intent(out) :: alpha

   if(.not.vsmCheckIso(visnum, niso)) call vsmError("error: vsmGetIsoAlpha")
   alpha = vsmVisParam(visnum)%isoparam(niso)%alpha
 end subroutine vsmGetIsoAlpha

 ! subroutines and functions for color slice
 subroutine vsmInitSliceparam
   integer :: i, k

   do k=1, vsmMaxVisKind
     do i=1, vsmMaxSlice
       vsmVisParam(k)%sliceparam(i)%sc(1:4) = 0.0_vdp
       vsmVisParam(k)%sliceparam(i)%alpha = 0.0
       vsmVisParam(k)%sliceparam(i)%nscal = -1
       vsmVisParam(k)%sliceparam(i)%amb = 0.0
       vsmVisParam(k)%sliceparam(i)%dif = 0.0
       vsmVisParam(k)%sliceparam(i)%spec = 0.0
       vsmVisParam(k)%sliceparam(i)%shin = 0.0
     end do
     vsmVisParam(k)%slice_num = 0
     vsmVisParam(k)%slice_solid = .false.
     vsmVisParam(k)%slice_trans = .false.
   end do
 end subroutine vsmInitSliceparam

 function vsmCheckSlice(vnum, nsli)
   integer, intent(in) :: vnum, nsli
   logical :: vsmCheckSlice

   if(vnum > vsmMaxVisKind .or. vnum < 1) then
     vsmCheckSlice = .false.
     return
   end if

   if(nsli < 1 .or. nsli > vsmVisParam(vnum)%slice_num) then
     vsmCheckSlice = .false.
     return
   end if
   vsmCheckSlice = .true.
 end function vsmCheckSlice

 subroutine vsmSetSliceParam(coef, nscal, alpha, amb, dif, spec, shin)
   real(kind=vdp), dimension(4), intent(in) :: coef
   integer, intent(in) :: nscal
   real, intent(in) :: alpha, amb, dif, spec, shin
   integer :: slice_num

   if(nscal > vsmMaxScalar .or. vsmNumViskind > vsmMaxVisKind) call vsmError('vsmSetSliceParam')
   slice_num = vsmVisParam(vsmNumViskind)%slice_num + 1
   if(slice_num > vsmMaxSlice) call vsmError('vsmSetSliceParam : exceed vsmMaxSlice')
   vsmVisParam(vsmNumViskind)%slice_num = slice_num
   vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%sc(1:4) = coef(1:4)

   vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%nscal = nscal

   call vsmCheckF01(alpha)
   vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%alpha = alpha
   call vsmCheckF01(amb); call vsmCheckF01(dif); call vsmCheckF01(spec); 
   vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%amb  = amb
   vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%dif  = dif
   vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%spec = spec
   vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%shin = shin
 end subroutine vsmSetSliceParam

 function vsmIsAboveSlice(visnum, nslice, x)
   integer, intent(in) :: nslice, visnum
   real(kind=vdp), dimension(3), intent(in) :: x
   real(kind=vdp) :: vsmIsAboveSlice
   integer :: i

   if(.not.vsmCheckSlice(visnum, nslice)) call vsmError("error: vsmIsAboveSlice")
     vsmIsAboveSlice = vsmVisParam(visnum)%sliceparam(nslice)%sc(1) * x(1) + &
       vsmVisParam(visnum)%sliceparam(nslice)%sc(2) * x(2) + &
       vsmVisParam(visnum)%sliceparam(nslice)%sc(3) * x(3) + &
       vsmVisParam(visnum)%sliceparam(nslice)%sc(4)
 end function vsmIsAboveSlice

 subroutine vsmGetSliceColor(visnum, nslice, alpha, amb, dif, spec, shin)
   integer, intent(in) :: nslice, visnum
   real, intent(out) :: alpha, amb, dif, spec, shin

   if(.not.vsmCheckSlice(visnum, nslice)) call vsmError("error: vsmGetSliceColor")
   alpha = vsmVisParam(visnum)%sliceparam(nslice)%alpha
   amb   = vsmVisParam(visnum)%sliceparam(nslice)%amb
   dif   = vsmVisParam(visnum)%sliceparam(nslice)%dif
   spec  = vsmVisParam(visnum)%sliceparam(nslice)%spec
   shin  = vsmVisParam(visnum)%sliceparam(nslice)%shin
 end subroutine vsmGetSliceColor

 subroutine vsmGetSliceAlpha(visnum, nslice, alpha)
   integer, intent(in) :: nslice, visnum
   real, intent(out) :: alpha

   if(.not.vsmCheckSlice(visnum, nslice)) call vsmError("error: vsmGetSliceAlpha")
   alpha = vsmVisParam(visnum)%sliceparam(nslice)%alpha
 end subroutine vsmGetSliceAlpha

 subroutine vsmGetSliceNorm(visnum, nslice, norm)
   integer, intent(in) :: nslice, visnum
   real(kind=vdp), dimension(3), intent(out) :: norm

   if(.not.vsmCheckSlice(visnum, nslice)) call vsmError("error: vsmGetSliceNorm")
   norm(1:3) = vsmVisParam(visnum)%sliceparam(nslice)%sc(1:3)
   call vsmNormVec(norm)
 end subroutine vsmGetSliceNorm

 subroutine vsmGetSliceNormf(visnum, nslice, norm)
   integer, intent(in) :: nslice, visnum
   real, dimension(3), intent(out) :: norm

   if(.not.vsmCheckSlice(visnum, nslice)) call vsmError("error: vsmGetSliceNormf")
   norm(1:3) = real(vsmVisParam(visnum)%sliceparam(nslice)%sc(1:3))
   call vsmNormVecF(norm)
 end subroutine vsmGetSliceNormf

 ! subroutines for volume rendering
 subroutine vsmInitVolparam
  integer :: i, k

  do k=1, vsmMaxVisKind
    do i=1, vsmMaxScalar
      vsmVisParam(k)%volparam(i)%nscal = -1
      vsmVisParam(k)%volparam(i)%amb  = 0.0
      vsmVisParam(k)%volparam(i)%dif  = 0.0
      vsmVisParam(k)%volparam(i)%spec = 0.0
      vsmVisParam(k)%volparam(i)%shin = 0.0
    end do
    vsmVisParam(k)%volren_num = 0
  end do
 end subroutine vsmInitVolparam

 function vsmCheckVolren(vnum, nvol)
   integer, intent(in) :: vnum, nvol
   logical :: vsmCheckVolren

   if(vnum > vsmMaxVisKind .or. vnum < 1) then
     vsmCheckVolren = .false.
     return
   end if

   if(nvol < 1 .or. nvol > vsmVisParam(vnum)%volren_num) then
     vsmCheckVolren = .false.
     return
   end if
   vsmCheckVolren = .true.
 end function vsmCheckVolren

 subroutine vsmSetVolren(nscal, amb, dif, spec, shin)
   integer, intent(in) :: nscal
   real, intent(in) :: amb, dif, spec, shin
   integer :: volren_num

   if(nscal > vsmMaxScalar .or. nscal < 1) call vsmError('vsmSetVolren')
   volren_num = vsmVisParam(vsmNumViskind)%volren_num + 1
   if(volren_num > vsmMaxVolren) call vsmError('vsmSetVolren : exceed vsmMaxVolren')
   vsmVisParam(vsmNumViskind)%volren_num = volren_num
   vsmVisParam(vsmNumViskind)%volparam(volren_num)%nscal = nscal

   call vsmCheckF01(amb); call vsmCheckF01(dif); call vsmCheckF01(spec); 
   vsmVisParam(vsmNumViskind)%volparam(volren_num)%amb  = amb
   vsmVisParam(vsmNumViskind)%volparam(volren_num)%dif  = dif
   vsmVisParam(vsmNumViskind)%volparam(volren_num)%spec = spec
   vsmVisParam(vsmNumViskind)%volparam(volren_num)%shin = shin
 end subroutine vsmSetVolren

 subroutine vsmGetVolColor(visnum, nvol, amb, dif, spec, shin)
   integer, intent(in) :: nvol, visnum
   real, intent(out) :: amb, dif, spec, shin

   if(.not.vsmCheckVolren(visnum, nvol)) call vsmError("error: vsmGetVolColor")
   amb   = vsmVisParam(visnum)%volparam(nvol)%amb
   dif   = vsmVisParam(visnum)%volparam(nvol)%dif
   spec  = vsmVisParam(visnum)%volparam(nvol)%spec
   shin  = vsmVisParam(visnum)%volparam(nvol)%shin
 end subroutine vsmGetVolColor

 ! common subroutines for ray casting and StreamLines
 ! vsmCalcIntvalU or vsmCalcIntvalR must be called when setting coordinates
 subroutine vsmCalcIntvalU(dx)
   real(kind=vdp), dimension(3), intent(in) :: dx
   real(kind=vdp) :: mind

   call vsmLogPrint("calcintvalu is called")
   mind = min(dx(1), dx(2), dx(3))
   vsmIntval%ray_interval = mind*vsmRayStride
   vsmIntval%dh = mind*vsmIntegStride
   if(vsmDbg) then
     call vsmLogPrint("ray_interval:", vsmIntval%ray_interval)
     call vsmLogPrint("dh:", vsmIntval%dh)
   end if
 end subroutine vsmCalcIntvalU

 subroutine vsmCalcIntvalR(coordx, coordy, coordz, nx)
   integer, dimension(3), intent(in) :: nx
   real(kind=vdp), dimension(nx(1)), intent(in) :: coordx
   real(kind=vdp), dimension(nx(2)), intent(in) :: coordy
   real(kind=vdp), dimension(nx(3)), intent(in) :: coordz
   real(kind=vdp) :: mind, d
   integer :: i

   mind = coordx(nx(1)) - coordx(1)
   do i=1, nx(1)-1
     d = coordx(i+1)-coordx(i)
     if( d < mind) then
        mind = d
     end if
   end do

   do i=1, nx(2)-1
     d = coordy(i+1)-coordy(i)
     if( d < mind) then
        mind = d
     end if
   end do

   do i=1, nx(3)-1
     d = coordz(i+1)-coordz(i)
     if( d < mind) then
        mind = d
     end if
   end do

   vsmIntval%ray_interval = mind*vsmRayStride
   vsmIntval%dh = mind*vsmIntegStride
 end subroutine vsmCalcIntvalR

 ! subrtouines for stream lines
 subroutine vsmInitSlineparam
   integer :: i, k

   do k=1, vsmMaxVisKind
     do i=1, vsmMaxVector
       vsmVisParam(k)%slineparam(i)%seed(1:3, 1:vsmMaxSeed) = 0.0_vdp
       vsmVisParam(k)%slineparam(i)%num_seed = 0
       vsmVisParam(k)%slineparam(i)%steps_remained(1:vsmMaxSeed) = vsmMaxSlineSteps
       vsmVisParam(k)%slineparam(i)%nvec = -1
       vsmVisParam(k)%slineparam(i)%amb(1:3)  = 0.0
       vsmVisParam(k)%slineparam(i)%dif(1:3)  = 0.0
       vsmVisParam(k)%slineparam(i)%spec(1:3) = 0.0
       vsmVisParam(k)%slineparam(i)%shin = 0.0
       vsmVisParam(k)%slineparam(i)%radius = 1.0_vdp
       vsmVisParam(k)%slineparam(i)%monocolor = .false.
     end do
     vsmVisParam(k)%sline_num = 0
   end do
 end subroutine vsmInitSlineparam

 function vsmCheckSline(vnum, nsline)
   integer, intent(in) :: vnum, nsline
   logical :: vsmCheckSline

   if(vnum > vsmMaxVisKind .or. vnum < 1) then
     vsmCheckSline = .false.
     return
   end if

   if(nsline < 1 .or. nsline > vsmVisParam(vnum)%sline_num) then
     vsmCheckSline = .false.
     return
   end if
   vsmCheckSline = .true.
 end function vsmCheckSline

 subroutine vsmAddSline
   integer :: sline_num

   sline_num = vsmVisParam(vsmNumViskind)%sline_num + 1
   if(sline_num > vsmMaxVector) call vsmError('vsmAddSline : exceed max_sline')
   vsmVisParam(vsmNumViskind)%sline_num = sline_num
 end subroutine vsmAddSline

 subroutine vsmSetSeedParam(x)
   real(kind=vdp), dimension(3), intent(in) :: x
   integer :: sline_num, num_seed

   sline_num = vsmVisParam(vsmNumViskind)%sline_num
   num_seed = vsmVisParam(vsmNumViskind)%slineparam(sline_num)%num_seed + 1
   if(num_seed > vsmMaxSeed) call vsmError('set seedparam : exceed vsmMaxSeed')
   vsmVisParam(vsmNumViskind)%slineparam(sline_num)%num_seed = num_seed
   vsmVisParam(vsmNumViskind)%slineparam(sline_num)%seed(1:3, num_seed) = x(1:3)
 end subroutine vsmSetSeedParam

 subroutine vsmSetSlineParam(nvec, mono, amb, dif, spec, shin, radius)
   integer, intent(in) :: nvec
   logical, intent(in) :: mono
   real, dimension(3), intent(in) :: amb, dif, spec
   real, intent(in) :: shin
   real(kind=vdp), intent(in) :: radius
   integer :: sline_num

   if(nvec > vsmMaxVector .or. nvec < 1) call vsmError('vsmSetSlineParam :: nvec')
   sline_num = vsmVisParam(vsmNumViskind)%sline_num
   vsmVisParam(vsmNumViskind)%slineparam(sline_num)%nvec = nvec
   call vsmCheckFA01(amb); call vsmCheckFA01(dif); call vsmCheckFA01(spec)
     if(mono) then
       vsmVisParam(vsmNumViskind)%slineparam(sline_num)%amb(1:3) = amb(1:3)
       vsmVisParam(vsmNumViskind)%slineparam(sline_num)%dif(1:3) = dif(1:3)
       vsmVisParam(vsmNumViskind)%slineparam(sline_num)%spec(1:3) = spec(1:3)
     else
       vsmVisParam(vsmNumViskind)%slineparam(sline_num)%amb(1) = amb(1)
       vsmVisParam(vsmNumViskind)%slineparam(sline_num)%dif(1) = dif(1)
       vsmVisParam(vsmNumViskind)%slineparam(sline_num)%spec(1) = spec(1)
     end if
   vsmVisParam(vsmNumViskind)%slineparam(sline_num)%shin = shin
   vsmVisParam(vsmNumViskind)%slineparam(sline_num)%radius = radius
   vsmVisParam(vsmNumViskind)%slineparam(sline_num)%monocolor = mono
 end subroutine vsmSetSlineParam

subroutine vsmGetSlineColor(visnum, nsline, amb, dif, spec, shin)
   integer, intent(in) :: nsline, visnum
   real, intent(out) :: amb, dif, spec, shin

   if(.not.vsmCheckSline(visnum, nsline)) call vsmError("error: vsmGetSlineColor")
   amb   = vsmVisParam(visnum)%slineparam(nsline)%amb(1)
   dif   = vsmVisParam(visnum)%slineparam(nsline)%dif(1)
   spec  = vsmVisParam(visnum)%slineparam(nsline)%spec(1)
   shin  = vsmVisParam(visnum)%slineparam(nsline)%shin
 end subroutine vsmGetSlineColor

 subroutine vsmGetMonoLineColor(visnum, nsline, amb, dif, spec, shin)
   integer, intent(in) :: nsline, visnum
   real, dimension(3), intent(out) :: amb, dif, spec
   real, intent(out) :: shin

   if(.not.vsmCheckSline(visnum, nsline)) call vsmError("error: vsmGetMonoLineColor")
   amb(1:3)   = vsmVisParam(visnum)%slineparam(nsline)%amb(1:3)
   dif(1:3)   = vsmVisParam(visnum)%slineparam(nsline)%dif(1:3)
   spec(1:3)  = vsmVisParam(visnum)%slineparam(nsline)%spec(1:3)
   shin  = vsmVisParam(visnum)%slineparam(nsline)%shin
 end subroutine vsmGetMonoLineColor

 ! subroutines for arrows
  subroutine vsmInitArrowparam
   integer :: i, k

   do k=1, vsmMaxVisKind
     do i=1, vsmMaxArrows
       vsmVisParam(k)%arrowparam(i)%sc(1:4) = 0.0_vdp
       vsmVisParam(k)%arrowparam(i)%seed(1:3, 1:vsmARMaxSeed) = 0.0_vdp
       vsmVisParam(k)%arrowparam(i)%num_seed = 0
       vsmVisParam(k)%arrowparam(i)%nvec = -1
       vsmVisParam(k)%arrowparam(i)%density = 1.0
       vsmVisParam(k)%arrowparam(i)%radius = 1.0
       vsmVisParam(k)%arrowparam(i)%length = 1.0
       vsmVisParam(k)%arrowparam(i)%mode = 0
       vsmVisParam(k)%arrowparam(i)%variable = .false.
     end do
     vsmVisParam(k)%arrow_num = 0
   end do
 end subroutine vsmInitArrowparam

 function vsmCheckArrow(vnum, narrow)
   integer, intent(in) :: vnum, narrow
   logical :: vsmCheckArrow

   if(vnum > vsmMaxVisKind .or. vnum < 1) then
     vsmCheckArrow = .false.
     return
   end if

   if(narrow < 1 .or. narrow > vsmVisParam(vnum)%arrow_num) then
     vsmCheckArrow = .false.
     return
   end if
   vsmCheckArrow = .true.
 end function vsmCheckArrow

 subroutine vsmAddArrow
   integer :: arrow_num

   arrow_num = vsmVisParam(vsmNumViskind)%arrow_num + 1
   if(arrow_num > vsmMaxArrows) call vsmError('vsmSetArrowParam')
   vsmVisParam(vsmNumViskind)%arrow_num = arrow_num
 end subroutine vsmAddArrow

 subroutine vsmSetArrowParam(coef, nvec, density, rad, length, mode, vari)
   real(kind=vdp), dimension(4), intent(in) :: coef
   integer, intent(in) :: nvec, mode
   logical, intent(in) :: vari
   real(kind=vdp), intent(in) :: density, rad, length
   integer :: arrow_num

   arrow_num = vsmVisParam(vsmNumViskind)%arrow_num
   if(nvec > vsmMaxVector .or. arrow_num > vsmMaxArrows) call vsmError('vsmSetArrowParam')
   vsmVisParam(vsmNumViskind)%arrowparam(arrow_num)%sc(1:4) = coef(1:4)
   vsmVisParam(vsmNumViskind)%arrowparam(arrow_num)%nvec = nvec
   vsmVisParam(vsmNumViskind)%arrowparam(arrow_num)%density = density
   vsmVisParam(vsmNumViskind)%arrowparam(arrow_num)%radius = rad
   vsmVisParam(vsmNumViskind)%arrowparam(arrow_num)%length = length
   vsmVisParam(vsmNumViskind)%arrowparam(arrow_num)%mode = mode
   vsmVisParam(vsmNumViskind)%arrowparam(arrow_num)%variable = vari
 end subroutine vsmSetArrowParam

 subroutine vsmSetArSeedParam(x)
   real(kind=vdp), dimension(3), intent(in) :: x
   integer :: ar_num, num_seed

   ar_num = vsmVisParam(vsmNumViskind)%arrow_num
   num_seed = vsmVisParam(vsmNumViskind)%arrowparam(ar_num)%num_seed + 1
   if(num_seed > vsmArMaxSeed) call vsmError('set seedparam : exceed vsmArMaxSeed')
   vsmVisParam(vsmNumViskind)%arrowparam(ar_num)%num_seed = num_seed
   vsmVisParam(vsmNumViskind)%arrowparam(ar_num)%seed(1:3, num_seed) = x(1:3)
 end subroutine vsmSetArSeedParam

 subroutine vsmGetArPlane(visnum, nslice, eq)
  real(kind=vdp), dimension(4), intent(out) :: eq
  integer, intent(in) :: nslice, visnum

  if(.not.vsmCheckArrow(visnum, nslice)) call vsmError("error: vsmGetArPlane")
  eq(1:4) = vsmVisParam(visnum)%arrowparam(nslice)%sc(1:4)
 end subroutine vsmGetArPlane

 subroutine vsmGetArMode(visnum, nslice, mode)
  integer, intent(out) :: mode
  integer, intent(in) :: nslice, visnum

  if(.not.vsmCheckArrow(visnum, nslice)) call vsmError("error: vsmGetArMode")
  mode = vsmVisParam(visnum)%arrowparam(nslice)%mode
 end subroutine vsmGetArMode

 subroutine vsmGetArParam(visnum, nslice, den, rad, length, vari)
  real(kind=vdp), intent(out) :: den, rad, length
  integer, intent(in) :: nslice, visnum
  logical,intent(out) :: vari

  if(.not.vsmCheckArrow(visnum, nslice)) call vsmError("error: vsmGetArParam")
  den = vsmVisParam(visnum)%arrowparam(nslice)%density
  rad = vsmVisParam(visnum)%arrowparam(nslice)%radius
  length = vsmVisParam(visnum)%arrowparam(nslice)%length
  vari = vsmVisParam(visnum)%arrowparam(nslice)%variable
 end subroutine vsmGetArParam

 ! subroutines and functions for particle visualization
 subroutine vsmInitPtclparam
  integer :: i, k

  do k=1, vsmMaxVisKind
    do i=1, vsmMaxPtcl
      vsmVisParam(k)%ptclparam(i)%radius = 0.0
      vsmVisParam(k)%ptclparam(i)%amb(1:3) = 0.0
      vsmVisParam(k)%ptclparam(i)%dif(1:3) = 0.0
      vsmVisParam(k)%ptclparam(i)%spec(1:3) = 0.0
      vsmVisParam(k)%ptclparam(i)%shin = 0.0
      vsmVisParam(k)%ptclparam(i)%monocolor = .false.
    end do
    vsmVisParam(k)%ptcl_num = 0
  end do
 end subroutine vsmInitPtclparam

 function vsmCheckPtcl(vnum, nptcl)
   integer, intent(in) :: vnum, nptcl
   logical :: vsmCheckPtcl

   if(vnum > vsmMaxVisKind .or. vnum < 1) then
     vsmCheckPtcl = .false.
     return
   end if

   if(nptcl < 1 .or. nptcl > vsmVisParam(vnum)%ptcl_num) then
     vsmCheckPtcl = .false.
     return
   end if
   vsmCheckPtcl = .true.
 end function vsmCheckPtcl

 subroutine vsmSetPtclParam(nptcl, mono, radius, amb, dif, spec, shin)
  real(kind=vdp), intent(in) ::radius
  real, dimension(3), intent(in) :: amb, dif, spec
  real, intent(in) :: shin
  integer, intent(in) :: nptcl
  logical, intent(in) :: mono
  integer :: ptcl_num

  ptcl_num = vsmVisParam(vsmNumViskind)%ptcl_num + 1
  if(ptcl_num > vsmMaxPtcl) call vsmError('vsmSetPtclParam')

  call vsmCheckFA01(amb); call vsmCheckFA01(dif); call vsmCheckFA01(spec)
  vsmVisParam(vsmNumViskind)%ptcl_num = ptcl_num
  vsmVisParam(vsmNumViskind)%ptclparam(ptcl_num)%nptcl = nptcl
  vsmVisParam(vsmNumViskind)%ptclparam(ptcl_num)%radius = radius
  vsmVisParam(vsmNumViskind)%ptclparam(ptcl_num)%amb(1:3) = amb(1:3)
  vsmVisParam(vsmNumViskind)%ptclparam(ptcl_num)%dif(1:3) = dif(1:3)
  vsmVisParam(vsmNumViskind)%ptclparam(ptcl_num)%spec(1:3) = spec(1:3)
  vsmVisParam(vsmNumViskind)%ptclparam(ptcl_num)%shin = shin
  vsmVisParam(vsmNumViskind)%ptclparam(ptcl_num)%monocolor = mono
 end subroutine vsmSetPtclParam

 subroutine vsmGetPtclCol(visnum, nptcl, amb, dif, spec, shin)
  integer, intent(in) :: nptcl, visnum
  real, dimension(3), intent(out) :: amb, dif, spec
  real, intent(out) :: shin

  if(.not.vsmCheckPtcl(visnum, nptcl)) call vsmError("error: vsmGetPtclCol")
  amb(1:3)  = vsmVisParam(visnum)%ptclparam(nptcl)%amb(1:3)
  dif(1:3)  = vsmVisParam(visnum)%ptclparam(nptcl)%dif(1:3)
  spec(1:3) = vsmVisParam(visnum)%ptclparam(nptcl)%spec(1:3)
  shin = vsmVisParam(visnum)%ptclparam(nptcl)%shin
 end subroutine vsmGetPtclCol

 subroutine vsmGetMPIVars(comm, rank, psize, ompn)
   integer, intent(out) :: comm
   integer, intent(out), optional :: rank
   integer, intent(out), optional :: psize
   integer, intent(out), optional :: ompn

   comm = vsmMPIVars%comm
   if(present(rank))  rank  = vsmMPIVars%myrank
   if(present(psize)) psize = vsmMPIVars%psize
   if(present(ompn))  ompn  = vsmMPIVars%omp_tn
 end subroutine vsmGetMPIVars

end module vsmParam

